/* -*- linux-c -*- */

/*
 *
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 */
/*
 * linux/drivers/evms/AIXlvm_vge.c
 *
 * EVMS AIX LVM Volume Group Emulator
 *
 *
 */

#define EVMS_DEBUG     1
#define EVMS_AIX_DEBUG 1

#define AIX_COMMON_SERVICES_MAJOR        0	// Required common services levels for the AIX kernel plugin
#define AIX_COMMON_SERVICES_MINOR        5	// These must be incremented if new function is added to common
#define AIX_COMMON_SERVICES_PATCHLEVEL   0	// services and the AIX kernel plugin uses the new function.
#define AIX_INCREMENT_REQUEST            1
#define AIX_DECREMENT_REQUEST           -1
#define AIX_RESYNC_BLOCKSIZE           512
#define AIX_SYNC_INCOMPLETE           0x01
#define AIX_SYNC_COMPLETE             0x00
#define AIX_MASTER                       0
#define AIX_SLAVE_1                      1
#define AIX_SLAVE_2                      2

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>

#include <linux/genhd.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/init.h>
#include <linux/slab.h>

#include <linux/evms/evms.h>
#include <linux/evms/evms_aix.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/locks.h>
#include <linux/delay.h>
#include <linux/reboot.h>
#include <linux/completion.h>
#include <linux/vmalloc.h>

#ifdef EVMS_AIX_DEBUG
static int AIX_volume_group_dump(void);
#endif

static struct aix_volume_group *AIXVolumeGroupList = NULL;
static struct evms_thread *AIX_mirror_read_retry_thread;
static struct evms_thread *AIX_mirror_resync_thread;
static struct evms_pool_mgmt *AIX_BH_list_pool = NULL;
static struct aix_mirror_bh *AIX_retry_list = NULL;
static struct aix_mirror_bh **AIX_retry_tail = NULL;
static spinlock_t AIX_retry_list_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t AIX_resync_list_lock = SPIN_LOCK_UNLOCKED;
static spinlock_t AIX_resync_pp_lock = SPIN_LOCK_UNLOCKED;
static int AIXResyncInProgress = FALSE;
static struct aix_resync_struct *AIX_resync_list = NULL;

// Plugin API prototypes

static void AIXiod(void *data);
static void AIXresync(void *data);
static int discover_aix(struct evms_logical_node **evms_logical_disk_head);
static int discover_volume_groups(struct evms_logical_node **);
static int discover_logical_volumes(void);
static int end_discover_aix(struct evms_logical_node **evms_logical_disk_head);
static void read_aix(struct evms_logical_node *node, struct buffer_head *bh);
static void write_aix(struct evms_logical_node *node, struct buffer_head *bh);
static int ioctl_aix(struct evms_logical_node *logical_node,
		     struct inode *inode,
		     struct file *file, unsigned int cmd, unsigned long arg);

static int aix_direct_ioctl(struct inode *inode,
			    struct file *file,
			    unsigned int cmd, unsigned long args);

static int AIX_remap_sector(struct evms_logical_node *node, u64 org_sector,	// logical sector to remap
			    u64 size,	// size (in sectors) of request to remap
			    u64 * new_sector,	// remapped sector
			    u64 * new_size,	// new size (in sectors)
			    struct partition_list_entry **partition,	// new node for which new_sector is relative
			    u32 * le, u32 * offset_in_le);

static int validate_build_volume_group_disk_info(struct evms_logical_node
						 *logical_node,
						 struct AIXlvm_rec *AIXlvm);

static int add_VG_data_to_VG_list(struct evms_logical_node *logical_node,
				  struct aix_volume_group *new_group,
				  short int pvNum);
static int add_PV_to_volume_group(struct aix_volume_group *group,
				  struct evms_logical_node *evms_partition,
				  int pvNum);
static struct aix_volume_group *AIX_create_volume_group(struct evms_logical_node
							*logical_node,
							struct AIXlvm_rec
							*AIXlvm);

static int AIX_update_volume_group(struct aix_volume_group *AIXVGLptr,
				   struct evms_logical_node *logical_node,
				   struct AIXlvm_rec *AIXlvm);

static int AIX_evms_cs_notify_lv_io_error(struct evms_logical_node *node);

static int AIX_pvh_data_posn(u32 vgda_psn, u32 * pvh_posn, struct partition_list_entry *partition, u32 numpvs);

static int AIX_resync_lv_mirrors(struct aix_logical_volume *volume, int force);

static int AIX_copy_on_read(struct aix_logical_volume *volume,
			    struct partition_list_entry *master_part,
			    struct partition_list_entry *slave1_part,
			    struct partition_list_entry *slave2_part,
			    u64 master_offset,
			    u64 slave1_offset,
			    u64 slave2_offset, u32 pe_size, int le);

static int export_volumes(struct evms_logical_node **evms_logical_disk_head);
static int lvm_cleanup(void);
static int AIX_copy_header_info(struct vg_header *AIXvgh,
				struct vg_header *AIXvgh2);
static int build_pe_maps(struct aix_volume_group *volume_group);

static struct aix_logical_volume *new_logical_volume(struct lv_entries
						     *AIXlvent,
						     struct aix_volume_group
						     *group, char *lv_name,
						     u32 stripesize);

static int check_log_volume_and_pe_maps(struct aix_volume_group *group);
static int check_volume_groups(void);
static int init_io_aix(struct evms_logical_node *node, int io_flag,	/* 0=read, 1=write */
		       u64 sect_nr,	/* disk LBA */
		       u64 num_sects,	/* # of sectors */
		       void *buf_addr);	/* buffer address */

static int delete_logical_volume(struct aix_logical_volume *volume);
static int delete_aix_node(struct evms_logical_node *logical_node);
static int deallocate_volume_group(struct aix_volume_group *group);

static void AIX_handle_read_mirror_drives(struct buffer_head *bh, int uptodate);

static void AIX_handle_write_mirror_drives(struct buffer_head *bh,
					   int uptodate);

static void aix_notify_cache_ctor(void *foo, kmem_cache_t * cachep,
				  unsigned long flags);

static void AIX_schedule_resync(struct aix_logical_volume *resync_volume,
				int force);
static struct aix_logical_volume *AIX_get_volume_data(char *object_name);

static void AIX_sync_mirrored_partitions(struct buffer_head *bh, int uptodate);

static int AIX_get_set_mirror_offset(struct aix_mirror_bh *tmp_bh,
				     int index, int offset);

static struct aix_mirror_bh *AIX_alloc_rbh(struct evms_logical_node *node,
					   struct buffer_head *bh,
					   u32 mirror_copies,
					   u32 le, u64 org_sector, int cmd);

static struct aix_mirror_bh *AIX_alloc_wbh(struct evms_logical_node *node,
					   struct evms_logical_node *node2,
					   struct evms_logical_node *node3,
					   struct buffer_head *bh,
					   u32 mirror_copies,
					   u32 le,
					   u64 new_sector2, u64 new_sector3);

static struct aix_mirror_bh *AIX_alloc_sbh(struct aix_logical_volume *volume,
					   struct partition_list_entry
					   *master_part,
					   struct partition_list_entry
					   *slave1_part,
					   struct partition_list_entry
					   *slave2_part, u64 master_offset,
					   u64 slave1_offset, u64 slave2_offset,
					   u32 pe_size);

static void AIX_free_headers(struct vg_header *AIXvgh,
			     struct vg_header *AIXvgh2,
			     struct vg_trailer *AIXvgt,
			     struct vg_trailer *AIXvgt2);

static int remove_group_from_list(struct aix_volume_group *group);

//****************************************************************************************************

/* END of PROTOTYES*/

#define GET_PHYSICAL_PART_SIZE(v1) (1 << v1)

#define COMPARE_TIMESTAMPS(t1, t2)	( (t1).tv_sec  == (t2).tv_sec && \
					  (t1).tv_nsec == (t2).tv_nsec )

#define COMPARE_UNIQUE_IDS(id1, id2)	( (id1).word1 == (id2).word1 && \
					  (id1).word2 == (id2).word2 && \
					  (id1).word3 == (id2).word3 && \
					  (id1).word4 == (id2).word4 )

#define SECTOR_IN_RANGE(s1, s2)  ((s2 > s1) && (s2 < s1 + AIX_RESYNC_BLOCKSIZE))

#define AIX_PV_STATE_VALID	         0	// Both VGDAs are valid and match.
#define AIX_PV_STATE_FIRST_VGDA		 1	// Only the first VGDA is valid.
#define AIX_PV_STATE_SECOND_VGDA	 2	// Only the second VGDA is valid.
#define AIX_PV_STATE_EITHER_VGDA	-1	// Both VGDAs are valid, but do not match each other.
#define AIX_PV_STATE_INVALID            -2	// We're in an invalid state but there's more PVs in this group

#ifndef EVMS_AIX_DEBUG
#define AIX_VOLUME_GROUP_DUMP()
#else
#define AIX_VOLUME_GROUP_DUMP() LOG_DEBUG("Called line:%d \n",__LINE__); \
                                AIX_volume_group_dump()
#endif

// Global LVM data structures

static struct evms_plugin_fops AIXlvm_fops = {
	.discover = discover_aix,
	.end_discover = end_discover_aix,
	.delete = delete_aix_node,
	.read = read_aix,
	.write = write_aix,
	.init_io = init_io_aix,
	.ioctl = ioctl_aix,
	.direct_ioctl = aix_direct_ioctl
};

static struct evms_plugin_header plugin_header = {
	.id = SetPluginID(IBM_OEM_ID,
			  EVMS_REGION_MANAGER,
			  EVMS_AIX_FEATURE_ID),
	.version = {
		    .major = 1,
		    .minor = 1,
		    .patchlevel = 2},
	.required_services_version = {
				      .major = AIX_COMMON_SERVICES_MAJOR,
				      .minor = AIX_COMMON_SERVICES_MINOR,
				      .patchlevel =
				      AIX_COMMON_SERVICES_PATCHLEVEL},
	.fops = &AIXlvm_fops
};

/*
 * Function: remap sector 
 *  Common function to remap volume lba to partition lba in appropriate PE
 */
static int
AIX_remap_sector(struct evms_logical_node *node, u64 org_sector,	// logical sector to remap
		 u64 size,	// size (in sectors) of request to remap
		 u64 * new_sector,	// remapped sector
		 u64 * new_size,	// new size (in sectors)
		 struct partition_list_entry **partition,	// new node for which new_sector is relative
		 u32 * le, u32 * offset_in_le)
{
	struct aix_logical_volume *volume;

	u32 sectors_per_stripe;
	u32 partition_to_use;
	u32 column;
	u32 stripe_in_column;

	u32 org_sector32;	// Until striping is 64-bit enabled.

	volume = (struct aix_logical_volume *) node->private;

#ifdef EVMS_DEBUG
	LOG_DEBUG("-- %s volume:%p lv:%d size:" PFU64 " Name:%s\n",
		  __FUNCTION__, volume, volume->lv_number, size, volume->name);
	LOG_DEBUG(" node %p node_name [%s] org_sector:" PFU64 "\n", node,
		  node->name, org_sector);
	LOG_DEBUG(" mirror_copies:%d volume->lv_size:" PFU64 "\n",
		  volume->mirror_copies, volume->lv_size);
#endif

	org_sector32 = org_sector;

	*(new_size) = size;

	// Check if volume is striped. Reset the size if the request
	// crosses a stripe boundary.
	if (volume->stripes > 1) {
#ifdef EVMS_DEBUG
		LOG_DEBUG(" *** STRIPED ***\n");
		LOG_DEBUG(" ------- volume->stripe_size:%d org_sector:%d volume_stripes:%d\n",
		     volume->stripe_size, org_sector32, volume->stripes);
#endif

		*(le) = org_sector >> volume->pe_size_shift;	// 64-bit safe
		*(offset_in_le) = org_sector & (volume->pe_size - 1);	// 64-bit safe

#ifdef EVMS_DEBUG
		LOG_DEBUG("OLD - le:%d -- offset_in_le:%d \n", *(le),
			  *(offset_in_le));
#endif

		sectors_per_stripe = volume->stripe_size / AIX_SECTOR_SIZE;
		partition_to_use =
		    (org_sector32 / sectors_per_stripe) % volume->stripes;
		stripe_in_column =
		    ((((org_sector32 / volume->stripe_size) / volume->stripes) *
		      volume->stripe_size) +
		     (org_sector32 % sectors_per_stripe));
		column =
		    ((org_sector32 / sectors_per_stripe) / volume->stripes) *
		    sectors_per_stripe;

#ifdef EVMS_DEBUG
		LOG_DEBUG("offset_in_le:%d org_sector:" PFU64
			  " pe_shift:%d stripe_shift:%d\n", *(offset_in_le),
			  org_sector, volume->pe_size_shift,
			  volume->stripe_size_shift);

		LOG_DEBUG(" org_sector:%d  sectors_per_stripe:%d partition_to_use:%d stripe_in_column:%d column:%d\n",
		     org_sector32, sectors_per_stripe, partition_to_use,
		     stripe_in_column, column);
		LOG_DEBUG(" offset_in_le + size:" PFU64
			  " volume->pe_size:%d volume->lv_size:" PFU64 "\n",
			  (*(offset_in_le) + size), volume->pe_size,
			  volume->lv_size);
#endif

		if (*(offset_in_le) + size > volume->pe_size) {
			*new_size = volume->pe_size - *(offset_in_le);
			LOG_DEBUG("  new_size " PFU64 "\n", *new_size);
		}

	}
	// Non-striped volume. Just find LE and offset. Reset the size
	// if the request crosses an LE boundary.
	else {
#ifdef EVMS_DEBUG
		LOG_DEBUG(" *** NON-STRIPED ***\n");
#endif

		*(le) = org_sector >> volume->pe_size_shift;	// 64-bit safe
		*(offset_in_le) = org_sector & (volume->pe_size - 1);	// 64-bit safe

	}

#ifdef EVMS_DEBUG
	LOG_DEBUG(" offset_in_le:%d org_sector:" PFU64 " shift:%d\n",
		  *(offset_in_le), org_sector, volume->pe_size_shift);

	if (*(le) >= volume->num_le) {
		LOG_DEBUG(" le Memory Overwrite !! le:%d vs volume->num_le:%d\n",
		     *(le), volume->num_le);
		return -EINVAL;
	}
#endif

	*(new_sector) = volume->le_to_pe_map[*(le)].pe_sector_offset + *(offset_in_le);
	*(partition) = volume->le_to_pe_map[*(le)].owning_pv;

#ifdef EVMS_DEBUG
	LOG_DEBUG(" new_sector:" PFU64 "\n", *(new_sector));
	LOG_DEBUG(" Owning Part %p\n", *(partition));
	LOG_DEBUG(" End %s\n", __FUNCTION__);
#endif

	return (0);
}

/*
 * Function: read_aix
 */
static void
read_aix(struct evms_logical_node *node, struct buffer_head *bh)
{
	struct partition_list_entry *partition;
	u64 org_sector;
	u64 new_sector;
	u64 new_size;
	struct aix_logical_volume *volume;
	struct aix_mirror_bh *tmp_bh;
	u32 le, offset_in_le, count;
	int flags = 0;

	volume = (struct aix_logical_volume *) node->private;
//#ifdef EVMS_DEBUG
//	LOG_DEBUG(" ***** %s ***** bh:%p volume->iter:%d\n", __FUNCTION__, bh,
//		  volume->mirror_iterations);
//#endif

#ifdef EVMS_DEBUG
	LOG_DEBUG(" node->total_vsectors:" PFU64 "\n", node->total_vsectors);
	LOG_DEBUG(" rsector:%lu rsize:%u node_flags:%u\n", bh->b_rsector,
		  bh->b_size, node->flags);
#endif

	// Check if I/O goes past end of logical volume.
	if (bh->b_rsector + (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) >
	    node->total_vsectors) {
		LOG_CRITICAL(" read_aix ERROR %d\n", __LINE__);
		buffer_IO_error(bh);
		return;
	}

	// Logical-to-physical remapping.
	if (AIX_remap_sector
	    (node, bh->b_rsector, (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT),
	     &new_sector, &new_size, &partition, &le, &offset_in_le)
	    || (!partition || !new_sector)) {
		LOG_CRITICAL(" read_aix bh: ERROR %d\n", __LINE__);
		buffer_IO_error(bh);
		return;
	}

	org_sector = bh->b_rsector;
	bh->b_rsector = new_sector;
	//bh->b_size    = new_size;  

#ifdef EVMS_DEBUG
	LOG_DEBUG(" read_aix Mirror_Copies:%d\n", volume->mirror_copies);
#endif

	if (volume->mirror_copies > AIX_DEFAULT_MIRRORING) {

		tmp_bh =
		    AIX_alloc_rbh(node, bh, 1, le, new_sector, AIX_LV_READ);

		if (!tmp_bh) {
			buffer_IO_error(bh);
			return;
		}

		if (volume->le_to_pe_map_mir1) {
			tmp_bh->mir_node1 =
			    volume->le_to_pe_map_mir1[le].owning_pv->
			    logical_node;
			tmp_bh->mir_sector1 =
			    volume->le_to_pe_map_mir1[le].pe_sector_offset +
			    offset_in_le;
		}

		if (volume->mirror_copies == AIX_MAX_MIRRORS) {
			tmp_bh->mir_node2 =
			    volume->le_to_pe_map_mir2[le].owning_pv->
			    logical_node;
			tmp_bh->mir_sector2 =
			    volume->le_to_pe_map_mir2[le].pe_sector_offset +
			    offset_in_le;
		}

		if (evms_cs_volume_request_in_progress
		    (tmp_bh->bh_req.b_rdev, AIX_INCREMENT_REQUEST, &count)) {
			buffer_IO_error(bh);
			return;
		}

		if (AIXResyncInProgress) {
			if (SECTOR_IN_RANGE
			    (tmp_bh->bh_req.b_rsector,
			     AIX_resync_list->master_offset)) {
				spin_lock_irqsave(&AIX_resync_list_lock, flags);
			}
		}

		R_IO(partition->logical_node, &tmp_bh->bh_req);

		if (AIXResyncInProgress) {
			if (SECTOR_IN_RANGE
			    (tmp_bh->bh_req.b_rsector,
			     AIX_resync_list->master_offset)) {
				spin_unlock_irqrestore(&AIX_resync_list_lock,
						       flags);
			}
		}

	} else {

		R_IO(partition->logical_node, bh);
	}

#ifdef EVMS_DEBUG
	LOG_DEBUG(" ***** %s ***** returning\n", __FUNCTION__);
#endif
	return;
}

/*
 * Function: write_aix
 */
static void
write_aix(struct evms_logical_node *node, struct buffer_head *bh)
{
	struct partition_list_entry *partition;
	u64 new_sector, new_sector2 = 0, new_sector3 = 0;
	u64 org_sector;
	u64 new_size;
	struct aix_logical_volume *volume;
	struct aix_mirror_bh *tmp_bh;
	struct evms_logical_node *node2 = NULL, *node3 = NULL;
	u32 le, offset_in_le, count;
	int flags = 0;

	volume = (struct aix_logical_volume *) node->private;

#ifdef EVMS_DEBUG
//	LOG_DEBUG(" ***** %s ***** bh:%p volume->iter:%d\n", __FUNCTION__, bh,
//		  volume->mirror_iterations);
	LOG_DEBUG(" write_aix rsector:%lu rsize:%u\n", bh->b_rsector,
		  bh->b_size);
	LOG_DEBUG(" write_aix total_sectors:" PFU64 "\n", node->total_vsectors);
#endif

	if (volume->lv_access & EVMS_LV_INCOMPLETE) {	//No writes allowed on incomplete volumes
		LOG_CRITICAL(" write_aix incomplete volume ERROR %d\n",
			     __LINE__);
		buffer_IO_error(bh);
		return;
	}

	// Check if I/O goes past end of logical volume.
	if (bh->b_rsector + (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) >
	    node->total_vsectors) {
		LOG_CRITICAL(" write_aix ERROR %d\n", __LINE__);
		buffer_IO_error(bh);
		return;
	}
	// Logical-to-Physical remapping
	if (AIX_remap_sector
	    (node, bh->b_rsector, (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT),
	     &new_sector, &new_size, &partition, &le, &offset_in_le)
	    || (!new_sector || !partition)) {
		LOG_CRITICAL(" write_aix ERROR %d\n", __LINE__);
		buffer_IO_error(bh);
		return;
	}

	org_sector = bh->b_rsector;
	bh->b_rsector = new_sector;
	//bh->b_size   = new_size;  

#ifdef EVMS_DEBUG
	LOG_DEBUG(" write_aix  Mirror_Copies:%d\n", volume->mirror_copies);
#endif

	if (volume->mirror_copies > AIX_DEFAULT_MIRRORING) {

		if (volume->le_to_pe_map_mir1) {
			new_sector2 =
			    volume->le_to_pe_map_mir1[le].pe_sector_offset +
			    offset_in_le;
			node2 =
			    volume->le_to_pe_map_mir1[le].owning_pv->
			    logical_node;
		}

		if (volume->mirror_copies == AIX_MAX_MIRRORS) {

			new_sector3 =
			    volume->le_to_pe_map_mir2[le].pe_sector_offset +
			    offset_in_le;
			node3 =
			    volume->le_to_pe_map_mir2[le].owning_pv->
			    logical_node;
		}

		tmp_bh =
		    AIX_alloc_wbh(partition->logical_node, node2, node3, bh, le,
				  volume->mirror_copies, new_sector2,
				  new_sector3);

		if (!tmp_bh) {
			buffer_IO_error(bh);
			return;
		}
		tmp_bh->node = node;

		tmp_bh = tmp_bh->mirror_bh_list;

		if (evms_cs_volume_request_in_progress
		    (tmp_bh->bh_req.b_rdev, AIX_INCREMENT_REQUEST, &count)) {
			buffer_IO_error(bh);
			// free memory here
			return;
		}

		if (AIXResyncInProgress) {
			if (SECTOR_IN_RANGE
			    (tmp_bh->bh_req.b_rsector,
			     AIX_resync_list->master_offset)) {
				spin_lock_irqsave(&AIX_resync_list_lock, flags);
			}
		}

		W_IO(tmp_bh->node, &tmp_bh->bh_req);

		if (AIXResyncInProgress) {
			if (SECTOR_IN_RANGE
			    (tmp_bh->bh_req.b_rsector,
			     AIX_resync_list->master_offset)) {
				spin_unlock_irqrestore(&AIX_resync_list_lock,
						       flags);
			}
		}

		tmp_bh = tmp_bh->next_r1;

		if (tmp_bh) {
			W_IO(tmp_bh->node, &tmp_bh->bh_req);
			tmp_bh = tmp_bh->next_r1;
		}

		if (tmp_bh) {
			W_IO(tmp_bh->node, &tmp_bh->bh_req);
		}

	} else {

		W_IO(partition->logical_node, bh);
	}

#ifdef EVMS_DEBUG
	LOG_DEBUG(" ***** %s returning *****\n", __FUNCTION__);
#endif
	return;
}

/*
 * Function: ioctl_aix
 *
 */
static int
ioctl_aix(struct evms_logical_node *logical_node,
	  struct inode *inode,
	  struct file *file, unsigned int cmd, unsigned long arg)
{
	struct aix_logical_volume *volume =
	    (struct aix_logical_volume *) (logical_node->private);
	int rc = 0;

	LOG_EXTRA(" Ioctl %u\n", cmd);

	switch (cmd) {

	case HDIO_GETGEO:
		{
			// Fixed geomerty for all LVM volumes 
			unsigned char heads = 64;
			unsigned char sectors = 32;
			long start = 0;
			struct hd_geometry *hd = (struct hd_geometry *) arg;
			short cylinders;
			cylinders = logical_node->total_vsectors;
			cylinders = (cylinders / heads) / sectors;

			if (hd == NULL) {
				return -EINVAL;
			}

			if (copy_to_user
			    ((char *) (&hd->heads), &heads, sizeof (heads)) != 0
			    || copy_to_user((char *) (&hd->sectors), &sectors,
					    sizeof (sectors)) != 0
			    || copy_to_user((short *) (&hd->cylinders),
					    &cylinders, sizeof (cylinders)) != 0
			    || copy_to_user((long *) (&hd->start), &start,
					    sizeof (start)) != 0) {
				return -EFAULT;
			}
		}
		break;

	case EVMS_QUIESCE_VOLUME:
		break;

	case EVMS_GET_DISK_LIST:
	case EVMS_CHECK_MEDIA_CHANGE:
	case EVMS_REVALIDATE_DISK:
	case EVMS_OPEN_VOLUME:
	case EVMS_CLOSE_VOLUME:
	case EVMS_CHECK_DEVICE_STATUS:
		{
			// These five ioctl all need to be broadcast to all PVs.
			struct aix_volume_group *group = volume->group;
			struct partition_list_entry *partition;
			for (partition = group->partition_list; partition;
			     partition = partition->next) {
				rc |=
				    IOCTL(partition->logical_node, inode, file,
					  cmd, arg);
			}
		}
		break;

	default:
		// Currently the VGE does not send any ioctl's down to the
		// partitions. Which partition would they go to?
		rc = -ENOTTY;
	}

	return rc;
}

/* Function: aix_direct_ioctl
 *
 *	This function provides a method for user-space to communicate directly
 *	with a plugin in the kernel.
 */
static int
aix_direct_ioctl(struct inode *inode,
		 struct file *file, unsigned int cmd, unsigned long args)
{
	struct aix_logical_volume *volume = NULL;
	struct evms_plugin_ioctl_pkt argument;
	int rc = 0;

	MOD_INC_USE_COUNT;
	LOG_DEBUG(" Function:%s cmd:%d \n", __FUNCTION__, cmd);

	// Copy user's parameters to kernel space
	if (copy_from_user
	    (&argument, (struct evms_plugin_ioctl *) args, sizeof (argument))) {
		MOD_DEC_USE_COUNT;
		return -EFAULT;
	}
	// Make sure this is supposed to be our ioctl.
	if (argument.feature_id != plugin_header.id) {
		MOD_DEC_USE_COUNT;
		return -EINVAL;
	}

	argument.feature_command = 1;

	switch (argument.feature_command) {

	case EVMS_AIX_RESYNC_MIRRORS:
		{
			struct aix_volume_resync_ioctl aix_lv_resync;

			if (copy_from_user
			    (&aix_lv_resync,
			     (struct aix_volume_resync_ioctl *) argument.
			     feature_ioctl_data, sizeof (aix_lv_resync))) {
				rc = -EINVAL;
				break;
			}

			volume = AIX_get_volume_data(aix_lv_resync.object_name);

			if (volume) {
				AIX_schedule_resync(volume, FALSE);
			} else {
				LOG_DEBUG
				    (" Function:%s object_name:%s -- no match found\n",
				     __FUNCTION__, aix_lv_resync.object_name);
				rc = -EINVAL;
			}

		}
		break;

	default:
		rc = -EINVAL;
		break;
	}

	argument.status = rc;
	copy_to_user((struct evms_plugin_ioctl *) args, &argument,
		     sizeof (argument));
	MOD_DEC_USE_COUNT;
	return rc;
}

/* Function: aix_direct_ioctl
 *
 *	This function provides a method for user-space to communicate directly
 *	with a plugin in the kernel.
 */
static struct aix_logical_volume *
AIX_get_volume_data(char *object_name)
{

	struct aix_volume_group *VG_ptr;
	struct aix_logical_volume *volume = NULL;
	int i;

	LOG_DEBUG(" Function:%s object_name:%s \n", __FUNCTION__, object_name);

	if (!object_name || !strlen(object_name)) {
		return NULL;
	}

	for (VG_ptr = AIXVolumeGroupList; VG_ptr; VG_ptr = VG_ptr->next) {
		for (i = 0; VG_ptr->volume_list[i]; i++) {
			if (!strcmp(VG_ptr->volume_list[i]->name, object_name)) {
				LOG_DEBUG
				    (" Function:%s FOUND!! volume_name:%s \n",
				     __FUNCTION__,
				     VG_ptr->volume_list[i]->name);
				volume = VG_ptr->volume_list[i];
				break;
			}
		}
	}

	if (!volume) {
		LOG_DEBUG(" Function:%s object_name:%s NOT FOUND !! volume:%p \n",
		     __FUNCTION__, object_name, volume);
	}

	return volume;
}

/*
 * Function: init_io_aix
 *
 */
static int
init_io_aix(struct evms_logical_node *node, int io_flag,	/* 0=read, 1=write */
	    u64 sect_nr,	/* disk LBA */
	    u64 num_sects,	/* # of sectors */
	    void *buf_addr)
{				/* buffer address */
	struct partition_list_entry *partition;
	u64 new_sector = 0;
	u64 new_size = 0;
	int rc = 0;
	u32 le, offset;

	LOG_DEBUG(" ************ init_io_aix() num_sects:" PFU64
		  " node:%p sect_nr:" PFU64 "\n", num_sects, node, sect_nr);

	// Init IO needs to deal with the possibility that a request can come
	// in that spans PEs or stripes. This is possible because there is no
	// limit on num_sects. To fix this, we loop through AIX_remap_sector and
	// INIT_IO until num_sects reaches zero.

	while (num_sects > 0) {

		if (AIX_remap_sector(node, sect_nr, num_sects, &new_sector, &new_size,
		     &partition, &le, &offset) || (!new_sector || !partition)) {
			LOG_CRITICAL("--- Error returned from AIX_remap_sector %d\n",
			     __LINE__);
			return -EIO;
		}

		LOG_DEBUG(" init_io_aix() line:%d logical_node:%p io_flag:%d new_sector:"
		     PFU64 " new_size:" PFU64 "\n", __LINE__,
		     partition->logical_node, io_flag, new_sector, new_size);

		rc = INIT_IO(partition->logical_node, io_flag, new_sector,
			     new_size, buf_addr);
		num_sects -= new_size;
		sect_nr += new_size;
		buf_addr = (void *) (((unsigned long) buf_addr) +
			      (unsigned long) (new_size << EVMS_VSECTOR_SIZE_SHIFT));
	}

	return rc;
}

/*
 * Function: AIXlvm_vge_init
 *
 */
int __init
AIXlvm_vge_init(void)
{

	LOG_DEBUG(" %s --------\n", __FUNCTION__);

	MOD_INC_USE_COUNT;
	return evms_cs_register_plugin(&plugin_header);	/* register with EVMS */
}

module_init(AIXlvm_vge_init);

/********** Required Plugin Functions **********/

/*
 * Function: discover_aix
 *
 *  This is the entry point into the LVM discovery process.
 */
static int
discover_aix(struct evms_logical_node **evms_logical_disk_head)
{
	int rc = 0, count = 0;

	MOD_INC_USE_COUNT;
	LOG_DEBUG("[%s] discover_volume_groups\n", __FUNCTION__);

	rc = discover_volume_groups(evms_logical_disk_head);

	if (rc) {
		LOG_ERROR("[%s] discover_volume_groups rc=%d\n", __FUNCTION__,rc);
	}

	if (AIXVolumeGroupList && !rc) {

		LOG_DEBUG("[%s] discover_logical_volumes\n", __FUNCTION__);

		rc = discover_logical_volumes();

		if (rc) {
			LOG_ERROR("[%s] discover_logical_volumes rc=%d\n",
				  __FUNCTION__, rc);
		}

		LOG_DEBUG("[%s] export_volumes\n", __FUNCTION__);

		count = export_volumes(evms_logical_disk_head);

		LOG_DEBUG("[%s] export_volumes count=%d\n", __FUNCTION__,
			  count);
	}

	MOD_DEC_USE_COUNT;
	return (count);
}

static int
discover_volume_groups(struct evms_logical_node **evms_logical_disk_head)
{
	struct evms_logical_node *logical_node;
	struct evms_logical_node *next_node;
	struct aix_ipl_rec_area *AIXpv;
	struct AIXlvm_rec *AIXlvm;	// Temp holder for the LVM on disk rec

	LOG_DEBUG(" Begin %s\n", __FUNCTION__);

	AIXpv = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXpv) {
		return -ENOMEM;
	}

	// We'll create at least one volume entry, if we don't find any AIX volumes we'll clean it up later

	AIXlvm = kmalloc(sizeof (struct AIXlvm_rec), GFP_KERNEL);
	if (!AIXlvm) {
		kfree(AIXpv);
		return -ENOMEM;
	}

	for (logical_node = *evms_logical_disk_head; logical_node;
	     logical_node = next_node) {

		// Grab the next list item in case we remove this partition from the global list.
		next_node = logical_node->next;

		// Read the first sector and see if it has a valid AIX PV signature.

		if (INIT_IO(logical_node, 0, 0, 1, AIXpv)) {
			// On an I/O error, continue on to the next
			// partition. The group that this partition
			// belongs to will be incomplete, but we still
			// need to discover any other groups.

			LOG_ERROR(" Error reading PV [%p]\n", logical_node);
			continue;
		}

		if (AIXpv->IPL_record_id == IPLRECID) {

			// This partition is definitely a PV,
			// but is it part of a valid VG?
			LOG_DEBUG(" DVG removing node from list logical_node %p\n",
			     logical_node);

			if (INIT_IO(logical_node, 0, PSN_LVM_REC, 1, AIXlvm)) {
				LOG_ERROR(" Error reading PV [%p]\n",logical_node);
				continue;
			}

			if (AIXlvm->lvm_id == AIX_LVM_LVMID) {

				if (validate_build_volume_group_disk_info(
					logical_node, AIXlvm)) {
					// Again, continue on and we'll
					// clean up later.
					continue;
				}

				evms_cs_remove_logical_node_from_list(
				    evms_logical_disk_head, logical_node);

			} else {
				LOG_DEBUG(" Found an AIX PV with no parent LVM (LVM ID: %d)\n",
				     AIXlvm->lvm_id);
				continue;
			}
		} else {
			LOG_DEBUG(" Found a PV not belonging to AIX [%p]\n",
				  logical_node);
		}
	}

	AIX_VOLUME_GROUP_DUMP();

	if (check_volume_groups()) {
		return -EINVAL;
	}

	kfree(AIXpv);
	kfree(AIXlvm);

	return 0;
}

/*
 * Function:  validate_build_volume_group_disk_info
 *
 *  Creates and validates the volume groups found on the disk structures.
 *  
 */
static int
validate_build_volume_group_disk_info(struct evms_logical_node *logical_node,
				      struct AIXlvm_rec *AIXlvm)
{

	struct aix_volume_group *AIXVGLptr = AIXVolumeGroupList;

	LOG_DEBUG(" VBVGDI pv_num:%d\n", AIXlvm->pv_num);

	while (AIXVGLptr) {
		if (COMPARE_UNIQUE_IDS(AIXlvm->vg_id, AIXVGLptr->vg_id)) {
			break;
		}
		AIXVGLptr = AIXVGLptr->next;	// There is more than one so walk the list 
	}

	if (!AIXVGLptr) {
		LOG_DEBUG(" VBVGDI AIXVGLptr:%p line:%d\n", AIXVGLptr,__LINE__);
		AIXVGLptr = AIX_create_volume_group(logical_node, AIXlvm);
        if (AIXVGLptr) {
            AIXVGLptr->next = AIXVolumeGroupList;
            AIXVolumeGroupList = AIXVGLptr;
        }
	} else {
		LOG_DEBUG(" VBVGDI Rediscover AIXVGLptr:%p line:%d\n",
			  AIXVGLptr, __LINE__);

		if (AIX_update_volume_group(AIXVGLptr, logical_node, AIXlvm)) {
			LOG_DEBUG
			    (" VBVGDI ERROR on Rediscover AIXVGLptr:%p  line:%d\n",
			     AIXVGLptr, __LINE__);
		}
	}

	if (!AIXVGLptr) {

		LOG_DEBUG(" VBVGDI AIXVGLptr:%p line:%d\n", AIXVGLptr,
			  __LINE__);
		LOG_DEBUG(" VBVGDI flags:%d\n", AIXVGLptr->flags);
		LOG_CRITICAL("Unable to allocate volume group data struct Volume Group Corruption !!\n");
		return -EINVAL;
	} else {

		LOG_DEBUG(" VBVGDI AIXVolumeGroupList:%p line:%d\n",
			  AIXVolumeGroupList, __LINE__);
		LOG_DEBUG(" VBVGDI AIXVGLptr:%p line:%d\n", AIXVGLptr,
			  __LINE__);
		LOG_DEBUG(" VBVGDI flags:%d\n", AIXVGLptr->flags);

		if (add_PV_to_volume_group(AIXVGLptr, logical_node, AIXlvm->pv_num)) {
			return -EINVAL;
		}
	}

	return 0;
}

/*
 * Function: add_VG_data_to_VG_list
 *
 *  Allocate space for a new LVM volume group and all of its sub-fields.
 *  Initialize the appropriate fields.
 */

static int
add_VG_data_to_VG_list(struct evms_logical_node *logical_node,
		       struct aix_volume_group *new_group, short int pvNum)
{
//	int pvh_pos;

//	struct pv_header *AIXpvh;

	// The array of pointer to the logical volumes.
	// Leave this allocation at the max permitted, the lv numbering may not be sequential so you may have gaps
	// in the array allocation i.e. 1,2,3,4,5,6,7,8,11,15,21,33 etc. even though you only have 12 LVs.

	LOG_DEBUG(" AVGDVGL Entering pvNum:%d vgda_PSN:%d\n", pvNum,
		  new_group->vgda_psn);

//	pvh_pos = AIX_PVH_DATA_PSN(new_group->vgda_psn, pvNum);

/*	AIXpvh = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXpvh) {
		return -ENOMEM;
	}

	memset(AIXpvh, 0, AIX_SECTOR_SIZE);

	LOG_DEBUG(" AVGDVGL pvh_pos:%d\n", pvh_pos);

	if (INIT_IO(logical_node, 0, pvh_pos, 1, AIXpvh)) {
		return -EIO;
	}

	LOG_DEBUG(" AVGDVGL AIXpvh->pv_num:%d\n", pvNum);
*/
	if (!new_group->volume_list) {
		new_group->volume_list =
		    kmalloc(LVM_MAXLVS * sizeof (struct aix_logical_volume *),
			    GFP_KERNEL);
		if (!new_group->volume_list) {
//			kfree(AIXpvh);
			return -ENOMEM;
		}
		memset(new_group->volume_list, 0,
		       (LVM_MAXLVS * sizeof (struct aix_logical_volume *)));
	}

	new_group->vg_id.word1 = new_group->AIXvgh->vg_id.word1;
	new_group->vg_id.word2 = new_group->AIXvgh->vg_id.word2;
	new_group->vg_id.word3 = new_group->AIXvgh->vg_id.word3;
	new_group->vg_id.word4 = new_group->AIXvgh->vg_id.word4;
//	new_group->numpvs = new_group->AIXvgh->numpvs;
//	new_group->numlvs = new_group->AIXvgh->numlvs;
//	new_group->lv_max = new_group->AIXvgh->maxlvs;
	new_group->pe_size = GET_PHYSICAL_PART_SIZE(new_group->AIXvgh->pp_size) /
				AIX_SECTOR_SIZE;

//	new_group->block_size = 0;
//	new_group->hard_sect_size = 0;
	new_group->flags |= AIX_VG_DIRTY;

//	kfree(AIXpvh);

	LOG_DEBUG(" AVGDVGL Vol Group ID %x\n", new_group->vg_id.word2);

	return 0;
}

/*
 * Function: add_PV_to_volume_group
 *
 *  Create a new partition_list_entry for the specified volume group.
 *  Initialize the new partition with the evms node and lvm pv information,
 *  and add the new partition to the group's list.
 */

static int
add_PV_to_volume_group(struct aix_volume_group *group,
		       struct evms_logical_node *evms_partition, int pvNum)
{
	struct partition_list_entry *new_partition;

	LOG_DEBUG(" APVVG Entering pvNum:%d\n", pvNum);

	group->flags |= AIX_VG_DIRTY;

	for (new_partition = group->partition_list; new_partition != NULL;
	     new_partition = new_partition->next) {
		if (new_partition->logical_node == evms_partition) {
			return 0;
		}
	}

	new_partition =
	    kmalloc(sizeof (struct partition_list_entry), GFP_KERNEL);
	if (!new_partition) {
		return -ENOMEM;
	}

	memset(new_partition, 0, sizeof (struct partition_list_entry));

	// Add this partition to this group's list.
	new_partition->logical_node = evms_partition;
	new_partition->pv_number = pvNum;

	if (evms_partition->hardsector_size > group->hard_sect_size) {
		group->hard_sect_size = evms_partition->hardsector_size;
	}
	if (evms_partition->block_size > group->block_size) {
		group->block_size = evms_partition->block_size;
	}

	// Add this partition to the beginning of its group's list.
	new_partition->next = group->partition_list;
	group->partition_list = new_partition;
	group->partition_count++;

	LOG_DEBUG(" APVVG partition_count:%d pv_num:%d\n",
		  group->partition_count, pvNum);

	return 0;
}

/****************************************************
*
*
*
*****************************************************/
static struct aix_volume_group *
AIX_create_volume_group(struct evms_logical_node *logical_node,
			struct AIXlvm_rec *AIXlvm)
{
	struct vg_header *AIXvgh = NULL, *AIXvgh2 = NULL;
	struct vg_trailer *AIXvgt = NULL, *AIXvgt2 = NULL;
	struct aix_volume_group *AIXVGLptr;

	AIXvgh = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgh) {
		return NULL;
	}

	AIXvgh2 = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgh2) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	AIXvgt = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgt) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	AIXvgt2 = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgt2) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	memset(AIXvgh, 0, AIX_SECTOR_SIZE);
	memset(AIXvgh2, 0, AIX_SECTOR_SIZE);
	memset(AIXvgt, 0, AIX_SECTOR_SIZE);
	memset(AIXvgt2, 0, AIX_SECTOR_SIZE);

	// First time thru we want to read this in, we may only have one PV in this group, all others 
	// may be corrupt, etc. If the info is clean we shouldn't get here.

	if (INIT_IO(logical_node, 0, AIXlvm->vgda_psn[0], 1, AIXvgh)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	if (INIT_IO(logical_node, 0, AIXlvm->vgda_psn[1], 1, AIXvgh2)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	if (INIT_IO(logical_node, 0, (AIXlvm->vgda_psn[0] + AIXlvm->vgda_len - 1), 1,
	     AIXvgt)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	if (INIT_IO(logical_node, 0, (AIXlvm->vgda_psn[1] + AIXlvm->vgda_len - 1), 1,
	     AIXvgt2)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}

	LOG_DEBUG("CVG AIXvgh->vgda_psn[%d]:%d\n", 0, AIXlvm->vgda_psn[0]);
	LOG_DEBUG("CVG AIXvgh->vgda_psn[%d]:%d\n", 1, AIXlvm->vgda_psn[1]);
	LOG_DEBUG("CVG AIXvgt psn[%d]:%d\n", 0,(AIXlvm->vgda_psn[0] + AIXlvm->vgda_len - 1));
	LOG_DEBUG("CVG AIXvgt psn[%d]:%d\n", 1,(AIXlvm->vgda_psn[1] + AIXlvm->vgda_len - 1));
	LOG_DEBUG("CVG Allocating AIXVGLptr:size:%d \n",(int) sizeof (struct aix_volume_group));

	AIXVGLptr = kmalloc(sizeof (struct aix_volume_group), GFP_KERNEL);
	if (!AIXVGLptr) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}
	memset(AIXVGLptr, 0, sizeof (struct aix_volume_group));

	AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;
	AIXVGLptr->flags |= AIX_VG_DIRTY;

	LOG_DEBUG("CVG AIXVGLptr:%p line %d\n", AIXVGLptr, __LINE__);

	AIXVGLptr->AIXvgh = kmalloc(sizeof (struct vg_header), GFP_KERNEL);
	if (!AIXVGLptr->AIXvgh) {
		kfree(AIXVGLptr);
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return NULL;
	}
	memset(AIXVGLptr->AIXvgh, 0, sizeof (struct vg_header));

	LOG_DEBUG("CVG COMP TS AIXVGLptr->CleanVGInfo:%d \n",
		  AIXVGLptr->CleanVGInfo);

	if (AIXVGLptr->CleanVGInfo == AIX_PV_STATE_INVALID) {
		if (COMPARE_TIMESTAMPS(AIXvgh->vg_timestamp, AIXvgt->timestamp)) {
			if (COMPARE_TIMESTAMPS
			    (AIXvgh2->vg_timestamp, AIXvgt2->timestamp)) {
				if (COMPARE_TIMESTAMPS
				    (AIXvgh->vg_timestamp,
				     AIXvgh2->vg_timestamp)) {
					// All timestamps match. Yea!
					AIXVGLptr->CleanVGInfo =
					    AIX_PV_STATE_VALID;
				} else {
					// Both VGDAs are good, but timestamps are
					// different. Can't tell yet which one is
					// correct. 
					AIXVGLptr->CleanVGInfo =
					    AIX_PV_STATE_EITHER_VGDA;
				}
			} else {
				// First VGDA is good, second is bad.
				AIXVGLptr->CleanVGInfo =
				    AIX_PV_STATE_FIRST_VGDA;
			}
		} else {
			if (COMPARE_TIMESTAMPS
			    (AIXvgh2->vg_timestamp, AIXvgt2->timestamp)) {
				// First VGDA is bad, second is good.
				AIXVGLptr->CleanVGInfo =
				    AIX_PV_STATE_SECOND_VGDA;
			} else if (AIXvgh->numpvs == 1) {	// We only have 1 PV in this group, mismatch or not this will have to do 
				AIXVGLptr->CleanVGInfo = AIX_PV_STATE_VALID;
			} else {
				// This should never happen.
				LOG_DEBUG("All four VG timestamps for %d are different. What happened?!?\n",
				     AIXVGLptr->vg_id.word2);
				AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;

			}
		}

		LOG_DEBUG("CVG SWITCH TS AIXVGLptr->CleanVGInfo:%d \n",
			  AIXVGLptr->CleanVGInfo);

		switch (AIXVGLptr->CleanVGInfo) {
		case AIX_PV_STATE_VALID:
		case AIX_PV_STATE_FIRST_VGDA:

			LOG_DEBUG("CVG SWITCH VALID %d size:%d\n",
				  AIXVGLptr->CleanVGInfo,
				  (int) sizeof (struct vg_header));

			AIX_copy_header_info(AIXVGLptr->AIXvgh, AIXvgh);	// Get the info. we need

			AIXVGLptr->vgda_psn = AIXlvm->vgda_psn[0];
			AIXVGLptr->vgda_len = AIXlvm->vgda_len;
			break;

		case AIX_PV_STATE_SECOND_VGDA:
			LOG_DEBUG("CVG SWITCH SECOND VGDA %d size:%d\n",
				  AIXVGLptr->CleanVGInfo,
				  (int) sizeof (struct vg_header));

			AIX_copy_header_info(AIXVGLptr->AIXvgh, AIXvgh2);	// Get the info. we need

			AIXVGLptr->vgda_psn = AIXlvm->vgda_psn[1];
			AIXVGLptr->vgda_len = AIXlvm->vgda_len;
			break;

		case AIX_PV_STATE_EITHER_VGDA:
			LOG_DEBUG("CVG SWITCH EITHER VGDA %d size:%d\n",
				  AIXVGLptr->CleanVGInfo,(int) sizeof (struct vg_header));
			if (COMPARE_UNIQUE_IDS(AIXvgh->vg_id, AIXvgh2->vg_id)) {

				AIX_copy_header_info(AIXVGLptr->AIXvgh, AIXvgh);	// Get the info. we need

				AIXVGLptr->vgda_psn = AIXlvm->vgda_psn[0];
				AIXVGLptr->vgda_len = AIXlvm->vgda_len;
			} else {
				AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;
				// Not sure where this PV belongs. It thinks it is
				// supposed to be in two different containers. We will
				// probably need to put this on a separate, temporary
				// list, and determine later which container is missing
				// a PV.
			}
			break;

		default:
			LOG_ERROR("Invalid PV state (%d) for %d\n",
				  AIXVGLptr->CleanVGInfo,
				  AIXVGLptr->vg_id.word2);
			AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;
			break;
		}

	}

    // Currently AIX Big VGDA is not supported - cleanup and return NULL so this VG doesn't get added

    if (AIXVGLptr->AIXvgh->bigvg != 0) {
        LOG_SERIOUS("Error creating Volume Group AIX Big VGDA is not currently supported\n");
        if (AIXVGLptr->AIXvgh) {
            kfree(AIXVGLptr->AIXvgh);
            AIXVGLptr->AIXvgh = NULL;
        }

        if (AIXVGLptr) {
            kfree(AIXVGLptr);
            AIXVGLptr = NULL;
        }

        AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
        return NULL;
    }

	add_VG_data_to_VG_list(logical_node, AIXVGLptr, AIXlvm->pv_num);

	AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);

	LOG_DEBUG("CVG Exiting CleanVGInfo:%d\n", AIXVGLptr->CleanVGInfo);

	return AIXVGLptr;
}

/****************************************************
*
*
*
*****************************************************/
static int
AIX_update_volume_group(struct aix_volume_group *AIXVGLptr,
			struct evms_logical_node *logical_node,
			struct AIXlvm_rec *AIXlvm)
{
	struct vg_header *AIXvgh = NULL, *AIXvgh2 = NULL;
	struct vg_trailer *AIXvgt = NULL, *AIXvgt2 = NULL;

	AIXvgh = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgh) {
		return -ENOMEM;
	}

	AIXvgh2 = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgh2) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	AIXvgt = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgt) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	AIXvgt2 = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXvgt2) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	// First time thru we want to read this in, we may only have one PV in this group, all others 
	// may be corrupt, etc. If the info is clean we shouldn't get here.

	if (INIT_IO(logical_node, 0, AIXlvm->vgda_psn[0], 1, AIXvgh)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	if (INIT_IO(logical_node, 0, AIXlvm->vgda_psn[1], 1, AIXvgh2)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	if (INIT_IO(logical_node, 0, (AIXlvm->vgda_psn[0] + AIXlvm->vgda_len - 1), 1,
	     AIXvgt)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	if (INIT_IO(logical_node, 0, (AIXlvm->vgda_psn[1] + AIXlvm->vgda_len - 1), 1,
	     AIXvgt2)) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}

	LOG_DEBUG("UVG AIXvgh->vgda_psn[%d]:%d\n", 0, AIXlvm->vgda_psn[0]);
	LOG_DEBUG("UVG AIXvgh->vgda_psn[%d]:%d\n", 1, AIXlvm->vgda_psn[1]);
	LOG_DEBUG("UVG AIXvgt psn[%d]:%d\n", 0,(AIXlvm->vgda_psn[0] + AIXlvm->vgda_len - 1));
	LOG_DEBUG("UVG AIXvgt psn[%d]:%d\n", 1,(AIXlvm->vgda_psn[1] + AIXlvm->vgda_len - 1));

	AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;
	AIXVGLptr->flags |= AIX_VG_DIRTY;

	LOG_DEBUG("UVG AIXVGLptr:%p line %d\n", AIXVGLptr, __LINE__);

	AIXVGLptr->AIXvgh = kmalloc(sizeof (struct vg_header), GFP_KERNEL);
	if (!AIXVGLptr->AIXvgh) {
		AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);
		return -ENOMEM;
	}
	memset(AIXVGLptr->AIXvgh, 0, sizeof (struct vg_header));

	LOG_DEBUG("UVG COMP TS AIXVGLptr->CleanVGInfo:%d \n",AIXVGLptr->CleanVGInfo);

	if (AIXVGLptr->CleanVGInfo == AIX_PV_STATE_INVALID) {
		if (COMPARE_TIMESTAMPS(AIXvgh->vg_timestamp, AIXvgt->timestamp)) {
			if (COMPARE_TIMESTAMPS
			    (AIXvgh2->vg_timestamp, AIXvgt2->timestamp)) {
				if (COMPARE_TIMESTAMPS
				    (AIXvgh->vg_timestamp,
				     AIXvgh2->vg_timestamp)) {
					// All timestamps match. Yea!
					AIXVGLptr->CleanVGInfo =
					    AIX_PV_STATE_VALID;
				} else {
					// Both VGDAs are good, but timestamps are
					// different. Can't tell yet which one is
					// correct. 
					AIXVGLptr->CleanVGInfo =
					    AIX_PV_STATE_EITHER_VGDA;
				}
			} else {
				// First VGDA is good, second is bad.
				AIXVGLptr->CleanVGInfo =
				    AIX_PV_STATE_FIRST_VGDA;
			}
		} else {
			if (COMPARE_TIMESTAMPS
			    (AIXvgh2->vg_timestamp, AIXvgt2->timestamp)) {
				// First VGDA is bad, second is good.
				AIXVGLptr->CleanVGInfo =
				    AIX_PV_STATE_SECOND_VGDA;
			} else if (AIXvgh->numpvs == 1) {	// We only have 1 PV in this group, mismatch or not this will have to do 
				AIXVGLptr->CleanVGInfo = AIX_PV_STATE_VALID;
			} else {
				// This should never happen.
				LOG_DEBUG
				    ("All four VG timestamps for %d are different. What happened?!?\n",
				     AIXVGLptr->vg_id.word2);
				AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;

			}
		}

		LOG_DEBUG("UVG SWITCH TS AIXVGLptr->CleanVGInfo:%d \n",
			  AIXVGLptr->CleanVGInfo);

		switch (AIXVGLptr->CleanVGInfo) {
		case AIX_PV_STATE_VALID:
		case AIX_PV_STATE_FIRST_VGDA:

			LOG_DEBUG("UVG SWITCH VALID %d size:%d\n",
				  AIXVGLptr->CleanVGInfo,
				  (int) sizeof (struct vg_header));

			AIX_copy_header_info(AIXVGLptr->AIXvgh, AIXvgh);	// Get the info. we need

			AIXVGLptr->vgda_psn = AIXlvm->vgda_psn[0];
			AIXVGLptr->vgda_len = AIXlvm->vgda_len;
			break;

		case AIX_PV_STATE_SECOND_VGDA:
			LOG_DEBUG("UVG SWITCH SECOND VGDA %d size:%d\n",
				  AIXVGLptr->CleanVGInfo,
				  (int) sizeof (struct vg_header));

			AIX_copy_header_info(AIXVGLptr->AIXvgh, AIXvgh2);	// Get the info. we need

			AIXVGLptr->vgda_psn = AIXlvm->vgda_psn[1];
			AIXVGLptr->vgda_len = AIXlvm->vgda_len;
			break;

		case AIX_PV_STATE_EITHER_VGDA:
			LOG_DEBUG("UVG SWITCH EITHER VGDA %d size:%d\n",
				  AIXVGLptr->CleanVGInfo,
				  (int) sizeof (struct vg_header));
			if (COMPARE_UNIQUE_IDS(AIXvgh->vg_id, AIXvgh2->vg_id)) {

				AIX_copy_header_info(AIXVGLptr->AIXvgh, AIXvgh);	// Get the info. we need

				AIXVGLptr->vgda_psn = AIXlvm->vgda_psn[0];
				AIXVGLptr->vgda_len = AIXlvm->vgda_len;
			} else {
				AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;
				// Not sure where this PV belongs. It thinks it is
				// supposed to be in two different containers. We will
				// probably need to put this on a separate, temporary
				// list, and determine later which container is missing
				// a PV.
			}
			break;

		default:
			LOG_ERROR("UVG Invalid PV state (%d) for %d\n",
				  AIXVGLptr->CleanVGInfo,
				  AIXVGLptr->vg_id.word2);
			AIXVGLptr->CleanVGInfo = AIX_PV_STATE_INVALID;
			break;
		}

	}

//	add_VG_data_to_VG_list(logical_node, AIXVGLptr, AIXlvm->pv_num);
	AIXVGLptr->flags |= AIX_VG_DIRTY;

	AIX_free_headers(AIXvgh, AIXvgh2, AIXvgt, AIXvgt2);

	LOG_DEBUG("UVG Exiting CleanVGInfo:%d\n", AIXVGLptr->CleanVGInfo);

	return 0;
}

/****************************************************
* Function: check_volume_groups
*
* We just want to make sure the volume groups have found
* all their drives.
*
* If not, we'll continue and build what we can
*****************************************************/
static int
check_volume_groups(void)
{
	struct aix_volume_group *group;
	struct aix_volume_group *next_group;
//	struct partition_list_entry *partitions;
//	int NumPVS = 0;

	LOG_DEBUG("CHVG Checking volume groups:\n");


	for (group = AIXVolumeGroupList; group; group = next_group) {
		next_group = group->next;

		if (group->flags & AIX_VG_DIRTY){
			if (group->AIXvgh->numlvs == 0) {
				remove_group_from_list(group);
				deallocate_volume_group(group);
			} else {
				if (group->partition_count != group->AIXvgh->numpvs) {
					group->flags |= AIX_VG_INCOMPLETE;
					LOG_ERROR("CHVG Found incomplete VG !! flags:%x\n",
						  group->flags);
					LOG_ERROR("CHVG Found %d PVs should have %d PVs\n",
						  group->partition_count, group->AIXvgh->numpvs);
				}
			}
		}
	}

	LOG_DEBUG("CHVG Finished Checking volume groups:\n");
	return 0;

}

/************************************************************************
 * Function: discover_logical_volumes
 *
 *  After all PVs have been claimed and added to the appropriate VG list,
 *  the volumes for each VG must be constructed.
 *
 *
 */
static int
discover_logical_volumes(void)
{

	struct aix_volume_group *AIXVGLPtr;
	struct aix_logical_volume *new_LV;
	struct partition_list_entry *partition;
	struct evms_logical_node *node;
	struct lv_entries *AIXlvent, *AIXlventHead;
	int j, lv_found, all_lvs_found, rc;
	struct namelist *AIXnamelist;
	char *NameBuffer;

	AIXlventHead =
	    kmalloc(MAX_SECTORS_LV_ENTRIES * AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXlventHead) {
		return -ENOMEM;
	}

	memset(AIXlventHead, 0, (MAX_SECTORS_LV_ENTRIES * AIX_SECTOR_SIZE));

	NameBuffer =
	    kmalloc(MAX_SECTORS_NAMELIST * AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!NameBuffer) {
		kfree(AIXlventHead);
		return -ENOMEM;
	}

	memset(NameBuffer, 0, (MAX_SECTORS_NAMELIST * AIX_SECTOR_SIZE));

	for (AIXVGLPtr = AIXVolumeGroupList; AIXVGLPtr;
		AIXVGLPtr = AIXVGLPtr->next ) {

		partition = AIXVGLPtr->partition_list;

		if (!(AIXVGLPtr->flags & AIX_VG_DIRTY)) {
			continue;
		}

		if (partition == NULL) {
			continue;
		}

		node = partition->logical_node;

		if (node == NULL) {
			continue;
		}

		LOG_DEBUG("DLV INIT_IO AIXNameList position:%d\n",
			  ((AIXVGLPtr->vgda_psn + AIXVGLPtr->vgda_len) - 1 -
			   MAX_SECTORS_NAMELIST));
		LOG_DEBUG("AIXVGLPTR:%p partition:%p node:%p \n", AIXVGLPtr,
			  partition, node);

		if (INIT_IO(node, 0,
		     ((AIXVGLPtr->vgda_psn + AIXVGLPtr->vgda_len) - 1 -
		      MAX_SECTORS_NAMELIST), MAX_SECTORS_NAMELIST,
		     NameBuffer)) {
			continue;
		}

		LOG_DEBUG("DLV INIT_IO AIXNameList\n");

		if (INIT_IO(node, 0, AIXVGLPtr->vgda_psn + PSN_LVE_REC,
		     MAX_SECTORS_LV_ENTRIES, AIXlventHead)) {
			continue;
		}
		AIXlvent = AIXlventHead;
		AIXnamelist = (struct namelist *) NameBuffer;

		LOG_DEBUG("DLV INIT_IO AIXlvent\n");
		// Search through the LV structs for valid LV entries
		// We're just going to search until all valid LVs are found
		// The max. allowable LVs is 256 and we want don't want to
		// search for 255 if only 8 are defined 1-8 however, there 
		// could be gaps in the LV numbering. i.e 1,2,3,4,5,6,7,8, 27,43, etc.

		for (j = 0, lv_found = 0, all_lvs_found = FALSE;
		     !all_lvs_found && j < LVM_MAXLVS; j++, AIXlvent++) {

			LOG_DEBUG(" ** DVIG:lv_size:%d lvname:[%s] j:%d lv_number:%d ** \n",
			     AIXlvent->num_lps, AIXnamelist->name[j], j,
			     AIXlvent->lvname);
			LOG_DEBUG(" DVIG:stripe_exp:%u stripesize:%u lv_status:%d\n",
			     AIXlvent->striping_width,
			     GET_PHYSICAL_PART_SIZE(AIXlvent->stripe_exp),
			     AIXlvent->lv_state);
			LOG_DEBUG(" DVIG Group:%x.Access:%x\n",
				  (unsigned int) AIXVGLPtr->vg_id.word2,
				  AIXlvent->permissions);
			LOG_DEBUG(" DVIG mirror:%d mirror_policy:%d mirwrt:%d \n",
			     AIXlvent->mirror, AIXlvent->mirror_policy,
			     AIXlvent->mirwrt_consist);

			// This is the same check we used in "diskedit" and "readdisk"
			if (AIXlvent->lv_state == 0 ||
			    AIXlvent->permissions > 0x10) {
				continue;
			}

       			lv_found++;
       			if (lv_found == AIXVGLPtr->AIXvgh->numlvs) {
       				all_lvs_found = TRUE;
       			}

       			LOG_DEBUG(" DVIG lv_found:%d all_lvs_found:%d \n",
       			     lv_found, all_lvs_found);

       			// Create a new logical volume and place it in the appropriate
       			// spot in this VG's volume list. For re-discovery, make sure
       			// this volume does not already exist.
       			if (!AIXVGLPtr->volume_list[AIXlvent->lvname]) {
       				new_LV =
       				    new_logical_volume(AIXlvent,
       						       AIXVGLPtr,
       						       AIXnamelist->
       						       name[j],
       						       GET_PHYSICAL_PART_SIZE
       						       (AIXlvent->
       							stripe_exp));
       				if (!new_LV) {
       					continue;
       				}
       				LOG_DEBUG(" DVIG Adding new logical volume %d to group:%x \n",
       				     new_LV->lv_number,AIXVGLPtr->vg_id.word2);

       				AIXVGLPtr->volume_list[new_LV->lv_number] = new_LV;
       			} else {
       				LOG_DEBUG("DVIG Updating Vol Exists\n");
       			}
		}

		// Build the le_to_pe_map for each volume that was discovered above.
		// This has to be done after all volumes in the group are discovered
		if ((rc = build_pe_maps(AIXVGLPtr))) {
			continue;
		}

		check_log_volume_and_pe_maps(AIXVGLPtr);
	}

	kfree(NameBuffer);
	kfree(AIXlventHead);

	return 0;
}

/*
 * Function: new_logical_volume
 *
 *  Allocate space for a new LVM logical volume, including space for the
 *  PE map 
 */
static struct aix_logical_volume *
new_logical_volume(struct lv_entries *AIXlvent,
		   struct aix_volume_group *volume_group,
		   char *lv_name, u32 stripesize)
{

	struct aix_logical_volume *new_volume;
	const char *name = "evms_AIXiod";
	const char *resync_name = "evms_AIXresync";

	LOG_DEBUG(" NLV: lv_number:%d lv_allocated_le:%d lv_size:%d\n",
		  AIXlvent->lvname, AIXlvent->num_lps,
		  AIXlvent->num_lps * volume_group->pe_size);

	// Allocate space for the new logical volume.
	new_volume = kmalloc(sizeof (struct aix_logical_volume), GFP_KERNEL);
	if (!new_volume) {
		return NULL;
	}
	memset(new_volume, 0, sizeof (struct aix_logical_volume));

	// Allocate space for the LE to PE mapping table
	// We add 1 for the allocated le to ease mapping later on, all AIX le are 1 based
	new_volume->le_to_pe_map =
	    kmalloc((AIXlvent->num_lps + 1) * sizeof (struct pe_table_entry),
		    GFP_KERNEL);
	if (!new_volume->le_to_pe_map) {
		delete_logical_volume(new_volume);
		return NULL;
	}

	memset(new_volume->le_to_pe_map, 0,
	       (AIXlvent->num_lps + 1) * sizeof (struct pe_table_entry));

	if (AIXlvent->mirror > AIX_DEFAULT_MIRRORING) {
		new_volume->le_to_pe_map_mir1 =
		    kmalloc((AIXlvent->num_lps +
			     1) * sizeof (struct pe_table_entry), GFP_KERNEL);
		if (!new_volume->le_to_pe_map_mir1) {
			delete_logical_volume(new_volume);
			return NULL;
		}
		memset(new_volume->le_to_pe_map_mir1, 0,
		       (AIXlvent->num_lps +
			1) * sizeof (struct pe_table_entry));
	}

	if (AIXlvent->mirror == AIX_MAX_MIRRORS) {
		new_volume->le_to_pe_map_mir2 =
		    kmalloc((AIXlvent->num_lps + 1) 
		    * sizeof (struct pe_table_entry), GFP_KERNEL);
		if (!new_volume->le_to_pe_map_mir2) {
			delete_logical_volume(new_volume);
			return NULL;
		}
		memset(new_volume->le_to_pe_map_mir2, 0,
		       (AIXlvent->num_lps +1)
		       * sizeof (struct pe_table_entry));
	}

	// Initialize the rest of the new volume.
	new_volume->lv_number = AIXlvent->lvname;
	new_volume->lv_size = AIXlvent->num_lps * (volume_group->pe_size);
	new_volume->lv_access = AIXlvent->permissions | EVMS_LV_NEW;	// All volumes start new.
	new_volume->lv_status = AIXlvent->lv_state;
	//new_volume->lv_minor          = MINOR(1);
	new_volume->mirror_copies = AIXlvent->mirror;
//	new_volume->mirror_iterations = AIX_DEFAULT_MIRRORING;
	new_volume->stripes = AIXlvent->striping_width;
	new_volume->stripe_size = stripesize;
	new_volume->stripe_size_shift = evms_cs_log2(stripesize);
	new_volume->pe_size = volume_group->pe_size;
	new_volume->pe_size_shift = evms_cs_log2(volume_group->pe_size);
	new_volume->num_le = AIXlvent->num_lps;
//	new_volume->new_volume = TRUE;
	new_volume->group = volume_group;

	volume_group->numlvs++;

	sprintf(new_volume->name, "aix/%s", lv_name);

	if (!AIX_BH_list_pool
	    && new_volume->mirror_copies > AIX_DEFAULT_MIRRORING) {

		// We only need the ReSync thread if we have at least one mirrored LV.
		// You can't ReSync a non-mirrored drive

		AIX_BH_list_pool =
		    evms_cs_create_pool(sizeof (struct aix_mirror_bh),
					"EVMS_AIX_BH", aix_notify_cache_ctor,
					NULL);
		if (!AIX_BH_list_pool) {
			return NULL;

			AIX_mirror_read_retry_thread =
			    evms_cs_register_thread(AIXiod, NULL, name);

			AIX_mirror_resync_thread =
			    evms_cs_register_thread(AIXresync, NULL,
						    resync_name);
		}
	}

	LOG_DEBUG("NLV lv_number:%d name:%s lv_size " PFU64 " \n",
		  new_volume->lv_number, new_volume->name, new_volume->lv_size);
	LOG_DEBUG("NLV stripe_size:%d stripe_size_shift:%d\n",
		  new_volume->stripe_size, new_volume->stripe_size_shift);

	return new_volume;
}

/* 
 * Function: aix_notify_cache_ctor
 * this function initializes the b_wait field in the buffer heads
 * in our private buffer head pool.
 */
static void
aix_notify_cache_ctor(void *foo, kmem_cache_t * cachep, unsigned long flags)
{
	if ((flags & (SLAB_CTOR_VERIFY | SLAB_CTOR_CONSTRUCTOR)) ==
	    SLAB_CTOR_CONSTRUCTOR) {
		struct aix_mirror_bh *rbh = (struct aix_mirror_bh *) foo;
		memset(rbh, 0, sizeof (struct aix_mirror_bh));
		atomic_set(&rbh->remaining, 0);
		init_waitqueue_head(&rbh->bh_req.b_wait);
	}
}

/*
 * Function: build_pe_maps
 *
 *  After all logical volumes have been discovered, the mappings from
 *  logical extents to physical extents must be constructed. Each PV
 *  contains a map on-disk of its PEs. Each PE map entry contains the
 *  logical volume number and the logical extent number on that volume.
 *  Our internal map is the reverse of this map for each volume, listing
 *  the PV node and sector offset for every logical extent on the volume.
 */
static int
    build_pe_maps(struct aix_volume_group *volume_group)
{
	struct partition_list_entry *partition;
	struct partition_list_entry *mirror_partition;
	struct pp_entries *AIXppent, *AIXppent_buff;
	struct pv_header *AIXpvh;
	u64 offset;
	u32 le_number;
	u32 j, pp_count, pvh_pos;
	u32 MirrorFound;
    u32 pvh_posn[LVM_MAXPVS];
    u32 rc;
#ifdef EVMS_DEBUG_MIRRORS
	u32 lv_found, all_lvs_found;
	u32 mirs = 0;
#endif

	LOG_DEBUG(" *** BPEM ***\n");
	// For every partition in this VG

	AIXppent_buff = kmalloc(AIX_SECTOR_SIZE * PHYS_VOL_OFFSET, GFP_KERNEL);
	if (!AIXppent_buff) {
		return -ENOMEM;
	}

	memset(AIXppent_buff, 0, AIX_SECTOR_SIZE * PHYS_VOL_OFFSET);
    memset(pvh_posn, 0, LVM_MAXPVS);

	AIXpvh = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXpvh) {
		kfree(AIXppent_buff);
		return -ENOMEM;
	}

	memset(AIXpvh, 0, AIX_SECTOR_SIZE);

	LOG_DEBUG(" BPEM AIXppent_buff:%d \n",
		  (AIX_SECTOR_SIZE * PHYS_VOL_OFFSET));

    // This next section is to calculate the sector spacing between PV info for the VG
    // AIX doesn't always space the info. the same. It could be 17 or 34 sectors apart
    // depending on the PE size selected.

    rc = AIX_pvh_data_posn(volume_group->vgda_psn, pvh_posn, volume_group->partition_list, volume_group->AIXvgh->numpvs);

    if (rc != 0) {
        kfree(AIXppent_buff);
        kfree(AIXpvh);
        return (rc);
    }

	for (partition = volume_group->partition_list; partition;
	    partition = partition->next) {

		LOG_DEBUG(" BPEM partition:%p next:%p\n", partition,
			  partition->next);

		pvh_pos = pvh_posn[partition->pv_number];

		LOG_DEBUG(" BPEM pvh_pos:%d pv_number:%d\n", pvh_pos, partition->pv_number);

		if (INIT_IO(partition->logical_node, 0, pvh_pos, 1, AIXpvh)) {
			kfree(AIXppent_buff);
			kfree(AIXpvh);
			return -EIO;
		}
		// For every entry in the PE map, calculate the PE's sector offset
		// and update the correct LV's PE map. LV number of 0 marks an unused PE.
		// For re-discovery, only compute entries for new volumes.

		if (INIT_IO(partition->logical_node, 0, pvh_pos, AIX_PVHPP_LENGTH,
			    AIXppent_buff)) {
			kfree(AIXppent_buff);
			kfree(AIXpvh);
			return -EIO;
		}

		AIXppent = AIXppent_buff;
		AIXppent++;

		pp_count = AIXpvh->pp_count;

        LOG_DEBUG("BPEM AIXpvh data: pp_count:%d psn_part1:%d pv_id1:%d pv_id2:%d pv_id3:%d pv_id4:%d pv_num:%d pv_state:%d vgdas:%d res1:%d res2:%d\n", AIXpvh->pp_count, 
                                     AIXpvh->psn_part1, 
                                     AIXpvh->pv_id.word1,
                                     AIXpvh->pv_id.word2,
                                     AIXpvh->pv_id.word3,
                                     AIXpvh->pv_id.word4,
                                     AIXpvh->pv_num,
                                    AIXpvh->pv_state, AIXpvh->pvnum_vgdas, AIXpvh->res1, AIXpvh->res2);

		LOG_DEBUG(" PE Map: volgrp:%x AIXpvh->pv_num:%d partition:%p next:%p lv_index:%d pp_count:%d\n",
		     volume_group->vg_id.word2, AIXpvh->pv_num, partition,
		     partition->next, AIXppent->lv_index, pp_count);

		for (j = 0; j < pp_count; j++,AIXppent++) {
			if (!AIXppent->lv_index || AIXppent->pp_state == AIX_LVM_LVUNDEF) {
				continue;
			}

			LOG_EXTRA(" -- pv:%x pp:%d st:%d nm:%s lv:%d lp:%d cp:%d fst v:%d fst p:%d snd v:%d snd p:%d \n",
			     volume_group->vg_id.word2, j + 1,
			     AIXppent->pp_state,
			     volume_group->volume_list[AIXppent->lv_index -1]->name,
			     AIXppent->lv_index, AIXppent->lp_num,
			     AIXppent->copy, AIXppent->fst_alt_vol,
			     AIXppent->fst_alt_part,
			     AIXppent->snd_alt_vol,
			     AIXppent->snd_alt_part);

			le_number = AIXppent->lp_num - 1;	// AIX lp's start @ 1, we want a 0 index
			offset = ((j * (volume_group->pe_size)) + AIXpvh->psn_part1);

			LOG_DEBUG(" PE Map: le_number:%d partition:%p lv_index:%d lv_name:%s\n",
			     le_number, partition, AIXppent->lv_index,
			     volume_group->volume_list[AIXppent->lv_index -1]->name);

			if (!volume_group->volume_list[AIXppent->lv_index - 1]) {
				LOG_SERIOUS("Failed attempt to access volume without memory allocation lv:%d\n",
				     AIXppent->lv_index - 1);
				continue;
			}

       			if (volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map
       			    && le_number <= volume_group->volume_list[AIXppent->lv_index - 1]->num_le) {

				volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map[le_number].owning_pv = partition;
				volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map[le_number].pe_sector_offset = offset;
				volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map[le_number].pp_state = AIXppent->pp_state;
       			}

       			if (volume_group->volume_list[AIXppent->lv_index -1]->mirror_copies >
       			    AIX_DEFAULT_MIRRORING) {

       				LOG_EXTRA(" PE Map: Mirror found lv:%d -- \n",
       				     AIXppent->lv_index);

       				for (mirror_partition = volume_group->partition_list,
					MirrorFound = FALSE;
					mirror_partition && !MirrorFound;
					mirror_partition = mirror_partition->next) {

       					if (mirror_partition->pv_number == AIXppent->fst_alt_vol) {

       						offset = (((AIXppent->fst_alt_part - 1) * (volume_group->pe_size)) + AIXpvh->psn_part1);

       						volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map_mir1[le_number].owning_pv = mirror_partition;
       						volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map_mir1[le_number].pe_sector_offset = offset;
       						volume_group->volume_list[AIXppent->lv_index -1]->le_to_pe_map_mir1[le_number].pp_state = AIXppent->pp_state;

       						LOG_EXTRA(" PE Map: mirror_partition:%p \n",
       						     mirror_partition);
       						LOG_EXTRA(" PE Map: mirror_sector_offet:%d\n",
       						     AIXppent->fst_alt_part);

       						MirrorFound = TRUE;
       					}
       				}

       				if (volume_group->volume_list[AIXppent->lv_index -1]->mirror_copies == AIX_MAX_MIRRORS) {

       					for (mirror_partition = volume_group->partition_list,
       					     MirrorFound = FALSE;
       					     mirror_partition && !MirrorFound;
       					     mirror_partition = mirror_partition->next) {

       						if (mirror_partition->pv_number == AIXppent->snd_alt_vol) {

       							offset = (((AIXppent->snd_alt_part - 1) * (volume_group->pe_size)) + AIXpvh->psn_part1);

       							volume_group->volume_list[AIXppent->lv_index-1]->le_to_pe_map_mir2[le_number].owning_pv = mirror_partition;
       							volume_group->volume_list[AIXppent->lv_index-1]->le_to_pe_map_mir2[le_number].pe_sector_offset = offset;
       							volume_group->volume_list[AIXppent->lv_index-1]->le_to_pe_map_mir2[le_number].pp_state = AIXppent->pp_state;

       							LOG_EXTRA(" PE Map: mirror_partition2:%p \n",
       							     mirror_partition);
       							LOG_EXTRA(" PE Map: mirror_sector_offet2:%d\n",
       							     AIXppent->snd_alt_part);

       							MirrorFound = TRUE;
       						}
       					}
       				}

       			}	// End of if mirroring is enabled 
		}
	}

//      LOG_EXTRA(" PE Map: PE maps:%d Mirror count:%d -- \n", lvs, mirs);

#ifdef EVMS_DEBUG_MIRRORS
	for (mirs = 0, lv_found = 0, all_lvs_found = FALSE;
	    !all_lvs_found && mirs < LVM_MAXLVS; mirs++) {

		if (volume_group->volume_list[mirs] != NULL) {
			if (volume_group->volume_list[mirs]->lv_status ==
			    LV_ACTIVE) {

				lv_found++;

				LOG_DEBUG(" PE Map: owning part lv %d -- %p\n",
					  mirs,
					  volume_group->volume_list[mirs]->
					  le_to_pe_map[0].owning_pv);
				if (volume_group->volume_list[mirs]->
				    mirror_copies > AIX_DEFAULT_MIRRORING) {
					LOG_DEBUG(" PE Map: mirror_partition lv %d -- %p \n",
					     mirs,
					     volume_group->volume_list[mirs]->
					     le_to_pe_map_mir1[0].owning_pv);
				}
				if (volume_group->volume_list[mirs]->
				    mirror_copies == AIX_MAX_MIRRORS) {
					LOG_DEBUG(" PE Map: mirror_partition lv %d -- %p \n",
					     mirs,
					     volume_group->volume_list[mirs]->
					     le_to_pe_map_mir2[0].owning_pv);
				}
			}
			if (lv_found == volume_group->AIXvgh->numlvs) {
				all_lvs_found = TRUE;
				LOG_DEBUG(" PE Map: all_lvs_found\n");
			}
		}
	}
#endif

	kfree(AIXpvh);
	kfree(AIXppent_buff);

	return 0;
}

/*
 * Function: check_log_volume_and_pe_maps
 *
 *  Make sure all volumes in this group have valid LE-to-PE maps.
 *  Any volume that doesn't is deleted. This is safe for re-discovery
 *  because only new volumes could have corrupted PE maps.
 */
static int
check_log_volume_and_pe_maps(struct aix_volume_group *group)
{
	struct aix_logical_volume *volume;
	int i, j, lv_found, all_lvs_found;

	LOG_DEBUG(" check_pe_map.\n");

	for (i = 0, all_lvs_found = FALSE, lv_found = 0;
	     !all_lvs_found && i < LVM_MAXLVS; i++) {
		if (!group->volume_list[i]) {
			LOG_DEBUG(" CPEM No Volume %d found \n", i);
			continue;
		}

		volume = group->volume_list[i];
		if (!volume->le_to_pe_map) {
			LOG_DEBUG(" CPEM Volume %s has no PE map.\n",
				  volume->name);
			delete_logical_volume(volume);
			continue;
		}

		LOG_DEBUG(" CPEM volume %s num_le: %d \n", volume->name,
			  volume->num_le);

		lv_found++;

		if (lv_found == group->AIXvgh->numlvs) {
			all_lvs_found = TRUE;
		}

		for (j = 0; j < volume->num_le; j++) {
			if (!volume->le_to_pe_map[j].owning_pv ||
			    !volume->le_to_pe_map[j].pe_sector_offset) {
				LOG_SERIOUS(" CPEM Volume (%s) incomplete PE map (LE %d) \n",
				     volume->name, j);
				volume->lv_access |= EVMS_LV_INCOMPLETE;
			}

			if (volume->mirror_copies > AIX_DEFAULT_MIRRORING) {
				if (!volume->le_to_pe_map_mir1[j].owning_pv ||
				    !volume->le_to_pe_map_mir1[j].
				    pe_sector_offset) {
					LOG_SERIOUS(" CPEM Volume (%s) incomplete PE mirror map 1 (LE %d) \n",
					     volume->name, j);
					volume->lv_access |= EVMS_LV_INCOMPLETE;
				}

				if (volume->mirror_copies == AIX_MAX_MIRRORS) {
					if (!volume->le_to_pe_map_mir2[j].
					    owning_pv
					    || !volume->le_to_pe_map_mir2[j].
					    pe_sector_offset) {
						LOG_SERIOUS(" CPEM Volume (%s) incomplete PE mirror map 2 (LE %d) \n",
						     volume->name, j);
						volume->lv_access |= EVMS_LV_INCOMPLETE;
					}
				}
			}
		}
	}

	LOG_EXTRA(" Leaving check_pe_map.\n");
	return 0;
}

/*
 * Function: export_volumes
 *
 *  The last thing this VGE must do is take each constructed volume and
 *  place it back on the evms logical partition list.
 */
static int
export_volumes(struct evms_logical_node **evms_partition_list)
{
	struct aix_volume_group *AIXVGLPtr;
	struct evms_logical_node *new_node;
	struct aix_logical_volume *volume;
	int j, lv_found, all_lvs_found;
	int count = 0;

	for (AIXVGLPtr = AIXVolumeGroupList; AIXVGLPtr; AIXVGLPtr = AIXVGLPtr->next) {

		if (!(AIXVGLPtr->flags & AIX_VG_DIRTY)) {
			LOG_DEBUG(" EV Existing group(%d), not dirty, skipping\n",
				  AIXVGLPtr->vg_id.word2);
			continue;
		}
       		LOG_DEBUG(" Exporting all new volumes numpvs:%d numlvs:%d \n",
       		     AIXVGLPtr->AIXvgh->numpvs, AIXVGLPtr->numlvs);

       		// Export every valid volume in the group. For re-discovery,
       		// make sure we are only exporting "new" volumes.

       		for (j = 0, all_lvs_found = FALSE, lv_found = 0;
       		     !all_lvs_found && j < LVM_MAXLVS; j++) {
       			if (AIXVGLPtr->volume_list[j] != NULL) {
       				if (AIXVGLPtr->volume_list[j]->lv_access & EVMS_LV_NEW) {

       					LOG_DEBUG(" EV Checking LV:[%d] volume:%p\n",
						     j,AIXVGLPtr->volume_list[j]);

       					volume = AIXVGLPtr->volume_list[j];
       					lv_found++;

       					if (lv_found == AIXVGLPtr->AIXvgh->numlvs) {
       						all_lvs_found = TRUE;
       					}
       					// For new volumes, create a new EVMS node and 
       					// initialize the appropriate fields.
       					if (evms_cs_allocate_logical_node(&new_node)) {
       						LOG_DEBUG(" Export Vol Error allocating node !!\n");
       						continue;
       					} else {
       						LOG_DEBUG(" EV Node allocated OK\n");
       					}

//       					volume->new_volume = 0;
       					volume->volume_node = new_node;
       					volume->lv_access &= (~EVMS_LV_NEW);
       					new_node->hardsector_size = AIXVGLPtr->hard_sect_size;
       					new_node->block_size = AIXVGLPtr->block_size;
       					new_node->plugin = &plugin_header;
       					new_node->private = volume;
       					new_node->total_vsectors = volume->lv_size;

       					LOG_DEBUG(" EV volume->name:[%s]\n",
       					     volume->name);

       					strncpy(new_node->name,volume->name,
       						EVMS_VOLUME_NAME_SIZE + 1);

       					// Is the volume read-only?
       					if (!(volume->lv_access & AIX_LV_WRITE)
       					    || volume->lv_access & EVMS_LV_INCOMPLETE)
       					{
       						new_node->flags |= EVMS_VOLUME_SET_READ_ONLY;
       						LOG_DEBUG(" EV Read Only volume->lv_access:%d\n",
       						     volume->lv_access);
       					}

       					evms_cs_add_logical_node_to_list(evms_partition_list,
       					     new_node);
       					count++;

       					LOG_DEBUG(" Exporting LVM volume %p new_node:%p ESD->volume_name[%s]\n",
       					     volume, new_node,new_node->name);
       				} else {
       					evms_cs_add_logical_node_to_list(evms_partition_list,
       					     AIXVGLPtr->volume_list[j]->volume_node);
       					count++;
       					LOG_DEBUG(" ELV vol_list[%d]%p\n", j,
       					     AIXVGLPtr->volume_list[j]);
       				}
       			} else {
       				LOG_DEBUG(" EV Checking LV:[%d] == NULL\n",j);
       			}
       		}	// end checking all lvs

		AIXVGLPtr->flags &= ~AIX_VG_DIRTY;
	}

	return count;
}

/*
 * Function: delete_logical_volume
 *
 *  This function deletes the in-memory representation of a single LVM
 *  logical volume, including its PE map and any snapshot data. It does
 *  not alter the parent volume group, except to remove this volume from
 *  its volume list.
 */
static int
delete_logical_volume(struct aix_logical_volume *volume)
{
	struct aix_volume_group *group = volume->group;

	LOG_DEBUG(" Deleting volume %s\n", volume->name);

	// Now free up all the memory. This includes the LE-to-PE map, any
	// mirror PEs, etc.
	if (volume->le_to_pe_map) {
		kfree(volume->le_to_pe_map);
		volume->le_to_pe_map = NULL;
	}

	if (volume->le_to_pe_map_mir1) {
		kfree(volume->le_to_pe_map_mir1);
		volume->le_to_pe_map_mir1 = NULL;
	}

	if (volume->le_to_pe_map_mir2) {
		kfree(volume->le_to_pe_map_mir2);
		volume->le_to_pe_map_mir2 = NULL;
	}
	// Remove this volume from the volume-group's list.
	if (group && group->volume_list[volume->lv_number] == volume) {
		group->volume_list[volume->lv_number] = NULL;
		group->numlvs--;
	}

	kfree(volume);

	return 0;
}

/* Function: remove_group_from_list
 *
 *	Remove an LVM volume group from the global LVM list.
 */
static int
remove_group_from_list(struct aix_volume_group *group)
{
	struct aix_volume_group **p_group;

	for (p_group = &AIXVolumeGroupList; *p_group;
	     p_group = &(*p_group)->next) {
		if (*p_group == group) {
			*p_group = (*p_group)->next;
			group->next = NULL;
			break;
		}
	}
	return 0;
}

/*
 * Function: delete_aix_node
 *
 *  This function deletes the in-memory representation of an LVM
 *  logical volume. Right now it makes a lot of assumptions about
 *  the data in the group not being corrupted. It would be possible
 *  to put in a lot of consistency checks before deleting everything
 *  to indicate if problems have occurred during the lifetime of the
 *  volume and its volume group.
 */
static int
delete_aix_node(struct evms_logical_node *logical_node)
{
	struct aix_logical_volume *volume =
	    (struct aix_logical_volume *) (logical_node->private);
	struct aix_volume_group *group = volume->group;

	if (delete_logical_volume(volume)) {
		return -EINVAL;
	}
	// If we just removed the last volume from this group, the entire group
	// can also be deleted.
	if (group && group->numlvs == 0) {
		remove_group_from_list(group);
		deallocate_volume_group(group);
	}
	// Free the logical node.
	evms_cs_deallocate_logical_node(logical_node);

	return 0;
}

/* Function: deallocate_volume_group
 *
 *  This function deletes the entire in-memory representation of an LVM
 *  volume group, including all partitions and logical volumes. If this
 *  group is on the VGE's volume group list, it is removed.
 */
static int
deallocate_volume_group(struct aix_volume_group *group)
{
	struct partition_list_entry *partition;
	struct partition_list_entry *next_part;
	int i;

	LOG_DEBUG(" Deleting volume group %x\n", group->vg_id.word2);

	// Delete all partitions from the group's list.
	for (partition = group->partition_list; partition;
	     partition = next_part) {

		next_part = partition->next;

		if (partition->logical_node) {
			// Send a delete command down to the partition manager.
			LOG_DEBUG(" Deleting PV %d from group %x\n",
				  partition->pv_number, group->vg_id.word2);
			DELETE(partition->logical_node);
		}
		kfree(partition);
	}

	// Delete all logical volumes, and the array of pointers.
	for (i = 0; i < LVM_MAXLVS; i++) {
		if (group->volume_list[i]) {
			delete_logical_volume(group->volume_list[i]);
		}
	}

	kfree(group);

	return 0;
}

/* Function: end_discover_aix
 *
 *	The discovery process at the region-manager level is now iterative,
 *	much like the EVMS feature level. To accomplish this correctly, and
 *	also to accomplish partial volume discovery, a second discover
 *	entry point is needed, so EVMS can tell the region managers that
 *	discovery is over, and to finish up any discovery that is not yet
 *	complete. When this function is called, it should be assumed that
 *	the node list has had nothing new added to it since the last call
 *	of the regular discover function. Therefore, when this function is
 *	called, we do not need to try to discovery any additional volume
 *	groups. We will, however, look for logical volumes once more. This
 *	gives us the ability to export (read-only) volumes that have
 *	partially corrupted LE maps due to missing PVs in their VG.
 */
static int
end_discover_aix(struct evms_logical_node **evms_logical_disk_head)
{

	int rc;

	MOD_INC_USE_COUNT;
	LOG_DEBUG("Final Discovery:\n");

	rc = discover_logical_volumes();

	if (!rc) {
		rc = export_volumes(evms_logical_disk_head);

		lvm_cleanup();
	}

	MOD_DEC_USE_COUNT;
	return rc;
}

/****************************************************
* Function: AIX_alloc_wbh
*
* Alloc any buffer heads from the pool and return a linked list
*
*
*****************************************************/
static struct aix_mirror_bh *
AIX_alloc_wbh(struct evms_logical_node *node,
	      struct evms_logical_node *node2,
	      struct evms_logical_node *node3,
	      struct buffer_head *bh,
	      u32 mirror_copies, u32 le, u64 new_sector2, u64 new_sector3)
{
	struct aix_mirror_bh *tmp_bh = NULL, *head_bh = NULL;
	int i;

	head_bh = evms_cs_allocate_from_pool(AIX_BH_list_pool, EVMS_BLOCKABLE);

	if (!head_bh) {
		LOG_SERIOUS("Unable to allocate memory for mirror pool line:%d\n",
		     __LINE__);
		return NULL;
	}

	head_bh->master_bh = bh;
	head_bh->mirror_bh_list = NULL;
	atomic_set(&head_bh->remaining, 0);

	for (i = AIX_DEFAULT_MIRRORING; i <= mirror_copies; i++) {

		tmp_bh =
		    evms_cs_allocate_from_pool(AIX_BH_list_pool,
					       EVMS_BLOCKABLE);
		if (!tmp_bh) {
			LOG_SERIOUS("Unable to allocate memory for mirror pool line:%d\n",
			     __LINE__);
			return NULL;
		}

		tmp_bh->next_r1 = head_bh->mirror_bh_list;
		head_bh->mirror_bh_list = tmp_bh;
		atomic_inc(&head_bh->remaining);

		memcpy(&tmp_bh->bh_req, bh, sizeof (struct buffer_head));
		atomic_set(&tmp_bh->remaining, 0);
		init_waitqueue_head(&tmp_bh->bh_req.b_wait);
		//tmp_bh->bh_req.b_size   = bh->b_size;

		switch (i) {

		case AIX_DEFAULT_MIRRORING:
			tmp_bh->node = node;
			tmp_bh->bh_req.b_rsector = bh->b_rsector;
			break;

		case AIX_FIRST_MIRROR:
			tmp_bh->node = node2;
			tmp_bh->bh_req.b_rsector = new_sector2;
			break;

		case AIX_MAX_MIRRORS:
			tmp_bh->node = node3;
			tmp_bh->bh_req.b_rsector = new_sector3;
			break;
		}

		tmp_bh->bh_req.b_end_io = AIX_handle_write_mirror_drives;	//setup callback routine 
		tmp_bh->bh_req.b_private = (void *) head_bh;

	}

	return head_bh;

}

/****************************************************
* Function: AIX_handle_write_mirror_drives
*
* Handles a write from a set of mirrored AIX LVs

*
*
*****************************************************/
static void
AIX_handle_write_mirror_drives(struct buffer_head *bh, int uptodate)
{
	struct aix_logical_volume *volume;
	struct evms_logical_node *node;
	struct aix_mirror_bh *tmp_bh = NULL, *tmp_bh2 = NULL;
	kdev_t tmp_b_rdev;
	u32 count, le = 0;

	tmp_bh = (struct aix_mirror_bh *) bh->b_private;
	tmp_b_rdev = tmp_bh->master_bh->b_rdev;
	node = tmp_bh->node;
	volume = (struct aix_logical_volume *) node->private;

	LOG_DEBUG("AHWMD node:%p bh_flags:%lu uptodate:%d mirror_copies:%d \n",
		  node, bh->b_state, uptodate, volume->mirror_copies);

	if (!uptodate) {
		le = tmp_bh->le;

		switch (tmp_bh->iteration) {
		case AIX_DEFAULT_MIRRORING:
			volume->le_to_pe_map[le].pp_state += AIX_LVM_LVSTALE;
			break;

		case AIX_FIRST_MIRROR:
			volume->le_to_pe_map_mir1[le].pp_state +=
			    AIX_LVM_LVSTALE;
			break;

		case AIX_MAX_MIRRORS:
			volume->le_to_pe_map_mir2[le].pp_state +=
			    AIX_LVM_LVSTALE;
			break;
		}

		AIX_evms_cs_notify_lv_io_error(node);
	}

	if (atomic_dec_and_test(&tmp_bh->remaining)) {
		tmp_bh->master_bh->b_end_io(tmp_bh->master_bh, uptodate);
		tmp_bh2 = tmp_bh->mirror_bh_list;
		evms_cs_deallocate_to_pool(AIX_BH_list_pool, tmp_bh);

		while (tmp_bh2) {
			tmp_bh = tmp_bh2->next_r1;
			evms_cs_deallocate_to_pool(AIX_BH_list_pool, tmp_bh2);
			tmp_bh2 = tmp_bh;
		}

		evms_cs_volume_request_in_progress(tmp_b_rdev,
						   AIX_DECREMENT_REQUEST,
						   &count);
	}

	return;
}

/****************************************************
* Function: AIX_alloc_rbh
*
* Alloc any buffer heads from the pool and return a linked list
*
*
*****************************************************/
static struct aix_mirror_bh *
AIX_alloc_rbh(struct evms_logical_node *node,
	      struct buffer_head *bh,
	      u32 mirror_copies, u32 le, u64 org_sector, int cmd)
{
	struct aix_mirror_bh *tmp_bh = NULL;

	tmp_bh = evms_cs_allocate_from_pool(AIX_BH_list_pool, EVMS_BLOCKABLE);

	if (!tmp_bh) {
		LOG_SERIOUS
		    ("Unable to allocate memory for mirror pool line:%d\n",
		     __LINE__);
		return NULL;
	}

	memcpy(&tmp_bh->bh_req, bh, sizeof (struct buffer_head));
	atomic_set(&tmp_bh->remaining, 0);
	tmp_bh->node = node;
	tmp_bh->master_bh = bh;
	tmp_bh->iteration = AIX_FIRST_MIRROR;
	//tmp_bh->eio.rsector     = eio->rsector;
	//tmp_bh->eio.rsize       = eio->rsize;
	tmp_bh->le = le;
	//tmp_bh->eio.bh          = &tmp_bh->bh_req;

	if (cmd == AIX_LV_READ) {
		tmp_bh->bh_req.b_end_io = AIX_handle_read_mirror_drives;	//setup callback routine 
	} else {
		tmp_bh->bh_req.b_end_io = AIX_sync_mirrored_partitions;	//setup callback routine 
	}

	tmp_bh->bh_req.b_private = (void *) tmp_bh;

	tmp_bh->cmd = cmd;
	tmp_bh->next_r1 = NULL;
	tmp_bh->node = node;

	return tmp_bh;

}

/****************************************************
* Function: AIX_reschedule_retry
*
* reschedule a read of one of our mirror copies
*
*
*****************************************************/
static void
AIX_reschedule_retry(struct aix_mirror_bh *aix_bh)
{
	unsigned long flags;

	spin_lock_irqsave(&AIX_retry_list_lock, flags);
	if (AIX_retry_list == NULL)
		AIX_retry_tail = &AIX_retry_list;
	*AIX_retry_tail = aix_bh;
	AIX_retry_tail = &aix_bh->next_r1;
	aix_bh->next_r1 = NULL;
	spin_unlock_irqrestore(&AIX_retry_list_lock, flags);
	evms_cs_wakeup_thread(AIX_mirror_read_retry_thread);
}

/****************************************************
* Function: AIX_handle_read_mirror_drives
*
* Handles a read from a set of mirrored AIX LVs

*
*
*****************************************************/
static void
AIX_handle_read_mirror_drives(struct buffer_head *bh, int uptodate)
{
	struct aix_logical_volume *volume;
	struct evms_logical_node *node;
	struct aix_mirror_bh *tmp_bh;
	kdev_t tmp_b_rdev;
	u32 count, le = 0;

	tmp_bh = (struct aix_mirror_bh *) bh->b_private;
	tmp_b_rdev = tmp_bh->master_bh->b_rdev;
	volume = (struct aix_logical_volume *) tmp_bh->node->private;
	node = tmp_bh->node;
	le = tmp_bh->le;

	LOG_DEBUG("AHRMD node:%p bh_flags:%lu uptodate:%d mirror_copies:%d \n",
		  node, bh->b_state, uptodate, volume->mirror_copies);

	switch (tmp_bh->iteration) {
	case AIX_DEFAULT_MIRRORING:
		count = volume->le_to_pe_map[le].pp_state;
		break;

	case AIX_FIRST_MIRROR:
		count = volume->le_to_pe_map[le].pp_state;
		break;

	case AIX_MAX_MIRRORS:
		count = volume->le_to_pe_map[le].pp_state;
		break;
	}

	if (count == (AIX_LVM_LVSTALE + AIX_LVM_LVDEFINED)) {
		uptodate = 0;
		count = 0;
	}

	if (!uptodate && tmp_bh->iteration < volume->mirror_copies) {
		AIX_evms_cs_notify_lv_io_error(node);
		AIX_reschedule_retry(tmp_bh);
	} else {
		tmp_bh->master_bh->b_end_io(tmp_bh->master_bh, uptodate);
		evms_cs_deallocate_to_pool(AIX_BH_list_pool, tmp_bh);
		evms_cs_volume_request_in_progress(tmp_b_rdev,
						   AIX_DECREMENT_REQUEST,
						   &count);

	}

	return;
}

/****************************************************
* This is a temporary function until a common EVMS
* notification function can be created.
*
*****************************************************/
static int
AIX_evms_cs_notify_lv_io_error(struct evms_logical_node *node)
{
	struct aix_logical_volume *volume;

	volume = (struct aix_logical_volume *) node->private;

	LOG_CRITICAL("Notify_ERROR !!  node:%p volume->lv_status:%d volume->name:[%s]\n",
	     node, volume->lv_status, volume->name);

	return 0;
}

/* Function: lvm_cleanup
 *
 *	This function runs through the entire lvm data structure, removing
 *	all items that are not needed at runtime. Currently, this is just the
 *	vg_disk_t structure and the pv_disk_t structure for each PV. Also, any
 *	groups that don't contain any volumes are deleted. All of the other
 *	volume_group, logical_volume and evms_logical_node structures will be
 *	kept around at run-time.
 */
static int
lvm_cleanup(void)
{
	struct aix_volume_group *group;

	group = AIXVolumeGroupList;

	while (group) {

		if (group->AIXvgh) {
			kfree(group->AIXvgh);
			group->AIXvgh = NULL;
		}

		group = group->next;
	}

	return 0;
}

/****************************************************
* Function: AIX_copy_header_info
*
* Copy the disk header info into the volume struct
* so we can use it later.
*
* 
*
*****************************************************/
static int
AIX_copy_header_info(struct vg_header *AIXvgh, struct vg_header *AIXvgh2)
{

	LOG_DEBUG("CHI  AIXvgh:%p AIXvgh2:%p\n", AIXvgh, AIXvgh2);

	if (AIXvgh) {

		AIXvgh->vg_timestamp.tv_sec = AIXvgh2->vg_timestamp.tv_sec;
		AIXvgh->vg_timestamp.tv_nsec = AIXvgh2->vg_timestamp.tv_nsec;
		AIXvgh->vg_id.word1 = AIXvgh2->vg_id.word1;
		AIXvgh->vg_id.word2 = AIXvgh2->vg_id.word2;
		AIXvgh->vg_id.word3 = AIXvgh2->vg_id.word3;
		AIXvgh->vg_id.word4 = AIXvgh2->vg_id.word4;
		AIXvgh->numlvs = AIXvgh2->numlvs;
		AIXvgh->maxlvs = AIXvgh2->maxlvs;
		AIXvgh->pp_size = AIXvgh2->pp_size;
		AIXvgh->numpvs = AIXvgh2->numpvs;
		AIXvgh->total_vgdas = AIXvgh2->total_vgdas;
		AIXvgh->vgda_size = AIXvgh2->vgda_size;
		AIXvgh->bigvg = AIXvgh2->bigvg;
		AIXvgh->quorum = AIXvgh2->quorum;
		AIXvgh->auto_varyon = AIXvgh2->auto_varyon;
		AIXvgh->checksum = AIXvgh2->checksum;
		AIXvgh->bigda_size = AIXvgh2->bigda_size;

	} else {
		return -ENOMEM;
	}

	LOG_DEBUG("Returning CHI  AIXvgh:%p AIXvgh2:%p\n", AIXvgh, AIXvgh2);

	return 0;
}

/****************************************************
* Function: AIX_free_header
*
* 
* 
* 
*
*****************************************************/
static void
AIX_free_headers(struct vg_header *AIXvgh, struct vg_header *AIXvgh2,
		 struct vg_trailer *AIXvgt, struct vg_trailer *AIXvgt2)
{

	if (AIXvgh) {
		kfree(AIXvgh);
		AIXvgh = NULL;
	}

	if (AIXvgh2) {
		kfree(AIXvgh2);
		AIXvgh2 = NULL;
	}

	if (AIXvgt) {
		kfree(AIXvgt);
		AIXvgt = NULL;
	}

	if (AIXvgt2) {
		kfree(AIXvgt2);
		AIXvgt2 = NULL;
	}

}

/****************************************************
* Function: AIXiod
*
* This is a kernel thread that handles read of mirrors
* This shouldn't ever run on a non-mirrored LV read
* 
*
*****************************************************/
static void
AIXiod(void *data)
{
	struct aix_mirror_bh *r1_bh;
	struct evms_logical_node *node;
	unsigned long flags;

	while (1) {

		spin_lock_irqsave(&AIX_retry_list_lock, flags);
		if (AIX_retry_list == NULL) {
			spin_unlock_irqrestore(&AIX_retry_list_lock, flags);
			break;
		}
		r1_bh = AIX_retry_list;
		AIX_retry_list = r1_bh->next_r1;
		spin_unlock_irqrestore(&AIX_retry_list_lock, flags);
		r1_bh->next_r1 = NULL;	// for mark

		switch (r1_bh->cmd) {
		case AIX_LV_READ:

			r1_bh->iteration++;
			LOG_DEBUG("Report from thread AIXiod READ\n");

			if (r1_bh->iteration == AIX_FIRST_MIRROR) {
				node = r1_bh->mir_node1;
				r1_bh->bh_req.b_rsector = r1_bh->mir_sector1;
			} else {
				node = r1_bh->mir_node2;
				r1_bh->bh_req.b_rsector = r1_bh->mir_sector2;
			}

			R_IO(node, &r1_bh->bh_req);

			break;

		default:
			LOG_DEBUG("AIXiod unknown cmd passed to thread:%d\n",
				  r1_bh->cmd);
			break;
		}

	}
	return;
}

/****************************************************
* Function: AIX_schedule_resync
*
* schedule a resync of one of our lv mirror copies
*
*
*****************************************************/
static void
AIX_schedule_resync(struct aix_logical_volume *resync_volume, int force)
{
	unsigned long flags;

	LOG_DEBUG("Function %s volume: %s \n", __FUNCTION__,
		  resync_volume->name);

	spin_lock_irqsave(&AIX_resync_list_lock, flags);

	if (!AIX_resync_list) {
		AIX_resync_list =
		    kmalloc(sizeof (struct aix_resync_struct), GFP_ATOMIC);
		if (!AIX_resync_list) {
			return;
		}
		memset(AIX_resync_list, 0, sizeof (struct aix_resync_struct));
	}

	AIX_resync_list->resync_vol = resync_volume;
	AIX_resync_list->next_resync_vol = NULL;

	spin_unlock_irqrestore(&AIX_resync_list_lock, flags);
	evms_cs_wakeup_thread(AIX_mirror_resync_thread);
}

/****************************************************
* Function: AIXresync
*
* This is a kernel thread that handles resync of mirrors
* This shouldn't ever run on a non-mirrored LV
* 
*
*****************************************************/
static void
AIXresync(void *data)
{

	struct aix_logical_volume *volume = NULL;
	int force = FALSE;	// Currently we don't force a resync of non-stale pe's

	if (AIX_resync_list == NULL) {
		LOG_ERROR("No Volumes on list to resync\n");
		return;
	}

	volume = AIX_resync_list->resync_vol;
	LOG_DEBUG("Function %s volume: %s \n", __FUNCTION__, volume->name);

	if (!volume) {
		LOG_ERROR("Invalid volume passed to sync\n");
		return;
	}

	if (AIXResyncInProgress) {
		LOG_ERROR("Unable to resync multiple LVs concurrently %s\n",
			  volume->name);
		return;
	}

	if (volume->mirror_copies == AIX_DEFAULT_MIRRORING) {
		LOG_ERROR("Unable to resync non-mirrored LV %s \n",
			  volume->name);
		return;
	}

	AIXResyncInProgress = TRUE;

	AIX_resync_lv_mirrors(volume, force);

	return;
}

/****************************************************
* Function: AIX_resync_lv_mirrors
*
* 
* 
* 
*
*****************************************************/
static int
AIX_resync_lv_mirrors(struct aix_logical_volume *volume, int force)
{

	int i;
	char pp_stale = FALSE;

	struct partition_list_entry *master_part = NULL;
	struct partition_list_entry *slave1_part = NULL;
	struct partition_list_entry *slave2_part = NULL;

	u64 master_offset = 0;
	u64 slave1_offset = 0;
	u64 slave2_offset = 0;

	LOG_DEBUG("Function %s volume: %s \n", __FUNCTION__, volume->name);

	for (i = 0; i < volume->num_le; i++, pp_stale = FALSE) {

		// We need to see which mirror has a valid non-stale copy.
		// The first non-stale copy will be our master and we'll
		// copy to the slave(s).

		if ((volume->le_to_pe_map[i].pp_state & AIX_LVM_LVSTALE)) {
			pp_stale = TRUE;
		}

		if (volume->le_to_pe_map_mir1 != NULL) {
			if ((volume->le_to_pe_map_mir1[i].
			     pp_state & AIX_LVM_LVSTALE)) {
				pp_stale = TRUE;
			}
		}

		if (volume->le_to_pe_map_mir2 != NULL) {
			if ((volume->le_to_pe_map_mir2[i].
			     pp_state & AIX_LVM_LVSTALE)) {
				pp_stale = TRUE;
			}
		}

		LOG_DEBUG("Function %s pp_stale:%d force:%d \n", __FUNCTION__,
			  pp_stale, force);

		if (pp_stale || force) {
			if (!(volume->le_to_pe_map[i].pp_state & AIX_LVM_LVSTALE)) {

				master_part = volume->le_to_pe_map[i].owning_pv;
				master_offset = volume->le_to_pe_map[i].pe_sector_offset;

				if (volume->le_to_pe_map_mir1 != NULL) {
					slave1_part = volume->le_to_pe_map_mir1[i].owning_pv;
					slave1_offset = volume->le_to_pe_map_mir1[i].pe_sector_offset;
				}

				if (volume->le_to_pe_map_mir2 != NULL) {
					slave2_part = volume->le_to_pe_map_mir2[i].owning_pv;
					slave2_offset = volume->le_to_pe_map_mir2[i].pe_sector_offset;
				}
			} else
			    if (!(volume->le_to_pe_map_mir1[i].pp_state & AIX_LVM_LVSTALE)) {
				master_part = volume->le_to_pe_map_mir1[i].owning_pv;
				master_offset = volume->le_to_pe_map_mir1[i].pe_sector_offset;

				if (volume->le_to_pe_map != NULL) {
					slave1_part = volume->le_to_pe_map[i].owning_pv;
					slave1_offset = volume->le_to_pe_map[i].pe_sector_offset;
				}

				if (volume->le_to_pe_map_mir2 != NULL) {
					slave2_part = volume->le_to_pe_map_mir2[i].owning_pv;
					slave2_offset = volume->le_to_pe_map_mir2[i].pe_sector_offset;
				}
			} else
			    if (!(volume->le_to_pe_map_mir2[i].pp_state & AIX_LVM_LVSTALE)) {
				master_part = volume->le_to_pe_map_mir2[i].owning_pv;
				master_offset = volume->le_to_pe_map_mir2[i].pe_sector_offset;

				if (volume->le_to_pe_map != NULL) {
					slave1_part = volume->le_to_pe_map[i].owning_pv;
					slave1_offset = volume->le_to_pe_map[i].pe_sector_offset;
				}

				if (volume->le_to_pe_map_mir1 != NULL) {
					slave2_part = volume->le_to_pe_map_mir1[i].owning_pv;
					slave2_offset = volume->le_to_pe_map_mir1[i].pe_sector_offset;
				}
			}

			if (AIX_copy_on_read(volume, master_part, slave1_part, slave2_part,
			     master_offset, slave1_offset, slave2_offset,
			     volume->pe_size, i)) {

				LOG_CRITICAL("ReSync of logical Volume %s FAILED !!\n",
				     volume->name);
				AIX_evms_cs_notify_lv_io_error(volume->
							       volume_node);
				break;
			}

		}

	}

	return 0;
}

/****************************************************
* Function: AIX_copy_on_read
*
* 
* 
* 
*
*****************************************************/
static int
AIX_copy_on_read(struct aix_logical_volume *volume,
		 struct partition_list_entry *master_part,
		 struct partition_list_entry *slave1_part,
		 struct partition_list_entry *slave2_part,
		 u64 master_offset,
		 u64 slave1_offset, u64 slave2_offset, u32 pe_size, int le)
{
	unsigned long flags;
	struct aix_mirror_bh *tmp_bh = NULL;

	// Check for valid partitions we need at least 2 good partitions so slave2 doesn't have to be valid

	if (!master_part || !slave1_part) {
		LOG_ERROR("Invalid partitions for resync master part:%p slave1_part:%p slave2_part:%p\n",
		     master_part, slave1_part, slave2_part);
		return -EINVAL;
	}

	LOG_DEBUG("Function %s volume:%s master_part:%d, slave1_part:%d, slave2_part:%d master_offset:"
	     PFU64 ", slave1_offset:" PFU64 " slave2_offset:" PFU64 ",  \n",
	     __FUNCTION__, volume->name, master_part->pv_number,
	     slave1_part->pv_number, slave2_part->pv_number, master_offset,
	     slave1_offset, slave2_offset);

	LOG_DEBUG("pe_size:%d le:%d\n", pe_size, le);

	tmp_bh =
	    AIX_alloc_sbh(volume, master_part, slave1_part, slave2_part,
			  master_offset, slave1_offset, slave2_offset, pe_size);

	if (!tmp_bh) {
		buffer_IO_error(&tmp_bh->bh_req);
		return -ENOMEM;
	}

/*	if (evms_cs_volume_request_in_progress
	    (tmp_bh->bh_req.b_rdev, AIX_INCREMENT_REQUEST, &count)) {
		buffer_IO_error(&tmp_bh->bh_req);
		return -EIO;
	} */

	spin_lock_irqsave(&AIX_resync_pp_lock, flags);

	LOG_DEBUG("Function:%s kicking off read node:%p\n", __FUNCTION__,
		  master_part->logical_node);

	R_IO(master_part->logical_node, &tmp_bh->bh_req);

	spin_unlock_irqrestore(&AIX_resync_pp_lock, flags);

	return 0;
}

/****************************************************
* Function: AIX_alloc_sbh
*
* Alloc any buffer heads from the pool and return a linked list
*
*
*****************************************************/
static struct aix_mirror_bh *
AIX_alloc_sbh(struct aix_logical_volume *volume,
	      struct partition_list_entry *master_part,
	      struct partition_list_entry *slave1_part,
	      struct partition_list_entry *slave2_part,
	      u64 master_offset,
	      u64 slave1_offset, u64 slave2_offset, u32 pe_size)
{
	struct aix_mirror_bh *tmp_bh = NULL, *head_bh = NULL;
	unsigned long flags;

	LOG_DEBUG("Function:%s Enter\n", __FUNCTION__);

	head_bh = evms_cs_allocate_from_pool(AIX_BH_list_pool, EVMS_BLOCKABLE);
	if (!head_bh) {
		LOG_SERIOUS
		    ("Unable to allocate memory for mirror pool line:%d\n",
		     __LINE__);
		return NULL;
	}
	// Update buffer so we block on a read/write on the normal IO path
	// if we're trying to sync the same sector on the disk
	// We don't want to block if it's different sectors

	spin_lock_irqsave(&AIX_resync_list_lock, flags);

	AIX_resync_list->master_part = master_part;
	AIX_resync_list->slave1_part = slave1_part;
	AIX_resync_list->slave2_part = slave2_part;
	AIX_resync_list->master_offset = master_offset;
	AIX_resync_list->slave1_offset = slave1_offset;
	AIX_resync_list->slave2_offset = slave2_offset;

	head_bh->bh_req.b_data = kmalloc(AIX_RESYNC_BLOCKSIZE + 1, GFP_NOIO);
	if (!head_bh->bh_req.b_data) {
		evms_cs_deallocate_to_pool(AIX_BH_list_pool, head_bh);
		LOG_SERIOUS
		    ("Unable to allocate memory for mirror pool line:%d\n",
		     __LINE__);
		return NULL;
	}

	memset(head_bh->bh_req.b_data, 0, AIX_RESYNC_BLOCKSIZE + 1);

	atomic_set(&head_bh->remaining, 0);
	head_bh->bh_req.b_rsector = master_offset;
	head_bh->bh_req.b_size = AIX_RESYNC_BLOCKSIZE;
	head_bh->sync_flag = AIX_SYNC_INCOMPLETE;
	head_bh->bh_req.b_end_io = AIX_sync_mirrored_partitions;
	head_bh->bh_req.b_page = virt_to_page(head_bh->bh_req.b_data);
	head_bh->bh_req.b_state = 0;
	set_bit(BH_Dirty, &head_bh->bh_req.b_state);
	set_bit(BH_Lock, &head_bh->bh_req.b_state);
	set_bit(BH_Req, &head_bh->bh_req.b_state);
	set_bit(BH_Mapped, &head_bh->bh_req.b_state);
	head_bh->master_bh = NULL;
	head_bh->mirror_bh_list = NULL;

	tmp_bh = evms_cs_allocate_from_pool(AIX_BH_list_pool, EVMS_BLOCKABLE);
	if (!tmp_bh) {
		LOG_SERIOUS
		    ("Unable to allocate memory for mirror pool line:%d\n",
		     __LINE__);
		return NULL;
	}

	head_bh->next_r1 = tmp_bh;
	memcpy(&tmp_bh->bh_req, head_bh, sizeof (struct buffer_head));
	atomic_set(&tmp_bh->remaining, 0);
	tmp_bh->bh_req.b_end_io = NULL;

	if (volume->mirror_copies == AIX_MAX_MIRRORS) {
		tmp_bh->next_r1 =
		    evms_cs_allocate_from_pool(AIX_BH_list_pool,
					       EVMS_BLOCKABLE);
		if (!tmp_bh->next_r1) {
			LOG_SERIOUS
			    ("Unable to allocate memory for mirror pool line:%d\n",
			     __LINE__);
			return NULL;
		}

		memcpy(&tmp_bh->next_r1->bh_req, head_bh,
		       sizeof (struct buffer_head));
		tmp_bh->next_r1->bh_req.b_end_io = NULL;
		atomic_set(&tmp_bh->next_r1->remaining, 0);
	}

	init_waitqueue_head(&head_bh->bh_req.b_wait);

	spin_unlock_irqrestore(&AIX_resync_list_lock, flags);

	LOG_DEBUG("Function:%s Exit head_bh:%p\n", __FUNCTION__, head_bh);

	return head_bh;
}

/****************************************************
* Function: AIX_sync_mirrored_partitions
*
* 
* 
* 
*
*****************************************************/
static void
AIX_sync_mirrored_partitions(struct buffer_head *bh, int uptodate)
{
	struct aix_logical_volume *volume = NULL;
	struct aix_mirror_bh *tmp_bh, *head_bh;

	head_bh = tmp_bh = (struct aix_mirror_bh *) bh->b_private;
	volume = (struct aix_logical_volume *) tmp_bh->node->private;

	LOG_DEBUG("Function:%s Enter uptodate:%d\n", __FUNCTION__, uptodate);

	if (!uptodate) {

		AIX_evms_cs_notify_lv_io_error(tmp_bh->node);
	}

	tmp_bh = head_bh->next_r1;

	LOG_DEBUG("Function:%s line:%d write to mirror:%p\n", __FUNCTION__,
		  __LINE__, tmp_bh);

	if (tmp_bh) {
		W_IO(tmp_bh->node, &tmp_bh->bh_req);
		AIX_get_set_mirror_offset(tmp_bh, AIX_SLAVE_1,
					  AIX_RESYNC_BLOCKSIZE);
	}

	tmp_bh = tmp_bh->next_r1;
	LOG_DEBUG("Function:%s line:%d write to mirror:%p\n", __FUNCTION__,
		  __LINE__, tmp_bh);

	if (tmp_bh) {
		W_IO(tmp_bh->node, &tmp_bh->bh_req);
		AIX_get_set_mirror_offset(tmp_bh, AIX_SLAVE_2,
					  AIX_RESYNC_BLOCKSIZE);
	}

	LOG_DEBUG("Function:%s line:%d read from  master:%p\n", __FUNCTION__,
		  __LINE__, head_bh);

	if (head_bh && head_bh->sync_flag) {
		AIX_get_set_mirror_offset(head_bh, AIX_MASTER,
					  AIX_RESYNC_BLOCKSIZE);
		if (head_bh->sync_flag == AIX_SYNC_INCOMPLETE) {
			R_IO(head_bh->node, &head_bh->bh_req);
		}
	}

	LOG_DEBUG("Function:%s line:%d head_bh->sync_flag:%d\n", __FUNCTION__,
		  __LINE__, head_bh->sync_flag);

	if (!head_bh->sync_flag) {
		tmp_bh = head_bh;
		head_bh = head_bh->next_r1;

		while (tmp_bh != NULL) {
			evms_cs_deallocate_to_pool(AIX_BH_list_pool, tmp_bh);
			tmp_bh = head_bh;
		}

		AIXResyncInProgress = FALSE;
/*		evms_cs_volume_request_in_progress(tmp_bh->bh_req.b_rdev,
						   AIX_DECREMENT_REQUEST,
						   &count); */

		if (AIX_resync_list) {
			kfree(AIX_resync_list);
		}
	}

	return;
}

/****************************************************
* Function: AIX_get_set_mirror_offset
*
* 
* 
* 
*
*****************************************************/
static int
AIX_get_set_mirror_offset(struct aix_mirror_bh *tmp_bh, int index, int offset)
{
	int flags;

	if (!tmp_bh) {
		return -EINVAL;
	}

	LOG_DEBUG("Function:%s Enter offset:%d\n", __FUNCTION__, offset);

	tmp_bh->bh_req.b_rsector += tmp_bh->bh_req.b_rsector + offset;

	if (tmp_bh->bh_req.b_rsector > tmp_bh->node->total_vsectors) {
		tmp_bh->sync_flag = AIX_SYNC_COMPLETE;
		return -EIO;
	}
	// Update buffer so we block on a read/write on the normal IO path
	// if we're trying to sync the same sector on the disk
	// We don't want to block if it's different sectors

	spin_lock_irqsave(&AIX_resync_list_lock, flags);

	if (AIX_resync_list->master_part->logical_node == tmp_bh->node) {
		AIX_resync_list->master_offset += offset;
	}

	if (AIX_resync_list->slave1_part->logical_node == tmp_bh->node) {
		AIX_resync_list->slave1_offset += offset;
	}

	if (AIX_resync_list->slave2_part->logical_node == tmp_bh->node) {
		AIX_resync_list->slave2_offset += offset;
	}

	spin_unlock_irqrestore(&AIX_resync_list_lock, flags);

	return 0;

}

static int AIX_pvh_data_posn(u32 vgda_psn, u32 * pvh_posn, struct partition_list_entry *partition, u32 numpvs)
{
    struct partition_list_entry * pv;
    struct pv_header            * AIXpvh;
    int posn = 0;
    int num_pps;
    int tmp,i;

    LOG_DEBUG("APDP - vgda_psn:%d numpvs:%d \n", vgda_psn, numpvs);

	AIXpvh = kmalloc(AIX_SECTOR_SIZE, GFP_KERNEL);
	if (!AIXpvh) {
		return -ENOMEM;
	}

    memset(AIXpvh, 0 , sizeof(struct pv_header));

    // Adjust this because when AIX VGs/Volumes are created on Intel platforms, the
    // pp_count could be anything since we don't give up the entire physical drive.
    // This is for calculation purposes only.

    pvh_posn[0] = 0;
    pv = partition;

    for (i = 1; i <= numpvs; i++) {
        for (pv = partition; pv->pv_number != i; pv = pv->next ); 

        LOG_DEBUG("APDP line:%d pp_count:%d \n", __LINE__,  AIXpvh->pp_count);

        num_pps = AIXpvh->pp_count;
        num_pps++; // Account for the pv_header on the front

        while ((num_pps * sizeof(struct pp_entries)) % AIX_SECTOR_SIZE) {
            LOG_EXTRA("num_pps:%d \n", num_pps);
            num_pps++;
        }

        tmp = (num_pps * sizeof(struct pp_entries)) / AIX_SECTOR_SIZE;

        LOG_DEBUG("APDP tmp:%d num_pps:%d \n", tmp,num_pps);

        posn = ((vgda_psn + PSN_PPH_OFFSET) + ((pv->pv_number -1) * tmp));

        pvh_posn[pv->pv_number] = posn;

        if (INIT_IO(pv->logical_node, 0, posn, 1, AIXpvh)) {
            kfree(AIXpvh);
            return -EIO;
        }

        pv = partition;
    }

    kfree(AIXpvh);

    return 0;
}

/****************************************************
* Function: AIX_volume_group_dump
*
* This is for debug purposes and will walk the volume group list
* and LV's within the volume groups
*
* It can be called at anytime however the output to the display is large
*
*****************************************************/
#ifdef EVMS_AIX_DEBUG
static int
AIX_volume_group_dump(void)
{
	struct aix_volume_group *AIXVGLDebugPtr;
	struct partition_list_entry *DebugPartitionList;
	struct aix_logical_volume *DebugLVList;
	int i;

	AIXVGLDebugPtr = AIXVolumeGroupList;

	if (!AIXVGLDebugPtr) {
		LOG_DEBUG("***********************************************\n");
		LOG_DEBUG("ERROR Nothing built in the list to check !!!   \n");
		LOG_DEBUG("***********************************************\n");
		return 0;
	}

	LOG_DEBUG("***********************************************    \n");
	LOG_DEBUG("Begin Volume Group Dump \n");
	LOG_DEBUG("***********************************************    \n");

	while (AIXVGLDebugPtr) {

		LOG_DEBUG("vg_number      %x\n", AIXVGLDebugPtr->vg_id.word2);
		LOG_DEBUG("numpsrtitions  %d\n", AIXVGLDebugPtr->partition_count);
		LOG_DEBUG("numlvs         %d\n", AIXVGLDebugPtr->numlvs);
		LOG_DEBUG("hard_sect_size %d\n", AIXVGLDebugPtr->hard_sect_size);
		LOG_DEBUG("block_size     %d\n", AIXVGLDebugPtr->block_size);
		LOG_DEBUG("flags          %d\n", AIXVGLDebugPtr->flags);
//		LOG_DEBUG("lv_max         %d\n", AIXVGLDebugPtr->lv_max);
		LOG_DEBUG("pe_size        %d\n", AIXVGLDebugPtr->pe_size);
		LOG_DEBUG("CleanVGInfo    %d\n", AIXVGLDebugPtr->CleanVGInfo);

		DebugPartitionList = AIXVGLDebugPtr->partition_list;

		LOG_DEBUG("********* Begin Volume Partition Dump ********* \n");

		if (!DebugPartitionList) {
			LOG_DEBUG("No partitions to check !!  \n");
		}

		while (DebugPartitionList) {
			LOG_DEBUG("logical_node       %p\n",
				  DebugPartitionList->logical_node);
			LOG_DEBUG("pv_number          %d\n",
				  DebugPartitionList->pv_number);
			LOG_DEBUG("block_size         %d\n",
				  DebugPartitionList->block_size);
			LOG_DEBUG("hard_sect_size     %d\n",
				  DebugPartitionList->hard_sect_size);
			LOG_DEBUG("-------------------------------------------------------------\n");
			DebugPartitionList = DebugPartitionList->next;
		}

		LOG_DEBUG("********* End Volume Partition Dump **********\n");

		LOG_DEBUG("********** Begin Logical Volume Partition Dump **********\n");

		DebugLVList = AIXVGLDebugPtr->volume_list[0];

		if (!DebugLVList) {
			LOG_DEBUG("No logical volumes to check !!  \n");
		}

		for (i = 0; i < LVM_MAXLVS && DebugLVList; i++) {

			DebugLVList = AIXVGLDebugPtr->volume_list[i];

			if (DebugLVList) {
				LOG_DEBUG("volume_list #    %d \n", i);
				LOG_DEBUG("lv_number        %d \n",
					  DebugLVList->lv_number);
				LOG_DEBUG("LV name          %s \n",
					  DebugLVList->name);
				LOG_DEBUG("lv_size          " PFU64 " \n",
					  DebugLVList->lv_size);
				LOG_DEBUG("lv_access        %d \n",
					  DebugLVList->lv_access);
				LOG_DEBUG("lv_status        %d \n",
					  DebugLVList->lv_status);
//				LOG_DEBUG("lv_minor         %d \n",
//					  DebugLVList->lv_minor);
				LOG_DEBUG("mirror_copies    %d \n",
					  DebugLVList->mirror_copies);
//				LOG_DEBUG("mirror_number    %d \n",
//					  DebugLVList->mirror_number);
				LOG_DEBUG("stripes          %d \n",
					  DebugLVList->stripes);
				LOG_DEBUG("stripe_size      %d \n",
					  DebugLVList->stripe_size);
				LOG_DEBUG("stripe_size_shift%d \n",
					  DebugLVList->stripe_size_shift);
				LOG_DEBUG("pe_size          %d \n",
					  DebugLVList->pe_size);
				LOG_DEBUG("pe_size_shift    %d \n",
					  DebugLVList->pe_size_shift);
				LOG_DEBUG("num_le           %d \n",
					  DebugLVList->num_le);
//				LOG_DEBUG("new_volume       %d \n",
//					  DebugLVList->new_volume);
				LOG_DEBUG("group            %p \n",
					  DebugLVList->group);
			}

		}

		AIXVGLDebugPtr = AIXVGLDebugPtr->next;

		LOG_DEBUG("********** End Logical Volume Partition Dump **********\n");

	}

	LOG_DEBUG("***********************************************\n");
	LOG_DEBUG("End Volume Group Dump                          \n");
	LOG_DEBUG("***********************************************\n");

	return 0;

}
#endif
