/* -*- linux-c -*- 
 *
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 *
 */
/*
 * linux/drivers/evms/drvlink.c

 *
 * EVMS Drive Linking Feature.
 *
 * This feature provides the ability to link multiple storage objects
 * together as a single virtual storage object.
 *
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/genhd.h>
#include <linux/blk.h>
#include <linux/init.h>
#include <linux/evms/evms.h>
#include <linux/evms/evms_drivelink.h>
#include <asm/uaccess.h>

#define LOG_PREFIX "drivelink: "

/* prototypes for mandatory plugin interface functions */
static int drivelink_discover(struct evms_logical_node **);
static int drivelink_delete(struct evms_logical_node *);
static void drivelink_read(struct evms_logical_node *, struct buffer_head *);
static void drivelink_write(struct evms_logical_node *, struct buffer_head *);
static int drivelink_ioctl(struct evms_logical_node *,
			   struct inode *,
			   struct file *, unsigned int, unsigned long);
static int drivelink_init_io(struct evms_logical_node *,
			     int, u64, u64, void *);

/* plugin function table definition */
static struct evms_plugin_fops fops = {
	.discover	= drivelink_discover,
	.delete		= drivelink_delete,
	.read		= drivelink_read,
	.write		= drivelink_write,
	.init_io	= drivelink_init_io,
	.ioctl		= drivelink_ioctl
};

/* plugin header definition */
static struct evms_plugin_header plugin_header = {
	.id = SetPluginID(IBM_OEM_ID,
			  EVMS_FEATURE,
			  EVMS_DRIVELINK_FEATURE_ID),
	.version = {
		.major		= 2,
		.minor		= 0,
		.patchlevel	= 2
	},
	.required_services_version = {
		.major		= 0,
		.minor		= 5,
		.patchlevel	= 0
	},
	.fops = &fops
};

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Discover function & Support routines            */
/********************************************************/

/**
 * le_feature_data_to_cpu:
 * @md: 	drivelink metadata
 * 
 * convert feature data from on-disk (Little Endian) format
 * to the native cpu endian format.
**/
static void
le_feature_data_to_cpu(struct evms_drivelink_metadata *md)
{
	int i;

	md->signature = le32_to_cpup(&md->signature);
	md->crc = le32_to_cpup(&md->crc);
	md->version.major = le32_to_cpup(&md->version.major);
	md->version.minor = le32_to_cpup(&md->version.minor);
	md->version.patchlevel = le32_to_cpup(&md->version.patchlevel);
	md->flags = le32_to_cpup(&md->flags);
	md->sequence_number = le64_to_cpup(&md->sequence_number);
	md->child_serial_number = le64_to_cpup(&md->child_serial_number);
	md->parent_serial_number = le64_to_cpup(&md->parent_serial_number);
	md->child_count = le64_to_cpup(&md->child_count);
	for (i = 0; i < EVMS_DRIVELINK_MAX_ENTRIES; i++) {
		struct evms_dl_ordering_table_entry *child_entry;

		child_entry = &md->ordering_table[i];
		child_entry->child_serial_number =
		    le64_to_cpup(&child_entry->child_serial_number);
		child_entry->child_vsize =
		    le64_to_cpup(&child_entry->child_vsize);
	}
}

/**
 * load_feature_data:	load a feature header from disk
 * @node: 	storage object
 * @md: 	ptr to drivelink metadata
 * 
 * loads and verifies redundant copies of drivelink metadata. @md is modified
 * and returned to the caller.
 * 
 * Return value: 0 on success
 *		 Otherwise error code
**/
static int
load_feature_data(struct evms_logical_node *node,
		  struct evms_drivelink_metadata **md)
{
	int i, rc = 0, rc_array[2] = { 0, 0 }, size_in_bytes;
	u64 real_metadata_size, feature_data_size;
	u64 starting_sector;
	struct evms_drivelink_metadata *cur_md, *md1, *md2 = NULL;
	char *location_name;

	/* verify the feature metadata size from the  */
	/* feature header agrees with the real size   */
	/* of the current metadata structure.         */
	real_metadata_size = evms_cs_size_in_vsectors(sizeof (**md));

	/* allocate a buffer large enough to hold all */
	/* sectors containing the feature's metadata  */
	size_in_bytes = real_metadata_size * EVMS_VSECTOR_SIZE;
	md1 = kmalloc(size_in_bytes, GFP_KERNEL);
	if (md1) {
		md2 = kmalloc(size_in_bytes, GFP_KERNEL);
		if (!md2) {
			kfree(md1);
			rc = -ENOMEM;
		}
	} else {
		rc = -ENOMEM;
	}
	if (!rc) {
		for (i = 0; i < 2; i++) {
			if (i == 0) {
				starting_sector =
				    node->feature_header->
				    feature_data1_start_lsn;
				feature_data_size =
				    node->feature_header->feature_data1_size;
				cur_md = md1;
				location_name = evms_primary_string;
			} else {
				starting_sector =
				    node->feature_header->
				    feature_data2_start_lsn;
				feature_data_size =
				    node->feature_header->feature_data2_size;
				cur_md = md2;
				location_name = evms_secondary_string;
			}
			/* check that real metadata size matches the  */
			/* feature data size                          */
			if (real_metadata_size != feature_data_size) {
				LOG_ERROR
				    ("%s feature data size("PFU64" bytes) doesn't match expected size("PFU64" bytes).\n",
				     location_name,
				     feature_data_size <<
				     EVMS_VSECTOR_SIZE_SHIFT,
				     real_metadata_size <<
				     EVMS_VSECTOR_SIZE_SHIFT);
				rc = -EINVAL;
				rc_array[i] = rc;
				continue;
			}
			/* load the node's feature data */
			rc = INIT_IO(node,
				     0,
				     starting_sector,
				     feature_data_size, cur_md);
			if (rc) {
				LOG_ERROR
				    ("error(%d) probing for %s feature data at sector("PFU64") on '%s'.\n",
				     rc, location_name, starting_sector,
				     node->name);
				rc_array[i] = rc;
				continue;
			}
			/* check for valid metadata signature */
			if (le32_to_cpup(&cur_md->signature) !=
			    EVMS_DRIVELINK_SIGNATURE) {
				rc = -ENODATA;
				LOG_SERIOUS
				    ("error(%d) invalid signature in %s feature data on '%s'\n",
				     rc, location_name, node->name);
				rc_array[i] = rc;
				continue;
			}
			/* validate feature data CRC */
			if (cur_md->crc != EVMS_MAGIC_CRC) {
				int org_crc, final_crc;
				org_crc = le32_to_cpup(&cur_md->crc);
				cur_md->crc = 0;
				final_crc =
				    evms_cs_calculate_crc(EVMS_INITIAL_CRC,
							  cur_md,
							  sizeof (*cur_md));
				if (final_crc != org_crc) {
					LOG_ERROR
					    ("CRC mismatch error [stored(%x), computed(%x)] in %s feature data on '%s'.\n",
					     org_crc, final_crc, location_name,
					     node->name);
					rc = -EINVAL;
					rc_array[i] = rc;
					continue;
				}
			} else {
				LOG_WARNING
				    ("CRC disabled in %s feature data on '%s'.\n",
				     location_name, node->name);
			}
			/* convert feature data from on-disk
			 * format (Little Endian) to native
			 * cpu endian format.
			 */
			le_feature_data_to_cpu(cur_md);
			/* check for valid structure version */
			rc = evms_cs_check_version(&metadata_ver,
						   &cur_md->version);
			if (rc) {
				LOG_SERIOUS
				    ("error(%d) obsolete version detected: actual(%d,%d,%d), requires(%d,%d,%d) in %s feature data on '%s'\n",
				     rc, cur_md->version.major,
				     cur_md->version.minor,
				     cur_md->version.patchlevel,
				     DRIVELINK_METADATA_MAJOR,
				     DRIVELINK_METADATA_MINOR,
				     DRIVELINK_METADATA_PATCHLEVEL,
				     location_name, node->name);
				rc_array[i] = rc;
			}
		}
		/* getting same return code for both copies? */
		if (rc_array[0] == rc_array[1]) {
			rc = rc_array[0];
			/* if no errors on both copies,
			 * check the sequence numbers.
			 * use the highest sequence number.
			 */
			if (!rc) {
				/* compare sequence numbers */
				if (md1->sequence_number ==
				    md2->sequence_number) {
					cur_md = md1;
				} else {
					LOG_WARNING
					    ("sequence number mismatches between front("PFU64") and rear("PFU64") feature data copies on node(%s)!\n",
					     md2->sequence_number,
					     md1->sequence_number, node->name);
					if (md1->sequence_number >
					    md2->sequence_number)
						cur_md = md1;
					else
						cur_md = md2;
					LOG_WARNING
					    ("using %s feature data copy!\n",
					     (cur_md ==
					      md1) ? evms_primary_string :
					     evms_secondary_string);
				}
			}
			/* getting different return codes for each copy */
		} else if (rc_array[0] == 0) {
			/* use 1st (rear) copy if its good */
			rc = 0;
			cur_md = md1;
		} else if (rc_array[1] == 0) {
			/* use 2nd (front) copy if its good */
			rc = 0;
			cur_md = md2;
		} else if ((rc_array[0] == -EINVAL) || (rc_array[1] == -EINVAL)) {
			/* fail if either give a fatal error */
			rc = -EINVAL;
			cur_md = NULL;
		}

		/* deallocate metadata buffers appropriately */
		if (rc || (cur_md == md1))
			kfree(md2);
		if (rc || (cur_md == md2))
			kfree(md1);

		/* save validated feature header pointer */
		if (!rc)
			*md = cur_md;
	}
	return (rc);
}

/**
 * find_parent_node_for_child_node: finds or creates a parent node for this child node
 * @child_node: 	input, child node
 * @md: 		input, on-disk metadata
 * @parent_node: 	output, parent node
 * @dl_private: 	output, runtime metadata
 * @discover_list: 	input/output, list of objects being discovered
 * 
 * finds or creates a parent node for the specified child node. if the parent node is
 * created, create and initialize the parent's private data area.
 * 
 * Return value: 0 on success
 *		 Otherwise error code.
**/
static int
find_parent_node_for_child_node(struct evms_logical_node *child_node,
				struct evms_drivelink_metadata *md,
				struct evms_logical_node **parent_node,
				struct runtime_data **dl_private,
				struct evms_logical_node **discover_list)
{
	int rc = 0, parent_found = FALSE;
	struct evms_logical_node *parent = NULL;
	struct runtime_data *rd = NULL;

	/* find the parent node for this child */
	for (parent = *discover_list; parent; parent = parent->next) {
		/* only parent nodes will have null feature headers */
		if (!parent->feature_header) {
			rd = (struct runtime_data *) parent->private;
			if (rd->parent_sn == md->parent_serial_number) {
				parent_found = TRUE;
				break;
			}
		}
	}
	/* if no parent node found, create it */
	if (parent_found == FALSE) {
		rc = evms_cs_allocate_logical_node(&parent);
		if (!rc) {
			/* transpose info from child to parent */
			parent->flags |= child_node->flags;
			strcpy(parent->name,
			       child_node->feature_header->object_name);
			/* copy evms system data to parent */
			parent->volume_info = child_node->volume_info;
			/* initialize the plugin id field */
			parent->plugin = &plugin_header;
			/* allocate parent's instance data */
			parent->private = kmalloc(sizeof(*rd), GFP_KERNEL);
			if (!parent->private)
				rc = -ENOMEM;
		}
		if (!rc) {
			/* initialize some instance data fields */
			rd = (struct runtime_data *) parent->private;
			rd->block_size = 0;
			rd->parent_sn = md->parent_serial_number;
			rd->child_count = md->child_count;
			/* allocate the child table */
			rd->child_table = kmalloc(sizeof(struct runtime_entry) *
						  rd->child_count, GFP_KERNEL);
			if (!rd->child_table)
				rc = -ENOMEM;
		}
		if (!rc) {
			memset(rd->child_table, 0,
				sizeof(struct runtime_entry) * rd->child_count);
			/* add the parent node to the discover list */
			rc = evms_cs_add_logical_node_to_list(discover_list,
							      parent);
			MOD_INC_USE_COUNT;
		}
		/* if any errors encountered, try to clean up */
		if (rc) {
			LOG_SERIOUS("find_parent_node: rc(%d) from '%s'\n",
				    rc, child_node->name);
			if (parent) {
				DELETE(parent);
				parent = NULL;
				rd = NULL;
			}
		}
	}

	*dl_private = rd;
	*parent_node = parent;

	return (rc);
}

/**
 * compute_child_index: compute the index for a specific child node
 * @node: 	the child node
 * @md: 	the drivelink on-disk metadata
 * 
 * compute and return and 0-based index value of this child node's position
 * in the parent node's ordering table.
 * 
 * Return value: -1 on error
 *		 otherwise the index of the specified child.
**/
static int
compute_child_index(struct evms_logical_node *node,
		    struct evms_drivelink_metadata *md)
{
	int i, position = -1;

	for (i = 0; i < md->child_count; i++) {
		if (md->ordering_table[i].child_serial_number ==
		    md->child_serial_number) {
			position = i;
			break;
		}
	}
	if (position == -1) {
		LOG_SERIOUS("%s: child not found from '%s'\n",
			    __FUNCTION__, node->name);
	}
	return (position);
}

/**
 * process_child_nodes: perform the discovery operation on each child node
 * @discover_list: 	the list of potential child objects
 * 
 * search the discovery list of drivelink child nodes. for each node found,
 * perform the discovery operation on it.
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
process_child_nodes(struct evms_logical_node **discover_list)
{
	int rc = 0, index = -1;
	struct evms_logical_node *node, *next_node, *parent;
	struct evms_drivelink_metadata *md;
	struct runtime_data *rd;
	struct runtime_entry *child_entry = NULL;

	for (node = *discover_list; node; node = next_node) {
		next_node = node->next;
		if ((!node->feature_header) ||
		    (node->feature_header->feature_id != plugin_header.id)) {
			continue;
		}

		rc = evms_cs_remove_logical_node_from_list(discover_list, node);
		if (rc)
			BUG();
		/* we need to load the feature data to   */
		/* find the parent's serial number this  */
		/* child node belongs to.                */
		md = NULL;
		rc = load_feature_data(node, &md);
		if (!rc) {
			/* find the parent node for this child */
			parent = NULL;
			rc = find_parent_node_for_child_node(node, md,
							     &parent, &rd,
							     discover_list);
		}
		if (!rc) {
			/* determine position of child in drive link object */
			index = compute_child_index(node, md);
			if (index == -1)
				rc = index;
		}
		if (!rc) {
			/* check for multiple child index requests */
			child_entry =
			    (struct runtime_entry *) &rd->child_table[index];
			/* check to see if this child index is 
			 * already in use.
			 */
			if (child_entry->child_node) {
				LOG_SERIOUS
				    ("attempt to put '%s' in child index(%d). Already occupied by '%s'.\n",
				     node->name, index,
				     child_entry->child_node->name);
				rc = -1;
			}
		}
		if (!rc) {
			/* fill in child info in parent */

			/* check the sector size for this node */
			if (node->hardsector_size > parent->hardsector_size)
				parent->hardsector_size = node->hardsector_size;
			/* check the block size for this node */
			if (node->block_size > parent->block_size)
				parent->block_size = node->block_size;
			/* set the child node */
			child_entry->child_node = node;
			/* set the metadata for this node */
			child_entry->child_metadata = md;
		}

		/* on error, clean up accordingly */
		if (rc) {
			if (md)
				kfree(md);
			LOG_SERIOUS("%s: rc(%d) from '%s'\n",
				    __FUNCTION__, rc, node->name);
			LOG_SERIOUS("deleting child node '%s'.\n", node->name);
			rc = DELETE(node);
			if (rc) {
				LOG_SERIOUS
				    ("error(%d) attempting to delete '%s'.\n",
				     rc, node->name);
			}
		}
	}

	/* errors are handled internal to this function */
	/* by deleting the failed node. This will get   */
	/* picked up by finalize_parent_nodes as a      */
	/* missing child node                           */
	return (0);
}

#define TEST_CHILD_PRESENCE		0
#define TEST_CHILD_COUNT		1
#define TEST_CHILD_PARENTS_SERIAL_NUM	2
#define TEST_CHILD_POSITION		3
#define TEST_CHILD_METADATA		4

/**
 * test_parent_node: verify that a parent is complete
 * @node: 	specified parent node
 * 
 * verify that the parent node has all of its child nodes accounted for.
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
test_parent_node(struct evms_logical_node *node)
{
	int i, rc = 0;
	struct runtime_data *rd;
	struct runtime_entry *child_entry;

	rd = (struct runtime_data *) node->private;
	for (i = 0; i < rd->child_count; i++) {
		child_entry = (struct runtime_entry *) &rd->child_table[i];

		/* insure each child entry is filled */
		if (!child_entry->child_node) {
			node->flags |=
			    EVMS_VOLUME_SET_READ_ONLY | EVMS_VOLUME_PARTIAL;
			LOG_ERROR("%s: missing child(%d).\n", __FUNCTION__, i);
		} else
			/* insure child count is the same */
			/* in each child's metadata       */
		if (child_entry->child_metadata->child_count != rd->child_count) {
			rc = -EVMS_FEATURE_FATAL_ERROR;
			LOG_ERROR("%s: child count wrong for node '%s'\n",
				  __FUNCTION__, node->name);
		} else
			/* insure parent serial number is    */
			/* the same in each child's metadata */
		if (child_entry->child_metadata->parent_serial_number !=
			    rd->parent_sn) {
			rc = -EVMS_FEATURE_FATAL_ERROR;
			LOG_ERROR
			    ("%s: incorrect [is("PFU64"), should be("PFU64")] child serial number for node '%s'\n",
			     __FUNCTION__,
			     child_entry->child_metadata->parent_serial_number,
			     rd->parent_sn, node->name);
		} else
			/* insure each is in the correct entry */
		if (child_entry->child_metadata->ordering_table[i].
			    child_serial_number !=
			    child_entry->child_metadata->child_serial_number) {
			rc = -EVMS_FEATURE_FATAL_ERROR;
			LOG_ERROR
			    ("%s: child reports different index for node '%s'\n",
			     __FUNCTION__, node->name);
		} else {
			struct runtime_entry *other_child_entry;
			int j, rc2;
			/* compare the children's metadata */

			/* look for another present child to 
			 * compare against.
			 */
			other_child_entry = NULL;
			for (j = 0; j < rd->child_count; j++) {
				/* skip comparing to ourselves */
				if (j == i) {
					continue;
				}
				/* is this child is present? */
				if (rd->child_table[j].child_node) {
					/* yes, use it */
					other_child_entry = &rd->child_table[j];
					break;
				}
			}
			/* if we can't find another valid
			 * child node's metadata to compare
			 * against, just skip this test.
			 */
			if (!other_child_entry) {
				continue;
			}
			rc2 =
			    memcmp(other_child_entry->child_metadata->
				   ordering_table,
				   child_entry->child_metadata->ordering_table,
				   sizeof (child_entry->child_metadata->
					   ordering_table));
			if (rc2) {
				rc = -EVMS_FEATURE_FATAL_ERROR;
				LOG_ERROR
				    ("%s: mismatching child metadata for nodes '%s' and '%s'\n",
				     __FUNCTION__,
				     rd->child_table[i - 1].child_node->name,
				     child_entry->child_node->name);
			}
		}
		/* stop if fatal error encountered */
		if (rc == -EVMS_FEATURE_FATAL_ERROR) {
			break;
		}
	}
	return (rc);
}

/**
 * perform_final_adjustments:	do final tweaks to parent node
 * @node: 	parent node
 * 
 * This function does the following:
 *           sets the vsize (in vsectors) field in each child node
 *           sets the voffset (in vsectors) field in each child node
 *           frees each child node's metadata
 *           sets the parent's total size field
**/
static void
perform_final_adjustments(struct evms_logical_node *node)
{
	int i;
	struct runtime_data *rd;
	struct runtime_entry *child_entry = NULL;
	struct evms_drivelink_metadata *ref_data = NULL;

	rd = (struct runtime_data *) node->private;
	/* find a valid copy of the ordering table.
	 * since all the ordering tables are the same
	 * we can just pick one to use for all the
	 * child computations.
	 */
	for (i = 0; i < rd->child_count; i++) {
		child_entry = (struct runtime_entry *) &rd->child_table[i];
		if (child_entry->child_node) {
			ref_data = child_entry->child_metadata;
			break;
		}
	}
	/* if we got this far, there should
	 * always be at least one valid child.
	 */
	if (!ref_data)
		BUG();
	/* compute the parent's usable size,
	 * and construct the table used to
	 * remap parent I/Os to child I/Os */
	for (i = 0; i < rd->child_count; i++) {
		child_entry = (struct runtime_entry *) &rd->child_table[i];
		/* set the LBA count for this child node */
		child_entry->vsize = ref_data->ordering_table[i].child_vsize;
		/* set the start LBA value for this child node */
		child_entry->voffset = node->total_vsectors;
		/* keep a running total of size in sectors */
		node->total_vsectors += child_entry->vsize;
		/* free the metadata for this child node */
		if (ref_data != child_entry->child_metadata) {
			kfree(child_entry->child_metadata);
		}
		child_entry->child_metadata = NULL;
		/* free the feature header for this child node */
		if (child_entry->child_node) {
			kfree(child_entry->child_node->feature_header);
			child_entry->child_node->feature_header = NULL;
		}
	}
	/* free the reference data */
	kfree(ref_data);
}

/**
 * finalize_parent_nodes: verify and prepare parent nodes
 * @discover_list: 	list of potential drivelink parent objects
 * 
 * verify the completeness of each parent node. if not complete, purge the in-memory
 * structs for this object and all its children. If complete, perform final tweaks
 * to allow this node to useable.
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
finalize_parent_nodes(struct evms_logical_node **discover_list)
{
	int rc = 0, rc2;
	struct evms_logical_node *node, *next_node;

	for (node = *discover_list; node; node = next_node) {
		next_node = node->next;
		/* only check parent nodes */
		if (!node->feature_header) {
			/* valid the children of this parent */
			rc = test_parent_node(node);
			if (!rc) {
				/* compute parent size and
				 * child remap table.
				 */
				perform_final_adjustments(node);
			} else {
				/* fatal error encountered. 
				 * cleanup from this node and
				 * delete it from memory.
				 */
				evms_cs_remove_logical_node_from_list
				    (discover_list, node);
				rc2 = DELETE(node);
				if (rc2) {
					LOG_SERIOUS
					    ("error(%d) attempting to delete '%s'.\n",
					     rc2, node->name);
				}
			}
		}
	}
	return (rc);
}

/**
 * drivelink_discover: discover drivelinked storage objects
 * @discover_list: 	the list of objects to inspect
 * 
 * perform the drivelink discover process on the objects in the discovery list
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
drivelink_discover(struct evms_logical_node **discover_list)
{
	int rc = 0;

	MOD_INC_USE_COUNT;
	rc = process_child_nodes(discover_list);
	if (!rc)
		rc = finalize_parent_nodes(discover_list);

	MOD_DEC_USE_COUNT;
	return (rc);
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Delete function                                 */
/********************************************************/

/**
 * drivelink_delete: purges a drivelink object and its children from memory
 * @node: 	the drivelink object to delete
 * 
 * purge the drivelink object, its private data, and all its children from memory.
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
drivelink_delete(struct evms_logical_node *node)
{
	int i, rc = 0;
	struct runtime_data *rd;
	struct runtime_entry *child_entry;

	LOG_DETAILS("deleting '%s'.\n", node->name);

	rd = (struct runtime_data *) node->private;
	if (rd) {
		for (i = 0; i < rd->child_count; i++) {
			child_entry = &rd->child_table[i];
			/* delete the child node */
			if (child_entry->child_node) {
				rc = DELETE(child_entry->child_node);
				if (rc)
					break;
				child_entry->child_node = NULL;
			}
			/* delete the child's metadata */
			if (child_entry->child_metadata) {
				kfree(child_entry->child_metadata);
				child_entry->child_metadata = NULL;
			}
		}
		if (!rc) {
			/* delete the child table */
			if (rd->child_table) {
				kfree(rd->child_table);
				rd->child_table = NULL;
			}
			/* delete the instance data */
			kfree(rd);
			node->private = NULL;
		}
	}
	if (!rc) {
		evms_cs_deallocate_logical_node(node);
		MOD_DEC_USE_COUNT;
	}

	return (rc);
}

/**
 * which_child: find the child node targetted by a IO to this drivelink object
 * @parent: 		parent drivelink object
 * @rsector: 		relative sector on the parent object
 * @max_io_sects: 	largest IO size on the child, starting from rsector position
 * 
 * This function find the child node a parent rsector maps to.
 * It then adjusts the rsector value to be child relative and
 * optionally computes the max # of sectors that can be access
 * from this starting point on the child. 
 *
 * Return value: 
 *   The child node, the child relative rsector and max io size are 
 * returned to the caller. On error, the returned child node will
 * be NULL.
**/
static struct evms_logical_node *
which_child(struct evms_logical_node *parent,
	    u64 * rsector, u64 * max_io_sects)
{
	int i;
	struct evms_logical_node *child = NULL;
	struct runtime_data *rd;
	struct runtime_entry *child_entry = NULL;

	rd = (struct runtime_data *) parent->private;
	for (i = 0; i < rd->child_count; i++) {
		child_entry = (struct runtime_entry *) &rd->child_table[i];

		if (*rsector >= child_entry->vsize) {
			*rsector -= child_entry->vsize;
		} else {
			/* get the child node */
			child = child_entry->child_node;
			/* compute the sector count if requested */
			if (max_io_sects)
				/* this is only used for INIT I/O
				 * to return the largest sector
				 * count size for this child based
				 * on first sector in the I/O.
				 */
				*max_io_sects = child_entry->vsize - *rsector;
			break;
		}
	}
	return (child);
}

/**
 * drivelink_io_error:	log an IO error for drivelink
 * @node: 	drivelink object
 * @bh: 	buffer head targetting this object
 * 
 * this function was primarily created because the function
 * buffer_IO_error is inline and kgdb doesn't allow breakpoints
 * to be set on inline functions. Since this was an error path
 * and not mainline, I decided to add a trace statement to help
 * report on the failing condition.
**/
static void
drivelink_io_error(struct evms_logical_node *node, int io_flag, struct buffer_head *bh)
{
	LOG_SERIOUS("%s error on '%s' remapping rsector("PFU64").\n",
		    (io_flag) ? "WRITE" : "READ",
		    node->name, (u64) bh->b_rsector);

	bh->b_end_io(bh, 0);
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Read function & Support routines                */
/********************************************************/

/**
 * drivelink_read: handles IO read operations to drivelink objects
 * @node: 	drivelink object
 * @bh: 	buffer head targetting this object
 * 
 * handles IO read operations to the drivelink objects. internally remaps the 
 * drivelink relative requests to the child relative requests and then routes 
 * it to the child for further processing.
**/
static void
drivelink_read(struct evms_logical_node *node, struct buffer_head *bh)
{
	struct evms_logical_node *child;
	u64 io_size, rsector;

	rsector = bh->b_rsector;
	child = which_child(node, &rsector, &io_size);
	if (child && ((bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) <= io_size)) {
		bh->b_rsector = rsector;
		R_IO(child, bh);
	} else {
		drivelink_io_error(node, READ, bh);
	}
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Write function & Support routines               */
/********************************************************/

/**
 * drivelink_read_write: handles IO write operations to drivelink objects
 * @node: 	drivelink object
 * @bh: 	buffer head targetting this object
 * 
 * handles IO write operations to the drivelink objects. internally remaps the 
 * drivelink relative requests to the child relative requests and then routes 
 * it to the child for further processing.
**/
static void
drivelink_write(struct evms_logical_node *node, struct buffer_head *bh)
{
	struct evms_logical_node *child;
	u64 io_size, rsector;

	rsector = bh->b_rsector;
	child = which_child(node, &rsector, &io_size);
	if (child && ((bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) <= io_size)) {
		bh->b_rsector = rsector;
		W_IO(child, bh);
	} else {
		drivelink_io_error(node, WRITE, bh);
	}
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Init I/O function                               */
/********************************************************/

/**
 * drivelink_init_io: performs synchronous IO to drivelink objects
 * @node: 	drivelink object
 * @io_flag: 	read/write flag
 * @sect_nr: 	starting sector, object relative (512 byte units)
 * @num_sects: 	count of sectors
 * @buf_addr: 	buffer address to read from/write to
 * 
 * This function must determine which child or children a
 * specified I/O request must be passed to. Also if, when,
 * and how a request must be broken up. 
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
drivelink_init_io(struct evms_logical_node *node, int io_flag,	
		  u64 sect_nr,
		  u64 num_sects,	
		  void *buf_addr)
{				
	int rc = 0;

	if (!node)
		rc = -EINVAL;
	else {
		u64 starting_sector, remaining_sectors;
		void *io_buf;
		struct runtime_data *rd;

		if ((sect_nr + num_sects) > node->total_vsectors) {
			LOG_SERIOUS
			    ("attempted out of bound("PFU64") %s on '%s' at sector("PFU64"), count("PFU64").\n",
			     node->total_vsectors, (io_flag) ? "WRITE" : "READ",
			     node->name, sect_nr, num_sects);
			rc = -EINVAL;
		} else {
			rd = (struct runtime_data *) node->private;
			/* make working copies of input parameters */
			starting_sector = sect_nr;
			remaining_sectors = num_sects;
			io_buf = buf_addr;
			/* loop until all I/O is performed */
			while (remaining_sectors) {
				u64 io_start, io_size;
				struct evms_logical_node *child;

				/* compute the child relative io_start
				 * and max io_size.
				 */
				io_start = starting_sector;
				child = which_child(node, &io_start, &io_size);
				/* adjust io_size based on
				 * original remaining sectors
				 * in this io.
				 */
				if (io_size > remaining_sectors)
					io_size = remaining_sectors;
				if (child) {
					rc = INIT_IO(child,
						     io_flag,
						     io_start, io_size, io_buf);
				} else {
					/* if partial volume, return 0's
					 * for missing children.
					 */
					if (io_flag == READ) {
						memset(io_buf, 0,
						       io_size <<
						       EVMS_VSECTOR_SIZE_SHIFT);
					}
				}
				if (!rc) {
					/* adjust working copies */
					starting_sector += io_size;
					remaining_sectors -= io_size;
					io_buf += io_size <<
					    EVMS_VSECTOR_SIZE_SHIFT;
				} else
					break;
			}
		}
	}

	return (rc);
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      IOCTL function & Support routines               */
/********************************************************/

/**
 * drivelink_ioctl_cmd_plugin_ioctl: drivelink support for the 'plugin ioctl' command
 * @node: 	drivelink object
 * @inode: 	VFS supplied parameter
 * @file: 	VFS supplied parameter
 * @cmd: 	the specific ioctl command
 * @arg: 	the specific ioctl arguments
 * 
 * this function handles 'plugin ioctl' commands. currently there is no specific
 * commands for this plugin. however, this plugin must broadcast some commands so
 * lower layers can receive them.
 * 
 * Return value: 0 on success
 *		 otherwise error code
**/
static int
drivelink_ioctl_cmd_plugin_ioctl(struct evms_logical_node *node,
				 struct inode *inode, struct file *file,
				 unsigned long cmd, unsigned long arg)
{
	int i, rc = 0;
	struct runtime_data *rd;
	struct evms_plugin_ioctl_pkt tmp, *user_parms;

	user_parms = (struct evms_plugin_ioctl_pkt *) arg;
	/* copy user's parameters to kernel space */
	if (copy_from_user(&tmp, user_parms, sizeof (tmp)))
		rc = -EFAULT;

	if (!rc) {
		rd = (struct runtime_data *) node->private;
		/* is this cmd targetted at this feature ? */
		if (tmp.feature_id == node->plugin->id) {
			switch (tmp.feature_command) {
			default:
				break;
			}
		} else {	/* broadcast this cmd to all children */
			for (i = 0; i < rd->child_count; i++) {
				struct evms_logical_node *child_node;

				child_node = rd->child_table[i].child_node;
				if (child_node) {
					rc = IOCTL(child_node, inode, file,
						   cmd, arg);
					if (rc)
						break;
				}
			}
		}
		/* copy info to userspace */
		if (copy_to_user(user_parms, &tmp, sizeof (tmp)))
			rc = -EFAULT;
	}
	return (rc);
}

/**
 * drivelink_ioctl_cmd_broadcast: broadcast ioctls to your children
 * @node: 	drivelink object
 * @inode: 	VFS supplied parameter
 * @file: 	VFS supplied parameter
 * @cmd: 	the specific ioctl command
 * @arg: 	the specific ioctl arguments
 * 
 * broadcast the specified ioctl command and arguments to all this objects
 * children. OR (logical opeation) the return values from all the children
 * and return the OR'd value to the caller.
 * 
 * Return value: 0 on success
 * 		 otherwise error code
**/
static int
drivelink_ioctl_cmd_broadcast(struct evms_logical_node *node,
			      struct inode *inode, struct file *file,
			      unsigned long cmd, unsigned long arg)
{
	int i, rc = 0;
	struct runtime_data *rd;

	rd = (struct runtime_data *) node->private;
	/* broadcast this cmd to all children */
	for (i = 0; i < rd->child_count; i++) {
		struct evms_logical_node *child_node;

		child_node = rd->child_table[i].child_node;
		if (child_node) {
			rc |= IOCTL(child_node, inode, file, cmd, arg);
		}
	}
	return (rc);
}

/**
 * drivelink_ioctl: main ioctl entry point and handler
 * @node:	drivelink object
 * @inode:	VFS supplied parameter
 * @file:	VFS supplied parameter
 * @cmd:	a specific ioctl command
 * @arg:	a specific ioctl argument
 *
 * handles specific ioctl command internally and routes other ioctls commands to
 * the appropriate entry points.
 *
 * Returns: 0 on success
 *	    otherwise error code
 **/
static int
drivelink_ioctl(struct evms_logical_node *node,
		struct inode *inode,
		struct file *file, unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	struct runtime_data *rd = NULL;
	struct hd_geometry hdgeo;

	if ((!node) || (!inode))
		rc = -EINVAL;

	if (!rc) {
		rd = (struct runtime_data *) node->private;
		switch (cmd) {
		case HDIO_GETGEO:
			hdgeo.heads = 255;
			hdgeo.sectors = 63;
			hdgeo.cylinders =
			    ((unsigned int) node->total_vsectors) /
			    hdgeo.heads / hdgeo.sectors;
			hdgeo.start = 0;
			if (copy_to_user((int *) arg, &hdgeo, sizeof (hdgeo)))
				rc = -EFAULT;
			break;
		case EVMS_QUIESCE_VOLUME:
		case EVMS_GET_DISK_LIST:
		case EVMS_CHECK_MEDIA_CHANGE:
		case EVMS_REVALIDATE_DISK:
		case EVMS_OPEN_VOLUME:
		case EVMS_CLOSE_VOLUME:
		case EVMS_CHECK_DEVICE_STATUS:
			rc = drivelink_ioctl_cmd_broadcast(node, inode, file,
							   cmd, arg);
			break;
		case EVMS_PLUGIN_IOCTL:
			rc = drivelink_ioctl_cmd_plugin_ioctl(node, inode, file,
							      cmd, arg);
			break;
		case EVMS_GET_BMAP:
			{
				struct evms_get_bmap_pkt *bmap;
				u64 io_start, io_size;
				struct evms_logical_node *child;

				bmap = (struct evms_get_bmap_pkt *) arg;
				io_start = bmap->rsector;
				child = which_child(node, &io_start, &io_size);
				if (child) {
					if (node->block_size !=
					    child->block_size) {
						bmap->status = -EPERM;
					} else {
						bmap->rsector = io_start;
						rc = IOCTL(child,
							   inode,
							   file, cmd, arg);
					}
				}
			}
			break;
		default:
			rc = -EINVAL;
			break;
		}
	}
	return (rc);
}

/********************************************************/
/* Required Module Entry Point:                         */
/*      drivelink_init                                  */
/********************************************************/

/**
 * drivelink_init: register this module for use within the EVMS framework
 * 
 * Return value: 0 on success
 *	         otherwise error code.
**/
int __init
drivelink_init(void)
{
	return evms_cs_register_plugin(&plugin_header);
}

/**
 * drivelink_exit: unregister this module from use within the EVMS framework
 * 
 * Return value: 0 on success
 *	         otherwise error code.
**/
void __exit
drivelink_exit(void)
{
	evms_cs_unregister_plugin(&plugin_header);
}

module_init(drivelink_init);
module_exit(drivelink_exit);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
