/* -*- linux-c -*- */
/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* linux/driver/evms/ldev_mgr.c
 *
 * EVMS - Local Device (Hard Drive) Manager
 *
 *  This plugin walks the gendisk list and creates logical disk structures for each
 *  local ide or scsi device.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/blk.h>		/* must be included by all block drivers */
#include <linux/genhd.h>
#include <linux/ide.h>
#include <linux/version.h>
#include "../scsi/scsi.h"
#include "../scsi/sd.h"
#include <linux/init.h>
#include <linux/evms/evms.h>
#include <linux/evms/ldev_mgr.h>

#define LOG_PREFIX "ldev_mgr: "

#define EVMS_LOCAL_DEVICE_MANAGER_ID    1

/**
 * struct ldev_private - private data used by this plugin
 * @major: major device number
 * @minor: minor device number
 * @bdev: block_device record for this device
 * @gd: gendisk entry for this device
 * @media_changed: media changed status field
 *
 * private data maintained for each device by this plugin
 **/
struct ldev_private {
	int major, minor;
	struct block_device *bdev;
	struct gendisk *gd;
	int media_changed;
};

/* prototypes for mandatory plugin interface functions */
static int discover_disks(struct evms_logical_node **);
static int ldev_mgr_delete(struct evms_logical_node *);
static void ldev_mgr_read(struct evms_logical_node *, struct buffer_head *);
static void ldev_mgr_write(struct evms_logical_node *, struct buffer_head *);
static int ldev_mgr_ioctl(struct evms_logical_node *,
			  struct inode *,
			  struct file *, unsigned int, unsigned long);
static int ldev_init_io(struct evms_logical_node *,
			int, u64, u64, void *);
static int ldev_mgr_direct_ioctl(struct inode *,
				 struct file *, unsigned int, unsigned long);

/* plugin function table definition */
static struct evms_plugin_fops fops = {
	.discover	= discover_disks,
	.delete		= ldev_mgr_delete,
	.read		= ldev_mgr_read,
	.write		= ldev_mgr_write,
	.init_io	= ldev_init_io,
	.ioctl		= ldev_mgr_ioctl,
	.direct_ioctl	= ldev_mgr_direct_ioctl
};

/* plugin header definition */
static struct evms_plugin_header plugin_header = {
	.id = SetPluginID(IBM_OEM_ID,
			  EVMS_DEVICE_MANAGER,
			  EVMS_LOCAL_DEVICE_MANAGER_ID),
	.version = {
		.major		= 1,
		.minor		= 1,
		.patchlevel	= 2
	},
	.required_services_version = {
		.major		= 0,
		.minor		= 5,
		.patchlevel	= 0
	},
	.fops = &fops
};

#define TYPE_NONE	0
#define TYPE_GENERIC	1
#define TYPE_IDE	2
#define TYPE_SCSI	3

#define INDEX_ALPHA	0
#define INDEX_NUMERIC	1

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Discover function & Support routines            */
/********************************************************/

#define MAX_NAME_BASE_SIZE	10
#define MAX_NAME_MODIFIER_SIZE	4
/**
 * struct blk_device_info - block device info
 * @devnode_name_base: base name (ie. hd or sd) for device
 * @null1: guaranteed end-of-string NULL
 * @devnode_name_modifier: name suffix (ie. ag for sdag) for device
 * @null2: guaranteed end-of-string NULL
 * @devnode_name_index: numeric device index (ie. 1 for hda1)
 * @devnode_name_type: indicates numeric or alpha modifier
 * @devnode_type: device type, IDE, SCSI, or GENERIC
 *
 * generic block device naming descriptor structure
 **/
struct blk_device_info {
	char devnode_name_base[MAX_NAME_BASE_SIZE];
	char null1;
	char devnode_name_modifier[MAX_NAME_MODIFIER_SIZE];
	char null2;
	int devnode_name_index;
	int devnode_name_type;
	int device_type;
};

static struct blk_device_info *blk_dev_info = NULL;

#define BLK_DEV_INFO(a,b,c,d,e) 				\
	strncpy(blk_dev_info[a].devnode_name_base, b, MAX_NAME_BASE_SIZE); 	\
	blk_dev_info[a].null1 = 0; 				\
	strncpy(blk_dev_info[a].devnode_name_modifier, c, MAX_NAME_MODIFIER_SIZE); 	\
	blk_dev_info[a].null2 = 0; 				\
	blk_dev_info[a].devnode_name_index = 0;  		\
	blk_dev_info[a].device_type = d;     			\
	blk_dev_info[a].devnode_name_type = e;

static void
init_blk_dev_info(struct blk_device_info *blk_dev_info)
{
	BLK_DEV_INFO(IDE0_MAJOR, "hd", "a", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE1_MAJOR, "hd", "c", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE2_MAJOR, "hd", "e", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE3_MAJOR, "hd", "g", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE4_MAJOR, "hd", "i", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE5_MAJOR, "hd", "k", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE6_MAJOR, "hd", "m", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE7_MAJOR, "hd", "o", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE8_MAJOR, "hd", "q", TYPE_IDE, INDEX_ALPHA);
	BLK_DEV_INFO(IDE9_MAJOR, "hd", "s", TYPE_IDE, INDEX_ALPHA);

	BLK_DEV_INFO(SCSI_DISK0_MAJOR, "sd", "a", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK1_MAJOR, "sd", "q", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK2_MAJOR, "sd", "ag", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK3_MAJOR, "sd", "aw", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK4_MAJOR, "sd", "bm", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK5_MAJOR, "sd", "cc", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK6_MAJOR, "sd", "cs", TYPE_SCSI, INDEX_ALPHA);
	BLK_DEV_INFO(SCSI_DISK7_MAJOR, "sd", "di", TYPE_SCSI, INDEX_ALPHA);

	BLK_DEV_INFO(XT_DISK_MAJOR, "xd", "a", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(CYCLADES_MAJOR, "double", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);

	BLK_DEV_INFO(MFM_ACORN_MAJOR, "mfm", "a", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(ACSI_MAJOR, "ad", "a", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(PS2ESDI_MAJOR, "ed", "a", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(40, "ez", "a", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(43, "nb", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(44, "ftl", "a", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(45, "pd", "a", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(47, "pf", "0", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(DAC960_MAJOR + 0, "rd/c0d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 1, "rd/c1d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 2, "rd/c2d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 3, "rd/c3d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 4, "rd/c4d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 5, "rd/c5d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 6, "rd/c6d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(DAC960_MAJOR + 7, "rd/c7d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);

	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR, "ida/c0d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR1, "ida/c1d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR2, "ida/c2d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR3, "ida/c3d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR4, "ida/c4d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR5, "ida/c5d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR6, "ida/c6d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);
	BLK_DEV_INFO(COMPAQ_SMART2_MAJOR7, "ida/c7d", "0", TYPE_GENERIC,
		     INDEX_NUMERIC);

	BLK_DEV_INFO(I2O_MAJOR + 0, "i2o/hd", "a", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 1, "i2o/hd", "q", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 2, "i2o/hd", "ag", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 3, "i2o/hd", "aw", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 4, "i2o/hd", "bm", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 5, "i2o/hd", "cc", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 6, "i2o/hd", "cs", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(I2O_MAJOR + 7, "i2o/hd", "di", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(92, "ppdd", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(93, "nftl", "a", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(DASD_MAJOR, "dasd", "a", TYPE_GENERIC, INDEX_ALPHA);
	BLK_DEV_INFO(MDISK_MAJOR, "mdisk", "a", TYPE_GENERIC, INDEX_ALPHA);

	BLK_DEV_INFO(96, "msd", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(97, "pktcdvd", "0", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(UBD_MAJOR, "ubd", "0", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(JSFD_MAJOR, "jsfd", "", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(101, "amiraid/ar", "0", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(104, "cciss/c0d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(105, "cciss/c1d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(106, "cciss/c2d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(107, "cciss/c3d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(108, "cciss/c4d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(108, "cciss/c5d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(110, "cciss/c6d", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(111, "cciss/c7d", "0", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(RAW_MAJOR, "raw", "0", TYPE_GENERIC, INDEX_NUMERIC);

	BLK_DEV_INFO(VXVM_MAJOR, "vx/dsk", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(VXDMP_MAJOR, "vx/dmp", "0", TYPE_GENERIC, INDEX_NUMERIC);
	BLK_DEV_INFO(LOOP_MAJOR, "loop", "0", TYPE_GENERIC, INDEX_NUMERIC);
}

static int
is_in_device_list(struct gendisk *gd, int major, int minor)
{
	int found, done, rc;
	struct evms_logical_node *device = NULL;
	struct ldev_private *ldev_prv;

	done = found = FALSE;
	while (done == FALSE) {
		rc = evms_cs_find_next_device(device, &device);
		if (rc || !device)
			done = TRUE;
		else {
			ldev_prv = device->private;
			if (ldev_prv->gd == gd)
				if (ldev_prv->major == major)
					if (ldev_prv->minor == minor)
						done = found = TRUE;
		}
	}
	return (found);
}

static void
build_devnode_name(char *name_buf, int major)
{
	char buf[11], *modifier, *buf_ptr;
	int int_mod, done;
	struct blk_device_info *bdi;

	bdi = &blk_dev_info[major];

	/* convert the base name modifier to an integer */
	modifier = bdi->devnode_name_modifier;
	int_mod = 0;
	while (*modifier) {
		if (bdi->devnode_name_type == INDEX_ALPHA) {
			int_mod *= 26;
			int_mod += *modifier - 'a';
		} else {
			int_mod *= 10;
			int_mod += *modifier - '0';
		}
		modifier++;
		if (*modifier) {
			int_mod++;
		}
	}
	/* add in device_index_value */
	int_mod += bdi->devnode_name_index;
	bdi->devnode_name_index++;

	/* convert integer modifier back to ALPHA/NUMERIC chars */
	memset(buf, 0, sizeof (buf));
	/* fill the buffer from the rear to front with the
	 * ascii version of the modifier, leaving space for
	 * NULL terminator at the end.
	 */
	buf_ptr = &buf[sizeof (buf) - 2];
	done = FALSE;
	do {
		if (bdi->devnode_name_type == INDEX_ALPHA) {
			*buf_ptr = (int_mod % 26) + 'a';
			int_mod /= 26;
		} else {
			*buf_ptr = (int_mod % 10) + '0';
			int_mod /= 10;
		}
		if (int_mod) {
			int_mod--;
		} else {
			done = TRUE;
		}
		buf_ptr--;
	} while (!done);

	/* find beginning of modifier in buffer */
	modifier = buf;
	while (!*modifier)
		modifier++;

	/* build the final device devnode name */
	sprintf(name_buf, "%s%s", bdi->devnode_name_base, modifier);
}

static int
ldev_mgr_lock_device(struct ldev_private *ldev_prv)
{
	int rc;
	struct block_device *bdev;

	bdev = bdget(MKDEV(ldev_prv->major, ldev_prv->minor));
	if (!bdev)
		return -ENOMEM;
	rc = blkdev_get(bdev, FMODE_READ | FMODE_WRITE, 0, BDEV_RAW);
	if (rc)
		return rc;
	ldev_prv->bdev = bdev;
	return 0;
}

static void
ldev_mgr_unlock_device(struct ldev_private *ldev_prv)
{
	struct block_device *bdev = ldev_prv->bdev;
	ldev_prv->bdev = NULL;
	if (!bdev) {
		LOG_ERROR("error: NULL bdev field detected!\n");
		BUG();
	}
	blkdev_put(bdev, BDEV_RAW);
}

#define DEVICE_KNOWN			1234
#define DEVICE_UNINITIALIZED		1235
#define DEVICE_MEDIA_NOT_PRESENT 	1236
static int
create_logical_disk(struct evms_logical_node **disk_list,
		    struct gendisk *gd, int device_index)
{
	int rc = 0, major, minor;
	struct evms_logical_node *new_disk = NULL;
	struct ldev_private *ldev_prv = NULL;
	char device_name[EVMS_VOLUME_NAME_SIZE + 1];

	major = gd->major;
	minor = device_index << gd->minor_shift;

	/* skip uninitialized devices */
	if (!blk_size[major])
		rc = DEVICE_UNINITIALIZED;
	else if (!blk_size[major][minor])
		rc = DEVICE_UNINITIALIZED;
	if (!rc) {
		/* construct the devnode name for this device */
		build_devnode_name(device_name, major);

		/* skip devices we already know about */
		if (is_in_device_list(gd, major, minor) == TRUE)
			rc = DEVICE_KNOWN;
	}
	/* allocate the new node */
	if (!rc) {
		rc = evms_cs_allocate_logical_node(&new_disk);
	}
	/* allocate new nodes's instance data */
	if (!rc) {
		ldev_prv = kmalloc(sizeof(struct ldev_private), GFP_KERNEL);
		if (!ldev_prv)
			rc = -ENOMEM;
	}
	/* initialize the new node */
	if (!rc) {
		memset(ldev_prv, 0, sizeof(struct ldev_private));
		new_disk->plugin = &plugin_header;

		/* initialize the instance data */
		new_disk->private = ldev_prv;
		ldev_prv->gd = gd;
		ldev_prv->major = major;
		ldev_prv->minor = minor;
		rc = ldev_mgr_lock_device(ldev_prv);
		if (rc) {
			LOG_ERROR("error(%d): unable to lock device(%d,%d)!\n",
				  rc, major, minor);
		}
	}
	if (!rc) {
		/* determine hardsector size */
		new_disk->hardsector_size = 512;
		if (hardsect_size[major]) {
			new_disk->hardsector_size = hardsect_size[major][minor];
		}
		/* save the block size */
		new_disk->block_size = 1024;
		if (blksize_size[major]) {
			new_disk->block_size = blksize_size[major][minor];
		}
		/* obtain the device size in sectors
		 *
		 * try 64bit size first, if that fails
		 * fall back on the 32bit size.
		 */
		/* try 64bit size */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
		rc = evms_cs_kernel_ioctl(new_disk, BLKGETSIZE64,
					  (ulong) & new_disk->total_vsectors);
		if (!rc) {
			/* convert bytes to 512 byte sectors */
			new_disk->total_vsectors >>= EVMS_VSECTOR_SIZE_SHIFT;
		} else
#endif
		{
			/* try 32bit size */
			ulong dev_size = 0;
			rc = evms_cs_kernel_ioctl(new_disk, BLKGETSIZE,
						  (ulong) & dev_size);
			new_disk->total_vsectors = dev_size;
		}
		if (!rc && !new_disk->total_vsectors) {
			rc = -ENOSPC;
		}
	}
	if (!rc) {
		/* remember removable devices */
		if (gd->flags)
			if (gd->flags[device_index] & GENHD_FL_REMOVABLE)
				new_disk->flags |= EVMS_DEVICE_REMOVABLE;

		/* save the devnode name for this device */
		strcpy(new_disk->name, device_name);

		/* register this device with evms */
		evms_cs_register_device(new_disk);
		MOD_INC_USE_COUNT;

		/* append this record the linked list */
		evms_cs_add_logical_node_to_list(disk_list, new_disk);
		LOG_DETAILS
		    ("added logical disk(%s) for physical disk(%u,%u,%s), size("PFU64") in 512 byte units\n",
		     new_disk->name, major, minor, new_disk->name,
		     new_disk->total_vsectors);

	}
	/* reset the "benign" error codes for the caller */
	switch (rc) {
	case DEVICE_UNINITIALIZED:
	case DEVICE_KNOWN:
	case DEVICE_MEDIA_NOT_PRESENT:
		rc = 0;
	case 0:
		break;
	default:
		LOG_ERROR
		    ("error(%d): creating logical disk for device(%d,%d).\n",
		     rc, major, minor);
		if (new_disk) {
			evms_cs_deallocate_logical_node(new_disk);
		}
		if (ldev_prv) {
			kfree(ldev_prv);
		}
		break;
	}
	return (rc);
}

static int
create_logical_generic_disks(struct evms_logical_node **disk_list,
			     struct gendisk *gd)
{
	int rc, i;

	/* This is a generic device */

	rc = 0;
	LOG_EXTRA("major name = %s\n", gd->major_name);
	LOG_EXTRA("number of real devices = %i\n", gd->nr_real);
	for (i = 0; i < gd->nr_real; i++) {
		LOG_EXTRA("device %d:\n", i);
		rc = create_logical_disk(disk_list, gd, i);
		if (rc)
			break;
	}
	return (rc);
}

static int
create_logical_ide_disks(struct evms_logical_node **disk_list,
			 struct gendisk *gd)
{
	int rc = 0, i;
	ide_hwif_t *ide_hwif;
	ide_drive_t *drive;

	/* This is an IDE device */
	LOG_EXTRA("found IDE major : %i - searching for disks\n", gd->major);

	ide_hwif = gd->real_devices;	/* IDE internal data */
	for (i = 0; i < MAX_DRIVES; i++) {
		drive = &(ide_hwif->drives[i]);
		if (drive->present && (drive->media == ide_disk)) {
			/* force the name index value on ide drives */
			blk_dev_info[gd->major].devnode_name_index = i;
			rc = create_logical_disk(disk_list, gd, i);
		}
		if (rc)
			break;
	}
	return (rc);
}

static int
create_logical_scsi_disks(struct evms_logical_node **disk_list,
			  struct gendisk *gd)
{
	int rc = 0, i;
	Scsi_Disk *SDisks;
	Scsi_Device *SDev;

	/* This is an SCSI device */
	LOG_EXTRA("found SCSI major : %i - searching for disks\n", gd->major);
	LOG_EXTRA("scsi: major name = %s\n", gd->major_name);
	LOG_EXTRA("scsi: number of real devices = %i\n", gd->nr_real);
	SDisks = gd->real_devices;	/* SCSI internal data */
	for (i = 0; i < gd->nr_real; i++) {
		SDev = SDisks[i].device;
		LOG_EXTRA
		    ("scsi: Channel = %i, Id = %i, Lun = %i, Capacity = %i\n",
		     SDev->channel, SDev->id, SDev->lun, SDisks[i].capacity);
		rc = create_logical_disk(disk_list, gd, i);
		if (rc)
			break;
	}
	return (rc);
}

static int
create_logical_disks(struct gendisk *gd, void *p_disk_list)
{
	int rc = 0;
	struct evms_logical_node **disk_list = p_disk_list;

	/* create logical disks from all IDE & SCSI devices */
	switch (blk_dev_info[gd->major].device_type) {
	case TYPE_IDE:
		rc = create_logical_ide_disks(disk_list, gd);
		break;
	case TYPE_SCSI:
		rc = create_logical_scsi_disks(disk_list, gd);
		break;
	case TYPE_GENERIC:
		rc = create_logical_generic_disks(disk_list, gd);
		break;
	default:
		LOG_EXTRA("unrecognized device major : %i\n", gd->major);
		break;
	}

	return (rc);
}

static int
discover_disks(struct evms_logical_node **disk_list)
{
	int rc = 0;

	MOD_INC_USE_COUNT;
	LOG_ENTRY_EXIT("%s Entry\n", __FUNCTION__);

	if (blk_dev_info == NULL) {
		/* allocate space for device info array */
		blk_dev_info = kmalloc(sizeof (struct blk_device_info)
				       * (MAX_BLKDEV + 1), GFP_KERNEL);
		if (blk_dev_info) {
			/* initialize device info array */
			memset(blk_dev_info, 0,
			       sizeof (struct blk_device_info) * (MAX_BLKDEV + 1));
			init_blk_dev_info(blk_dev_info);
		} else {
			rc = -ENOMEM;
		}
	}
	if (!rc)
		/* create logical disks from the raw devices */
		rc = walk_gendisk(create_logical_disks, disk_list);

	/* free blk_dev_info table and null the ptr to it */
	kfree(blk_dev_info);
	blk_dev_info = NULL;

	LOG_ENTRY_EXIT("%s Exit\n", __FUNCTION__);
	MOD_DEC_USE_COUNT;
	return (rc);
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Delete function                                 */
/********************************************************/

static int
ldev_mgr_delete(struct evms_logical_node *disk)
{
	struct ldev_private *ldev_prv;

	/* reset any evms volume related info from
	 * the device node, because we can't predict
	 * how this node will be used in the future.
	 */

	/* removed the feature header if its been used
	 */
	if (disk->feature_header) {
		kfree(disk->feature_header);
		disk->feature_header = NULL;
	}
	/* remove the volume_info structure and flag
	 * if this has been used directly by an evms
	 * feature.
	 */
	evms_cs_deallocate_volume_info(disk);
	/* reset the flags field to the appropriate state
	 */
	disk->flags &= ~EVMS_VOLUME_FLAG;

	/* disk nodes only get deleted when:
	 * 1)  there are no references to the disk node
	 *      in memory.
	 * 2)  the device is removable
	 * 3)  the device reported a media change
	 *
	 * All three of these conditions must be true
	 * before the disk node can be deleted. 
	 * evms_check_for_device_changes should set
	 * and ensure these conditions before issuing
	 * deletes.
	 *
	 * Newly installed removable media will be
	 * picked up in this modules discover code.
	 *
	 * OR disk nodes can will be deleted if the
	 * devices they represent go away, for example
	 * in the case of a hotunplugged device or a
	 * required driver having been unloaded.
	 */
	if (disk->flags & (EVMS_MEDIA_CHANGED | EVMS_DEVICE_UNAVAILABLE)) {
		LOG_DETAILS("deleting '%s'.\n", disk->name);

		evms_cs_unregister_device(disk);
		MOD_DEC_USE_COUNT;
		ldev_prv = disk->private;
		ldev_mgr_unlock_device(ldev_prv);
		if (ldev_prv) {
			kfree(ldev_prv);
		}
		evms_cs_deallocate_logical_node(disk);
	}
	return 0;
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Read function                                   */
/********************************************************/

/* 
 * function: ldev_mgr_io_error
 * 
 * this function was primarily created because the function
 * buffer_IO_error is inline and kgdb doesn't allow breakpoints
 * to be set on inline functions. Since this was an error path
 * and not mainline, I decided to add a trace statement to help
 * report on the failing condition.
 *
 */
static void
ldev_mgr_io_error(struct evms_logical_node *disk, int io_flag, struct buffer_head *bh, int rc)
{
	if (rc == -EOVERFLOW) {
		LOG_SERIOUS
		    ("attempt to %s beyond boundary("PFU64") on (%s), rsector(%ld).\n",
		     (io_flag) ? "WRITE" : "READ", disk->total_vsectors - 1,
		     disk->name, bh->b_rsector);
	} else if (rc == -ENXIO) {
		LOG_SERIOUS("attempt to access a non-existent device(%s).\n",
			    disk->name);
	}
	bh->b_end_io(bh, 0);
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Read function                                   */
/********************************************************/

static void
ldev_mgr_read(struct evms_logical_node *disk, struct buffer_head *bh)
{
	int rc = 0;
	request_queue_t *q;
	struct ldev_private *ldev_prv;

	ldev_prv = disk->private;
	if (bh->b_rsector + (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) <=
	    disk->total_vsectors) {
		bh->b_rdev = MKDEV(ldev_prv->major, ldev_prv->minor);
		q = blk_get_queue(bh->b_rdev);
		if (q) {
			disk->flags &= ~EVMS_DEVICE_UNAVAILABLE;
			q->make_request_fn(q, READ, bh);
			return;
		} else {
			rc = -ENXIO;
			disk->flags |= EVMS_DEVICE_UNAVAILABLE;
		}
	} else {
		rc = -EOVERFLOW;
	}
	if (rc) {
		ldev_mgr_io_error(disk, READ, bh, rc);
	}
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Write function                                  */
/********************************************************/

static void
ldev_mgr_write(struct evms_logical_node *disk, struct buffer_head *bh)
{
	int rc = 0;
	request_queue_t *q;
	struct ldev_private *ldev_prv;

	ldev_prv = disk->private;
	if (bh->b_rsector + (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) <=
	    disk->total_vsectors) {
		bh->b_rdev = MKDEV(ldev_prv->major, ldev_prv->minor);
		q = blk_get_queue(bh->b_rdev);
		if (q) {
			disk->flags &= ~EVMS_DEVICE_UNAVAILABLE;
			q->make_request_fn(q, WRITE, bh);
			return;
		} else {
			rc = -ENXIO;
			disk->flags |= EVMS_DEVICE_UNAVAILABLE;
		}
	} else {
		rc = -EOVERFLOW;
	}
	if (rc) {
		ldev_mgr_io_error(disk, WRITE, bh, rc);
	}
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      Init_io function & Support routines             */
/********************************************************/

/*
 * function: allocate_bh
 *
 * This function obtains a buffer head from the private 
 * buffer head pool (pre-allocated at EVMS initial 
 * discovery time). 
 *
 * NOTE: All access to the buffer head pool are protected 
 * by a private spinlock.
 *
 */
static inline struct buffer_head *
allocate_bh(void)
{
	struct buffer_head *bh =
	    evms_cs_allocate_from_pool(evms_bh_pool, FALSE);
	if (bh) {
		init_waitqueue_head(&bh->b_wait);
	}
	return (bh);
}

/*
 * function: deallocate_bh
 *
 * This function returns a buffer head to the private 
 * buffer head pool (pre-allocated at EVMS initial 
 * discovery time). 
 *
 * NOTE: All access to the buffer head pool are protected 
 * by a private spinlock.
 *
 */
static inline void
deallocate_bh(struct buffer_head *bh)
{
	evms_cs_deallocate_to_pool(evms_bh_pool, bh);
}

/* this is the buffer head control block structure definition */
typedef struct bh_cb_s {
	int rc;
	atomic_t blks_allocated;
	wait_queue_head_t cb_wait;
} bh_cb_t;

/*
 * function: __wait_on_bh_cb
 *
 * This is a worker function to wait_on_bh_cb.
 * This function waits for a set of private buffer heads
 * associated to the specified buffer head control block
 * to return from I/O completion. On completion of the
 * last buffer head, the calling function is awakened
 * and continues running.
 *
 * This is the worker function to the function wait_on_bh_cb.
 *
 */
static void
__wait_on_bh_cb(bh_cb_t * bh_cb)
{
	struct task_struct *tsk = current;
	DECLARE_WAITQUEUE(wait, tsk);

	add_wait_queue(&bh_cb->cb_wait, &wait);
	do {
		run_task_queue(&tq_disk);
		set_task_state(tsk, TASK_UNINTERRUPTIBLE);
		if (!atomic_read(&bh_cb->blks_allocated))
			break;
		schedule();
	} while (atomic_read(&bh_cb->blks_allocated));
#ifdef O1_SCHEDULER
	set_task_state(tsk, TASK_RUNNING);
#else
	tsk->state = TASK_RUNNING;
#endif
	remove_wait_queue(&bh_cb->cb_wait, &wait);
}

/*
 * function: wait_on_bh_cb
 *
 * This function waits for a set of private buffer heads
 * associated to the specified buffer head control block
 * to return from I/O completion. On completion of the
 * last buffer head, the calling function is awakened
 * and continues running.
 *
 */
static void
wait_on_bh_cb(bh_cb_t * bh_cb)
{
	if (atomic_read(&bh_cb->blks_allocated))
		__wait_on_bh_cb(bh_cb);
	else
		/* if we ended up with no buffer heads on
		 * this pass, lets wait a until a few buffer
		 * heads have been freed and try again. This
		 * should provide a reasonable delay.
		 */
		schedule();
}

/*
 * function: end_bh_cb_io
 *
 * This is the I/O completion function that is called for
 * each private buffer head obtained from the buffer head 
 * pool. Control is return thru this routine so we can track
 * all outstanding requests to know when to awaken the caller,
 * and to regain control after all I/Os have been performed.
 *
 */
static void
end_bh_cb_io_sync(struct buffer_head *bh, int uptodate)
{
	bh_cb_t *bh_cb = (bh_cb_t *) bh->b_private;

	/* record that errors occurred */
	if (!uptodate) {
		bh_cb->rc = -EIO;
	}
	mark_buffer_uptodate(bh, uptodate);
	unlock_buffer(bh);

	deallocate_bh(bh);
	atomic_dec(&bh_cb->blks_allocated);
	if (!atomic_read(&bh_cb->blks_allocated))
		if (waitqueue_active(&bh_cb->cb_wait))
			wake_up(&bh_cb->cb_wait);
}

/*
 * function: ldev_partial_sector_init_io
 *
 * This function is a support function for ldev_init_io,
 * which handles the cases of performing I/O to only a part
 * of non-standard sized hardsector. This function is not 
 * designed to be called directly, but via ldev_init_io.
 *
 */
static int
ldev_partial_sector_init_io(struct evms_logical_node *node,
			    int io_flag,
			    bh_cb_t * bh_cb,
			    u64 next_lsn,
			    u64 sector_lsn,
			    u64 io_size,
			    void *bufptr, unsigned char **sector_buf)
{
	int rc = 0;
	struct ldev_private *ldev_prv = node->private;
	kdev_t dev = MKDEV(ldev_prv->major, ldev_prv->minor);
	struct buffer_head *bh;

	if (*sector_buf == NULL) {
		/* allocate buffer for incoming sector */
		*sector_buf = kmalloc(node->hardsector_size, GFP_KERNEL);
		if (!*sector_buf)
			return -ENOMEM;
	}
	/* allocate a buffer head from the pool */
	while ((bh = allocate_bh()) == NULL)
		/* yielding the cpu is playing it
		 * safe. it might be wiser to just
		 * spin. requires more thought.
		 */
		schedule();

	/* set up the buffer head for this sector */
	bh->b_end_io = end_bh_cb_io_sync;
	bh->b_size = node->hardsector_size;
	bh->b_rdev = dev;
	bh->b_rsector = next_lsn - sector_lsn;
	bh->b_data = *sector_buf;
	bh->b_page = virt_to_page(*sector_buf);	/* this isn't handling the case of a block with more than 1 sector, that spans pages */
	bh->b_state = 0;
	set_bit(BH_Dirty, &bh->b_state);
	set_bit(BH_Lock, &bh->b_state);
	set_bit(BH_Req, &bh->b_state);
	set_bit(BH_Mapped, &bh->b_state);
	bh->b_private = (void *) bh_cb;
	atomic_inc(&bh_cb->blks_allocated);

	/* drive the buffer head down   */
	/* to the device                */
	generic_make_request(READ, bh);

	/* wait for all bh's I/O's to end */
	wait_on_bh_cb(bh_cb);

	/* copy data to/from user */
	if (io_flag != WRITE)
		/* READ */
		memcpy(bufptr,
		       *sector_buf + (sector_lsn << EVMS_VSECTOR_SIZE_SHIFT),
		       io_size << EVMS_VSECTOR_SIZE_SHIFT);
	else {
		/* WRITE */
		memcpy(*sector_buf + (sector_lsn << EVMS_VSECTOR_SIZE_SHIFT),
		       bufptr, io_size << EVMS_VSECTOR_SIZE_SHIFT);

		/* allocate a buffer head from the pool */
		while ((bh = allocate_bh()) == NULL)
			/* yielding the cpu is playing it
			 * safe. it might be wiser to just
			 * spin. requires more thought.
			 */
			schedule();

		/* set up the buffer head for this sector */
		bh->b_end_io = end_bh_cb_io_sync;
		bh->b_size = node->hardsector_size;
		bh->b_rdev = dev;
		bh->b_rsector = next_lsn - sector_lsn;
		bh->b_data = *sector_buf;
		bh->b_page = virt_to_page(*sector_buf);	/* this isn't handling the case of a block with more than 1 sector, that spans pages */
		bh->b_state = 0;
		set_bit(BH_Dirty, &bh->b_state);
		set_bit(BH_Lock, &bh->b_state);
		set_bit(BH_Req, &bh->b_state);
		set_bit(BH_Mapped, &bh->b_state);
		bh->b_private = (void *) bh_cb;
		atomic_inc(&bh_cb->blks_allocated);

		/* drive the buffer head down   */
		/* to the device                */
		generic_make_request(WRITE, bh);

		/* wait for all bh's I/O's to end */
		wait_on_bh_cb(bh_cb);
	}
	return (rc);
}

/*
 * function: ldev_init_io
 *
 * This function provides support for synchronous I/O 
 * operations to the underlying devices. These I/O 
 * operations are NOT buffered in any way including the 
 * operating system's buffer cache.
 *
 * This function can work with any hardsector size that
 * is a power of 2.
 *
 * node           : logical node of the target logical disk
 * io_flag        : 0 = read, 1 = write, 2 = read-a-head
 * starting_lsn   : the 0-based (disk relative) logical
 *		  :  (512 byte) sector number (lsn) 
 * num_lsns       : the total number of lsns in this I/O
 * bufptr         : address of the memory to read/write the data
 *
 */
static int
ldev_init_io(struct evms_logical_node *node,
	     int io_flag,
	     u64 starting_lsn, u64 num_lsns, void *bufptr)
{
	int rc = 0, lsns_per_hardsector, lsns_per_blocksize;
	unchar *sector_buf = NULL, *cur_bufptr;
	u64 next_lsn, remaining_lsns, sector_lsn;
	struct ldev_private *ldev_prv = node->private;
	kdev_t dev = MKDEV(ldev_prv->major, ldev_prv->minor);
	bh_cb_t bh_cb;

	LOG_EVERYTHING
	    ("%s Entry: Disk(%u,%u), ioflag(%u), start_lsn("PFU64"), num_lsns("PFU64"), bufptr(0x%p)\n",
	     __FUNCTION__, ldev_prv->major, ldev_prv->minor, io_flag,
	     starting_lsn, num_lsns, bufptr);

	/* check for valid device */
	if (!blk_size[ldev_prv->major][ldev_prv->minor]) {
		node->flags |= EVMS_DEVICE_UNAVAILABLE;
		return (-ENXIO);
	}
	/* check for 0 length request */
	if (num_lsns == 0) {
		LOG_ERROR("%s: error requesting 0 sectors.\n", __FUNCTION__);
		return (-EINVAL);
	}
	/* check for out of bound request */
	if ((starting_lsn + num_lsns) > node->total_vsectors) {
		LOG_ERROR
		    ("%s: attempted %s beyond logical disk boundary("PFU64" LSNs), requesting LSN("PFU64"), total LSNs("PFU64").\n",
		     __FUNCTION__, (io_flag == WRITE) ? "WRITE" : "READ",
		     node->total_vsectors, starting_lsn, num_lsns);
		return (-EINVAL);
	}
	/* check for invalid io_flag value */
	switch (io_flag) {
	case READ:		/* read...   */
	case WRITE:		/* write...  */
	case READA:		/* reada...  */
		break;
	default:
		return (-EINVAL);
	}

	/* compute some per device info once up-front */
	lsns_per_hardsector = node->hardsector_size / EVMS_VSECTOR_SIZE;
	lsns_per_blocksize = node->block_size / EVMS_VSECTOR_SIZE;

	/* initialize the buffer head control block */
	memset(&bh_cb, 0, sizeof (bh_cb_t));
	init_waitqueue_head(&bh_cb.cb_wait);
	atomic_set(&bh_cb.blks_allocated, 0);

	/* only update the local copy of variables */
	cur_bufptr = bufptr;
	next_lsn = starting_lsn;
	remaining_lsns = num_lsns;

	/* check for a mid-sector starting offset
	 *
	 * if found, perform I/O on part of that
	 * sector
	 */
	sector_lsn = next_lsn & (lsns_per_hardsector - 1);
	if (sector_lsn) {
		u64 io_size;

		/* determine bytes in IO to this sector */
		io_size = lsns_per_hardsector - sector_lsn;
		if (io_size > remaining_lsns)
			io_size = remaining_lsns;

		/* perform the partial sector io */
		rc = ldev_partial_sector_init_io(node, io_flag, &bh_cb,
						 next_lsn,
						 sector_lsn, io_size,
						 cur_bufptr, &sector_buf);

		if (!rc) {
			/* update progress in local variables */
			cur_bufptr += io_size << EVMS_VSECTOR_SIZE_SHIFT;
			next_lsn += io_size;
			remaining_lsns -= io_size;
		}
	}

	/* continue if no errors found */
	if (!rc) {
		/* perform I/O on all the complete sectors
		 * in this request.
		 *
		 * loop until there are no more complete sectors
		 * to process.
		 */
		while (remaining_lsns >= lsns_per_hardsector) {
			/* this inner loop attempts to drive as many
			 * bytes (in sector size multiples) down to 
			 * the device as possible using the available
			 * buffer heads in the pool.
			 */
			while (remaining_lsns >= lsns_per_hardsector) {
				struct buffer_head *bh;

				/* allocate a buffer head from the pool */
				bh = allocate_bh();
				if (bh == NULL)
					break;

				/* set up the buffer head for this I/O */
				bh->b_end_io = end_bh_cb_io_sync;
				bh->b_size =
				    (remaining_lsns >= lsns_per_blocksize) ?
				    node->block_size : node->hardsector_size;
				bh->b_data = cur_bufptr;
				bh->b_rdev = dev;
				bh->b_rsector = next_lsn;
				bh->b_page = virt_to_page(cur_bufptr);	/* this isn't handling the case of a block with more than 1 sector, that spans pages */
				bh->b_state = 0;
				set_bit(BH_Dirty, &bh->b_state);
				set_bit(BH_Lock, &bh->b_state);
				set_bit(BH_Req, &bh->b_state);
				set_bit(BH_Mapped, &bh->b_state);
				bh->b_private = (void *) &bh_cb;
				atomic_inc(&bh_cb.blks_allocated);

				/* drive the buffer head down   */
				/* to the device                */
				generic_make_request(io_flag, bh);

				/* update progress in local variables */
				cur_bufptr += bh->b_size;
				next_lsn +=
				    bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT;
				remaining_lsns -=
				    bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT;
			}
			/* wait for all bh's I/O's to end */
			wait_on_bh_cb(&bh_cb);
		}
	}

	/* continue if no errors found */
	if (!rc)
		/* check for a mid-sector ending offset
		 *
		 * if found, perform I/O on part of that
		 * sector
		 */
		if (remaining_lsns)
			/* perform the partial sector io */
			rc = ldev_partial_sector_init_io(node, io_flag, &bh_cb,
							 next_lsn,
							 0, remaining_lsns,
							 cur_bufptr,
							 &sector_buf);

	/* free the sector buffer if it was allocated */
	if (sector_buf)
		kfree(sector_buf);

	/* coalesce return codes */
	rc |= bh_cb.rc;

	LOG_EVERYTHING("%s Exit: rc(%u)\n", __FUNCTION__, rc);

	return (rc);
}

static int
ldev_mgr_direct_ioctl(struct inode *inode,
		      struct file *file, unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	struct ldev_private *ldev_prv;
	struct evms_plugin_ioctl_pkt tmp, *user_parms;
	struct ldev_plugin_ioctl pi_data;
	struct evms_logical_node *disk;

	MOD_INC_USE_COUNT;

	user_parms = (struct evms_plugin_ioctl_pkt *) arg;
	/* copy user's parameters to kernel space */
	if (copy_from_user(&tmp, user_parms, sizeof (tmp)))
		rc = -EFAULT;

	if (!rc) {
		/* validate its meant for us */
		if (tmp.feature_id != plugin_header.id) {
			rc = -EINVAL;
		}
	}

	if (!rc) {
		/* copy feature ioctl data to kernel space */
		if (copy_from_user(&pi_data, tmp.feature_ioctl_data,
				   sizeof (pi_data))) {
			rc = -EFAULT;
		}
	}

	if (!rc) {
		/* find the disk node specified by the disk_handle */
		int done = FALSE;
		disk = NULL;
		while (!done) {
			rc = evms_cs_find_next_device(disk,
						      &disk);
			if (rc) {
				break;
			}
			if (!disk) {
				rc = -ENODATA;
				break;
			}
			if (disk ==
			    DEV_HANDLE_TO_NODE(pi_data.disk_handle)) {
				done = TRUE;
			}
		}
	}

	if (!rc) {
		/* perform feature command */
		ldev_prv = (struct ldev_private *) disk->private;
		switch (tmp.feature_command) {
			kdev_t save_dev;
		case LDEV_MGR_BROADCAST_IOCTL_CMD:
			save_dev = inode->i_rdev;
			inode->i_rdev =
			    MKDEV(ldev_prv->major, ldev_prv->minor);
			rc = ldev_prv->bdev->bd_op->ioctl(inode, file,
							  pi_data.cmd,
							  pi_data.arg);
			inode->i_rdev = save_dev;
			break;
		default:
			rc = -EINVAL;
			break;
		}
	}

	/* return status value */
	tmp.status = rc;
	copy_to_user((struct evms_plugin_ioctl_pkt *) arg, &tmp, sizeof (tmp));
	MOD_DEC_USE_COUNT;
	return rc;
}

/********************************************************/
/* Required Plugin Function Table Entry Point:          */
/*      IOCTL function & Support routines               */
/********************************************************/

static int
ldev_mgr_ioctl(struct evms_logical_node *disk,
	       struct inode *inode,
	       struct file *file, unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	struct ldev_private *ldev_prv = disk->private;
	kdev_t save_dev;
	struct block_device *save_bdev;

	if (!inode || !disk)
		return -EINVAL;

	save_dev = inode->i_rdev;
	inode->i_rdev = MKDEV(ldev_prv->major, ldev_prv->minor);
	save_bdev = inode->i_bdev;
	inode->i_bdev = ldev_prv->bdev;
	/* check device availability */
	if (!blk_get_queue(MKDEV(ldev_prv->major, ldev_prv->minor))) {
		disk->flags |= EVMS_DEVICE_UNAVAILABLE;
	}
	switch (cmd) {
	case EVMS_QUIESCE_VOLUME:
	case EVMS_PLUGIN_IOCTL:
		break;
	case EVMS_GET_BMAP:
		{
			struct evms_get_bmap_pkt *bmap =
			    (struct evms_get_bmap_pkt *) arg;
			bmap->dev = MKDEV(ldev_prv->major, ldev_prv->minor);
			bmap->status = 0;
		}
		break;
	case EVMS_OPEN_VOLUME:
		if (disk->flags & EVMS_DEVICE_UNAVAILABLE) {
			rc = -ENXIO;
		} else {
			rc = ldev_prv->bdev->bd_op->open(inode, file);
		}
		break;
	case EVMS_CLOSE_VOLUME:
		if (disk->flags & EVMS_DEVICE_UNAVAILABLE) {
			rc = -ENXIO;
		} else {
			rc = ldev_prv->bdev->bd_op->release(inode, file);
		}
		break;
	case EVMS_CHECK_MEDIA_CHANGE:
		if (disk->flags & EVMS_DEVICE_UNAVAILABLE) {
			rc = -ENXIO;
		} else {
			/* once we detect that media changed
			 * is 'set', don't send any more ioctls
			 * down to the device, until the
			 * media change has been 'reset' by a
			 * revalidate disk ioctl. when already
			 * 'set', just return a 1 w/o actually
			 * performing another ioctl call to the
			 * device.
			 */
			if (ldev_prv->media_changed == TRUE) {
				rc = 1;
				break;
			}
			rc = ldev_prv->bdev->bd_op->
			    check_media_change(MKDEV
					       (ldev_prv->major,
						ldev_prv->minor));
			if (rc == 1) {
				ldev_prv->media_changed = TRUE;
				disk->flags |= EVMS_MEDIA_CHANGED;
			}
		}
		break;
	case EVMS_REVALIDATE_DISK:
		if (disk->flags & EVMS_DEVICE_UNAVAILABLE) {
			rc = -ENXIO;
		} else {
			/* don't actually send this ioctl down
			 * to the device, until we know that
			 * previous check media change ioctl
			 * has occurred.
			 *
			 * when we do actually send the ioctl
			 * down, reset the local media_changed
			 * flag.
			 */
			if (ldev_prv->media_changed == FALSE)
				break;
			rc = ldev_prv->bdev->bd_op->
			    revalidate(MKDEV
				       (ldev_prv->major, ldev_prv->minor));
			ldev_prv->media_changed = FALSE;
		}
		break;
	case EVMS_GET_DISK_LIST:
		rc = evms_cs_add_item_to_list((struct evms_list_node **) arg,
					      disk);
		if (rc > 0)
			rc = 0;
		break;
	case EVMS_CHECK_DEVICE_STATUS:
		if (arg) {
			int *status = (int *) arg;
			*status |= disk->flags;
		}
		break;
	case EVMS_UPDATE_DEVICE_INFO:
		/* determine hardsector size */
		disk->hardsector_size = 512;
		if (hardsect_size[ldev_prv->major]) {
			disk->hardsector_size = hardsect_size[ldev_prv->major][ldev_prv->minor];
		}
		/* save the block size */
		disk->block_size = 1024;
		if (blksize_size[ldev_prv->major]) {
			disk->block_size = blksize_size[ldev_prv->major][ldev_prv->minor];
		}
		/* device size in sectors
		 *
		 * try 64bit size first, if that fails
		 * fall back on the 32bit size.
		 */
		/* try 64bit size */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,18)
		rc = evms_cs_kernel_ioctl(disk, BLKGETSIZE64,
					  (ulong) & disk->total_vsectors);
		if (!rc) {
			/* convert bytes to 512 byte sectors */
			disk->total_vsectors >>= EVMS_VSECTOR_SIZE_SHIFT;
		} else 
#endif
			{
			/* try 32bit size */
			ulong dev_size = 0;
			rc = evms_cs_kernel_ioctl(disk, BLKGETSIZE,
						  (ulong) & dev_size);
			disk->total_vsectors = dev_size;
		}
		break;
	default:
		if (disk->flags & EVMS_DEVICE_UNAVAILABLE) {
			rc = -ENXIO;
		} else {
			rc = ldev_prv->bdev->bd_op->ioctl(inode, file, cmd,
							  arg);
		}
		break;
	}
	inode->i_bdev = save_bdev;
	inode->i_rdev = save_dev;

	return (rc);
}

/********************************************************/
/* Required Module Entry Point:                         */
/*      ldev_mgr_init                                   */
/********************************************************/

static int __init
ldev_mgr_init(void)
{
	return evms_cs_register_plugin(&plugin_header);
}

static void __exit
ldev_mgr_exit(void)
{
	evms_cs_unregister_plugin(&plugin_header);
}

module_init(ldev_mgr_init);
module_exit(ldev_mgr_exit);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
