/*
   linear.c : Multiple Devices driver for Linux
              Copyright (C) 1994-96 Marc ZYNGIER
	      <zyngier@ufr-info-p7.ibp.fr> or
	      <maz@gloups.fdn.fr>

   Linear mode management functions.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <linux/module.h>
#include <linux/evms/evms_md.h>
#include <linux/evms/evms_linear.h>
#include <linux/slab.h>


#define MAJOR_NR MD_MAJOR
#define MD_DRIVER
#define MD_PERSONALITY

#define LOG_PREFIX "md linear: "
static int linear_run (mddev_t *mddev)
{
	linear_conf_t *conf;
	struct linear_hash *table;
	mdk_rdev_t *rdev;
	int size, i, j, nb_zone;
	unsigned int curr_offset;

	MOD_INC_USE_COUNT;

	conf = kmalloc (sizeof (*conf), GFP_KERNEL);
	if (!conf)
		goto out;
	mddev->private = conf;

	if (evms_md_check_ordering(mddev)) {
		printk("linear: disks are not ordered, aborting!\n");
		goto out;
	}

	/*
	 * Find the smallest device.
	 */

	conf->smallest = NULL;
	curr_offset = 0;
	ITERATE_RDEV_ORDERED(mddev,rdev,j) {
		dev_info_t *disk = conf->disks + j;
		disk->node = rdev->node;
		disk->dev = rdev->dev;
		disk->size = rdev->size;
		disk->offset = curr_offset;

		curr_offset += disk->size;

		if (!conf->smallest || (disk->size < conf->smallest->size))
			conf->smallest = disk;
	}

	nb_zone = conf->nr_zones = evms_md_size[mdidx(mddev)] / conf->smallest->size + 
		((evms_md_size[mdidx(mddev)] % conf->smallest->size) ? 1 : 0);
  
	conf->hash_table = kmalloc (sizeof (struct linear_hash) * nb_zone,
					GFP_KERNEL);
	if (!conf->hash_table)
		goto out;

	/*
	 * Here we generate the linear hash table
	 */
	table = conf->hash_table;
	i = 0;
	size = 0;
	for (j = 0; j < mddev->nb_dev; j++) {
		dev_info_t *disk = conf->disks + j;

		if (size < 0) {
			table[-1].dev1 = disk;
		}
		size += disk->size;

		while (size>0) {
			table->dev0 = disk;
			table->dev1 = NULL;
			size -= conf->smallest->size;
			table++;
		}
	}
	if (table-conf->hash_table != nb_zone)
		BUG();
	LOG_DETAILS("%s: nr_zones=%d, smallest=%lu\n",
		    __FUNCTION__, conf->nr_zones, conf->smallest->size);
	return 0;

out:
	if (conf)
		kfree(conf);
	MOD_DEC_USE_COUNT;
	return 1;
}

static int linear_stop (mddev_t *mddev)
{
	linear_conf_t *conf = mddev_to_conf(mddev);
  
	kfree(conf->hash_table);
	kfree(conf);

	MOD_DEC_USE_COUNT;

	return 0;
}

/*
 * Function: linear_map
 */
static int linear_map(
	mddev_t *mddev,
	struct evms_logical_node **node,
	struct buffer_head *bh)
{
	linear_conf_t *conf = mddev_to_conf(mddev);
	struct linear_hash *hash;
	dev_info_t *tmp_dev;
	unsigned long block;

	block = (bh->b_rsector >> 1);
	hash = conf->hash_table + (block / conf->smallest->size);
	if (block >= (hash->dev0->size + hash->dev0->offset)) {
		if (!hash->dev1) {
			LOG_ERROR("%s: hash->dev1==NULL for block %ld\n", __FUNCTION__, block);
			return -ENXIO;
		}
		tmp_dev = hash->dev1;
	} else
		tmp_dev = hash->dev0;
    
	if ( (block + (bh->b_size >> 10)) > (tmp_dev->size + tmp_dev->offset)
				|| block < tmp_dev->offset) {
		LOG_ERROR("%s: Block %ld out of bounds on node %s size %ld offset %ld\n",
			  __FUNCTION__,
			   block,
			   tmp_dev->node->name,
			   tmp_dev->size,
			   tmp_dev->offset);
		return -ENXIO;
	}
	bh->b_rsector -= (tmp_dev->offset << 1);
	*node = tmp_dev->node;
	return 0;
}

static void linear_read(
	struct evms_logical_node *md_node,
	struct buffer_head *bh)
{
	mddev_t *mddev = EVMS_MD_NODE_TO_MDDEV(md_node);
	struct evms_logical_node *node;

	if (evms_md_check_boundary(md_node, bh)) return;

	if (!linear_map(mddev, &node, bh)) {
		R_IO(node, bh);
	} else {
		bh->b_end_io(bh, 0);
	}
}

static void linear_write(
	struct evms_logical_node *md_node,
	struct buffer_head *bh)
{
	mddev_t *mddev = EVMS_MD_NODE_TO_MDDEV(md_node);
	struct evms_logical_node *node;

	if (evms_md_check_boundary(md_node, bh)) return;

	if (!linear_map(mddev, &node, bh)) {
		W_IO(node, bh);
	} else {
		bh->b_end_io(bh, 0);
	}
}

static int linear_status (char *page, mddev_t *mddev)
{
	int sz = 0;

#undef MD_DEBUG
#ifdef MD_DEBUG
	int j;
	linear_conf_t *conf = mddev_to_conf(mddev);
  
	sz += sprintf(page+sz, "      ");
	for (j = 0; j < conf->nr_zones; j++)
	{
		sz += sprintf(page+sz, "[%s",
			partition_name(conf->hash_table[j].dev0->dev));

		if (conf->hash_table[j].dev1)
			sz += sprintf(page+sz, "/%s] ",
			  partition_name(conf->hash_table[j].dev1->dev));
		else
			sz += sprintf(page+sz, "] ");
	}
	sz += sprintf(page+sz, "\n");
#endif
	sz += sprintf(page+sz, " %dk rounding", mddev->chunk_size/1024);
	return sz;
}

static int linear_evms_ioctl (
	mddev_t 	* mddev,
	struct inode 	* inode,
	struct file 	* file, 
	unsigned int 	cmd,
	unsigned long 	arg)
{
	int rc = 0;
	struct evms_logical_node *node;

	switch (cmd) {
		case EVMS_GET_BMAP:
		{
			struct evms_get_bmap_pkt *bmap = (struct evms_get_bmap_pkt *)arg;
			struct buffer_head *bh = 
				evms_cs_allocate_from_pool(evms_bh_pool, FALSE);
			if (bh) {
				bh->b_rsector = (unsigned long)bmap->rsector;
				bh->b_size = node->block_size;
				rc = linear_map(mddev, &node, bh);
				if (!rc) {
					bmap->rsector = (u64)bh->b_rsector;
					if (node)
						rc = IOCTL(node, inode, file, cmd, arg);
					else
						rc = -ENODEV;
				}
				evms_cs_deallocate_to_pool(evms_bh_pool, bh);
			} else
				rc = -ENOMEM;
			break;
		}

		default:
			rc = -EINVAL;
	}
	return rc;
}

static mdk_personality_t linear_personality = {
	.name		= "evms_linear",
	.read		= linear_read,
	.write		= linear_write,
	.run		= linear_run,
	.stop		= linear_stop,
	.status		= linear_status,
	.evms_ioctl	= linear_evms_ioctl
};

static int md__init linear_init (void)
{
	return evms_register_md_personality (LINEAR, &linear_personality);
}

static void linear_exit (void)
{
	evms_unregister_md_personality (LINEAR);
}


module_init(linear_init);
module_exit(linear_exit);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
