/*
 *
 *   Copyright (c) International Business Machines Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/*
 * linux/drivers/evms/os2lvm_vge.c
 *
 * EVMS OS/2 LVM Emulator
 *
 * This Volume Group Emulator will take the type 0x35 partitions created by
 *  OS/2 versions 4.5 and later and build them into volumes.  It emulates
 *  the Drive Linking and Bad Block Relocation features and therefore
 *  provides binary compatibility with the OS/2 version.  Of course, if
 *  you select to mkfs a file system OS/2 doesn't support, you're on your
 *  own...
 *
 * Since OS/2 LVM volumes can only exist on DOS-style partitioned disks,
 *  this VGE has a dependency on dospart.c to report a list of the
 *  candidate partitions.  This module will then take the appropriate partitions
 *  from the list and use them to build the OS/2-style volumes.
 *
 * Change Activity:
 *
 *   7/01/2001  John Stiles  getting started.
 *   9/14/2001  John Stiles  original version.
 *  11/01/2001  John Stiles  new naming scheme.
 *  11/21/2001  John Stiles  i/o path changes.
 */

#define EVMS_DEBUG 1
#define EVMS_OS2_DEBUG 1

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/config.h>
#include <linux/genhd.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/init.h>
#include <linux/evms/evms.h>
#include <linux/evms/evms_os2.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>

#define LOG_PREFIX "os2lvm: "

// Global Structure and Type definitions
struct transfer_record {
	int Write_Flag;		/* 0 = read, 1 = write */
	struct os2_dl_entry *Partition_Data;
	struct buffer_head *bh;
	struct transfer_record *next;
};

struct tracking_record {	/* structure used to track IO requests that must be broken into two pieces due to drive linking */
	unsigned int io_in_progress;
	int up_to_date;
	struct buffer_head *org_bh;	/* Original IO */
	struct buffer_head *link1_bh;	/* First child. */
	struct os2_dl_entry *link1_data;
	struct transfer_record *link1_transfer_rec;
	int link1_bbr_attempted;
	struct buffer_head *link2_bh;	/* Second child */
	struct os2_dl_entry *link2_data;
	struct transfer_record *link2_transfer_rec;
	int link2_bbr_attempted;
};

// Prototypes for local VGE functions
static int discover_os2lvm_partitions(struct evms_logical_node **);
static struct evms_logical_node *find_os2_volume(u32);
static int add_os2link(struct os2_dl_entry *,
		       struct evms_logical_node *);
static struct os2_dl_entry
    *find_link_data(struct os2_dl_entry **, u32);
static int find_drive_link(struct evms_logical_node *,
			   struct os2_dl_entry **, u64 *, u64 *);
static int validate_signaturesector(struct evms_logical_node *,
				    LVM_Signature_Sector *, u32);
static int validate_drivelinksector(void *, int, u32);
static int validate_bbrtablesector(void *, int, u32);
static u32 check_for_os2_bbr_relocations(char *);
static int check_os2_volumes(struct evms_logical_node **);
static int OS2_ioctl_cmd_broadcast(struct evms_logical_node *node,
				   struct inode *inode, struct file *file,
				   unsigned long cmd, unsigned long arg);
static int os2_ioctl_cmd_plugin_ioctl(struct evms_logical_node *node,
				      struct inode *inode, struct file *file,
				      unsigned long cmd, unsigned long arg);
static void BBR_Worker(void *);
static void OS2_BBR_Write_Callback(struct transfer_record * Transfer_Record,
				   struct buffer_head *bh,
				   int uptodate, int *redrive);
static void BBR_Transfer_IO(struct transfer_record * Transfer_Record);
static void OS2_DL_Callback(struct buffer_head *bh, int uptodate);
static int Sector_Is_Remapped(struct os2_dl_entry * io_dlentry,
			      u64 Source_Sector, u64 * Replacement_Sector);
static void Invalidate_Mapping(struct os2_dl_entry * io_dlentry,
			       u64 Source_Sector,
			       int Replacement_Sector_Is_Bad);
static int Create_New_BBR_Table_Entry(struct os2_dl_entry *
				      io_dlentry, u64 starting_lsn,
				      unsigned int count, void *buffer);
static void Clone_Bufferhead(struct buffer_head *Source,
			     struct buffer_head *Child);

// Prototypes for local memory allocation/deallocation functions
static struct os2_dl_entry *new_os2_drive_link(LVM_Signature_Sector *,
							 struct
							 evms_logical_node *);
static char *new_os2_link_data(u32, u32, u32, struct evms_logical_node *);
static char *new_os2_bbr_data(u32, u32, u32, struct evms_logical_node *);
static struct evms_logical_node *new_os2volume(u32, char *);
static int delete_os2lvm_volume(struct evms_logical_node *);
static int delete_os2_drive_link(struct os2_dl_entry *, int);

// Prototypes for Function Table interface
static int discover_os2lvm(struct evms_logical_node **);
static int delete_os2lvm(struct evms_logical_node *);
static void read_os2lvm(struct evms_logical_node *, struct buffer_head *);
static void write_os2lvm(struct evms_logical_node *, struct buffer_head *);
static int init_io_os2lvm(struct evms_logical_node *, int, u64, u64, void *);
static int ioctl_os2lvm(struct evms_logical_node *, struct inode *,
			struct file *, unsigned int, unsigned long);
static int do_os2_bbr_io(struct os2_dl_entry *, int, u64, u64,
			 void *);

// Global data structures
static struct evms_logical_node *os2lvm_nodes = NULL;
static struct evms_thread *BBR_Worker_Thread = NULL;
static spinlock_t BBR_Queue_Lock = SPIN_LOCK_UNLOCKED;
static const char *BBR_Worker_Name = "evms_os2_bbr_io";
static struct transfer_record *BBR_IO_List_Head = NULL;
static struct transfer_record *BBR_IO_List_Tail = NULL;
static struct evms_pool_mgmt *BBR_Transfer_Pool = NULL;
static char *BBR_Transfer_Pool_Name = "OS-2 Transfer Pool";
static char *DL_Tracking_Pool_Name = "OS-2 Tracking Pool";
static struct evms_pool_mgmt *DL_Tracking_Pool = NULL;

// Required plug-in Function Table definition
static struct evms_plugin_fops function_table = {
	.discover	= discover_os2lvm,
	.delete		= delete_os2lvm,
	.read		= read_os2lvm,
	.write		= write_os2lvm,
	.init_io	= init_io_os2lvm,
	.ioctl		= ioctl_os2lvm
};

// Required plug-in Header definition
static struct evms_plugin_header plugin_header = {
	.id = SetPluginID(IBM_OEM_ID,
			  EVMS_REGION_MANAGER,
			  2),
	.version = {
		.major		= 1,
		.minor		= 1,
		.patchlevel	= 2
	},
	.required_services_version = {
		.major		= EVMS_COMMON_SERVICES_MAJOR,
		.minor		= EVMS_COMMON_SERVICES_MINOR,
		.patchlevel	= EVMS_COMMON_SERVICES_PATCHLEVEL
	},
	.fops = &function_table
};

//  Required Plugin Functions

/*
 * Function:  discover_os2lvm
 *
 *      This is the entry point into the discovery process.
 */
static int
discover_os2lvm(struct evms_logical_node **evms_partition_list)
{
	int rc;

	MOD_INC_USE_COUNT;

	if (!BBR_Transfer_Pool) {
		BBR_Transfer_Pool =
		    evms_cs_create_pool(sizeof (struct transfer_record),
					BBR_Transfer_Pool_Name, NULL, NULL);
		if (!BBR_Transfer_Pool) {
			MOD_DEC_USE_COUNT;
			return -ENOMEM;
		}
	}

	if (!DL_Tracking_Pool) {
		DL_Tracking_Pool =
		    evms_cs_create_pool(sizeof (struct tracking_record),
					DL_Tracking_Pool_Name, NULL, NULL);
		if (!DL_Tracking_Pool) {
			MOD_DEC_USE_COUNT;
			return -ENOMEM;
		}
	}

	rc = discover_os2lvm_partitions(evms_partition_list);

	if (!rc) {
		rc = check_os2_volumes(evms_partition_list);
	}

	MOD_DEC_USE_COUNT;
	return rc;
}

/*
 * Function:  delete_os2lvm
 *
 *      This is the entry point for deleting a node.
 */
static int
delete_os2lvm(struct evms_logical_node *logical_node)
{
	LOG_EXTRA("Deleting volume: %s\n", logical_node->name);

	return delete_os2lvm_volume(logical_node);
}

/*
 * Function:  read_os2lvm
 */
static void
read_os2lvm(struct evms_logical_node *node, struct buffer_head *bh)
{
	int rc;
	u64 sector_count;
	u64 rsector;
	struct buffer_head *Link1 = NULL;
	struct buffer_head *Link2 = NULL;
	struct tracking_record *Tracking_Record = NULL;
	struct os2_dl_entry *cur_dlentry = NULL;
	struct transfer_record *Transfer_Record;

	rsector = bh->b_rsector;
	sector_count = bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT;
	rc = find_drive_link(node, &cur_dlentry, &rsector, &sector_count);
	bh->b_rsector = rsector;
	switch (rc) {
	case 1:
		if (cur_dlentry->bbr_is_active) {
			Transfer_Record = evms_cs_allocate_from_pool(BBR_Transfer_Pool, 1);	/* Block until we get a transfer record. */
			/* Transfer the IO to the BBR Worker Thread. */
			Transfer_Record->Write_Flag = 0;
			Transfer_Record->Partition_Data = cur_dlentry;
			Transfer_Record->bh = bh;
			Transfer_Record->next = NULL;
			BBR_Transfer_IO(Transfer_Record);
		} else
			R_IO(cur_dlentry->link_partition, bh);
		break;
	case 2:
		/* We must split the IO.  Duplicate the buffer head twice and allocate the tracking record. */
		Tracking_Record = evms_cs_allocate_from_pool(DL_Tracking_Pool, 1);	/* Block until we get a tracking record. */
		Link1 = evms_cs_allocate_from_pool(evms_bh_pool, 1);
		Link2 = evms_cs_allocate_from_pool(evms_bh_pool, 1);

		/* Initialize the tracking record so we can associate the two new I/Os with the original. */
		Tracking_Record->io_in_progress = 2;
		Tracking_Record->up_to_date = 0;
		Tracking_Record->org_bh = bh;

		/* Create the I/O to the first link. */
		Clone_Bufferhead(bh, Link1);
		Link1->b_private = Tracking_Record;
		Link1->b_end_io = OS2_DL_Callback;
		Link1->b_size = sector_count << EVMS_VSECTOR_SIZE_SHIFT;
		Tracking_Record->link1_bh = Link1;
		Tracking_Record->link1_data = cur_dlentry;
		Tracking_Record->link1_bbr_attempted = 0;
		Tracking_Record->link1_transfer_rec = NULL;

		/* Create the I/O to the second link */
		Clone_Bufferhead(bh, Link2);
		Link2->b_private = Tracking_Record;
		Link2->b_end_io = OS2_DL_Callback;
		Link2->b_data += sector_count << EVMS_VSECTOR_SIZE_SHIFT;
		Link2->b_rsector = 0;
		Link2->b_size =
		    bh->b_size - (sector_count << EVMS_VSECTOR_SIZE_SHIFT);
		Tracking_Record->link2_bh = Link2;
		Tracking_Record->link2_data = cur_dlentry->next;
		Tracking_Record->link2_bbr_attempted = 0;
		Tracking_Record->link2_transfer_rec = NULL;

		/* Process the I/O to the first link. */
		if (cur_dlentry->bbr_is_active) {
			Transfer_Record = evms_cs_allocate_from_pool(BBR_Transfer_Pool, 1);	/* Block until we get a transfer record. */
			/* Transfer the IO to the BBR Worker Thread. */
			Transfer_Record->Write_Flag = 0;
			Transfer_Record->Partition_Data = cur_dlentry;
			Transfer_Record->bh = Tracking_Record->link1_bh;
			Transfer_Record->next = NULL;
			BBR_Transfer_IO(Transfer_Record);
		} else
			R_IO(cur_dlentry->link_partition,
			     Tracking_Record->link1_bh);

		/* Process the I/O to the second link. */
		cur_dlentry = cur_dlentry->next;
		if (cur_dlentry->bbr_is_active) {
			Transfer_Record = evms_cs_allocate_from_pool(BBR_Transfer_Pool, 1);	/* Block until we get a transfer record. */
			/* Transfer the IO to the BBR Worker Thread. */
			Transfer_Record->Write_Flag = 0;
			Transfer_Record->Partition_Data = cur_dlentry;
			Transfer_Record->bh = Tracking_Record->link2_bh;
			Transfer_Record->next = NULL;
			BBR_Transfer_IO(Transfer_Record);
		} else
			R_IO(cur_dlentry->link_partition,
			     Tracking_Record->link2_bh);

		break;
	default:
		LOG_SERIOUS("READ error, request exceeds volume size.\n");
		bh->b_end_io(bh, 0);
		break;
	}
}

/*
 * Function:  write_os2lvm
 */
static void
write_os2lvm(struct evms_logical_node *node, struct buffer_head *bh)
{
	int rc;
	u64 rsector;
	u64 sector_count;
	struct buffer_head *Link1 = NULL;
	struct buffer_head *Link2 = NULL;
	struct tracking_record *Tracking_Record = NULL;
	struct os2_dl_entry *cur_dlentry = NULL;
	struct transfer_record *Transfer_Record;

	rsector = bh->b_rsector;
	sector_count = bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT;
	rc = find_drive_link(node, &cur_dlentry, &rsector, &sector_count);
	bh->b_rsector = rsector;
	switch (rc) {
	case 1:
		/* Set up a Transfer Record.  If there are Bad Blocks on the partition that this I/O is 
		   directed to, then we will need the Transfer Record to put the I/O in the queue for the 
		   BBR Worker Thread.  If there are no bad blocks, then we will need the Transfer Record 
		   for the OS2_BBR_Write_Callback function.  This function expects the Transfer Record to 
		   be pre-allocated and available because it is running on an interrupt thread and should
		   not do memory allocation.  If there is an error during the write, then the 
		   OS2_BBR_Write_Callback function will use the Transfer Record to transfer the I/O
		   to the BBR worker thread for further processing.  If there are no errors during the I/O,
		   then the OS2_BBR_Write_Callback will deallocate the Transfer Record.                     */
		Transfer_Record = evms_cs_allocate_from_pool(BBR_Transfer_Pool, 1);	/* Block until we get a transfer record. */
		Transfer_Record->Write_Flag = 1;
		Transfer_Record->Partition_Data = cur_dlentry;
		Transfer_Record->bh = bh;
		Transfer_Record->next = NULL;
		if (cur_dlentry->bbr_is_active) {
			/* Transfer the IO to the BBR Worker Thread. */
			BBR_Transfer_IO(Transfer_Record);
		} else {
			evms_cs_register_for_end_io_notification
			    (Transfer_Record, bh, OS2_BBR_Write_Callback);
			W_IO(cur_dlentry->link_partition, bh);
		}
		break;
	case 2:
		/* We must split the IO.  Duplicate the buffer head twice and allocate the tracking record. */
		Tracking_Record = evms_cs_allocate_from_pool(DL_Tracking_Pool, 1);	/* Block until we get a tracking record. */
		Link1 = evms_cs_allocate_from_pool(evms_bh_pool, 1);
		Link2 = evms_cs_allocate_from_pool(evms_bh_pool, 1);

		/* Initialize the tracking record so we can associate the two new I/Os with the original. */
		Tracking_Record->io_in_progress = 2;
		Tracking_Record->up_to_date = 0;
		Tracking_Record->org_bh = bh;

		/* Create the I/O to the first link. */
		Clone_Bufferhead(bh, Link1);
		Link1->b_private = Tracking_Record;
		Link1->b_end_io = OS2_DL_Callback;
		Link1->b_size = sector_count << EVMS_VSECTOR_SIZE_SHIFT;
		Tracking_Record->link1_bh = Link1;
		Tracking_Record->link1_data = cur_dlentry;

		/* Create the I/O to the second link */
		Clone_Bufferhead(bh, Link2);
		Link2->b_private = Tracking_Record;
		Link2->b_end_io = OS2_DL_Callback;
		Link2->b_data += sector_count << EVMS_VSECTOR_SIZE_SHIFT;
		Link2->b_rsector = 0;
		Link2->b_size =
		    bh->b_size - (sector_count << EVMS_VSECTOR_SIZE_SHIFT);
		Tracking_Record->link2_bh = Link2;
		Tracking_Record->link2_data = cur_dlentry->next;

		Transfer_Record = evms_cs_allocate_from_pool(BBR_Transfer_Pool, 1);	/* Block until we get a transfer record. */
		Transfer_Record->Write_Flag = 1;
		Transfer_Record->Partition_Data = cur_dlentry;
		Transfer_Record->bh = Tracking_Record->link1_bh;
		Transfer_Record->next = NULL;
		Tracking_Record->link1_transfer_rec = Transfer_Record;
		/* Process the I/O to the first link. */
		if (cur_dlentry->bbr_is_active) {
			/* Transfer the IO to the BBR Worker Thread. */
			Tracking_Record->link1_bbr_attempted = 1;
			BBR_Transfer_IO(Transfer_Record);
		} else {
			Tracking_Record->link1_bbr_attempted = 0;
			W_IO(cur_dlentry->link_partition,
			     Tracking_Record->link1_bh);
		}

		/* Process the I/O to the second link. */
		cur_dlentry = cur_dlentry->next;
		Transfer_Record = evms_cs_allocate_from_pool(BBR_Transfer_Pool, 1);	/* Block until we get a transfer record. */
		Transfer_Record->Write_Flag = 1;
		Transfer_Record->Partition_Data = cur_dlentry;
		Transfer_Record->bh = Tracking_Record->link2_bh;
		Transfer_Record->next = NULL;
		Tracking_Record->link2_transfer_rec = Transfer_Record;
		if (cur_dlentry->bbr_is_active) {
			/* Transfer the IO to the BBR Worker Thread. */
			Tracking_Record->link2_bbr_attempted = 1;
			BBR_Transfer_IO(Transfer_Record);
		} else {
			Tracking_Record->link2_bbr_attempted = 0;
			W_IO(cur_dlentry->link_partition,
			     Tracking_Record->link2_bh);
		}

		break;
	default:
		LOG_SERIOUS("WRITE error, request exceeds volume size.\n");
		bh->b_end_io(bh, 0);
		break;
	}
}

static int
os2_ioctl_cmd_plugin_ioctl(struct evms_logical_node *node,
			   struct inode *inode,
			   struct file *file,
			   unsigned long cmd, unsigned long arg)
{
	int rc = 0;
	os2_volume_runtime_entry_t *Node_Data;
	struct os2_dl_entry *curlink, *nextlink;
	struct evms_plugin_ioctl_pkt tmp, *user_parms;

	user_parms = (struct evms_plugin_ioctl_pkt *) arg;
	/* copy user's parameters to kernel space */
	if (copy_from_user(&tmp, user_parms, sizeof (tmp)))
		rc = -EFAULT;

	if (!rc) {
		Node_Data = (os2_volume_runtime_entry_t *) node->private;
		/* is this cmd targetted at this feature ? */
		if (tmp.feature_id == node->plugin->id) {
			switch (tmp.feature_command) {
			default:
				break;
			}
		} else {	/* broadcast this cmd to all children */
			curlink = Node_Data->drive_link;

			/* broadcast this cmd to all children */
			while (curlink) {
				nextlink = curlink->next;

				rc = IOCTL(curlink->link_partition, inode, file,
					   cmd, arg);

				if (rc) {
					break;
				}
				curlink = nextlink;
			}

		}
		/* copy info to userspace */
		if (copy_to_user(user_parms, &tmp, sizeof (tmp)))
			rc = -EFAULT;
	}
	return (rc);
}

static int
OS2_ioctl_cmd_broadcast(struct evms_logical_node *node,
			struct inode *inode,
			struct file *file, unsigned long cmd, unsigned long arg)
{
	int rc = 0;
	os2_volume_runtime_entry_t *Node_Data;
	struct os2_dl_entry *curlink, *nextlink;

	Node_Data = (os2_volume_runtime_entry_t *) node->private;
	curlink = Node_Data->drive_link;

	/* broadcast this cmd to all children */
	while (curlink) {
		nextlink = curlink->next;

		rc |= IOCTL(curlink->link_partition, inode, file, cmd, arg);

		curlink = nextlink;
	}

	return (rc);
}

/*
 * Function:  ioctl_os2lvm
 */
static int
ioctl_os2lvm(struct evms_logical_node *logical_node,
	     struct inode *inode,
	     struct file *file, unsigned int cmd, unsigned long arg)
{
	int rc = 0;
	u64 Sectors_Per_Cylinder;
	u64 Total_Sectors;
	struct evms_logical_node *partition_node;

	partition_node =
	    ((os2_volume_runtime_entry_t *) logical_node->private)->drive_link->
	    link_partition;

	if (!inode)
		return -EINVAL;

	LOG_EVERYTHING("Ioctl %d\n", cmd);

	switch (cmd) {
	case HDIO_GETGEO:
		{
			// Return fake geometry
			struct hd_geometry *hd = (struct hd_geometry *) arg;
			short cylinders;
			unsigned char heads = 255;
			unsigned char sectors =
			    OS2LVM_SYNTHETIC_SECTORS_PER_TRACK;
			long start = 0;

			/* OS/2 always created a fake geometry using the maximum cylinder size. */
			Sectors_Per_Cylinder = heads * sectors;
			for (cylinders = 0, Total_Sectors = 0;
			     Total_Sectors <
			     ((os2_volume_runtime_entry_t *) logical_node->
			      private)->size_in_sectors; cylinders++)
				Total_Sectors += Sectors_Per_Cylinder;

			cylinders--;

			if (copy_to_user
			    ((short *) (&hd->cylinders), &cylinders,
			     sizeof (cylinders))
			    || copy_to_user((char *) (&hd->heads), &heads,
					    sizeof (heads))
			    || copy_to_user((char *) (&hd->sectors), &sectors,
					    sizeof (sectors))
			    || copy_to_user((long *) (&hd->start), &start,
					    sizeof (start))) {
				return -EFAULT;
			}
		}
		break;

	case EVMS_GET_BMAP:
		// No kernel images allowed on OS/2 volumes right now.
		rc = -EINVAL;
		break;

	case EVMS_QUIESCE_VOLUME:
	case EVMS_GET_DISK_LIST:
	case EVMS_CHECK_MEDIA_CHANGE:
	case EVMS_REVALIDATE_DISK:
	case EVMS_OPEN_VOLUME:
	case EVMS_CLOSE_VOLUME:
	case EVMS_CHECK_DEVICE_STATUS:
		rc = OS2_ioctl_cmd_broadcast(logical_node, inode, file, cmd,
					     arg);
		break;
	case EVMS_PLUGIN_IOCTL:
		rc = os2_ioctl_cmd_plugin_ioctl(logical_node, inode, file, cmd,
						arg);
		break;
	default:
		rc = -EINVAL;
		break;
	}

	return rc;
}

/*
 * Function:  init_io_os2lvm
 */
static int
init_io_os2lvm(struct evms_logical_node *node, int io_flag,	/* 0=read, 1=write   */
	       u64 sect_nr,	/* disk LBA          */
	       u64 num_sects,	/* # of sectors      */
	       void *buf_addr)
{				/* buffer address    */
	int rc = 0;
	u64 sector_count;
	struct evms_logical_node *partition_node;
	struct os2_dl_entry *cur_dlentry = NULL;

	sector_count = num_sects;
	rc = find_drive_link(node, &cur_dlentry, &sect_nr, &sector_count);
	switch (rc) {
	case 1:
		partition_node = cur_dlentry->link_partition;
		if (cur_dlentry->bbr_is_active)
			rc = do_os2_bbr_io(cur_dlentry, io_flag, sect_nr,
					   num_sects, buf_addr);
		else {
			rc = INIT_IO(partition_node, io_flag, sect_nr,
				     num_sects, buf_addr);
			if (rc && io_flag) {
				cur_dlentry->bbr_is_active = 1;
				rc = do_os2_bbr_io(cur_dlentry, io_flag,
						   sect_nr, num_sects,
						   buf_addr);
			}
		}
		break;
	case 2:
		partition_node = cur_dlentry->link_partition;
		if (cur_dlentry->bbr_is_active)
			rc = do_os2_bbr_io(cur_dlentry, io_flag, sect_nr,
					   sector_count, buf_addr);
		else {
			rc = INIT_IO(partition_node, io_flag, sect_nr,
				     sector_count, buf_addr);
			if (rc && io_flag) {
				cur_dlentry->bbr_is_active = 1;
				rc = do_os2_bbr_io(cur_dlentry, io_flag,
						   sect_nr, sector_count,
						   buf_addr);
			}
		}

		if (!rc) {
			cur_dlentry = cur_dlentry->next;
			partition_node = cur_dlentry->link_partition;
			num_sects -= sector_count;
			buf_addr += sector_count << OS2_SECTOR_SHIFT;
			rc = 1;
			if (cur_dlentry->bbr_is_active)
				rc = do_os2_bbr_io(cur_dlentry, io_flag, 0,
						   num_sects, buf_addr);
			else {
				rc = INIT_IO(partition_node, io_flag, 0,
					     num_sects, buf_addr);
				if (rc && io_flag) {
					cur_dlentry->bbr_is_active = 1;
					rc = do_os2_bbr_io(cur_dlentry, io_flag,
							   0, num_sects,
							   buf_addr);
				}

			}
		}
		break;
	default:
		LOG_SERIOUS("INITIO error, request exceeds volume size.\n");
		break;
	}

	return rc;
}

/*
 * Function:  do_os2_bbr_io
 *
 *      Check the Bad Block Relocation list for relocated sectors.  If any are found,
 *       this function will do the i/o directly.
 *      Return values:  0 == i/o done,  1 == unable to complete i/o
 */
static int
do_os2_bbr_io(struct os2_dl_entry * io_dlentry, int rw,	/* 0=read, 1=write  */
	      u64 starting_lsn,	/* disk LBA         */
	      u64 count,	/* # of sectors     */
	      void *buffer)
{				/* buffer address   */
	u64 lsn, remapped_lsn;
	int rc;

	// For each sector in this request, check if this sector has already
	// been remapped. If so, process all previous sectors in this request,
	// followed by the remapped sector. Then reset the starting lsn and
	// count and keep going with the rest of the request as if it were
	// a whole new request.
	for (lsn = 0; lsn < count; lsn++) {
		remapped_lsn = starting_lsn + lsn;
		rc = Sector_Is_Remapped(io_dlentry, remapped_lsn,
					&remapped_lsn);
		if (rc) {
			// Process all sectors in the request up to this one.
			if (lsn > 0) {
				rc = INIT_IO(io_dlentry->link_partition, rw,
					     starting_lsn, lsn, buffer);
				if (rc) {
					/* If this is a read, then we are done. */
					if (!rw) {
						return 1;
					}

					/* Since this was a write, we must see if we can remap the bad sector to a replacement sector. */
					if (!Create_New_BBR_Table_Entry
					    (io_dlentry, starting_lsn, lsn,
					     buffer)) {
						/* We were unable to remap the bad sector(s) in the I/O.  We can not complete the I/O. */
						return 1;
					}
				}
				buffer += (lsn * OS2_BYTES_PER_SECTOR);
			}
			// Process the remapped sector.
			rc = INIT_IO(io_dlentry->link_partition, rw,
				     remapped_lsn, 1, buffer);
			if (rc) {
				/* If this is a read, then we are done. */
				if (!rw) {
					return 1;
				}

				/* Get the original sector that was remapped. */
				remapped_lsn = starting_lsn + lsn;

				/* Invalidate the current remapping. */
				Invalidate_Mapping(io_dlentry, remapped_lsn, 1);

				/* Try to remap the bad sector to another replacement sector. */
				if (!Create_New_BBR_Table_Entry
				    (io_dlentry, remapped_lsn, 1, buffer)) {
					/* We were unable to remap the bad sector(s) in the I/O.  We can not complete the I/O. */
					return 1;
				}

			}

			buffer += OS2_BYTES_PER_SECTOR;

			starting_lsn += (lsn + 1);
			count -= (lsn + 1);
			lsn = -1;
		}

	}

	/* Are there any sectors left to process? */
	if (count > 0) {
		rc = INIT_IO(io_dlentry->link_partition, rw, starting_lsn,
			     count, buffer);
		if (rc) {
			/* If this is a read, then we are done. */
			if (!rw) {
				return 1;
			}

			/* Since this was a write, we must see if we can remap the bad sector to a replacement sector. */
			if (!Create_New_BBR_Table_Entry
			    (io_dlentry, starting_lsn, count, buffer)) {
				/* We were unable to remap the bad sector(s) in the I/O.  We can not complete the I/O. */
				return 1;
			}

		}

	}

	return 0;
}

/*
 * Function: os2lvm_vge_init
 */
int __init
os2lvm_vge_init(void)
{
	/* Should I be allocating the pools and BBR Worker Thread here? */
	return evms_cs_register_plugin(&plugin_header);	/* register with EVMS */
}

void __exit
os2lvm_vge_exit(void)
{
	/* BUGBUG - Is there where I need to kill the BBR Worker Thread and free any memory I am still holding? */

	evms_cs_unregister_plugin(&plugin_header);
}

module_init(os2lvm_vge_init);
module_exit(os2lvm_vge_exit);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

// Local VGE Functions

/*
 * Function:  discover_os2lvm_partitions
 *
 *     Examine the list of logical partitions.  Any type 0x35 partition that contains
 *      a valid OS/2 signature sector is consumed and added to the appropriate logical
 *      volume.
 */
static int
discover_os2lvm_partitions(struct evms_logical_node **evms_partition_list)
{
	struct evms_logical_node *evms_partition;
	struct evms_logical_node *next_partition;
	struct evms_logical_node *new_volume;
	u64 sectornum = 0;
	u32 volumeserial;
	char *sigsect;
	char *volumename;
	char driveletter[8];
	LVM_Signature_Sector *sigsector;
	struct os2_dl_entry *new_dlentry;

	LOG_ENTRY_EXIT("Discovering OS/2 Logical Volumes\n");
	sigsect = kmalloc(OS2_BYTES_PER_SECTOR, GFP_KERNEL);
	if (!sigsect) {
		LOG_SERIOUS("Could not allocate Signature sector data\n");
		return -ENOMEM;
	}

	for (evms_partition = *evms_partition_list; evms_partition;
	     evms_partition = next_partition) {
		// Save the next node. We may remove this one from the list.
		next_partition = evms_partition->next;

		// The node must not have the OS/2 vge id.
		if (evms_partition->plugin->id == plugin_header.id) {
			continue;
		}

		LOG_EXTRA("Examining partition serial %s\n",
			  evms_partition->name);

		// Have to go to the last accessible sector of the partition and
		//  read it in.  It should be the LVM Signature Sector.
		sectornum = evms_partition->total_vsectors - 1;
		if (INIT_IO(evms_partition, 0, sectornum, 1, sigsect)) {
			// On an I/O error, continue on to the next partition.
			// This means that the volume it belongs to will be incomplete
			//  and later deleted in the completeness check.
			LOG_SERIOUS("I/O error on Signature sector read\n");
			continue;
		}
		sigsector = (LVM_Signature_Sector *) sigsect;

		// Validate the Signature Sector
		if (validate_signaturesector
		    (evms_partition, sigsector, OS2_BYTES_PER_SECTOR)) {
			LOG_EXTRA("Signature sector is not valid\n");
			continue;
		}
// Bugbug - At this point, we have validated an OS/2 LVM Signature Sector.  However, if the partition
// is not marked as a type 0x35, then this Signature Sector may be erroneous.  The problem here is that
// there is currently no way to find out if this partition was marked as a type 0x35.  Also, if we 
// should reject this partition due to some problem with the drive linking or BBR metadata, should we
// leave the partition in the evms partition list or not?  If the partition was marked as a type 0x35
// and the Signature Sector was valid, then I would say that we should remove it from the evms partition
// partition list.  If the partition is not marked as a type 0x35 but the Signature Sector is valid, then
// we could have a stray Signature Sector, in which case the partition should remain in the evms partition
// list.  The OS/2 LVM Signature Sector does have additional information that could be used to resolve
// this issue, such as the starting LBA of the partition that the Signature Sector belongs to, but
// we can not get the starting LBA of the partition to compare against.  If we leave the partition in
// the evms partition list when we should not, then an extraneous compatibility volume could result.
		// Build the Metadata for this partition
		if (!
		    (new_dlentry =
		     new_os2_drive_link(sigsector, evms_partition))) {
			continue;
		}
		// Search for the parent Volume for this partition
		volumeserial = sigsector->Volume_Serial_Number;
		if (!(new_volume = find_os2_volume(volumeserial))) {

			// If not found, allocate a new Volume
			LOG_EVERYTHING("Parent not found, allocate new.\n");
			if (sigsector->Drive_Letter != '\0') {
				driveletter[0] = sigsector->Drive_Letter;
				driveletter[1] = '\0';
				volumename = driveletter;
			} else
				volumename = sigsector->Volume_Name;

			if (!
			    (new_volume =
			     new_os2volume(volumeserial, volumename))) {
				delete_os2_drive_link(new_dlentry, 0);
				new_dlentry = NULL;
				continue;
			}
		}
		// Now remove the partition from the List
		evms_cs_remove_logical_node_from_list(evms_partition_list,
						      evms_partition);

		if (((os2_volume_runtime_entry_t *) new_volume->private)->
		    complete) {
			// Volume is complete, delete this duplicate
			delete_os2_drive_link(new_dlentry, 0);
			LOG_EVERYTHING("Deleting duplicate node.\n");
			((os2_volume_runtime_entry_t *) new_volume->private)->Export_Needed = 1;	//We must export this volume again!
		} else		/* Add this partition to its parent Volume */
			add_os2link(new_dlentry, new_volume);

	}

	kfree(sigsect);
	LOG_ENTRY_EXIT("Finished Discovering OS/2 Logical Volumes\n");

	return 0;
}

/*
 * Function:  find_os2_volume
 *
 *      Search for the OS/2 volume that matches the volume serial.
 */
static struct evms_logical_node *
find_os2_volume(u32 volumeserial)
{
	os2_volume_runtime_entry_t *cur_volume;
	struct evms_logical_node *cur_node;

	cur_node = os2lvm_nodes;

	while (cur_node) {
		cur_volume = (os2_volume_runtime_entry_t *) cur_node->private;
		if (cur_volume->Volume_Serial_Number == volumeserial) {
			LOG_EVERYTHING("%s: found volser match.\n",
				       __FUNCTION__);
			return cur_node;
		}
		LOG_EVERYTHING("%s: volser does not match.\n", __FUNCTION__);
		cur_node = cur_volume->next_os2lvm_node;
	}

	return NULL;
}

/*
 * Function:  add_os2link
 *
 *      Add the Drive Link metadata to the parent OS/2 volume.
 */
static int
add_os2link(struct os2_dl_entry * newlink,
	    struct evms_logical_node *parent_volume)
{
	os2_volume_runtime_entry_t *parent_metadata =
	    (os2_volume_runtime_entry_t *) parent_volume->private;
	struct os2_dl_entry *curlink =
	    parent_metadata->drive_link, *nextlink;

	if (curlink) {
		nextlink = curlink->next;
		while (nextlink) {
			curlink = nextlink;
			nextlink = curlink->next;
		}
		curlink->next = newlink;
	} else {
		parent_metadata->drive_link = newlink;
	}
	parent_metadata->drive_link_count++;
	parent_metadata->size_in_sectors += newlink->sector_count;
	parent_volume->total_vsectors += newlink->sector_count;
	return 0;
}

/*
 * Function:  find_link_data
 *
 *      Find the Drive Link metadata that matches the partition serial number.
 *       Remove it from the link_list passed in.
 */
static struct os2_dl_entry *
find_link_data(struct os2_dl_entry ** link_list, u32 partitionser)
{
	struct os2_dl_entry *curlink = *link_list, *prevlink = NULL;

	while (curlink) {
		if (curlink->partition_serial == partitionser) {
			if (prevlink) {
				prevlink->next = curlink->next;
			} else {
				*link_list = curlink->next;
			}
			curlink->next = NULL;
			return curlink;
		}
		prevlink = curlink;
		curlink = prevlink->next;
	}

	return NULL;
}

/*
 * Function:  find_drive_link
 *
 *      Walk the linked list of drive links to find the proper
 *       target partition.  Returns the metadata associated with
 *       the drive link.
 *      Return values:  1 == data contained in 1 partition, 2 == data crosses 2 partitions,
 *                      0 == target partition not found
 */
static int
find_drive_link(struct evms_logical_node *node,
		struct os2_dl_entry ** dlentry,
		u64 * sector, u64 * num_sectors)
{
	u64 last_link_sector, cur_last_sector;
	struct os2_dl_entry *curlink =
	    ((os2_volume_runtime_entry_t *) node->private)->drive_link,
	    *nextlink;

	while (curlink) {
		nextlink = curlink->next;
		last_link_sector =
		    curlink->start_sector + curlink->sector_count;
		if (*sector < last_link_sector) {
			*dlentry = curlink;
			cur_last_sector = *sector + *num_sectors;
			*sector -= curlink->start_sector;
			LOG_EVERYTHING
			    ("I/O start_RBA == "PFU64" , sector_count == "PFU64"\n",
			     *sector, *num_sectors);
			if (cur_last_sector <= last_link_sector)
				return 1;
			else {
				if ((*dlentry)->next)
					*num_sectors -=
					    cur_last_sector - last_link_sector;
				else
					return 0;
			}
			return 2;
		}

		curlink = nextlink;
	}

	return 0;
}

// Allocation/Deallocation Functions

/*
 * Function:  new_os2_drive_link
 *
 *      Allocate space for a new OS/2 drive link structure.
 *        Initialize the appropriate fields.
 *        Note:  since the BBR info applies to each link, the BBR structures
 *               are also initialized here.
 */
static struct os2_dl_entry *
new_os2_drive_link(LVM_Signature_Sector * signature_sector,
		   struct evms_logical_node *evms_partition)
{
	int i;
	u32 feature, feature_size, sectoroffset;
	struct os2_dl_entry *new_dlentry;

	new_dlentry =
	    kmalloc(sizeof (struct os2_dl_entry), GFP_KERNEL);
	if (!new_dlentry) {
		LOG_SERIOUS("Could not allocate drivelink metadata\n");
		return NULL;
	}
	memset(new_dlentry, 0, sizeof (struct os2_dl_entry));
	new_dlentry->sector_count =
	    signature_sector->Partition_Size_To_Report_To_User;
	new_dlentry->partition_serial =
	    signature_sector->partition_serial;
	new_dlentry->bbr_is_active = 0;	// initialize to not active
	new_dlentry->link_partition = evms_partition;
	init_MUTEX(&(new_dlentry->bbr_table_lock));

	sectoroffset = signature_sector->Partition_Start;
	LOG_EVERYTHING("Partition Start is at LBA %i\n", sectoroffset);
	for (i = 0; i < OS2LVM_MAX_FEATURES_PER_VOLUME; i++) {
		feature = signature_sector->LVM_Feature_Array[i].Feature_ID;
		if (feature) {
			feature_size =
			    signature_sector->LVM_Feature_Array[i].
			    Feature_Data_Size;
			LOG_EVERYTHING("Entry %d in Feature Table is valid,\n",
				       i + 1);
			LOG_EVERYTHING("Feature Data size is %i sectors.\n",
				       feature_size);
			if (feature == DRIVE_LINKING_FEATURE_ID) {
				if (!new_dlentry->link_data) {
					new_dlentry->dl_lsn1 =
					    signature_sector->
					    LVM_Feature_Array[i].
					    Location_Of_Primary_Feature_Data -
					    sectoroffset;
					new_dlentry->dl_lsn2 =
					    signature_sector->
					    LVM_Feature_Array[i].
					    Location_Of_Secondary_Feature_Data -
					    sectoroffset;
					new_dlentry->link_data =
					    new_os2_link_data(new_dlentry->
							      dl_lsn1,
							      new_dlentry->
							      dl_lsn2,
							      feature_size,
							      evms_partition);
					if (new_dlentry->link_data == NULL) {
						delete_os2_drive_link
						    (new_dlentry, 0);
						new_dlentry = NULL;
					}
				} else {
					LOG_WARNING
					    ("os2lvm_vge: Drive Linking Feature encountered twice in the same Feature Array!\n");
					delete_os2_drive_link(new_dlentry, 0);
					new_dlentry = NULL;
				}
			} else if (feature == BBR_FEATURE_ID) {
				if (!new_dlentry->bbr_data) {
					new_dlentry->bbr_lsn1 =
					    signature_sector->
					    LVM_Feature_Array[i].
					    Location_Of_Primary_Feature_Data;
					new_dlentry->bbr_lsn2 =
					    signature_sector->
					    LVM_Feature_Array[i].
					    Location_Of_Secondary_Feature_Data;
					new_dlentry->bbr_feature_size =
					    feature_size;
					new_dlentry->bbr_data =
					    new_os2_bbr_data(new_dlentry->
							     bbr_lsn1,
							     new_dlentry->
							     bbr_lsn2,
							     feature_size,
							     evms_partition);
					if (new_dlentry->bbr_data == NULL) {
						delete_os2_drive_link
						    (new_dlentry, 0);
						new_dlentry = NULL;
					} else if (signature_sector->
						   LVM_Feature_Array[i].
						   Feature_Active) {
						new_dlentry->bbr_is_active =
						    check_for_os2_bbr_relocations
						    (new_dlentry->bbr_data);
					}
				} else {
					LOG_WARNING
					    ("os2lvm_vge: BBR Feature encountered twice in the same Feature Array!\n");
					delete_os2_drive_link(new_dlentry, 0);
					new_dlentry = NULL;
				}
			} else {
				LOG_WARNING
				    ("os2lvm_vge: Unknown Feature entry %d found.\n",
				     feature);
				delete_os2_drive_link(new_dlentry, 0);
				new_dlentry = NULL;
			}

			if (signature_sector->LVM_Feature_Array[i].
			    Feature_Active) {
				LOG_EVERYTHING("Feature is active.\n");
			}
		}
	}

	if (new_dlentry &&
	    ((!new_dlentry->bbr_data) || (!new_dlentry->link_data))
	    ) {
		LOG_WARNING("os2lvm_vge: Incomplete Feature Data found.\n");
		delete_os2_drive_link(new_dlentry, 0);
		new_dlentry = NULL;
	}
	return new_dlentry;
}

/*
 * Function:  new_os2_link_data
 *
 *      Allocate space for OS/2 drive link information.
 *      Read in and validate the information from disk.
 *      Note:  assumes 512 byte sectors.
 */
static char *
new_os2_link_data(u32 linksector1,
		  u32 linksector2,
		  u32 linknumsectors, struct evms_logical_node *link_partition)
{
	char *new_data1;	/* Buffer used to hold the primary copy of the drive linking data. */
	char *new_data2;	/* Buffer used to hold the secondary copy of the drive linking data. */
	char *p1;		/* Used to access individual sectors of data within new_data1. */
	char *p2;		/* Used to access individual sectors of data within new_data2. */
	int memsize = linknumsectors * OS2_BYTES_PER_SECTOR;
	u32 i, seq1, seq2;

	/* Allocate Memory for the buffers to hold the drive linking data. */
	LOG_EVERYTHING("Drive Linking Feature entry found.\n");
	new_data1 = kmalloc(memsize, GFP_KERNEL);
	if (!new_data1) {
		LOG_SERIOUS("Could not allocate Primary Link data\n");
		return NULL;
	}
	new_data2 = kmalloc(memsize, GFP_KERNEL);
	if (!new_data2) {
		LOG_SERIOUS("Could not allocate Secondary Link data\n");
		kfree(new_data1);
		return NULL;
	}

	LOG_EVERYTHING("Primary Feature Data starts at RBA %i\n", linksector1);
	LOG_EVERYTHING("Secondary Feature Data starts at RBA %i\n",
		       linksector2);

	/* Read the drive linking data into memory. */
	if (INIT_IO(link_partition, 0, linksector1, linknumsectors, new_data1)) {
		LOG_SERIOUS("I/O error reading Primary Feature Data.\n");
		seq1 = 0;
		p1 = NULL;
	} else {
		/* Set up access to the buffer.  Extract the Master Sequence Number from the buffer. */
		p1 = new_data1;
		seq1 = ((struct link_table_first_sector *) p1)->Sequence_Number;
	}

	if (INIT_IO(link_partition, 0, linksector2, linknumsectors, new_data2)) {
		LOG_SERIOUS("I/O error reading Secondary Feature Data.\n");
		seq2 = 0;
		p2 = NULL;
	} else {
		/* Set up access to the second buffer.  Extract its copy of the Master Sequence Number. */
		p2 = new_data2;
		seq2 = ((struct link_table_sector *) p2)->Sequence_Number;
	}

	/* Validate both copies of the drive linking data one sector at a time. */
	for (i = 0; i < linknumsectors;
	     i++, p1 += OS2_BYTES_PER_SECTOR, p2 += OS2_BYTES_PER_SECTOR) {
		if ((seq1 > 0)
		    && validate_drivelinksector((struct link_table_sector *) p1, i,
						seq1)) {
			LOG_SERIOUS
			    ("The primary copy of the drive link data is invalid!  Sector %i is not valid\n",
			     i);
			seq1 = 0;
		}

		if ((seq2 > 0)
		    && validate_drivelinksector((struct link_table_sector *) p2, i,
						seq2)) {
			LOG_SERIOUS
			    ("The secondary copy of the drive link data is invalid!  Sector %i is not valid\n",
			     i);
			seq2 = 0;
		}

	}

	LOG_EVERYTHING("Primary Feature Data sequence # %i\n", seq1);
	LOG_EVERYTHING("Secondary Feature Data sequence # %i\n", seq2);

	/* Choose which copy of the drive linking data to use.  If both sequence numbers are 0, then both copies
	   of the drive linking data are bad.  If both are equal and non-zero, then both copies are good and it
	   really doesn't matter which one you choose.  Otherwise, choose the copy with the highest sequence number. */
	if (seq2 > seq1) {
		kfree(new_data1);
		return new_data2;
	} else {
		kfree(new_data2);
		if (!seq1) {
			kfree(new_data1);
			new_data1 = NULL;
		}
	}
	return new_data1;
}

/*
 * Function:  new_os2_bbr_data
 *
 *      Allocate space for OS/2 bad block relocation information.
 *      Read in and validate the information from disk.
 *      Note:  assumes 512 byte sectors.
 */
static char *
new_os2_bbr_data(u32 bbrsector1,
		 u32 bbrsector2,
		 u32 bbrnumsectors, struct evms_logical_node *bbr_partition)
{
	char *new_data1;	/* Buffer to hold the primary copy of the BBR data. */
	char *new_data2;	/* Buffer to hold the secondary copy of the BBR data. */
	char *p1;		/* Used to examine the individual sectors of BBR data within new_data1. */
	char *p2;		/* Used to examine the individual sectors of BBR data within new_data2. */
	int memsize = bbrnumsectors * OS2_BYTES_PER_SECTOR;
	u32 i, seq1, seq2;

	LOG_EVERYTHING("BBR Feature entry found.\n");

	/* Allocate memory for the buffers. */
	new_data1 = kmalloc(memsize, GFP_KERNEL);
	if (!new_data1) {
		LOG_SERIOUS("Could not allocate Primary BBR data\n");
		return NULL;
	}
	new_data2 = kmalloc(memsize, GFP_KERNEL);
	if (!new_data2) {
		LOG_SERIOUS("Could not allocate Secondary BBR data\n");
		kfree(new_data1);
		return NULL;
	}

	LOG_EVERYTHING("Primary Feature Data starts at RBA %i\n", bbrsector1);
	LOG_EVERYTHING("Secondary Feature Data starts at RBA %i\n", bbrsector2);

	/* Read in both copies of the BBR data. */
	if (INIT_IO(bbr_partition, 0, bbrsector1, bbrnumsectors, new_data1)) {
		LOG_SERIOUS("I/O error reading Primary Feature Data.\n");
		seq1 = 0;
		p1 = NULL;
	} else {
		/* Establish access to the first sector of the BBR data.  Extract the Master Sequence Number
		   for this copy of the BBR data.                                                             */
		p1 = new_data1;
		seq1 = ((LVM_BBR_Table_First_Sector *) p1)->Sequence_Number;
	}

	if (INIT_IO(bbr_partition, 0, bbrsector2, bbrnumsectors, new_data2)) {
		LOG_SERIOUS("I/O error reading Secondary Feature Data.\n");
		seq2 = 0;
		p2 = NULL;
	} else {
		/* Establish access to the first sector of the second copy of the BBR data.  Extract the 
		   Master Sequence Number for this copy of the BBR data.                                   */
		p2 = new_data2;
		seq2 = ((LVM_BBR_Table_Sector *) p2)->Sequence_Number;
	}

	/* Validate both copies of the BBR Data, one sector at a time. */
	for (i = 0; i < bbrnumsectors;
	     i++, p1 += OS2_BYTES_PER_SECTOR, p2 += OS2_BYTES_PER_SECTOR) {
		if ((seq1 > 0) && validate_bbrtablesector(p1, i, seq1)) {
			LOG_SERIOUS
			    ("The primary BBR data is invalid!  Sector %i is not valid\n",
			     i);
			seq1 = 0;
		}

		if ((seq2 > 0) && validate_bbrtablesector(p2, i, seq2)) {
			LOG_SERIOUS
			    ("The secondary BBR data is invalid!  Sector %i is not valid\n",
			     i);
			seq2 = 0;
		}

	}

	LOG_EVERYTHING("Primary Feature Data sequence # %i\n", seq1);
	LOG_EVERYTHING("Secondary Feature Data sequence # %i\n", seq2);

	/* Choose which copy of the BBR Data to use based upon the sequence number.  If both sequence numbers
	   are 0, then there is no valid BBR data.  If both are non-zero and equal, then it really doesn't 
	   matter which copy is used.  Otherwise, choose the copy with the highest sequence number.            */
	if (seq2 > seq1) {
		kfree(new_data1);
		return new_data2;
	} else {
		kfree(new_data2);
		if (!seq1) {
			kfree(new_data1);
			new_data1 = NULL;
		}
	}
	return new_data1;
}

/*
 * Function:  new_os2volume
 *
 *      Allocate space for a new OS/2 logical volume.
 *      Initialize the appropriate fields.
 */
static struct evms_logical_node *
new_os2volume(u32 volumeserial, char *volume_name)
{
	struct evms_logical_node *new_node;
	os2_volume_runtime_entry_t *cur_volume;

	if (evms_cs_allocate_logical_node(&new_node)) {
		LOG_SERIOUS("Could not allocate new volume\n");
		return NULL;
	}
	new_node->private =
	    kmalloc(sizeof (os2_volume_runtime_entry_t), GFP_KERNEL);
	if (!new_node->private) {
		LOG_SERIOUS("Could not allocate volume metadata\n");
		evms_cs_deallocate_logical_node(new_node);
		return NULL;
	}
	memset(new_node->private, 0, sizeof (os2_volume_runtime_entry_t));
	new_node->plugin = &plugin_header;
	new_node->system_id = LVM_PARTITION_INDICATOR;
	sprintf(new_node->name, "os2/%s", volume_name);
	cur_volume = (os2_volume_runtime_entry_t *) new_node->private;
	cur_volume->Volume_Serial_Number = volumeserial;
	cur_volume->Export_Needed = 1;

	if (os2lvm_nodes == NULL)
		os2lvm_nodes = new_node;

	// This is the first node discovered. Start the BBR thread.
	if (!BBR_Worker_Thread) {
		BBR_Worker_Thread =
		    evms_cs_register_thread(BBR_Worker, NULL, BBR_Worker_Name);
		if (!BBR_Worker_Thread) {
			kfree(new_node->private);
			evms_cs_deallocate_logical_node(new_node);
			os2lvm_nodes = NULL;
			return NULL;
		}
	} else {
		cur_volume =
		    (os2_volume_runtime_entry_t *) os2lvm_nodes->private;
		while (cur_volume->next_os2lvm_node)
			cur_volume =
			    (os2_volume_runtime_entry_t *) cur_volume->
			    next_os2lvm_node->private;
		cur_volume->next_os2lvm_node = new_node;
	}

	MOD_INC_USE_COUNT;

	return new_node;
}

/*
 * Function:  delete_os2lvm_volume
 *
 *      This function deletes the in-memory representation of an OS/2
 *      logical volume.
 */
static int
delete_os2lvm_volume(struct evms_logical_node *logical_node)
{
	struct os2_dl_entry *curdrvlink =
	    ((os2_volume_runtime_entry_t *) logical_node->private)->drive_link,
	    *nextdrvlink;
	os2_volume_runtime_entry_t *cur_volume, *next_volume;

	while (curdrvlink) {
		nextdrvlink = curdrvlink->next;
		delete_os2_drive_link(curdrvlink, 1);
		curdrvlink = nextdrvlink;
	}

	cur_volume = (os2_volume_runtime_entry_t *) os2lvm_nodes->private;
	if (os2lvm_nodes == logical_node)
		os2lvm_nodes = cur_volume->next_os2lvm_node;
	else {
		while (cur_volume->next_os2lvm_node) {
			next_volume =
			    (os2_volume_runtime_entry_t *) cur_volume->
			    next_os2lvm_node->private;
			if (cur_volume->next_os2lvm_node == logical_node) {
				cur_volume->next_os2lvm_node =
				    next_volume->next_os2lvm_node;
				break;
			}
		}
	}

	if (os2lvm_nodes == NULL) {
		// Just deleted the last os2 node. Stop the BBR thread.
		if (BBR_Worker_Thread) {
			evms_cs_unregister_thread(BBR_Worker_Thread);
			BBR_Worker_Thread = NULL;
		}
	}

	kfree(logical_node->private);
	evms_cs_deallocate_logical_node(logical_node);

	MOD_DEC_USE_COUNT;

	return 0;
}

/*
 * Function:  delete_os2_drive_link
 *
 *      This function deletes the drive link runtime structure and any
 *       other structures it points to.
 */
static int
delete_os2_drive_link(struct os2_dl_entry * drive_link,
		      int delete_link_partition)
{
	if (drive_link->link_data)
		kfree(drive_link->link_data);
	if (drive_link->bbr_data)
		kfree(drive_link->bbr_data);
	if (delete_link_partition)
		DELETE(drive_link->link_partition);
	kfree(drive_link);

	return 0;
}

// Consistency Checking Functions

/*
 * Function:  validate_signaturesector
 *
 *      This function checks the OS/2 LVM Signature Sector
 */
static int
validate_signaturesector(struct evms_logical_node *evms_partition,
			 LVM_Signature_Sector * signature_sector,
			 u32 sectorsize)
{
	u32 crc_hold, crc_new;

	/* In order for a signature sector to be considered valid, its signature and CRC must
	   be correct.  Also, OS/2 stores the starting LBA of the partition and the size of
	   the partition that this signature sector corresponds to.  These should be checked
	   as well.  However, since the starting LBA of the partition that this belongs to is
	   not available to us as part of an struct evms_logical_node, we can only check the size
	   of the partition against what is stored in the signature sector.                    */

	/* The signature used is in two parts.  Test the first part. */
	if (signature_sector->LVM_Signature1 != OS2LVM_PRIMARY_SIGNATURE) {
		LOG_EVERYTHING("Primary LVM Signature failed.\n");
		return 1;
	}

	/* Test the second part of the signature. */
	if (signature_sector->LVM_Signature2 != OS2LVM_SECONDARY_SIGNATURE) {
		LOG_EVERYTHING("Secondary LVM Signature failed.\n");
		return 1;
	}

	/* Calculate the CRC and compare it against the stored CRC. */
	crc_hold = signature_sector->Signature_Sector_CRC;
	signature_sector->Signature_Sector_CRC = 0;
	crc_new =
	    evms_cs_calculate_crc(EVMS_INITIAL_CRC, (void *) signature_sector,
				  sectorsize);
	if (crc_hold != crc_new) {
		LOG_EVERYTHING("Signature sector crc failed.\n");
		LOG_EVERYTHING("sector_crc == %x , calc_crc == %x \n", crc_hold,
			       crc_new);
		return 1;
	}
	// The partition size must == that found in the Signature Sector
	if (evms_partition->total_vsectors !=
	    signature_sector->Partition_Sector_Count) {
		LOG_EXTRA("Partition size is not valid\n");
		return 1;
	}

	return 0;
}

/*
 * Function:  validate_drivelinksector
 *
 *      This function checks the OS/2 LVM Drivelink Feature Sector
 */
static int
validate_drivelinksector(void *Sector_To_Validate,
			 int Sector_Index, u32 Master_Sequence_Number)
{
	u32 crc_hold, crc_new;
	struct link_table_first_sector *First_Sector =
	    (struct link_table_first_sector *) Sector_To_Validate;
	struct link_table_sector *Link_Sector =
	    (struct link_table_sector *) Sector_To_Validate;

	/* The OS/2 drive linking data covers several sectors.  The format of the first sector is slightly
	   different from the following sectors because it contains additional information about how many
	   drive links are actually in use.  The following sectors just contain portions of the drive link
	   table.  Each sector of OS/2 drive linking data contains a signature, crc, and sequence number
	   which must be validated.                                                                         */

	if (Sector_Index == 0) {

		/* Link Table Master Signature Check */
		if (LINK_TABLE_MASTER_SIGNATURE !=
		    First_Sector->Link_Table_Signature) {
			LOG_EVERYTHING
			    ("Link Table Master Signature Test failed.\n");
			return 1;
		}

		/* We will NOT check the sequence number here as the first sector of drive link data is the
		   source of the Master_Sequence_Number which was passed in to us.                           */

		/* Set up for the CRC Check */
		crc_hold = First_Sector->Link_Table_CRC;
		First_Sector->Link_Table_CRC = 0;
	} else {
		/* Link Table Internal Signature Check */
		if (LINK_TABLE_SIGNATURE != Link_Sector->Link_Table_Signature) {
			LOG_EVERYTHING
			    ("Link Table Internal Signature Test failed.\n");
			return 1;
		}

		/* Check the sequence number. */
		if (Master_Sequence_Number != Link_Sector->Sequence_Number) {
			LOG_EVERYTHING
			    ("Link Table Internal Sequence Number Test failed.\n");
			return 1;
		}

		/* Set up for the CRC Check */
		crc_hold = Link_Sector->Link_Table_CRC;
		Link_Sector->Link_Table_CRC = 0;
	}

	crc_new =
	    evms_cs_calculate_crc(EVMS_INITIAL_CRC, Sector_To_Validate,
				  OS2_BYTES_PER_SECTOR);
	if (crc_hold != crc_new) {
		LOG_EVERYTHING("Link Table crc failed.\n");
		LOG_EVERYTHING("sector_crc == %x , calc_crc == %x \n", crc_hold,
			       crc_new);
		return 1;
	}

	return 0;
}

/*
 * Function:  validate_bbrtablesector
 *
 *      This function checks the OS/2 LVM Bad Block Relocation Feature Sector
 */
static int
validate_bbrtablesector(void *Sector_To_Validate,
			int Sector_Index, u32 Master_Sequence_Number)
{
	u32 crc_hold, crc_new;
	LVM_BBR_Table_First_Sector *First_Sector =
	    (LVM_BBR_Table_First_Sector *) Sector_To_Validate;
	LVM_BBR_Table_Sector *BBR_Sector =
	    (LVM_BBR_Table_Sector *) Sector_To_Validate;

	/* The OS/2 bad block relocation (BBR) data covers several sectors.  The format of the first sector 
	   is different from the following sectors because it contains additional information about how many
	   relocations are actually in use and the size and location of the block of replacement sectors.  
	   The following sectors just contain portions of the BBR remap table.  Each sector of OS/2 BBR data 
	   contains a signature, crc, and sequence number which must be validated.                             */

	if (Sector_Index == 0) {

		/* BBR Table Master Signature Check */
		if (BBR_TABLE_MASTER_SIGNATURE != First_Sector->Signature) {
			LOG_EVERYTHING
			    ("BBR Table Master Signature Test failed.\n");
			return 1;
		}

		/* We will NOT check the sequence number here as the first sector of BBR data is the
		   source of the Master_Sequence_Number which was passed in to us.                      */

		/* Set up for the CRC Check */
		crc_hold = First_Sector->CRC;
		First_Sector->CRC = 0;

	} else {
		/* BBR Table Internal Signature Check */
		if (BBR_TABLE_SIGNATURE != BBR_Sector->Signature) {
			LOG_EVERYTHING
			    ("BBR Table Internal Signature Test failed.\n");
			return 1;
		}

		/* Check the sequence number. */
		if (Master_Sequence_Number != BBR_Sector->Sequence_Number) {
			LOG_EVERYTHING
			    ("BBR Table Internal Sequence Number Test failed.\n");
			return 1;
		}

		/* Set up for the CRC Check */
		crc_hold = BBR_Sector->CRC;
		BBR_Sector->CRC = 0;
	}

	crc_new =
	    evms_cs_calculate_crc(EVMS_INITIAL_CRC, Sector_To_Validate,
				  OS2_BYTES_PER_SECTOR);
	if (crc_hold != crc_new) {
		LOG_EVERYTHING("BBRTable crc failed.\n");
		LOG_EVERYTHING("sector_crc == %x , calc_crc == %x \n", crc_hold,
			       crc_new);
		return 1;
	}

	return 0;
}

/*
 * Function:  check_for_os2_bbr_relocations
 *
 *      This function checks the OS/2 LVM Bad Block Relocation Tables
 *       for any active relocation sectors.  The bbr table is reformatted in memory
 *       to make searches faster.
 *      Return values:  0 == no active relocations, 1 == contains active relocations
 */
static u32
check_for_os2_bbr_relocations(char *bbr_data_ptr)
{
	LVM_BBR_Feature *feature_data = (LVM_BBR_Feature *) bbr_data_ptr;

	if (feature_data->control.Table_Entries_In_Use) {
		LOG_EVERYTHING("There are %d active relocations.\n",
			       feature_data->control.Table_Entries_In_Use);
		return 1;
	}

	return 0;
}

/*
 * Function:  check_os2_volumes
 *
 *      This function performs a consistency check on all existing OS/2
 *        Logical Volumes.  The list of constituent partitions ( links )
 *        is checked and ordered according to the Link Table.  If any link
 *        is missing or inconsistent, the entire volume will be deleted.
 */
static int
check_os2_volumes(struct evms_logical_node **node_list)
{
	os2_volume_runtime_entry_t *cur_volume;
	os2_volume_runtime_entry_t *previous_volume;
	struct evms_logical_node *cur_node;
	struct evms_logical_node *previous_node = NULL;
	struct os2_dl_entry *link_list, *link_hold;
	struct link_table_first_sector *psector1;
	int i, rc = 0;
	u32 numlinks, countlinks, linkser;
	u32 Master_Sequence_Number;	/* Used to check whether or not all of the copies of Drive Linking data match. */
	u64 partition_offset;
	char *sect_ptr;

	LOG_ENTRY_EXIT("Checking OS/2 Logical Volumes\n");

	cur_node = os2lvm_nodes;

	while (cur_node) {
		cur_volume = (os2_volume_runtime_entry_t *) cur_node->private;
		link_list = NULL;
		if (!cur_volume->complete) {	/* need to verify this one  */
			cur_volume->complete = 1;
			LOG_EVERYTHING("Checking volume %s\n", cur_node->name);

			// Reset fields for sort operation
			cur_volume->size_in_sectors = 0;
			numlinks = cur_volume->drive_link_count;
			cur_volume->drive_link_count = 0;
			cur_node->total_vsectors = 0;
			link_list = cur_volume->drive_link;
			cur_volume->drive_link = NULL;

			// Access the link data to order the drive links
			psector1 =
			    (struct link_table_first_sector *) link_list->
			    link_data;
			Master_Sequence_Number = psector1->Sequence_Number;

			if (numlinks != psector1->Links_In_Use) {
				LOG_SERIOUS
				    ("Link Count mismatch vol=%i, table=%i\n",
				     numlinks, psector1->Links_In_Use);
				cur_volume->complete = 0;
				countlinks = 0;
			} else {
				if (numlinks > LINKS_IN_FIRST_SECTOR) {
					countlinks = LINKS_IN_FIRST_SECTOR;
					numlinks -= LINKS_IN_FIRST_SECTOR;
				} else {
					countlinks = numlinks;
					numlinks = 0;
				}

			}

			partition_offset = 0;
			for (i = 0;
			     (i < countlinks) && (cur_volume->complete == 1);
			     i++) {
				linkser =
				    psector1->Link_Table[i].
				    partition_serial;
				if ((link_hold =
				     find_link_data(&link_list, linkser))) {
					// Add this partition to its parent Volume
					add_os2link(link_hold, cur_node);
					LOG_EVERYTHING
					    ("Link start_RBA == "PFU64" , sector_count == "PFU64"\n",
					     partition_offset,
					     link_hold->sector_count);
					link_hold->start_sector =
					    partition_offset;
					partition_offset +=
					    link_hold->sector_count;
				} else {
					LOG_SERIOUS
					    ("Link Table entry %i metadata missing\n",
					     i);
					cur_volume->complete = 0;
					break;
				}
			}

			sect_ptr = (char *) psector1;

			while (numlinks && (cur_volume->complete == 1)) {
				if (numlinks > LINKS_IN_NEXT_SECTOR) {
					countlinks = LINKS_IN_NEXT_SECTOR;
					numlinks -= LINKS_IN_NEXT_SECTOR;
				} else {
					countlinks = numlinks;
					numlinks = 0;
				}
				sect_ptr += OS2_BYTES_PER_SECTOR;
				if (Master_Sequence_Number !=
				    ((struct link_table_sector *) sect_ptr)->
				    Sequence_Number) {
					cur_volume->complete = 0;
					LOG_SERIOUS
					    ("Bad Sequence Number for Drive Linking Metadata!\n");
				} else {
					for (i = 0; i < countlinks; i++) {
						linkser =
						    ((struct link_table_sector *)
						     sect_ptr)->Link_Table[i].
						    partition_serial;
						if ((link_hold =
						     find_link_data(&link_list,
								    linkser))) {
							// Add this partition to its parent Volume
							add_os2link(link_hold,
								    cur_node);
							LOG_EVERYTHING
							    ("Link start_RBA == "PFU64" , sector_count == "PFU64"\n",
							     partition_offset,
							     link_hold->
							     sector_count);
							link_hold->
							    start_sector =
							    partition_offset;
							partition_offset +=
							    link_hold->
							    sector_count;
						} else {
							LOG_SERIOUS
							    ("Link Table entry %i metadata missing\n",
							     i);
							cur_volume->complete =
							    0;
							break;
						}
					}
				}
			}
		}

		/* If the volume is complete we can export it for use. */
		if (cur_volume->complete && (link_list == NULL)) {

			// Link new volume into the node list
			if (cur_volume->Export_Needed &&
			    (!evms_cs_add_logical_node_to_list
			     (node_list, cur_node))
			    ) {
				rc++;
				cur_volume->Export_Needed = 0;
			}

			previous_node = cur_node;
			cur_node = cur_volume->next_os2lvm_node;
		} else {
			/* Remove the volume from os2lvm_nodes list and delete it. */
			if (previous_node != NULL) {

				previous_volume =
				    (os2_volume_runtime_entry_t *)
				    previous_node->private;
				previous_volume->next_os2lvm_node =
				    cur_volume->next_os2lvm_node;
				cur_volume->next_os2lvm_node = NULL;

				delete_os2lvm_volume(cur_node);

				cur_node = previous_volume->next_os2lvm_node;
			} else {
				previous_node = cur_volume->next_os2lvm_node;
				delete_os2lvm_volume(cur_node);
				cur_node = previous_node;
				previous_node = NULL;
				os2lvm_nodes = cur_node;
			}

			/* If any items remain in link_list, delete those as well. */
			while (link_list) {
				link_hold = link_list->next;
				delete_os2_drive_link(link_list, 1);
				link_list = link_hold;
			}

		}

	}

	LOG_ENTRY_EXIT("Finished Checking OS/2 Logical Volumes\n");

	return rc;
}

/* BBR_Transfer_IO
 *
 *	Transfer the responsibility for completing the specified IO from
 *      the thread that requested it to the BBR Worker Thread
 */
static void
BBR_Transfer_IO(struct transfer_record * Transfer_Record)
{
	unsigned long flags;
	int Wake_Worker_Thread = 0;	/* Assume that the worker is already awake. */

	spin_lock_irqsave(&BBR_Queue_Lock, flags);

	/* The BBR IO List is a singly linked list.  BBR_IO_List_Head points
	   to the first item in the list, and BBR_IO_List_Tail points to the
	   last item in the list.                                            */
	Transfer_Record->next = NULL;
	if (!BBR_IO_List_Tail) {	/* Empty list */
		BBR_IO_List_Head = Transfer_Record;
		Wake_Worker_Thread = 1;	/* Wake up the worker thread. */
	} else			/* Items already in the list. */
		BBR_IO_List_Tail->next = Transfer_Record;

	BBR_IO_List_Tail = Transfer_Record;

	spin_unlock_irqrestore(&BBR_Queue_Lock, flags);
	if (Wake_Worker_Thread)
		evms_cs_wakeup_thread(BBR_Worker_Thread);

	return;
}

/* OS2_DL_Callback
 * 
 * This is the callback function used when an I/O request has to be broken 
 * into two parts because it crosses a drive link boundary.
 *
 */
static void
OS2_DL_Callback(struct buffer_head *bh, int uptodate)
{

	struct tracking_record *Tracking_Record;
	struct buffer_head *Original;

	Tracking_Record = bh->b_private;

	/* Is this a read or a write? */
	if (Tracking_Record->link1_transfer_rec ||
	    Tracking_Record->link2_transfer_rec) {
		/* We have a write here.  Was it successful? */
		if (!uptodate) {
			/* Have we tried BBR yet? */
			if ((bh == Tracking_Record->link1_bh) &&
			    (!Tracking_Record->link1_bbr_attempted)) {
				/* Attempt BBR. */
				BBR_Transfer_IO(Tracking_Record->
						link1_transfer_rec);
				Tracking_Record->link1_bbr_attempted = 1;
				return;
			} else if ((bh == Tracking_Record->link2_bh) &&
				   (!Tracking_Record->link2_bbr_attempted)) {
				/* Attempt BBR. */
				BBR_Transfer_IO(Tracking_Record->
						link2_transfer_rec);
				Tracking_Record->link2_bbr_attempted = 1;
				return;
			}

		}

	}

	Tracking_Record->io_in_progress -= 1;
	if (Tracking_Record->io_in_progress) {
		Tracking_Record->up_to_date = uptodate;
	}
	Original = Tracking_Record->org_bh;

	if (!Tracking_Record->io_in_progress) {
		uptodate &= Tracking_Record->up_to_date;
		/* If this is a write, then Transfer Records will have been set up for both Link1 and Link2.
		   If the transfer records were used because of BBR, then the BBR worker thread will have
		   disposed of the transfer records.  If the transfer records were not used, then we must
		   dispose of them here to prevent memory leaks.                                             */
		if (Tracking_Record->link1_transfer_rec &&
		    (!Tracking_Record->link1_bbr_attempted)) {
			evms_cs_deallocate_to_pool(BBR_Transfer_Pool,
						   Tracking_Record->
						   link1_transfer_rec);
		}
		if (Tracking_Record->link2_transfer_rec &&
		    (!Tracking_Record->link2_bbr_attempted)) {
			evms_cs_deallocate_to_pool(BBR_Transfer_Pool,
						   Tracking_Record->
						   link2_transfer_rec);
		}
		evms_cs_deallocate_to_pool(evms_bh_pool,
					   Tracking_Record->link1_bh);
		evms_cs_deallocate_to_pool(evms_bh_pool,
					   Tracking_Record->link2_bh);
		evms_cs_deallocate_to_pool(DL_Tracking_Pool, Tracking_Record);
		Original->b_end_io(Original, uptodate);
	}

	return;
}

/* OS2_BBR_Write_Callback
 *
 *	This is the callback for normal write requests. Check for an error
 *	during the I/O, and send to the worker thread for processing if necessary.
 */
static void
OS2_BBR_Write_Callback(struct transfer_record * Transfer_Record,
		       struct buffer_head *bh, int uptodate, int *redrive)
{
	if (!uptodate) {
		BBR_Transfer_IO(Transfer_Record);
		*redrive = TRUE;
	} else {
		evms_cs_deallocate_to_pool(BBR_Transfer_Pool, Transfer_Record);
	}

	return;
}

/* Worker thread to handle:

   I/O to drive/partitions/objects where bad blocks are known to exist
   I/O to drive/partition/object where a new bad block has been discovered and the I/O must be redriven.
   
*/
static void
BBR_Worker(void *Not_Used)
{
	unsigned long flags;
	struct transfer_record *Current_IO;
	int complete;

	for (;;) {
		// Process bbr_io_list, one entry at a time.
		spin_lock_irqsave(&BBR_Queue_Lock, flags);

		/* Is there any work for us? */
		if (!BBR_IO_List_Head) {
			spin_unlock_irqrestore(&BBR_Queue_Lock, flags);
			break;	/* List empty - nothing to do. */
		}

		/* Get the IO to perform. */
		Current_IO = BBR_IO_List_Head;
		BBR_IO_List_Head = Current_IO->next;
		if (!BBR_IO_List_Head)
			BBR_IO_List_Tail = BBR_IO_List_Head;

		spin_unlock_irqrestore(&BBR_Queue_Lock, flags);

		/* Now lets process the I/O request. */
		complete = do_os2_bbr_io(Current_IO->Partition_Data,
					 Current_IO->Write_Flag,
					 Current_IO->bh->b_rsector,
					 Current_IO->bh->
					 b_size >> EVMS_VSECTOR_SIZE_SHIFT,
					 Current_IO->bh->b_data);

		/* We need to do the callback. */
		Current_IO->bh->b_end_io(Current_IO->bh, (complete == 0));

		/* Now cleanup */
		evms_cs_deallocate_to_pool(BBR_Transfer_Pool, Current_IO);
	}

	return;			/* Go to sleep. */

}

/* 
 * Sector_Is_Remapped
 *
 * This function returns 1 if the specified sector has been remapped, 0 if it has not
 *
 * If the sector has been remapped, then the new sector is returned in Replacement_Sector
 *
 */
static int
Sector_Is_Remapped(struct os2_dl_entry * io_dlentry,
		   u64 Source_Sector, u64 * Replacement_Sector)
{
	LVM_BBR_Feature *Feature_Data =
	    (LVM_BBR_Feature *) io_dlentry->bbr_data;
	unsigned int Sector_Index;	/* The BBR Table is spread across several sectors.  This tracks which sector we are looking at. */
	unsigned int BBR_Table_Index;	/* This tracks the actual entry in the BBR Table that we are examining. */
	unsigned int BBR_Table_Entries_In_Use =
	    Feature_Data->control.Table_Entries_In_Use;
	struct bbr_table_entry * table_entry;
	unsigned int guard1;

	/* Default value is no remap. */
	*Replacement_Sector = Source_Sector;

	do {
		guard1 = io_dlentry->guard1;	/* Lamport's Theorem */

		for (BBR_Table_Index = 0;
		     BBR_Table_Index < BBR_Table_Entries_In_Use;
		     BBR_Table_Index++) {
			Sector_Index =
			    BBR_Table_Index / BBR_TABLE_ENTRIES_PER_SECTOR;
			table_entry =
			    &(Feature_Data->remap[Sector_Index].
			      BBR_Table[BBR_Table_Index -
					(Sector_Index *
					 BBR_TABLE_ENTRIES_PER_SECTOR)]);
			if (table_entry->BadSector == (u32)Source_Sector) {
				*Replacement_Sector =
				    (u64)table_entry->ReplacementSector;
				break;
			}
		}

	} while (guard1 != io_dlentry->guard2);	/* Lamport's Theorem */

	if (*Replacement_Sector != Source_Sector)
		return 1;
	else
		return 0;
}

/*
 * Invalidate_Mapping
 *
 * This function either frees a replacement sector to be reused, or it 
 * marks the replacement sector as bad.
 *
 */
static void
Invalidate_Mapping(struct os2_dl_entry * dlentry,
		   u64 Source_Sector, int Replacement_Sector_Is_Bad)
{
	LVM_BBR_Feature *Feature_Data = (LVM_BBR_Feature *) dlentry->bbr_data;
	unsigned int Sector_Index;	/* The BBR Table is spread across several sectors.  This tracks which sector we are looking at. */
	unsigned int BBR_Table_Index;	/* This tracks the actual entry in the BBR Table that we are examining. */
	unsigned int BBR_Table_Entries_In_Use =
	    Feature_Data->control.Table_Entries_In_Use;
	struct bbr_table_entry * table_entry = NULL;

	/* Lock for the BBR Table. */
	down(&(dlentry->bbr_table_lock));

	/* Find the entry to invalidate. */
	for (BBR_Table_Index = 0; BBR_Table_Index < BBR_Table_Entries_In_Use;
	     BBR_Table_Index++) {
		Sector_Index = BBR_Table_Index / BBR_TABLE_ENTRIES_PER_SECTOR;
		table_entry =
		    &(Feature_Data->remap[Sector_Index].
		      BBR_Table[BBR_Table_Index -
				(Sector_Index * BBR_TABLE_ENTRIES_PER_SECTOR)]);
		if (table_entry->BadSector == Source_Sector) {
			break;
		}
	}

	/* Now that we have found the entry, we must invalidate it. */
	if (Replacement_Sector_Is_Bad) {
		table_entry->BadSector = (u32) - 1;
	}
	/* OS/2 supported a method for clearing out bad block remappings if the filesystem on the volume supported
	   the tracking of bad blocks.  We don't support that under Linux, so there is no else case here.           */

	/* Unlock the BBR Table */
	up(&(dlentry->bbr_table_lock));

	return;
}

/*
 * Create_New_struct bbr_table_entry
 *
 * Finds bad blocks within the range specified, allocates replacement sectors,
 * writes the data to the replacement sectors, and updates the BBR metadata on
 * disk to reflect the new mapping.  Returns 1 if successful, 0 otherwise.
 *
 */
static int
Create_New_BBR_Table_Entry(struct os2_dl_entry * dlentry,
			   u64 starting_lsn, unsigned int count, void *buffer)
{
	u64 lsn;
	struct bbr_table_entry *Table_Entry;
	unsigned int Sector_Index;
	unsigned int Table_Index;
	int rc;
	int rc2;
	u32 New_Sequence_Number;
	LVM_BBR_Feature *BBR_Data = (LVM_BBR_Feature *) dlentry->bbr_data;

	for (lsn = starting_lsn; lsn < (starting_lsn + count); lsn++) {
		rc = INIT_IO(dlentry->link_partition, 1, lsn, 1, buffer);
		while (rc) {

			/* Lock for the BBR Table. */
			down(&(dlentry->bbr_table_lock));

			/* Increment the second guard value. This will cause those reading the BBR Table to spin. */
			dlentry->guard2++;

			/* Ensure that the bbr active flag is set. */
			dlentry->bbr_is_active = 1;

			/* Allocate a replacement sector */
			if (BBR_Data->control.Table_Entries_In_Use <
			    BBR_Data->control.Table_Size) {
				Sector_Index =
				    BBR_Data->control.Table_Entries_In_Use /
				    BBR_TABLE_ENTRIES_PER_SECTOR;
				Table_Index =
				    BBR_Data->control.Table_Entries_In_Use %
				    BBR_TABLE_ENTRIES_PER_SECTOR;
				BBR_Data->control.Table_Entries_In_Use =
				    BBR_Data->control.Table_Entries_In_Use + 1;
				Table_Entry =
				    (struct bbr_table_entry *) & (BBR_Data->
							   remap[Sector_Index].
							   BBR_Table
							   [Table_Index]);
				Table_Entry->BadSector = lsn;
			} else {
				/* There are no more replacement sectors available!  Time to bail ... */
				up(&(dlentry->bbr_table_lock));
				return 0;
			}

			/* Now that we have a replacement sector, increment the first guard value.  This will free any 
			   threads reading the BBR Table.                                                                */
			dlentry->guard1++;

			/* Release the lock now that we have a replacement sector. */
			up(&(dlentry->bbr_table_lock));

			/* Test the replacement sector. */
			rc = INIT_IO(dlentry->link_partition, 1,
				     Table_Entry->ReplacementSector, 1, buffer);
			if (rc) {
				/* The replacement sector was bad.  Lets mark it bad in the table and try again. */
				Table_Entry->BadSector = (u32) - 1;
			}

		}		/* End of processing for the current sector. */

	}			/* end of loop to test each sector in the I/O and remap any bad ones found. */

	/* Need to write the modified BBR Table back to disk.  This includes updating the sequence numbers and CRCs. */

	/* Lock for the BBR Table. */
	down(&(dlentry->bbr_table_lock));

	/* Increment the sequence numbers. */
	New_Sequence_Number = BBR_Data->control.Sequence_Number + 1;
	BBR_Data->control.Sequence_Number = New_Sequence_Number;
	for (Sector_Index = 0;
	     Sector_Index < BBR_Data->control.Sectors_Per_Table;
	     Sector_Index++) {
		BBR_Data->remap[Sector_Index].Sequence_Number =
		    New_Sequence_Number;
	}

	/* Calculate the new CRC values. */
	BBR_Data->control.CRC = 0;
	BBR_Data->control.CRC =
	    evms_cs_calculate_crc(EVMS_INITIAL_CRC, &(BBR_Data->control),
				  OS2_BYTES_PER_SECTOR);
	for (Sector_Index = 0;
	     Sector_Index < BBR_Data->control.Sectors_Per_Table;
	     Sector_Index++) {
		BBR_Data->remap[Sector_Index].CRC = 0;
		BBR_Data->remap[Sector_Index].CRC =
		    evms_cs_calculate_crc(EVMS_INITIAL_CRC,
					  &(BBR_Data->remap[Sector_Index]),
					  OS2_BYTES_PER_SECTOR);
	}

	/* Now we must write the table back to the partition from whence it came. */

	/* Write the first copy. */
	rc = INIT_IO(dlentry->link_partition, 1, dlentry->bbr_lsn1,
		     dlentry->bbr_feature_size, BBR_Data);

	/* Write the second copy. */
	rc2 =
	    INIT_IO(dlentry->link_partition, 1, dlentry->bbr_lsn2,
		    dlentry->bbr_feature_size, BBR_Data);

	/* If both copies failed to reach the disk, then fail the I/O. */
	if (rc && rc2) {
		rc = 0;
	} else
		rc = 1;

	/* Unlock the BBR Table */
	up(&(dlentry->bbr_table_lock));

	/* Indicate success. */
	return rc;
}

/*
 * Clone_Bufferhead
 *
 * Prepares a usable copy of an existing bufferhead.
 *
 */
static void
Clone_Bufferhead(struct buffer_head *Source, struct buffer_head *Child)
{
	Child->b_next = NULL;
	Child->b_blocknr = Source->b_blocknr;
	Child->b_size = Source->b_size;
	Child->b_list = BUF_LOCKED;
	Child->b_dev = Source->b_dev;
	atomic_set(&Child->b_count, 0);
	atomic_set(&Child->b_count, atomic_read(&Source->b_count));
	Child->b_rdev = Source->b_rdev;
	Child->b_state = Source->b_state;
	Child->b_flushtime = 0;
	Child->b_next_free = NULL;
	Child->b_prev_free = NULL;
	Child->b_this_page = (struct buffer_head *) 1;
	Child->b_reqnext = NULL;
	Child->b_pprev = NULL;
	Child->b_data = Source->b_data;
	Child->b_page = Source->b_page;
	Child->b_end_io = Source->b_end_io;
	Child->b_private = Source->b_private;
	Child->b_rsector = Source->b_rsector;
	Child->b_inode_buffers.next = NULL;
	Child->b_inode_buffers.prev = NULL;
	return;
}
