/* -*- linux-c -*- */
/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * linux/drivers/evms/snapshot.c
 *
 * EVMS SnapShot Feature.
 *
 * This feature provides the ability to Snapshot ANY existing EVMS volume
 * (including compatibility) to a new EVMS volume that is created when the
 * SnapShot is enabled. This feature will appear in the call stack for both
 * the original and the snapshot volume.
 */

#define LOG_PREFIX "snapshot: "

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/compiler.h>
#include <linux/mempool.h>
#include <linux/proc_fs.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <linux/evms/evms.h>
#include <linux/evms/evms_snapshot.h>

static struct proc_dir_entry * snap_proc = NULL;
static unsigned int snapshot_count = 0; /* Number of active snapshots and originals. */

/* Memory pools. */
static kmem_cache_t * snap_page_slab = NULL;
static mempool_t * snap_page_pool = NULL;
static kmem_cache_t * snap_buffer_slab = NULL;
static mempool_t * snap_buffer_pool = NULL;
static kmem_cache_t * snap_async_org_io_slab = NULL;
static mempool_t * snap_async_org_io_pool = NULL;
static kmem_cache_t * snap_async_snap_io_slab = NULL;
static mempool_t * snap_async_snap_io_pool = NULL;
static kmem_cache_t * snap_hash_entry_slab = NULL;
static mempool_t * snap_hash_entry_pool = NULL;

#ifdef SNAPSHOT_DEBUG
static struct async_org_io * debug_async_org_io_list = NULL;
static spinlock_t debug_async_org_io_list_lock = SPIN_LOCK_UNLOCKED;
#endif

/* API prototypes */
static int snap_discover_volumes(struct evms_logical_node ** evms_node_list);
static int snap_delete_volume(struct evms_logical_node * node);
static void snap_read(struct evms_logical_node * node,
		      struct buffer_head * bh);
static void snap_write(struct evms_logical_node * node,
		       struct buffer_head * bh);
static int snap_init_io(struct evms_logical_node * node, int rw,
			u64 sect_nr, u64 num_sects, void * buf_addr);
static int snap_ioctl(struct evms_logical_node * node,
		      struct inode * inode, struct file * file,
		      unsigned int cmd, unsigned long arg);

/* Other functions that require prototypes. */
static int add_snapshot(struct evms_logical_node * node,
			struct snapshot_metadata * metadata,
			struct evms_logical_node ** evms_node_list);
static int snap_proc_read(char * page, char ** start, off_t off,
			  int count, int * eof, void * data);
static void snapshot_do_rollback(void * volume);
static void snap_async_io_thread(void * volume);
void snap_read_chunk_cb(struct buffer_head * bh, int uptodate);
void snap_write_chunk_cb(struct buffer_head * bh, int uptodate);
void snap_cow_table_cb(struct buffer_head * bh, int uptodate);

/* Snapshot plugin's function table and header. */
static struct evms_plugin_fops function_table = {
	.discover	= snap_discover_volumes,
	.delete		= snap_delete_volume,
	.read		= snap_read,
	.write		= snap_write,
	.init_io	= snap_init_io,
	.ioctl		= snap_ioctl
};

static struct evms_plugin_header plugin_header = {
	.id = SetPluginID(IBM_OEM_ID,
			  EVMS_ASSOCIATIVE_FEATURE,
			  EVMS_SNAPSHOT_FEATURE_ID),
	.version = {
		.major		= EVMS_SNAPSHOT_VERSION_MAJOR,
		.minor		= EVMS_SNAPSHOT_VERSION_MINOR,
		.patchlevel	= EVMS_SNAPSHOT_VERSION_PATCHLEVEL
	},
	.required_services_version = {
		.major		= EVMS_COMMON_SERVICES_MAJOR,
		.minor		= EVMS_COMMON_SERVICES_MINOR,
		.patchlevel	= EVMS_COMMON_SERVICES_PATCHLEVEL
	},
	.fops = &function_table
};

/**
 * convert_metadata - Perform endian conversion on a metadata sector.
 * @metadata: snapshot metadata sector
 **/
static void convert_metadata(struct snapshot_metadata * metadata)
{
	metadata->signature		= le32_to_cpup(&metadata->signature);
	metadata->CRC			= le32_to_cpup(&metadata->CRC);
	metadata->version.major		= le32_to_cpup(&metadata->version.major);
	metadata->version.minor		= le32_to_cpup(&metadata->version.minor);
	metadata->version.patchlevel	= le32_to_cpup(&metadata->version.patchlevel);
	metadata->flags			= le32_to_cpup(&metadata->flags);
	metadata->original_size		= le64_to_cpup(&metadata->original_size);
	metadata->lba_of_COW_table	= le64_to_cpup(&metadata->lba_of_COW_table);
	metadata->lba_of_first_chunk	= le64_to_cpup(&metadata->lba_of_first_chunk);
	metadata->chunk_size		= le32_to_cpup(&metadata->chunk_size);
	metadata->total_chunks		= le32_to_cpup(&metadata->total_chunks);
}

static void *slab_pool_alloc(int gfp_mask, void * data)
{
	return kmem_cache_alloc(data, gfp_mask);
}

static void slab_pool_free(void * ptr, void * data)
{
	kmem_cache_free(data, ptr);
}

/**
 * allocate_snapshot_hash_entry
 * @volume:	Snapshot volume to get a new entry for.
 * @org_chunk:	Number of original chunk.
 * @snap_chunk:	Number of remap chunk.
 * @chunk_state: see SNAP_CHUNK_*
 *
 * Get a snapshot_hash_entry from the pool and initialize. Accessing the
 * free_hash_list is safe, since we only call this function while holding
 * the snap_semaphore.
 */
static struct snapshot_hash_entry *
allocate_snapshot_hash_entry(struct snapshot_volume * volume,
			     u64 org_chunk,
			     u64 snap_chunk,
			     u32 chunk_state)
{
	struct snapshot_hash_entry * hash_entry;

	hash_entry = volume->free_hash_list;
	if (hash_entry) {
		volume->free_hash_list = hash_entry->next;
		hash_entry->org_chunk = org_chunk;
		hash_entry->snap_chunk = snap_chunk;
		hash_entry->chunk_state = chunk_state;
		hash_entry->snap_io = NULL;
		hash_entry->next = NULL;
		hash_entry->prev = NULL;
		spin_lock_init(&hash_entry->chunk_state_lock);
	} else {
		/* Should never happen, since hash entries are max
		 * allocated at discovery time.
		 */
		BUG();
	}

	return hash_entry;
}

/**
 * insert_snapshot_hash_entry
 *
 * Insert a new entry into a snapshot hash chain, immediately following the
 * specified entry. This function should not be used to add an entry into an
 * empty list, or as the first entry in an existing list. For that case, use
 * insert_snapshot_map_entry_at_head().
 */
static int insert_snapshot_hash_entry(struct snapshot_hash_entry * entry,
				      struct snapshot_hash_entry * base)
{
	entry->next = base->next;
	entry->prev = base;
	base->next = entry;
	if ( entry->next ) {
		entry->next->prev = entry;
	}
	return 0;
}

/**
 * insert_snapshot_hash_entry_at_head
 *
 * Insert a new entry into a snapshot chain as the first entry in the chain.
 */
static int insert_snapshot_hash_entry_at_head(struct snapshot_hash_entry * entry,
					      struct snapshot_hash_entry ** head)
{
	entry->next = *head;
	entry->prev = NULL;
	*head = entry;
	if ( entry->next ) {
		entry->next->prev = entry;
	}
	return 0;
}

/**
 * set_snapshot_flags
 * @snap_node:
 * @set_flag:	Flags to turn "on" in the metadata sector.
 * @unset_flag: Flags to turn "off" in the metadata sector.
 *
 * Set the flags field in the metadata and write the metadata sector to
 * the snapshot volume. The node passed in to this function should be the
 * "lower" of the snapshot nodes, meaning the one consumed by the snapshot
 * plugin, not the one exported from the plugin.
 *
 * Appropriate values for the two flag parameters are:
 *   EVMS_SNAPSHOT_DISABLED
 *   EVMS_SNAPSHOT_FULL
 *   EVMS_SNAPSHOT_ROLLBACK
 *   EVMS_SNAPSHOT_ROLLBACK_COMP
 */
static int set_snapshot_flags(struct evms_logical_node * snap_node,
			      u32 set_flag,
			      u32 unset_flag)
{
	unsigned char data[EVMS_VSECTOR_SIZE] = {0};
	struct snapshot_metadata * metadata = (struct snapshot_metadata*)data;

	/* Read the metadata sector */
	if ( INIT_IO(snap_node, 0, snap_node->total_vsectors-3, 1, data) ) {
		return -EIO;
	}

	/* Set the appropriate flags. Do endian conversion on the fly. */
	metadata->flags |= cpu_to_le32p(&set_flag);
	metadata->flags &= ~(cpu_to_le32p(&unset_flag));
	metadata->CRC = 0;
	metadata->CRC = cpu_to_le32(evms_cs_calculate_crc(EVMS_INITIAL_CRC,
							  metadata,
							  sizeof(struct snapshot_metadata)));

	/* Write the metadata sector back to the volume. */
	if ( INIT_IO(snap_node, 1, snap_node->total_vsectors-3, 1, data) ) {
		return -EIO;
	}

	return 0;
}

/**
 * disable_snapshot
 */
static void disable_snapshot(struct snapshot_volume * snap_volume,
			     int update_metadata)
{
	LOG_ERROR("Disabling snapshot volume '%s'.\n",
		snap_volume->exported_node->name);
	snap_volume->flags |= EVMS_SNAPSHOT_DISABLED;
	if ( update_metadata ) {
		set_snapshot_flags(snap_volume->logical_node,
				   EVMS_SNAPSHOT_DISABLED, 0);
	} else {
		snap_volume->flags |= EVMS_SNAPSHOT_DISABLED_PENDING;
		evms_cs_wakeup_thread(snap_volume->snapshot_org->async_io_thread);
	}
}

/**
 * snap_discover_volumes
 *
 * Inspect the global node list, looking for volumes with a valid
 * snapshot metadata sector.
 */
static int snap_discover_volumes(struct evms_logical_node ** evms_node_list)
{
	struct evms_logical_node * node, * next_node;
	struct snapshot_metadata * metadata = NULL;
	int org_crc, final_crc, rc = 0;

	MOD_INC_USE_COUNT;

	/* A buffer for reading the metadata. */
	metadata = kmalloc(EVMS_VSECTOR_SIZE, GFP_KERNEL);
	if (!metadata) {
		MOD_DEC_USE_COUNT;
		return -ENOMEM;
	}

	/* Check every node on the discovery list. */
	for ( node = *evms_node_list; node && !rc; node = next_node ) {
		next_node = node->next;

		/* This node must not be one we put back on the list already,
		 * and must have a feature header with snapshot's ID.
		 */
		if ( node->plugin->id == plugin_header.id ||
		     ! node->feature_header ||
		     node->feature_header->feature_id != plugin_header.id ) {
			continue;
		}

		/* Read third-to-last sector for the snapshot metadata. */
		rc = INIT_IO(node, 0, node->total_vsectors-3, 1, metadata);
		if (rc) {
			LOG_ERROR("IO error reading sector "PFU64" on '%s'.\n",
				node->total_vsectors-3, node->name);
			rc = -EVMS_FEATURE_FATAL_ERROR;
			evms_cs_remove_logical_node_from_list(evms_node_list,
							      node);
			DELETE(node);
			continue;
		}

		/* Check for a valid snapshot signature. */
		if ( le32_to_cpup(&metadata->signature) !=
		     EVMS_SNAPSHOT_SIGNATURE ) {
			continue;
		}
		evms_cs_remove_logical_node_from_list(evms_node_list, node);

		/* Check for a valid CRC. */
		org_crc = le32_to_cpup(&metadata->CRC);
		metadata->CRC = 0;
		final_crc = evms_cs_calculate_crc(EVMS_INITIAL_CRC, metadata,
						  sizeof(struct snapshot_metadata));
		if ( final_crc != org_crc ) {
			LOG_ERROR("CRC error in feature data on '%s'.\n",
				  node->name);
			rc = -EVMS_FEATURE_FATAL_ERROR;
			DELETE(node);
			continue;
		}

		/* Check for correct metadata version. */
		convert_metadata(metadata);
		if ( metadata->version.major > plugin_header.version.major ) {
			LOG_ERROR("ERROR: unsuppoprted metadata version on '%s'.\n",
				node->name);
			rc = -EVMS_FEATURE_FATAL_ERROR;
			DELETE(node);
			continue;
		}

		rc = add_snapshot(node, metadata, evms_node_list);
	}

	kfree(metadata);
	MOD_DEC_USE_COUNT;
	return rc;
}

/**
 * check_quiesce
 *
 * Make sure an original volume and all of its snapshots are quiesced.
 */
static int check_quiesce(struct snapshot_volume * org_volume)
{
	struct snapshot_volume * next_vol;
	
	for ( next_vol = org_volume;
	      next_vol;
	      next_vol = next_vol->snapshot_next ) {
		if ( ! (next_vol->flags & EVMS_SNAPSHOT_QUIESCED) ) {
			LOG_ERROR("Can't delete snapshot, volume '%s' not quiesced.\n",
				next_vol->logical_node->name);
			return -EBUSY;
		}
	}
	return 0;
}

/**
 * remove_snapshot_from_chain
 *
 * Remove the specified snapshot volume from its original's chain of snapshots.
 */
static int remove_snapshot_from_chain(struct snapshot_volume * snap_volume)
{
	struct snapshot_volume ** p_volume;

	if ( snap_volume->snapshot_org ) {
		down_write(&snap_volume->snapshot_org->snap_semaphore);
		for ( p_volume = &snap_volume->snapshot_org->snapshot_next;
		      *p_volume;
		      p_volume = &(*p_volume)->snapshot_next ) {
			if ( *p_volume == snap_volume ) {
				*p_volume = (*p_volume)->snapshot_next;
				break;
			}
		}
		up_write(&snap_volume->snapshot_org->snap_semaphore);
	}
	snap_volume->snapshot_org = NULL;
	snap_volume->snapshot_next = NULL;
	return 0;
}

/**
 * delete_snapshot_hash_chain
 *
 * Delete all items in a single chain in the hash table.
 */
static int delete_snapshot_hash_chain(struct snapshot_hash_entry * head)
{
	struct snapshot_hash_entry * next;

	while (head) {
		next = head->next;
		mempool_free(head, snap_hash_entry_pool);
		head = next;
	}
	return 0;
}		

/**
 * snapshot_delete_pools
 *
 * Delete all memory pools after all snapshots have been deleted.
 * Also shutdown the daemon thread.
 */
static void snapshot_delete_pools(void)
{
	/* The pool of data pages. */
	if (snap_page_slab) {
		if (snap_page_pool) {
			mempool_destroy(snap_page_pool);
			snap_page_pool = NULL;
		}
		kmem_cache_destroy(snap_page_slab);
		snap_page_slab = NULL;
	}

	/* The pool of snap_io_buffer's. */
	if (snap_buffer_slab) {
		if (snap_buffer_pool) {
			mempool_destroy(snap_buffer_pool);
			snap_buffer_pool = NULL;
		}
		kmem_cache_destroy(snap_buffer_slab);
		snap_buffer_slab = NULL;
	}

	/* The pool of async_org_io's. */
	if (snap_async_org_io_slab) {
		if (snap_async_org_io_pool) {
			mempool_destroy(snap_async_org_io_pool);
			snap_async_org_io_pool = NULL;
		}
		kmem_cache_destroy(snap_async_org_io_slab);
		snap_async_org_io_slab = NULL;
	}

	/* The pool of async_snap_io's. */
	if (snap_async_snap_io_slab) {
		if (snap_async_snap_io_pool) {
			mempool_destroy(snap_async_snap_io_pool);
			snap_async_snap_io_pool = NULL;
		}
		kmem_cache_destroy(snap_async_snap_io_slab);
		snap_async_snap_io_slab = NULL;
	}

	/* The pool of hash table entries. */
	if (snap_hash_entry_slab) {
		if (snap_hash_entry_pool) {
			mempool_destroy(snap_hash_entry_pool);
			snap_hash_entry_pool = NULL;
		}
		kmem_cache_destroy(snap_hash_entry_slab);
		snap_hash_entry_slab = NULL;
	}
}

/**
 * snapshot_create_pools
 *
 * Allocate all of the memory pools when the first snapshot is created.
 * Also start up the daemon thread for processing async I/O's.
 */
static int snapshot_create_pools(void)
{
	/* Pool of data pages. */
	if (!snap_page_slab) {
		snap_page_slab = kmem_cache_create("snap_page_slab",
						   PAGE_SIZE, 0,
						   SLAB_HWCACHE_ALIGN,
						   NULL, NULL);
		if (snap_page_slab) {
			snap_page_pool = mempool_create(1, slab_pool_alloc,
							slab_pool_free,
							snap_page_slab);
		}
	}

	/* Pool of snap_io_buffer's. */
	if (!snap_buffer_slab) {
		snap_buffer_slab = kmem_cache_create("snap_bh_slab",
						     sizeof(struct snap_io_buffer),
						     0, SLAB_HWCACHE_ALIGN,
						     NULL, NULL);
		if (snap_buffer_slab) {
			snap_buffer_pool = mempool_create(1, slab_pool_alloc,
							  slab_pool_free,
							  snap_buffer_slab);
		}
	}

	/* Pool of async_org_io's. */
	if (!snap_async_org_io_slab) {
		snap_async_org_io_slab = kmem_cache_create("async_org_io_slab",
							   sizeof(struct async_org_io),
							   0, SLAB_HWCACHE_ALIGN,
							   NULL, NULL);
		if (snap_async_org_io_slab) {
			snap_async_org_io_pool = mempool_create(1, slab_pool_alloc,
								slab_pool_free,
								snap_async_org_io_slab);
		}
	}

	/* Pool of async_snap_io's. */
	if (!snap_async_snap_io_slab) {
		snap_async_snap_io_slab = kmem_cache_create("async_snap_io_slab",
							    sizeof(struct async_snap_io),
							    0, SLAB_HWCACHE_ALIGN,
							    NULL, NULL);
		if (snap_async_snap_io_slab) {
			snap_async_snap_io_pool = mempool_create(1, slab_pool_alloc,
								 slab_pool_free,
								 snap_async_snap_io_slab);
		}
	}

	/* Pool of hash table entries. */
	if (!snap_hash_entry_slab) {
		snap_hash_entry_slab = kmem_cache_create("snap_hash_slab",
							 sizeof(struct snapshot_hash_entry),
							 0, SLAB_HWCACHE_ALIGN,
							 NULL, NULL);
		if (snap_hash_entry_slab) {
			snap_hash_entry_pool = mempool_create(1, slab_pool_alloc,
							      slab_pool_free,
							      snap_hash_entry_slab);
		}
	}

	if ( ! snap_page_slab          || ! snap_page_pool          ||
	     ! snap_buffer_slab        || ! snap_buffer_pool        ||
	     ! snap_async_org_io_slab  || ! snap_async_org_io_pool  ||
	     ! snap_async_snap_io_slab || ! snap_async_snap_io_pool ||
	     ! snap_hash_entry_slab    || ! snap_hash_entry_pool    ) {
		LOG_CRITICAL("No memory available to create snapshot pools.\n");
		snapshot_delete_pools();
		return -ENOMEM;
	}

	return 0;
}

/**
 * snap_delete_volume
 *
 * Delete the in-memory representation of a volume. The specified node
 * can actually be either a snapshot or an original. Deleting a snapshot
 * causes it to be removed from its original's chain of snapshots.
 *
 * For async snapshots, we will need to flush the COW table and mark the
 * snapshot clean in the metadata.
 */
static int snap_delete_volume(struct evms_logical_node * node)
{
	struct snapshot_volume * volume = node->private;
	struct snapshot_volume * org_volume = volume->snapshot_org;
	struct snapshot_volume * next_vol;
	int i, rc = 0;

	/* Don't delete a snapshot that's rolling back. */
	if ( volume->flags & EVMS_SNAPSHOT_ROLLBACK &&
	     ! (volume->flags & EVMS_SNAPSHOT_DISABLED) ) {
		LOG_ERROR("Can't delete '%s' during snapshot rollback.",
			  node->name);
		return -EBUSY;
	}

	/* Delete the instance data. */
	if (volume) {
		if ( volume->flags & EVMS_SNAPSHOT ) {
			/* This node is a snapshot. Check that this snapshot and
			 * its original have been quiesced. For async snapshots,
			 * make sure there are no outstanding remaps in
			 * progress. Then remove it from the original's chain of
			 * snapshots.
			 */
			if ( ! (volume->flags & EVMS_SNAPSHOT_QUIESCED) ) {
				LOG_ERROR("Can't delete snapshot, snapshot volume '%s' not quiesced.\n",
					  volume->logical_node->name);
				return -EBUSY;
			}
			if ( org_volume &&
			     ! (org_volume->flags & EVMS_SNAPSHOT_QUIESCED) ) {
				LOG_ERROR("Can't delete snapshot, original volume '%s' not quiesced.\n",
					  org_volume->logical_node->name);
				return -EBUSY;
			}

			remove_snapshot_from_chain(volume);

			/* If we just deleted the only/last snapshot for this
			 * original, the original will not be modified. It is
			 * the engine's responsibility to delete the original
			 * and rediscover in order to clear it of its snapshot
			 * information. Even if that doesn't happen, the state
			 * of the kernel will still be safe. I/O's coming into
			 * this plugin for the original will just be passed
			 * down without any other action or modification.
			 */

			/* Unregister the proc-fs entry for this node. */
			if (snap_proc) {
				remove_proc_entry(node->volume_info->volume_name,
						  snap_proc);
			}
		} else {
			/* This is an original. It's the engine's responsibility
			 * to delete all snapshots before deleting an original.
			 * Otherwise, a snapshot could be left pointing to an
			 * original that no longer exists. Thus, we just need to
			 * make sure there are no snapshots in the chain.
			 */
			rc = check_quiesce(volume);
			if (rc) {
				return rc;
			}

			/* Shut down the async I/O thread. */
			if (volume->async_io_thread) {
				evms_cs_unregister_thread(volume->async_io_thread);
			}

			/* Loop through all snapshots left on this original,
			 * and NULL out their org pointer, in case they don't
			 * get deleted.
			 */
			for ( next_vol = volume->snapshot_next; next_vol;
			      next_vol = next_vol->snapshot_next ) {
				next_vol->snapshot_org = NULL;
			}
		}

		/* Free up all memory used by the instance data, including
		 * the underlying node, the hash table, and the data buffer.
		 */
		if (volume->logical_node) {
			rc = DELETE(volume->logical_node);
			if (rc) {
				return rc;
			}
		}
		if (volume->snapshot_map) {
			/* Delete all of the hash chains,
			 * then the actual table.
			 */
			for ( i = 0; i < volume->hash_table_size; i++ ) {
				delete_snapshot_hash_chain(volume->snapshot_map[i]);
			}
			delete_snapshot_hash_chain(volume->free_hash_list);
			vfree(volume->snapshot_map);
		}
		if (volume->chunk_data_buffer) {
			kfree(volume->chunk_data_buffer);
		}
		if (volume->rollback_thread) {
			evms_cs_unregister_thread(volume->rollback_thread);
		}

		kfree(volume);
	}

	evms_cs_deallocate_logical_node(node);
	snapshot_count--;

	/* If there are no more snapshot objects, free up the memory pools. */
	if ( snapshot_count == 0 ) {
		snapshot_delete_pools();
	}

	MOD_DEC_USE_COUNT;

	return 0;
}

/**
 * search_snapshot_hash_chain
 *
 * Search the hash chain that is anchored at the specified head pointer. If the
 * chunk number is found, a pointer to that entry in the chain is set, and a 1
 * is returned. If the chunk is not found, a pointer to the previous entry is
 * set and 0 is returned. If the return pointer is NULL, this means either the
 * list is empty, or the specified sector should become the first list item.
 */
static int search_snapshot_hash_chain(u64 chunk,
				      struct snapshot_hash_entry * head,
				      struct snapshot_hash_entry ** result)
{
	struct snapshot_hash_entry * curr = head;
	struct snapshot_hash_entry * prev = head;
	while ( curr && curr->org_chunk < chunk ) {
		prev = curr;
		curr = curr->next;
	}
	if (!curr) {
		/* Either an empty chain or went off the end of the chain. */
		*result = prev;
		return 0;
	} else if ( curr->org_chunk != chunk ) {
		*result = curr->prev;
		return 0;
	} else {
		*result = curr;
		return 1;
	}
}

/**
 * snapshot_remap_chunk
 *
 * Perform a sector remap on a snapshot volume. This should be called from the
 * I/O read path, It first determines the base sector of the chunk containing
 * the specified sector, and saves the remainder. Then it performs a search
 * through the snapshot map for the specified volume. If a match is found, the
 * sector number is changed to the new value. If no match is found, the value
 * is left the same, meaning the read should proceed down the original volume.
 */
static int snapshot_remap_chunk(struct snapshot_volume * snap_volume,
				struct buffer_head * bh)
{
	struct snapshot_hash_entry * result;
	u64 chunk, sector = bh->b_rsector;
	unsigned long remainder, hash_value;
	unsigned long flags, queued = FALSE;

	remainder = sector & (u64)(snap_volume->chunk_size - 1);
	chunk = sector >> snap_volume->chunk_shift;
	hash_value = ((unsigned long)chunk) % snap_volume->hash_table_size;

	if ( search_snapshot_hash_chain(chunk,
					snap_volume->snapshot_map[hash_value],
					&result) ) {
		bh->b_rsector = (result->snap_chunk << snap_volume->chunk_shift)
				 + remainder;
		if ( result->chunk_state != SNAP_CHUNK_COPIED ) {
			/* If this chunk is in the middle of being copied,
			 * place this request on the pending list.
			 */
			spin_lock_irqsave(&result->chunk_state_lock, flags);
			if ( result->chunk_state != SNAP_CHUNK_COPIED ) {
				bh->b_reqnext = result->snap_io->pending_reads;
				result->snap_io->pending_reads = bh;
				if (!result->snap_io->dev) {
					result->snap_io->dev = bh->b_rdev;
				}
				evms_cs_volume_request_in_progress(result->snap_io->dev,
								   +1, NULL);
				queued = TRUE;
			}
			spin_unlock_irqrestore(&result->chunk_state_lock, flags);
		}

		if (queued) {
			return -1;
		} else {
			return 1;
		}
	}
	return 0;
}

/**
 * snap_read
 */
static void snap_read(struct evms_logical_node * node,
		      struct buffer_head * bh)
{
	struct snapshot_volume * volume = node->private;
	u64 alignment;
	int rc;

	/* Size check. */
	if ( bh->b_rsector + (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) >
	     node->total_vsectors ) {
		bh->b_end_io(bh, 0);
		return;
	}

	/* Can't read if rollback is in progress. */
	if ( volume->flags & EVMS_SNAPSHOT_ROLLBACK ) {
		LOG_ERROR("Cannot read from snapshot '%s' during rollback.\n",
			  volume->logical_node->name);
		bh->b_end_io(bh, 0);
		return;
	}

	/* On a read to the original, we can just pass it through completely
	 * untouched. Only reads to the snapshot can be remapped.
	 */
	if ( volume->flags & EVMS_SNAPSHOT_ORG ) {
		R_IO(volume->logical_node, bh);
		return;
	}

	/* Lock the snapshot before processing the request. */
	down_read(&volume->snap_semaphore);

	/* Make sure the snapshot is not full/disabled, and that
	 * the original is present.
	 */
	if ( (volume->flags & (EVMS_SNAPSHOT_DISABLED|EVMS_SNAPSHOT_FULL)) ||
	     (! volume->snapshot_org) ) {
		bh->b_end_io(bh, 0);
		up_read(&volume->snap_semaphore);
		return;
	}

	/* Check for unaligned I/O. This is mostly to prevent XFS from
	 * sending a request that spans a chunk.
	 */
	alignment = bh->b_rsector;
	alignment <<= EVMS_VSECTOR_SIZE_SHIFT;
	if ( unlikely(alignment & (bh->b_size - 1)) ) {
		LOG_ERROR("Unaligned request [rsector(%lx), size(%x)] rejected on snapshot %s.\n",
			  bh->b_rsector, bh->b_size, node->name);
		bh->b_end_io(bh, 0);
		up_read(&volume->snap_semaphore);
		return;
	}

	/* Check if this sector has been remapped. */
	rc = snapshot_remap_chunk(volume, bh);
	if ( rc > 0 ) {
		/* Sector was remapped. Send IO to the snapshot. */
		up_read(&volume->snap_semaphore);
		R_IO(volume->logical_node, bh);
	} else if ( rc < 0 ) {
		/* Sector was remapped but queued to be driven later. */
		up_read(&volume->snap_semaphore);
	} else  {
		/* Has not been remapped. Send IO to the original. */
		R_IO(volume->snapshot_org->logical_node, bh);
		up_read(&volume->snap_semaphore);
	}
}


/********** Asynchronous Snapshot I/O Code **********/


/**
 * snap_deallocate_buffer
 */
static void snap_deallocate_buffer(struct snap_io_buffer * buf,
				   int deallocate_page)
{
	if (buf) {
		if (deallocate_page) {
			mempool_free(buf->bh->b_data, snap_page_pool);
		}
		mempool_free(buf, snap_buffer_pool);
	}
}

/**
 * snap_allocate_buffer
 *
 * Allocate a snap_io_buffer and a data page from the respective memory
 * pools. Initialize as appropriate.
 */
static struct snap_io_buffer * snap_allocate_buffer(int allocate_page)
{
	struct snap_io_buffer * buf;
	struct buffer_head * bh;

	/* Grab a snap_io_buffer from the pool. */
	buf = mempool_alloc(snap_buffer_pool, GFP_NOIO);
	if (!buf) {
		return NULL;
	}
	memset(buf, 0, sizeof(struct snap_io_buffer));
	bh = buf->bh = &buf->_bh;

	/* Grab a data page from the pool. */
	if (allocate_page) {
		bh->b_data = mempool_alloc(snap_page_pool, GFP_NOIO);
		if (!bh->b_data) {
			snap_deallocate_buffer(buf, FALSE);
			return NULL;
		}
		bh->b_page = virt_to_page(bh->b_data);
	}

	/* Initialize the rest of the buffer. */
	bh->b_size = PAGE_SIZE;
	bh->b_list = BUF_LOCKED;
	atomic_set(&bh->b_count, 1);
	bh->b_this_page = (struct buffer_head *)1;
	bh->b_private = buf;
	set_bit(BH_Dirty, &bh->b_state);
	set_bit(BH_Lock, &bh->b_state);
	set_bit(BH_Req, &bh->b_state);
	set_bit(BH_Mapped, &bh->b_state);
	set_bit(BH_Uptodate, &bh->b_state);
	init_waitqueue_head(&bh->b_wait);
	INIT_LIST_HEAD(&buf->chunk_write_list);

	return buf;
}

/**
 * snap_deallocate_buffer_list
 *
 * Free each buffer in the specified list.
 */
static void snap_deallocate_buffer_list(struct snap_io_buffer * buf_list,
					int deallocate_pages)
{
	struct snap_io_buffer * buf, * buf_next;

	for ( buf = buf_list; buf; buf = buf_next ) {
		buf_next = buf->buffer_next;
		snap_deallocate_buffer(buf, deallocate_pages);
	}
}

/**
 * snap_allocate_buffer_list
 *
 * Allocate a list of snap_io_buffer's which will be used to copy a chunk
 * from the original to the snapshot.
 */
static struct snap_io_buffer *
snap_allocate_buffer_list(unsigned int count,
			  u64 starting_lba,
			  void (*callback)(struct buffer_head *, int),
			  int allocate_pages)
{
	struct snap_io_buffer * buf, * head = NULL;
	struct snap_io_buffer ** tail = &head;
	unsigned int i;

	for ( i = 0; i < count; i++ ) {
		/* Get a buffer from the pool. */
		buf = snap_allocate_buffer(allocate_pages);
		if (!buf) {
			snap_deallocate_buffer_list(head, allocate_pages);
			return NULL;
		}

		/* Set the callback function and the sector value. */
		buf->bh->b_end_io = callback;
		buf->bh->b_rsector = starting_lba + i *
				      (PAGE_SIZE >> EVMS_VSECTOR_SIZE_SHIFT);

		/* Add this buffer to the list to return. */
		*tail = buf;
		tail = &buf->buffer_next;
	}

	return head;
}

/**
 * deallocate_async_snap_io
 */
static void deallocate_async_snap_io(struct async_snap_io * async_snap_io)
{
	DEBUG_CHECK_SNAP_IO(async_snap_io);

	snap_deallocate_buffer(async_snap_io->cow_table_buffer, TRUE);
	snap_deallocate_buffer_list(async_snap_io->copy_buffers, FALSE);
	mempool_free(async_snap_io, snap_async_snap_io_pool);
}

/**
 * allocate_async_snap_io
 * @snap_volume:	The snapshot volume this chunk belongs to.
 * @hash_entry:		The entry in the hash table representing this chunk.
 * @async_org_io:	The parent async I/O structure.
 * @snap_chunk_lba:	The starting LBA on the snapshot for this chunk.
 * @buffer_count:	The number of buffers needed to copy this chunk.
 *
 * Allocate an async_snap_io structure from the pool and initialize.
 * Create a list of buffer heads to use for the copy.
 */
static struct async_snap_io *
allocate_async_snap_io(struct snapshot_volume * snap_volume,
		       struct snapshot_hash_entry * hash_entry,
		       struct async_org_io * async_org_io,
		       u64 snap_chunk_lba,
		       unsigned int buffer_count)
{
	struct async_snap_io * async_snap_io;

	async_snap_io = mempool_alloc(snap_async_snap_io_pool, GFP_NOIO);
	if (async_snap_io) {
		memset(async_snap_io, 0, sizeof(struct async_snap_io));
		async_snap_io->snap_volume = snap_volume;
		async_snap_io->hash_table_entry = hash_entry;
		async_snap_io->org_io = async_org_io;
		INIT_LIST_HEAD(&async_snap_io->snap_pending_io_list);
		INIT_LIST_HEAD(&async_snap_io->cow_write_list);
		atomic_set(&async_snap_io->write_count, buffer_count);

		async_snap_io->cow_table_buffer = snap_allocate_buffer(TRUE);
		if (async_snap_io->cow_table_buffer) {
			/* The buffer for the COW table needs to be adjusted. */
			struct snap_io_buffer * buf = async_snap_io->cow_table_buffer;
			buf->bh->b_size = EVMS_VSECTOR_SIZE;
			buf->bh->b_end_io = snap_cow_table_cb;
			buf->buffer_private = async_snap_io;

			async_snap_io->copy_buffers =
				snap_allocate_buffer_list(buffer_count,
							  snap_chunk_lba,
							  snap_write_chunk_cb,
							  FALSE);
			if (!async_snap_io->copy_buffers) {
				deallocate_async_snap_io(async_snap_io);
				async_snap_io = NULL;
			}
		} else {
			deallocate_async_snap_io(async_snap_io);
			async_snap_io = NULL;
		}
	}
	return async_snap_io;
}

/**
 * deallocate_async_org_io
 */
static void deallocate_async_org_io(struct async_org_io * async_org_io)
{
	DEBUG_REMOVE_ORG_IO_FROM_LIST(async_org_io);

	snap_deallocate_buffer_list(async_org_io->copy_buffers, TRUE);
	mempool_free(async_org_io, snap_async_org_io_pool);
}

/**
 * allocate_async_org_io
 *
 * Allocate an async_org_io structure from the pool and initialize.
 * Create a list of buffer heads to use for the copy.
 */
static struct async_org_io *
allocate_async_org_io(struct snapshot_volume * org_volume,
		      u64 org_chunk_lba,
		      unsigned int buffer_count)
{
	struct async_org_io * async_org_io;

	async_org_io = mempool_alloc(snap_async_org_io_pool, GFP_NOIO);
	if (async_org_io) {
		DEBUG_ADD_ORG_IO_TO_LIST(async_org_io);

		memset(async_org_io, 0, sizeof(struct async_org_io));
		async_org_io->org_volume = org_volume;
		spin_lock_init(&async_org_io->pending_writes_lock);
		INIT_LIST_HEAD(&async_org_io->org_pending_io_list);
		atomic_set(&async_org_io->copy_count, 0);
		atomic_set(&async_org_io->ref_count, 1);

		async_org_io->copy_buffers =
			snap_allocate_buffer_list(buffer_count,
						  org_chunk_lba,
						  snap_read_chunk_cb,
						  TRUE);
		if (!async_org_io->copy_buffers) {
			deallocate_async_org_io(async_org_io);
			async_org_io = NULL;
		}
	}
	return async_org_io;
}

/**
 * deallocate_async_io
 *
 * This function deletes the entire async I/O structure, including the
 * async_org_io, all async_snap_io's, and all buffer heads and pages.
 */
static void deallocate_async_io(struct async_org_io * async_org_io)
{
	struct async_snap_io * async_snap_io, * next_snap_io;

	for ( async_snap_io = async_org_io->snap_io_list;
	      async_snap_io;
	      async_snap_io = next_snap_io ) {
		next_snap_io = async_snap_io->snap_io_list_next;
		deallocate_async_snap_io(async_snap_io);
	}
	deallocate_async_org_io(async_org_io);
}

/**
 * process_org_pending_io_list
 *
 * Grab the first item from the org_pending_io_list and send all
 * waiting write requests to the original.
 */
static void process_org_pending_io_list(struct snapshot_volume * org_volume,
					int * done)
{
	struct async_org_io * async_org_io;
	struct buffer_head * bh;
	unsigned long flags;

	spin_lock_irqsave(&org_volume->org_pending_io_list_lock, flags);
	if ( list_empty(&org_volume->org_pending_io_list) ) {
		spin_unlock_irqrestore(&org_volume->org_pending_io_list_lock,
				       flags);
		*done = TRUE;
	} else {
		async_org_io = ORG_PENDING_IO_ENTRY(org_volume->org_pending_io_list.next);
		list_del(&async_org_io->org_pending_io_list);
		spin_unlock_irqrestore(&org_volume->org_pending_io_list_lock,
				       flags);

		for ( bh = async_org_io->pending_writes; bh;
		      bh = async_org_io->pending_writes ) {
			async_org_io->pending_writes = bh->b_reqnext;
			bh->b_reqnext = NULL;
			W_IO(async_org_io->org_volume->logical_node, bh);
			evms_cs_volume_request_in_progress(async_org_io->dev,
							   -1, NULL);
		}

		if ( atomic_dec_and_test(&async_org_io->ref_count) ) {
			deallocate_async_io(async_org_io);
		}
		*done = FALSE;
	}
}


/**
 * process_snap_pending_io_list
 *
 * Grab the first item from the snap_pending_io_list and send
 * all waiting read and write requests to the snapshot.
 */
static void process_snap_pending_io_list(struct snapshot_volume * org_volume,
					 int * done)
{
	struct async_snap_io * async_snap_io;
	struct buffer_head * bh;
	unsigned long flags;

	spin_lock_irqsave(&org_volume->snap_pending_io_list_lock, flags);
	if ( list_empty(&org_volume->snap_pending_io_list) ) {
		spin_unlock_irqrestore(&org_volume->snap_pending_io_list_lock,
				       flags);
	} else {
		async_snap_io = SNAP_PENDING_IO_ENTRY(org_volume->snap_pending_io_list.next);
		list_del(&async_snap_io->snap_pending_io_list);
		spin_unlock_irqrestore(&org_volume->snap_pending_io_list_lock,
				       flags);

		/* Reads */
		for ( bh = async_snap_io->pending_reads; bh;
		      bh = async_snap_io->pending_reads ) {
			async_snap_io->pending_reads = bh->b_reqnext;
			bh->b_reqnext = NULL;
			R_IO(async_snap_io->snap_volume->logical_node, bh);
			evms_cs_volume_request_in_progress(async_snap_io->dev,
							   -1, NULL);
		}
		/* Writes */
		for ( bh = async_snap_io->pending_writes; bh;
		      bh = async_snap_io->pending_writes ) {
			async_snap_io->pending_writes = bh->b_reqnext;
			bh->b_reqnext = NULL;
			W_IO(async_snap_io->snap_volume->logical_node, bh);
			evms_cs_volume_request_in_progress(async_snap_io->dev,
							   -1, NULL);
		}

		if ( atomic_dec_and_test(&async_snap_io->org_io->ref_count) ) {
			deallocate_async_io(async_snap_io->org_io);
		}
		*done = FALSE;
	}
}

/**
 * process_chunk_write_list
 *
 * Grab the first item from the chunk_write_list and send down
 * writes to each snapshot of this original.
 */
static void process_chunk_write_list(struct snapshot_volume * org_volume,
				     int * done)
{
	struct snap_io_buffer * buf, * snap_buf;
	struct async_snap_io * async_snap_io;
	unsigned long flags;

	spin_lock_irqsave(&org_volume->chunk_write_list_lock, flags);
	if ( list_empty(&org_volume->chunk_write_list) ) {
		spin_unlock_irqrestore(&org_volume->chunk_write_list_lock,
				       flags);
	} else {
		buf = CHUNK_WRITE_ENTRY(org_volume->chunk_write_list.next);
		list_del(&buf->chunk_write_list);
		spin_unlock_irqrestore(&org_volume->chunk_write_list_lock,
				       flags);

		for ( snap_buf = buf->copy_next; snap_buf;
		      snap_buf = snap_buf->copy_next ) {
			async_snap_io = snap_buf->buffer_private;
			W_IO(async_snap_io->snap_volume->logical_node,
			     snap_buf->bh);
		}
		*done = FALSE;
	}
}

/**
 * write_cow_table
 *
 * On S/390 machines, the hardsector size is usually 4k, and the driver won't
 * accept I/O requests that are less than 4k in size. Thus, the COW table
 * cannot be written as a single sector. We must first read in the entire
 * 4k hardsector, overlay the 512 byte COW table, and then write out the entire
 * 4k again.
 *
 * On machines with hardsector size of 512, the COW table write will be
 * processed just as it was before.
 *
 * If an error occurs in this function, we will send down the buffer as as
 * read instead of a write. This will ensure that the callback function still
 * runs and cleans up the async_io structures and releases all pending I/Os.
 */
static inline void write_cow_table(struct snapshot_volume * snap,
				   struct buffer_head * bh)
{
	if ( snap->logical_node->hardsector_size > bh->b_size ) {
		u64 offset;
		u8 * buffer;
		unsigned short b_size = bh->b_size;
		int rc;

		offset = bh->b_rsector & ((snap->logical_node->hardsector_size >>
					   EVMS_VSECTOR_SIZE_SHIFT) - 1);
		bh->b_rsector -= offset;
		bh->b_size = snap->logical_node->hardsector_size;

		/* Need a buffer to temporarily hold the COW table sector. */
		buffer = kmalloc(b_size, GFP_NOIO);
		if (!buffer) {
			disable_snapshot(snap, TRUE);
			R_IO(snap->logical_node, bh);
			return;
		}
		memcpy(buffer, bh->b_data, b_size);
		
		/* Read in the entire hardsector from disk. */
		rc = INIT_IO(snap->logical_node, READ, bh->b_rsector,
			     snap->logical_node->hardsector_size >>
			     EVMS_VSECTOR_SIZE_SHIFT,
			     bh->b_data);
		if (rc) {
			disable_snapshot(snap, TRUE);
			R_IO(snap->logical_node, bh);
			return;
		}

		/* Copy the COW table back into the buffer. */
		memcpy(bh->b_data + (offset << EVMS_VSECTOR_SIZE_SHIFT),
		       buffer, b_size);
	}

	W_IO(snap->logical_node, bh);
}

/**
 * process_cow_table_write_lists
 */
static void process_cow_table_write_lists(struct snapshot_volume * org_volume,
					  int * done)
{
	struct snapshot_volume * snap_volume;
	struct async_snap_io * async_snap_io, * async_snap_io2;
	struct list_head * lh;
	unsigned long flags;

	/* Check the chunk_write_list for each snapshot on this original. */
	down_read(&org_volume->snap_semaphore);
	for ( snap_volume = org_volume->snapshot_next;
	      snap_volume;
	      snap_volume = snap_volume->snapshot_next ) {

		/* While we are here, see if the DISABLED bit needs to
		 * be written to disk.
		 */
		if ( snap_volume->flags & EVMS_SNAPSHOT_DISABLED &&
		     snap_volume->flags & EVMS_SNAPSHOT_DISABLED_PENDING ) {
			disable_snapshot(snap_volume, TRUE);
			snap_volume->flags &= ~EVMS_SNAPSHOT_DISABLED_PENDING;
		}

		spin_lock_irqsave(&snap_volume->cow_table_write_list_lock,
				  flags);

		if ( list_empty(&snap_volume->cow_table_write_list) ) {
			spin_unlock_irqrestore(&snap_volume->cow_table_write_list_lock,
					       flags);
			continue;
		}

		/* Check for an in-flight COW-table-write. */
		async_snap_io = COW_WRITE_ENTRY(snap_volume->cow_table_write_list.next);
		if ( atomic_read(&async_snap_io->write_count) != 0 ) {
			spin_unlock_irqrestore(&snap_volume->cow_table_write_list_lock,
					       flags);
			continue;
		}

		/* See if there are any COW-table-writes that can be skipped. */
		list_for_each(lh, &snap_volume->cow_table_write_list) {
			/* No need to check the first list element, since
			 * we've already examined it.
			 */
			if ( lh->prev != &snap_volume->cow_table_write_list ) {
				async_snap_io = COW_WRITE_ENTRY(lh);
				async_snap_io2 = COW_WRITE_ENTRY(lh->prev);
				if ( atomic_read(&async_snap_io->write_count) != 0 ) {
					async_snap_io = async_snap_io2;
					break;
				}
				if ( async_snap_io->cow_table_buffer->bh->b_rsector !=
				     async_snap_io2->cow_table_buffer->bh->b_rsector ) {
					async_snap_io = async_snap_io2;
					break;
				}
			}
		}

		/* We have the buffer to send down. Now mark all
		 * previous COW-table buffers as in-flight.
		 */
		list_for_each(lh, &snap_volume->cow_table_write_list) {
			async_snap_io2 = COW_WRITE_ENTRY(lh);
			atomic_dec(&async_snap_io2->write_count);
			if ( async_snap_io2 == async_snap_io ) {
				break;
			}
			else {
				DEBUG_INC_COW_TABLE_OVERLAPS(snap_volume);
			}
		}

		spin_unlock_irqrestore(&snap_volume->cow_table_write_list_lock, flags);

		/* Write the COW table. */
		DEBUG_INC_COW_TABLE_WRITES(snap_volume);
		write_cow_table(snap_volume, async_snap_io->cow_table_buffer->bh);

		*done = FALSE;
	}
	up_read(&org_volume->snap_semaphore);
}

/**
 * snap_async_io_thread
 *
 * This is the async I/O thread function. It processes requests from four
 * lists, which are embedded in the original volume structure passed to the
 * thread.
 *
 * The first list, org_pending_io_list, contains async_org_io's, each of which
 * contain a list of write requests to the original volume that are waiting on
 * the completion of a chunk copy.
 *
 * The second list, snap_pending_io_list, contains async_snap_io's, each of
 * which contain a list of read requests and a list of write requests to the
 * snapshot volume that are waiting on the completion of a chunk copy.
 *
 * The third list, chunk_write_list, contains buffers that were used to read
 * part of a chunk from the original volume. Those buffers are linked to other
 * buffers which are used to write the same part of that chunk to the snapshot.
 *
 * The fourth list is actually the list of snapshots for this original. Each
 * snapshot then has a list of COW-table buffers that have to be written. The
 * processing of this list is optimized to eliminate unnecessary, overlapping
 * writes of the COW table.
 *
 * The loop will continue as long as there is an item on at least one of
 * the four lists. When they are all empty, the loop exits and the thread
 * goes back to sleep.
 */
static void snap_async_io_thread(void * volume)
{
	struct snapshot_volume * org_volume = volume;
	int done = FALSE;

	while (!done) {
		process_org_pending_io_list(org_volume, &done);

		process_snap_pending_io_list(org_volume, &done);

		process_chunk_write_list(org_volume, &done);

		process_cow_table_write_lists(org_volume, &done);
	}

	run_task_queue(&tq_disk);
}

/**
 * schedule_org_pending_io
 *
 * Place the async_org_io on the thread's processing list.
 */
static void schedule_org_pending_io(struct async_org_io * async_org_io)
{
	struct snapshot_volume * org_volume = async_org_io->org_volume;
	unsigned long flags;

	spin_lock_irqsave(&org_volume->org_pending_io_list_lock, flags);
	list_add_tail(&async_org_io->org_pending_io_list,
		      &org_volume->org_pending_io_list);
	spin_unlock_irqrestore(&org_volume->org_pending_io_list_lock, flags);
	evms_cs_wakeup_thread(org_volume->async_io_thread);
}

/**
 * schedule_snap_pending_io
 *
 * Place the async_snap_io on the thread's processing list.
 */
static void schedule_snap_pending_io(struct async_snap_io * async_snap_io)
{
	struct snapshot_volume * org_volume = async_snap_io->org_io->org_volume;
	unsigned long flags;

	spin_lock_irqsave(&org_volume->snap_pending_io_list_lock, flags);
	list_add_tail(&async_snap_io->snap_pending_io_list,
		      &org_volume->snap_pending_io_list);
	spin_unlock_irqrestore(&org_volume->snap_pending_io_list_lock, flags);
	evms_cs_wakeup_thread(org_volume->async_io_thread);
}

/**
 * schedule_chunk_write
 *
 * Place the buffer on the chunk_write_list for the thread to process. This
 * list uses the chunk_write_list field in the snap_io_buffer.
 */
static void schedule_chunk_write(struct snap_io_buffer * buf)
{
	struct async_org_io * org_io = buf->buffer_private;
	struct snapshot_volume * org_volume = org_io->org_volume;
	unsigned long flags;

	spin_lock_irqsave(&org_volume->chunk_write_list_lock, flags);
	list_add_tail(&buf->chunk_write_list, &org_volume->chunk_write_list);
	spin_unlock_irqrestore(&org_volume->chunk_write_list_lock, flags);
	evms_cs_wakeup_thread(org_volume->async_io_thread);
}

/**
 * schedule_cow_table_write
 *
 * Place the async_snap_io on the thread's processing list.
 */
static void schedule_cow_table_write(struct async_snap_io * async_snap_io)
{
	struct snapshot_volume * snap_volume = async_snap_io->snap_volume;
	unsigned long flags;

	spin_lock_irqsave(&snap_volume->cow_table_write_list_lock, flags);
	list_add_tail(&async_snap_io->cow_write_list,
		      &snap_volume->cow_table_write_list);
	spin_unlock_irqrestore(&snap_volume->cow_table_write_list_lock, flags);
}

/**
 * snap_read_chunk_cb
 *
 * This is the callback function for reading chunks from the original.
 * When each read completes, we have to decrement the read_count in the
 * async_org_io. If this count reaches zero, we can decrement the
 * chunk_lock's in all of the hash entries, and send the original write
 * request down. Finally, send this buffer head over to the thread to
 * send the writes down to the snapshots.
 */
void snap_read_chunk_cb(struct buffer_head * bh,
			int uptodate)
{
	struct snap_io_buffer * buf = bh->b_private;

	if (!uptodate) {
		/* Error reading the chunk. Disable all snapshots on this org. */
		struct async_org_io * async_org_io = buf->buffer_private;
		struct snapshot_volume * snap_volume;
		LOG_ERROR("Error reading chunk from original '%s'.\n",
			  async_org_io->org_volume->exported_node->name);
		for ( snap_volume = async_org_io->org_volume->snapshot_next;
		      snap_volume;
		      snap_volume = snap_volume->snapshot_next ) {
			disable_snapshot(snap_volume, FALSE);
		}
	}

	schedule_chunk_write(buf);
}

/**
 * snap_write_chunk_cb
 *
 * This is the callback function for writing chunks to the snapshot. When
 * each write completes, decrement the write_count in the async_snap_io.
 * If this count reaches zero, decrement the chunk_lock in the hash entry,
 * and decrement the remap_count in the async_org_io. If the remap_count
 * reaches zero, then everybody is done, and we can free up the entire
 * async_io structure.
 */
void snap_write_chunk_cb(struct buffer_head * bh,
			 int uptodate)
{
	struct snap_io_buffer * buf = bh->b_private;
	struct async_snap_io * async_snap_io = buf->buffer_private;

	if (!uptodate) {
		/* Error writing chunk. Disable this snapshot. */
		LOG_ERROR("Error writing chunk to snapshot '%s'.\n",
			async_snap_io->snap_volume->exported_node->name);
		disable_snapshot(async_snap_io->snap_volume, FALSE);
	}

	atomic_dec(&async_snap_io->write_count);
	evms_cs_wakeup_thread(async_snap_io->org_io->org_volume->async_io_thread);
}

/**
 * snap_cow_table_cb
 *
 * This is the callback function for writing out the COW table.
 */
void snap_cow_table_cb(struct buffer_head * bh,
		       int uptodate)
{
	struct snap_io_buffer * buf = bh->b_private;
	struct async_snap_io * async_snap_io = buf->buffer_private;
	struct async_snap_io * async_snap_io2;
	struct async_org_io * async_org_io;
	struct snapshot_volume * snap_volume = async_snap_io->snap_volume;
	struct list_head * lh, * tmp;
	unsigned long flags, flags2;

	if (!uptodate) {
		/* Error writing the COW table sector. Disable the snapshot. */
		struct snapshot_volume * snap_volume = buf->buffer_private;
		LOG_ERROR("Error writing COW table to snapshot '%s'.\n",
			snap_volume->exported_node->name);
		disable_snapshot(snap_volume, FALSE);
	}

	spin_lock_irqsave(&snap_volume->cow_table_write_list_lock, flags);

	list_for_each_safe(lh, tmp, &snap_volume->cow_table_write_list) {
		list_del(lh);
		async_snap_io2 = COW_WRITE_ENTRY(lh);
		async_org_io = async_snap_io2->org_io;

		/* Mark the chunk as copied in the hash table. */
		spin_lock_irqsave(&async_snap_io2->hash_table_entry->chunk_state_lock,
				  flags2);
		async_snap_io2->hash_table_entry->chunk_state = SNAP_CHUNK_COPIED;
		async_snap_io2->hash_table_entry->snap_io = NULL;
		spin_unlock_irqrestore(&async_snap_io2->hash_table_entry->chunk_state_lock,
				       flags2);

		/* Release any pending I/Os waiting on this chunk. */
		schedule_snap_pending_io(async_snap_io2);
		if ( atomic_dec_and_test(&async_org_io->copy_count) ) {
			schedule_org_pending_io(async_org_io);
		}

		if ( async_snap_io2 == async_snap_io ) {
			break;
		}
	}

	spin_unlock_irqrestore(&snap_volume->cow_table_write_list_lock, flags);
}

/**
 * snap_queue_original_request
 *
 * An existing remap was found for the chunk for this write request.
 * If the chunk has been fully copied, then the request can go through
 * normally. If the chunk is still being processed, this request must
 * be queued up to be driven after the chunk has been copied.
 */
static void snap_queue_original_request(struct snapshot_volume * snap_volume,
					struct buffer_head * org_bh,
					struct snapshot_hash_entry * target_entry,
					u64 remainder,
					int * queued_org_bh,
					int write_to_snapshot)
{
	struct async_org_io * org_io;
	unsigned long flags, flags2;

	if (write_to_snapshot) {
		org_bh->b_rsector = (target_entry->snap_chunk <<
				     snap_volume->chunk_shift) +
				    remainder;
	}

	if ( ! *queued_org_bh &&
	     target_entry->chunk_state != SNAP_CHUNK_COPIED ) {
		spin_lock_irqsave(&target_entry->chunk_state_lock, flags);
		if (write_to_snapshot) {
			/* A write to the snapshot. */
			if ( target_entry->chunk_state != SNAP_CHUNK_COPIED ) {
				org_bh->b_reqnext =
					target_entry->snap_io->pending_writes;
				target_entry->snap_io->pending_writes = org_bh;
				if (!target_entry->snap_io->dev) {
					target_entry->snap_io->dev =
						org_bh->b_rdev;
				}
				evms_cs_volume_request_in_progress(target_entry->snap_io->dev,
								   +1, NULL);
				*queued_org_bh = TRUE;
			}
		} else {
			/* A write to the original. */
			if ( target_entry->chunk_state != SNAP_CHUNK_COPIED ) {
				org_io = target_entry->snap_io->org_io;
				spin_lock_irqsave(&org_io->pending_writes_lock,
						  flags2);
				org_bh->b_reqnext = org_io->pending_writes;
				org_io->pending_writes = org_bh;
				if (!org_io->dev) {
					org_io->dev = org_bh->b_rdev;
				}
				spin_unlock_irqrestore(&org_io->pending_writes_lock,
						       flags2);
				evms_cs_volume_request_in_progress(org_io->dev,
								   +1, NULL);
				*queued_org_bh = TRUE;
			}
		}
		spin_unlock_irqrestore(&target_entry->chunk_state_lock, flags);
	}
}

/**
 * snapshot_copy_1
 *
 * Check this snapshot node to see if the given sector/chunk has been
 * remapped yet. If it hasn't, create a new hash table entry, update the
 * in-memory COW table, write the COW table to disk if it is full, and
 * then start the process of copying the chunk from the original to the
 * snapshot.
 */
static int snapshot_copy_1(struct snapshot_volume * snap_volume,
			   struct buffer_head * org_bh,
			   struct async_org_io ** async_org_io,
			   int * queued_org_bh,
			   int write_to_snapshot)
{
	struct snapshot_volume * org_volume = snap_volume->snapshot_org;
	struct snapshot_hash_entry * target_entry, * new_map_entry;
	struct snap_io_buffer * cow_buf, *buf1, *buf2;
	struct async_snap_io * async_snap_io;
	u64 org_sector = org_bh->b_rsector;
	u64 org_chunk_lba, snap_chunk_lba;
	u64 alignment;
	u64 chunk, remainder;
	unsigned long hash_value, buffer_count, sectors_in_chunk;

	/* Grab the read-lock when checking for an existing remap. */
	down_read(&snap_volume->snap_semaphore);

	/* Make sure the snapshot has not been disabled. */
	if ( snap_volume->flags & (EVMS_SNAPSHOT_DISABLED|EVMS_SNAPSHOT_FULL) ||
	     ! org_volume ) {
		up_read(&snap_volume->snap_semaphore);
		return -ENOSPC;
	}

	/* Check for unaligned I/O. This is mostly to prevent XFS from
	 * sending a request that spans a chunk.
	 */
	alignment = org_sector << EVMS_VSECTOR_SIZE_SHIFT;
	if ( unlikely(alignment & (org_bh->b_size - 1)) ) {
		LOG_ERROR("Unaligned request [rsector(%lx), size(%x)] rejected on snapshot %s.\n",
			  org_bh->b_rsector, org_bh->b_size,
			  snap_volume->logical_node->name);
		if (!write_to_snapshot) {
			disable_snapshot(snap_volume, TRUE);
		}
		up_read(&snap_volume->snap_semaphore);
		return -EINVAL;
	}

	/* Search the hash table to see if this sector has already been
	 * remapped on this snapshot.
	 */
	chunk = org_sector >> snap_volume->chunk_shift;
	remainder = org_sector & (u64)(snap_volume->chunk_size - 1);
	hash_value = (unsigned long)chunk % snap_volume->hash_table_size;

	if ( search_snapshot_hash_chain(chunk,
					snap_volume->snapshot_map[hash_value],
					&target_entry) ) {
		/* Chunk is already remapped. If the remap is still in progress,
		 * queue up this request to be handled later. If the remap is
		 * complete, we can just keep going.
		 */
		up_read(&snap_volume->snap_semaphore);
		snap_queue_original_request(snap_volume, org_bh,
					    target_entry, remainder,
					    queued_org_bh, write_to_snapshot);
		return 0;
	}

	/* Convert to a write-lock and check again for a remap.
	 * (Same search and check as just before).
	 */
	up_read(&snap_volume->snap_semaphore);
	down_write(&snap_volume->snap_semaphore);
	if ( search_snapshot_hash_chain(chunk,
					snap_volume->snapshot_map[hash_value],
					&target_entry) ) {
		/* Chunk is already remapped. If the remap is still in progress,
		 * queue up this request to be handled later. If the remap is
		 * complete, we can just keep going.
		 */
		up_write(&snap_volume->snap_semaphore);
		snap_queue_original_request(snap_volume, org_bh,
					    target_entry, remainder,
					    queued_org_bh, write_to_snapshot);
		return 0;
	}

	/* Is there enough room left on this snapshot to remap this chunk? */
	if ( snap_volume->next_free_chunk >= snap_volume->num_chunks ) {
		/* Once the snapshot becomes full, further writes to the
		 * original can't be remapped, and thus this snapshot
		 * will become "corrupted".
		 */
		snap_volume->flags |= EVMS_SNAPSHOT_FULL;
		set_snapshot_flags(snap_volume->logical_node,
				   EVMS_SNAPSHOT_FULL, EVMS_SNAPSHOT_DISABLED);
		up_write(&snap_volume->snap_semaphore);
		return -ENOSPC;
	}

	/* Create and initialize a new hash table entry for the new remap.
	 * The value SNAP_CHUNK_COPYING indicates that this chunk still has to
	 * be read from the original and written to the snapshot.
	 */
	new_map_entry = allocate_snapshot_hash_entry(snap_volume,
						     chunk,
						     snap_volume->next_free_chunk,
						     SNAP_CHUNK_COPYING);
	if (!new_map_entry) {
		/* Can't get memory for map entry. Disable this snapshot. */
		LOG_ERROR("Memory error allocating hash table entry for snapshot '%s'.\n",
			  snap_volume->exported_node->name);
		disable_snapshot(snap_volume, TRUE);
		up_write(&snap_volume->snap_semaphore);
		return -ENOMEM;
	}

	/* Add the entry to the hash table. */
	if (target_entry) {	
		insert_snapshot_hash_entry(new_map_entry, target_entry);
	} else {
		insert_snapshot_hash_entry_at_head(new_map_entry,
						   &(snap_volume->snapshot_map[hash_value]));
	}

	/* Calculate the number of buffers that will be needed to copy this
	 * chunk, and the starting LBAs for both the org and the snap.
	 */
	org_chunk_lba = chunk * org_volume->chunk_size;
	snap_chunk_lba = snap_volume->next_free_chunk * org_volume->chunk_size;
	snap_volume->next_free_chunk++;
	sectors_in_chunk = min(((u64)org_volume->chunk_size),
			       org_volume->logical_node->total_vsectors -
			       org_chunk_lba);
	buffer_count = (sectors_in_chunk +
			(PAGE_SIZE >> EVMS_VSECTOR_SIZE_SHIFT) - 1) /
		       (PAGE_SIZE >> EVMS_VSECTOR_SIZE_SHIFT);

	/* Create the parent async_org_io structure if it hasn't been done yet. */
	if (!*async_org_io) {
		*async_org_io = allocate_async_org_io(org_volume,
						      org_chunk_lba,
						      buffer_count);
		if (!*async_org_io) {
			// BUGBUG: Disable the snapshot?
			BUG();
		}

		/* If we are only reading a partial chunk from the original,
		 * may need to readjust the size in the last buffer.
		 */
		if ( (sectors_in_chunk < org_volume->chunk_size) &&
		     (sectors_in_chunk &
		      ((PAGE_SIZE >> EVMS_VSECTOR_SIZE_SHIFT) - 1)) ) {
			for ( buf1 = (*async_org_io)->copy_buffers;
			      buf1->buffer_next;
			      buf1 = buf1->buffer_next ) {
				;
			}
			buf1->bh->b_size = (sectors_in_chunk <<
					    EVMS_VSECTOR_SIZE_SHIFT) &
					   (PAGE_SIZE - 1);
		}
	}

	/* Create an async_snap_io structure for this snapshot and attach to
	 * the org io structure.
	 */
	async_snap_io = allocate_async_snap_io(snap_volume, new_map_entry,
					       *async_org_io, snap_chunk_lba,
					       buffer_count);
	if (!async_snap_io) {
		// BUGBUG: Disable the snapshot?
		BUG();
	}

	/* Fill in the next entry in the COW table. Copy the COW table to the
	 * buffer to be written out later.
	 */
	snap_volume->cow_table[snap_volume->next_cow_entry] = cpu_to_le64p(&chunk);
	snap_volume->next_cow_entry++;
	cow_buf = async_snap_io->cow_table_buffer;
	cow_buf->bh->b_rdev = org_bh->b_rdev;
	cow_buf->bh->b_rsector = snap_volume->current_cow_sector;
	memcpy(cow_buf->bh->b_data, snap_volume->cow_table, EVMS_VSECTOR_SIZE);

	/* If the COW table is full, reinitialize for the next sector. */
	if ( snap_volume->next_cow_entry >= (EVMS_VSECTOR_SIZE/sizeof(u64)) ) {
		snap_volume->next_cow_entry = 0;
		snap_volume->current_cow_sector++;
		memset(snap_volume->cow_table, 0xff, EVMS_VSECTOR_SIZE);
	}

	/* Attach the original buffer head, if it hasn't been queued
	 * already on a different copy.
	 */
	if (!*queued_org_bh) {
		org_bh->b_reqnext = NULL;
		if (write_to_snapshot) {
			/* Write to the snapshot. Attach to the async_snap_io. */
			org_bh->b_rsector = (new_map_entry->snap_chunk <<
					     snap_volume->chunk_shift) +
					    remainder;
			async_snap_io->pending_writes = org_bh;
			async_snap_io->dev = org_bh->b_rdev;
		} else {
			/* Write to the original. Attatch to the async_org_io. */
			(*async_org_io)->pending_writes = org_bh;
			(*async_org_io)->dev = org_bh->b_rdev;
		}
		evms_cs_volume_request_in_progress(org_bh->b_rdev, +1, NULL);
		*queued_org_bh = TRUE;
	}

	/* Point the hash table entry at this async_snap_io. Then add this
	 * async_snap_io to the list in the async_org_io, as well as to the
	 * list in the snapshot volume.
	 */
	new_map_entry->snap_io = async_snap_io;

	async_snap_io->snap_io_list_next = (*async_org_io)->snap_io_list;
	(*async_org_io)->snap_io_list = async_snap_io;
	atomic_inc(&(*async_org_io)->copy_count);
	atomic_inc(&(*async_org_io)->ref_count);

	schedule_cow_table_write(async_snap_io);

	/* Parallel walk through the copy_buffer's in the org and the snap,
	 * updating all necessary pointers and lists.
	 */
	for ( buf1 = (*async_org_io)->copy_buffers,
	      buf2 = async_snap_io->copy_buffers;
	      buf1 && buf2;
	      buf1 = buf1->buffer_next, buf2 = buf2->buffer_next ) {
		buf2->copy_next = buf1->copy_next;
		buf2->buffer_private = async_snap_io;
		buf2->bh->b_rdev = org_bh->b_rdev;
		buf2->bh->b_data = buf1->bh->b_data;
		buf2->bh->b_page = buf1->bh->b_page;

		buf1->bh->b_rdev = org_bh->b_rdev;
		buf1->copy_next = buf2;
		buf1->buffer_private = *async_org_io;
	}

	/* We're done modifying snapshot volume info, so we can release the
	 * lock. We can't start any reads until all snapshots for this original
	 * have been checked. Return and start the reads later.
	 */
	up_write(&snap_volume->snap_semaphore);

	return 0;
}

/**
 * snapshot_copy_data
 */
static void snapshot_copy_data(struct snapshot_volume * org_volume,
			       struct buffer_head * org_bh)
{
	struct snapshot_volume * snap_volume, * next_volume;
	struct async_org_io * async_org_io = NULL;
	struct snap_io_buffer * buf;
	int queued_org_bh = FALSE;

	/* Check each snapshot on this original
	 * to see which one's need a remap.
	 */
	for ( snap_volume = org_volume->snapshot_next;
	      snap_volume; snap_volume = next_volume ) {
		next_volume = snap_volume->snapshot_next;
		snapshot_copy_1(snap_volume, org_bh, &async_org_io,
				&queued_org_bh, FALSE);
	}

	if (async_org_io) {
		/* One or more snapshots need a remap. The async_io structures
		 * have been built. Now we just need to run through them and
		 * start all of the reads.
		 */
		for ( buf = async_org_io->copy_buffers;
		      buf; buf = buf->buffer_next ) {
			R_IO(org_volume->logical_node, buf->bh);
		}
	} else if (!queued_org_bh) {
		/* None of the snapshots needed a remap, and we didn't have to
		 * queue this request to be processed later due to a copy in
		 * progress. The write can be sent down normally.
		 */
		W_IO(org_volume->logical_node, org_bh);
	}
}

/**
 * writeable_snapshot_copy_data
 */
static void writeable_snapshot_copy_data(struct snapshot_volume * snap_volume,
					 struct buffer_head * org_bh)
{
	struct snapshot_volume * org_volume = snap_volume->snapshot_org;
	struct async_org_io * async_org_io = NULL;
	struct snap_io_buffer * buf;
	int rc, queued_org_bh = FALSE;

	rc = snapshot_copy_1(snap_volume, org_bh, &async_org_io,
			     &queued_org_bh, TRUE);
	if ( rc < 0 ) {
		org_bh->b_end_io(org_bh, 0);
		return;
	}

	if (async_org_io) {
		/* Need to remap this chunk to the snapshot. The async_io
		 * structures have been built. Just need to run through them
		 * and start all of the reads.
		 */
		for ( buf = async_org_io->copy_buffers; buf;
		      buf = buf->buffer_next ) {
			R_IO(org_volume->logical_node, buf->bh);
		}
	} else if (!queued_org_bh) {
		/* No remap. The write can be sent down immediately. */
		W_IO(snap_volume->logical_node, org_bh);
	}
}

/**
 * snap_write
 */
static void snap_write(struct evms_logical_node * node,
		       struct buffer_head * bh)
{
	struct snapshot_volume * volume = node->private;

	/* Size check. */
	if ( bh->b_rsector + (bh->b_size >> EVMS_VSECTOR_SIZE_SHIFT) >
	     node->total_vsectors) {
		bh->b_end_io(bh, 0);
		return;
	}

	/* Can't write if rollback is in progress. */
	if ( volume->flags & EVMS_SNAPSHOT_ROLLBACK ) {
		LOG_ERROR("Cannot write to snapshot '%s' during rollback.\n",
			  volume->logical_node->name);
		bh->b_end_io(bh, 0);
		return;
	}

	if ( volume->flags & EVMS_SNAPSHOT ) {
		/* Snapshot. */
		if ( volume->flags & EVMS_SNAPSHOT_WRITEABLE ) {
			writeable_snapshot_copy_data(volume, bh);
		} else {
			bh->b_end_io(bh, 0);
		}
	} else {
		/* Original. */
		snapshot_copy_data(volume, bh);
	}
}

/**
 * snap_ioctl
 */
static int snap_ioctl(struct evms_logical_node * logical_node,
		      struct inode * inode,
		      struct file * file,
		      unsigned int cmd,
		      unsigned long arg)
{
	struct snapshot_volume * volume = logical_node->private;
	struct evms_quiesce_vol_pkt * quiesce;
	struct evms_plugin_ioctl_pkt pkt, * user_pkt;
	int percent_full, rc = 0;

	switch (cmd) {
	case EVMS_QUIESCE_VOLUME:
		quiesce = (struct evms_quiesce_vol_pkt*)arg;
		if (quiesce->command) {
			/* Quiesce */
			volume->flags |= EVMS_SNAPSHOT_QUIESCED;
		} else {
			/* Un-quiesce */
			volume->flags &= ~EVMS_SNAPSHOT_QUIESCED;
		}
		break;

	case EVMS_GET_BMAP:
		if ( volume->flags & EVMS_SNAPSHOT_ORG ) {
			rc = IOCTL(volume->logical_node, inode, file, cmd, arg);
		} else {
			rc = -EINVAL;
		}
		break;

	case EVMS_PLUGIN_IOCTL:
		user_pkt = (struct evms_plugin_ioctl_pkt *)arg;

		/* Copy user's parameters to kernel space. */
		if ( copy_from_user(&pkt, user_pkt, sizeof(pkt)) ) {
			rc = -EFAULT;
			break;
		}

		if ( pkt.feature_id != logical_node->plugin->id ) {
			/* This ioctl is not targetted at snapshotting, so
			 * broadcast the command to all children.
			 */
			rc = IOCTL(logical_node, inode, file, cmd, arg);
			break;
		}

		switch (pkt.feature_command) {
		case SNAPSHOT_QUERY_PERCENT_FULL:
			if ( volume->flags & EVMS_SNAPSHOT_FULL ) {
				percent_full = -1;
			} else if ( volume->flags & EVMS_SNAPSHOT_DISABLED ) {
				percent_full = -2;
			} else {
				percent_full = (volume->next_free_chunk * 100) /
						volume->num_chunks;
			}
			rc = copy_to_user(pkt.feature_ioctl_data,
					  &percent_full,
					  sizeof(percent_full));
			break;

		case SNAPSHOT_START_ROLLBACK:
			if ( volume->flags & EVMS_SNAPSHOT_FULL ) {
				rc = -ENOSPC;
			} else if ( volume->flags & EVMS_SNAPSHOT_DISABLED ) {
				rc = -EIO;
			} else if ( ! (volume->flags & EVMS_SNAPSHOT) ) {
				rc = -EINVAL;
			} else {
				set_snapshot_flags(volume->logical_node,
						   EVMS_SNAPSHOT_ROLLBACK, 0);
			}
			break;

		case SNAPSHOT_CHECK_STATE:
			rc = copy_to_user(pkt.feature_ioctl_data,
					  &volume->flags,
					  sizeof(volume->flags));
			break;

		default:
			rc = -EINVAL;
		}
		break;

	case EVMS_CHECK_MEDIA_CHANGE:
	case EVMS_REVALIDATE_DISK:
	case EVMS_GET_DISK_LIST:
	case EVMS_CHECK_DEVICE_STATUS:
		/* Broadcast these to all children. */
		if ( ! (volume->flags & EVMS_SNAPSHOT_ORG) ) {
			volume = volume->snapshot_org;
		}
		while (volume) {
			rc = IOCTL(volume->logical_node, inode, file, cmd, arg);
			volume = volume->snapshot_next;
		}
		break;

	case EVMS_OPEN_VOLUME:
		/* Disallow opens on rollback in progress.
		 * Otherwise fall through.
		 */
		if ( volume->flags & EVMS_SNAPSHOT_ROLLBACK) {
			LOG_ERROR("Cannot open snapshot volume '%s' during rollback\n",
				  volume->logical_node->name);
			rc = -EBUSY;
			break;
		}

	default:
		rc = IOCTL(volume->logical_node, inode, file, cmd, arg);

	}
	return rc;
}

/**
 * snap_init_io
 */
static int snap_init_io(struct evms_logical_node * node,
			int rw,
			u64 sect_nr,
			u64 num_sects,
			void * buf_addr)
{
	struct snapshot_volume * volume = node->private;

	/* No init io access to snapshot, and no writes allowed to original
	 * since they would not be snapshotted.
	 */
	if ( rw || (volume->flags & EVMS_SNAPSHOT) ) {
		return -EINVAL;
	}
	return INIT_IO(volume->logical_node, rw,
		       sect_nr, num_sects, buf_addr);
}

/**
 * add_cow_entry_to_snapshot_map
 *
 * This function takes a cow table entry (from the on-disk data), and
 * converts it into an appropriate entry for the snapshot map, and
 * inserts it into the appropriate map for the specified volume.
 */
static int add_cow_entry_to_snapshot_map(u64 org_chunk,
					 u64 snap_chunk,
					 struct snapshot_volume * volume)
{
	struct snapshot_hash_entry * new_entry, * target_entry;
	unsigned long hash_value;

	new_entry = allocate_snapshot_hash_entry(volume, org_chunk,
						 snap_chunk, SNAP_CHUNK_COPIED);
	if (!new_entry) {
		return -ENOMEM;
	}

	hash_value = (long)org_chunk % volume->hash_table_size;
	if ( search_snapshot_hash_chain(org_chunk,
					volume->snapshot_map[hash_value],
					&target_entry) ) {
		/* A duplicate mapping was found. This should never happen. */
	} else {
		if (target_entry) {
			insert_snapshot_hash_entry(new_entry, target_entry);
		} else {
			insert_snapshot_hash_entry_at_head(new_entry,	
							   &(volume->snapshot_map[hash_value]));
		}
	}
	return 0;
}

/**
 * build_snapshot_maps
 *
 * Construct the initial hash table state based on
 * existing COW tables on the disk.
 */
static int build_snapshot_maps(struct snapshot_volume * volume)
{
	int rc = 0;
	int done = FALSE;
	while (!done) {
		/* Read in one sector's worth of COW tables. */
		if ( INIT_IO(volume->logical_node, 0,
			     volume->current_cow_sector, 1,
			     volume->cow_table) ) {
			return -EIO;
		}

		/* Translate every valid COW table entry into
		 * a snapshot map entry.
		 */
		for ( volume->next_cow_entry = 0;
		      volume->next_cow_entry < (EVMS_VSECTOR_SIZE/sizeof(u64)) &&
		      volume->cow_table[volume->next_cow_entry] != 0xffffffffffffffff;
		      volume->next_cow_entry++, volume->next_free_chunk++ ) {
			rc = add_cow_entry_to_snapshot_map(le64_to_cpup(&volume->cow_table[volume->next_cow_entry]),
							   volume->next_free_chunk,
							   volume);
			if (rc) {
				return(rc);
			}
		}

		/* Move on to the next sector if necessary. */
		if ( volume->next_cow_entry ==
		     (EVMS_VSECTOR_SIZE/sizeof(u64)) ) {
			volume->current_cow_sector++;
		} else {
			done = TRUE;
		}
	}
	return 0;
}

/**
 * initialize_snapshot_node
 */
static int initialize_snapshot_node(struct evms_logical_node * snap_node,
				    struct evms_logical_node * new_snap_node,
				    struct evms_logical_node * org_node,
				    struct snapshot_metadata * metadata)
{
	struct snapshot_volume * snap_volume;
	struct snapshot_hash_entry * new_entry;
	int i, rc = 0;

	/* Instance data for the snapshot. */
	snap_volume = kmalloc(sizeof(struct snapshot_volume), GFP_KERNEL);
	if (!snap_volume) {
		set_snapshot_flags(snap_node, EVMS_SNAPSHOT_DISABLED, 0);
		snap_delete_volume(new_snap_node);
		DELETE(snap_node);
		return -ENOMEM;
	}
	memset(snap_volume, 0, sizeof(struct snapshot_volume));

	/* Initialize the snapshot node. */
	new_snap_node->total_vsectors = org_node->total_vsectors;
	new_snap_node->plugin = &plugin_header;
	new_snap_node->private = snap_volume;	
	new_snap_node->flags = snap_node->flags |
				(org_node->flags & (EVMS_DEVICE_REMOVABLE | EVMS_VOLUME_PARTIAL)) |
				((metadata->flags & EVMS_SNAPSHOT_WRITEABLE) ? 0 : EVMS_VOLUME_READ_ONLY);
	new_snap_node->hardsector_size = snap_node->hardsector_size;
	new_snap_node->block_size = snap_node->block_size;
	new_snap_node->system_id = EVMS_SNAPSHOT_SIGNATURE;
	new_snap_node->volume_info = snap_node->volume_info;
	/* Get the new node's name from the consumed node's feature header. */
	strcpy(new_snap_node->name, snap_node->feature_header->object_name);

	/* Initialize the private data. */
	snap_volume->logical_node = snap_node;
	snap_volume->exported_node = new_snap_node;
	init_rwsem(&snap_volume->snap_semaphore);
	snap_volume->chunk_size = metadata->chunk_size;
	snap_volume->chunk_shift = evms_cs_log2((u64)metadata->chunk_size);
	snap_volume->num_chunks = metadata->total_chunks;
	snap_volume->current_cow_sector = metadata->lba_of_COW_table;
	snap_volume->hash_table_size = metadata->total_chunks / MAX_HASH_CHAIN_ENTRIES + 1;
	snap_volume->flags = EVMS_SNAPSHOT |
			     (metadata->flags & EVMS_SNAPSHOT_WRITEABLE) |
			     (metadata->flags & EVMS_SNAPSHOT_ASYNC);
	INIT_LIST_HEAD(&snap_volume->cow_table_write_list);
	spin_lock_init(&snap_volume->cow_table_write_list_lock);

#ifdef SNAPSHOT_DEBUG
	atomic_set(&snap_volume->cow_table_writes, 0);
	atomic_set(&snap_volume->cow_table_overlaps, 0);
#endif

	if ( metadata->flags & EVMS_SNAPSHOT_ROLLBACK ) {

		/* Buffer for reading rollback data. */
		snap_volume->chunk_data_buffer = kmalloc(SNAPSHOT_CHUNK_BUFFER_SIZE <<
							 EVMS_VSECTOR_SIZE_SHIFT,
							 GFP_KERNEL);
		if (!snap_volume->chunk_data_buffer) {
			disable_snapshot(snap_volume, TRUE);
			snap_delete_volume(new_snap_node);
			return -ENOMEM;
		}

		/* Create the rollback thread. */
		snap_volume->rollback_thread =
			evms_cs_register_thread(snapshot_do_rollback,
						snap_volume,
						"evms_snapshot_rollback");
		if (!snap_volume->rollback_thread){
			LOG_SERIOUS("Could not start rollback thread for snapshot '%s'.\n",
				    snap_node->name);
			disable_snapshot(snap_volume, TRUE);
			snap_delete_volume(new_snap_node);
			return -ENOMEM;
		}
	} else {
		/* Snapshot hash table. */
		snap_volume->snapshot_map = vmalloc(snap_volume->hash_table_size *
						    sizeof(struct snapshot_hash_entry*));
		if (!snap_volume->snapshot_map) {
			disable_snapshot(snap_volume, TRUE);
			snap_delete_volume(new_snap_node);
			return -ENOMEM;
		}
		memset(snap_volume->snapshot_map, 0,
		       snap_volume->hash_table_size *
		       sizeof(struct snapshot_hash_entry*));

		/* Pre-allocate all of the hash entries we will need and
		 * store them in the free list in the volume.
		 */
		for ( i = 0; i < snap_volume->num_chunks; i++ ) {
			new_entry = mempool_alloc(snap_hash_entry_pool,
						  GFP_KERNEL);
			if (!new_entry) {
				disable_snapshot(snap_volume, TRUE);
				snap_delete_volume(new_snap_node);
				return -ENOMEM;
			}
			new_entry->next = snap_volume->free_hash_list;
			snap_volume->free_hash_list = new_entry;
		}

		rc = build_snapshot_maps(snap_volume);
		if (rc) {
			disable_snapshot(snap_volume, TRUE);
			snap_delete_volume(new_snap_node);
			return rc;
		}
	}

	return 0;
}

/**
 * initialize_original_node
 */
static int initialize_original_node(struct evms_logical_node * snap_node,
				    struct evms_logical_node * new_snap_node,
				    struct evms_logical_node * org_node,
				    struct evms_logical_node * new_org_node)
{
	struct snapshot_volume * snap_volume = new_snap_node->private;
	struct snapshot_volume * org_volume;

	/* Instance data for the original. */
	org_volume = kmalloc(sizeof(struct snapshot_volume), GFP_KERNEL);
	if (!org_volume) {
		disable_snapshot(snap_volume, TRUE);
		snap_delete_volume(new_snap_node);
		snap_delete_volume(new_org_node);
		return -ENOMEM;
	}
	memset(org_volume, 0, sizeof(struct snapshot_volume));

	/* Initialize the new node. */
	new_org_node->total_vsectors = org_node->total_vsectors;
	new_org_node->plugin = &plugin_header;
	new_org_node->private = org_volume;	
	new_org_node->flags = org_node->flags |
			      (snap_node->flags &
			       (EVMS_DEVICE_REMOVABLE | EVMS_VOLUME_PARTIAL));
	new_org_node->hardsector_size = org_node->hardsector_size;
	new_org_node->block_size = org_node->block_size;
	new_org_node->system_id = EVMS_ORIGINAL_SIGNATURE;
	new_org_node->volume_info = org_node->volume_info;
	/* Must reuse the original node's name. */
	strcpy(new_org_node->name, org_node->name);

	/* Initialize the private data. */
	org_volume->logical_node = org_node;
	org_volume->exported_node = new_org_node;
	init_rwsem(&org_volume->snap_semaphore);
	org_volume->chunk_size = snap_volume->chunk_size;
	org_volume->chunk_shift = snap_volume->chunk_shift;
	org_volume->flags = EVMS_SNAPSHOT_ORG |
			    (snap_volume->flags & EVMS_SNAPSHOT_ASYNC);
	INIT_LIST_HEAD(&org_volume->chunk_write_list);
	spin_lock_init(&org_volume->chunk_write_list_lock);
	INIT_LIST_HEAD(&org_volume->org_pending_io_list);
	spin_lock_init(&org_volume->org_pending_io_list_lock);
	INIT_LIST_HEAD(&org_volume->snap_pending_io_list);
	spin_lock_init(&org_volume->snap_pending_io_list_lock);

	/* Start the async I/O thread for this original. */
	org_volume->async_io_thread =
		evms_cs_register_thread(snap_async_io_thread, org_volume,
					"evms_async_snapshot");
	if (!org_volume->async_io_thread) {
		disable_snapshot(snap_volume, TRUE);
		snap_delete_volume(new_snap_node);
		snap_delete_volume(new_org_node);
		return -ENOMEM;
	}

	return 0;
}

/**
 * add_snapshot
 *
 * Initializes a snapshot instance and exports an evms_logical_node to
 * the global list.
 */
static int add_snapshot(struct evms_logical_node * snap_node,
			struct snapshot_metadata * metadata,
			struct evms_logical_node ** evms_node_list)
{
	struct evms_logical_node * new_snap_node;
	struct evms_logical_node * new_org_node;
	struct evms_logical_node * org_node;
	struct snapshot_volume * snap_volume;
	struct snapshot_volume * org_volume;
	struct snapshot_volume * tmp_volume;
	int rc = 0;

	/* Make sure the snapshot is not full or disabled. */
	if ( metadata->flags & (EVMS_SNAPSHOT_DISABLED | EVMS_SNAPSHOT_FULL) ) {
		LOG_WARNING("Error: Snapshot %s discovered as disabled/full.\n",
			    snap_node->name);
		LOG_WARNING("       Deleting from further use.\n");
		DELETE(snap_node);
		return -ENOSPC;
	}

	/* Inspect the global list until a node is found with the name of
	 * this snapshot's original. There can only be one original for
	 * each snapshot.
	 */
	for ( org_node = *evms_node_list;
	      org_node && strncmp(EVMS_GET_NODE_NAME(org_node),
				  metadata->original_volume,
				  EVMS_VOLUME_NAME_SIZE);
	      org_node = org_node->next ) {
		;
	}
	if (!org_node) {
		/* No original was found. Disable and delete the snapshot. */
		LOG_ERROR("Error: No original found for snapshot %s, looking for %s\n",
			  snap_node->name, metadata->original_volume);
		set_snapshot_flags(snap_node, EVMS_SNAPSHOT_DISABLED, 0);
		DELETE(snap_node);
		return -ENODEV;
	}

	LOG_DEBUG("Adding snapshot for '%s'\n", org_node->name);

	/* We found the original on the list. Verify the size to be sure the
	 * name didn't change for compatibility. For non-512-byte hardsector
	 * sizes, round down org node to a hardsector multiple to be the same
	 * as what was stored in the metadata.
	 */
	if ( (org_node->total_vsectors &
	      (~((org_node->hardsector_size/EVMS_VSECTOR_SIZE)-1))) !=
	     metadata->original_size ) {
		/* The snapshot no longer points at a valid original.
		 * Disable and delete the snapshot.
		 */
		LOG_ERROR("Error: Original volume size does not match for snapshot '%s'!\n",
			  snap_node->name);
		LOG_ERROR("         volume=%s: org_size="PFU64", current size="PFU64"\n",
			  org_node->name, metadata->original_size,
			  org_node->total_vsectors);
		set_snapshot_flags(snap_node, EVMS_SNAPSHOT_DISABLED, 0);
		DELETE(snap_node);
		return -ENODEV;
	}

	/* New EVMS node for the snapshot. */
	if ( evms_cs_allocate_logical_node(&new_snap_node) ) {
		set_snapshot_flags(snap_node, EVMS_SNAPSHOT_DISABLED, 0);
		DELETE(snap_node);
		return -ENOMEM;
	}

	MOD_INC_USE_COUNT;
	snapshot_count++;

	snapshot_create_pools();

	rc = initialize_snapshot_node(snap_node, new_snap_node,
				      org_node, metadata);
	if (rc) {
		return rc;
	}
	snap_volume = new_snap_node->private;

	/* Check to see if the node we found is one we put back on the list due
	 * to another snapshot of the original, if so then don't allocate a new
	 * node and volume info, just get the old one.
	 */
	if ( org_node->plugin->id != plugin_header.id ) {

		/* New EVMS node for the original. */
		if ( evms_cs_allocate_logical_node(&new_org_node) ) {
			disable_snapshot(snap_volume, TRUE);
			snap_delete_volume(new_snap_node);
			return -ENOMEM;
		}

		MOD_INC_USE_COUNT;
		snapshot_count++;

		rc = initialize_original_node(snap_node, new_snap_node,
					      org_node, new_org_node);
		if (rc) {
			return rc;
		}
		org_volume = new_org_node->private;

		/* Remove the original volume from the global list, then
		 * add the new version of the original to the global list.
		 */
		evms_cs_remove_logical_node_from_list(evms_node_list, org_node);
		evms_cs_add_logical_node_to_list(evms_node_list, new_org_node);
	} else {
		/* There is already at least one snapshot for this original. */
		new_org_node = org_node;
		org_volume = new_org_node->private;
		org_node = org_volume->logical_node;

		/* Make sure this snapshot matches the current
		 * chunk size if we have async snapshots.
		 */
		if ( snap_volume->chunk_size != org_volume->chunk_size ) {
			LOG_ERROR("Cannot add snapshot '%s' with chunk size %u to original '%s' with chunk size %u.\n",
				  new_snap_node->name, snap_volume->chunk_size,
				  new_org_node->name, org_volume->chunk_size);
			disable_snapshot(snap_volume, TRUE);
			snap_delete_volume(new_snap_node);
			return -EINVAL;
		}

		/* If the new snapshot is Removable or Partial, propogate
		 * the flags to the original and all other snapshots.
		 */
		for ( tmp_volume = org_volume;
		      tmp_volume;
		      tmp_volume = tmp_volume->snapshot_next) {
			tmp_volume->exported_node->flags |=
				(snap_node->flags &
				 (EVMS_DEVICE_REMOVABLE | EVMS_VOLUME_PARTIAL));
		}
	}

	/* Create a proc-fs entry for this snapshot. */
	if (snap_proc) {
		create_proc_read_entry(snap_node->feature_header->volume_name,
				       S_IFREG, snap_proc,
				       snap_proc_read, new_snap_node);
	}

	/* Insert the new snapshot at the start of the original's chain. */
	down_write(&org_volume->snap_semaphore);
	snap_volume->snapshot_next = org_volume->snapshot_next;
	org_volume->snapshot_next = snap_volume;
	snap_volume->snapshot_org = org_volume;
	up_write(&org_volume->snap_semaphore);

	/* Place the new snapshot on the global list. */
	evms_cs_add_logical_node_to_list(evms_node_list, new_snap_node);

	if ( metadata->flags & EVMS_SNAPSHOT_ROLLBACK ) {
		org_volume->flags |= EVMS_SNAPSHOT_ROLLBACK;
		snap_volume->flags |= EVMS_SNAPSHOT_ROLLBACK;
		evms_cs_wakeup_thread(snap_volume->rollback_thread);
	}

	return 0;
}

/**
 * do_rollback
 */
void snapshot_do_rollback(void * volume)
{
	struct snapshot_volume * snap_volume = volume;
	struct snapshot_volume * org_volume = snap_volume->snapshot_org;
	u32 io_size = snap_volume->chunk_size;
	u32 sectors = io_size;
	int done = FALSE;
	int i, iterations = 1;

	evms_cs_invalidate_volume(org_volume->exported_node);
	evms_cs_invalidate_volume(snap_volume->exported_node);

	/* Safety to start at chunk 0. */
	snap_volume->next_free_chunk = 0;
	while (!done) {

		if ( SNAPSHOT_CHUNK_BUFFER_SIZE < snap_volume->chunk_size ) {
			iterations = snap_volume->chunk_size /
				     org_volume->chunk_size;
			sectors = io_size = org_volume->chunk_size;
		}

		/* Read in one sector's worth of COW tables. */
		if ( INIT_IO(snap_volume->logical_node, 0,
			     snap_volume->current_cow_sector, 1,
			     snap_volume->cow_table) ) {
			LOG_ERROR("Error reading COW table from snapshot during rollback, aborting rollback\n");
			return;
		}

		/* Translate every valid COW table entry into
		 * a snapshot map entry.
		 */
		for ( snap_volume->next_cow_entry = 0;
		      snap_volume->next_cow_entry <
		       (EVMS_VSECTOR_SIZE/sizeof(u64)) &&
		      snap_volume->cow_table[snap_volume->next_cow_entry] !=
		       0xffffffffffffffff;
		      snap_volume->next_cow_entry++,
		      snap_volume->next_free_chunk++ ) {
			for ( i = 0; i < iterations; i++ ) {

				/* Don't go off the end of the original. */
				if ( io_size >
				     org_volume->logical_node->total_vsectors -
				      (snap_volume->cow_table[snap_volume->next_cow_entry] *
				       snap_volume->chunk_size + i * io_size) ) {
					sectors = org_volume->logical_node->total_vsectors -
						  (snap_volume->cow_table[snap_volume->next_cow_entry] *
						  snap_volume->chunk_size + i * io_size);
				}

				/* Read the chunk from the snapshot volume. */
				if ( INIT_IO(snap_volume->logical_node, READ,
					     (snap_volume->next_free_chunk *
					      snap_volume->chunk_size +
					      i*io_size),
					     sectors,
					     snap_volume->chunk_data_buffer) ) {
					LOG_ERROR("Error reading chunk %u from snapshot '%s'. Continuing.\n",
						  snap_volume->next_free_chunk,
						  snap_volume->logical_node->name);
				}

				/* Write the chunk to the original volume. */
				if ( INIT_IO(org_volume->logical_node, WRITE,
					     snap_volume->cow_table[snap_volume->next_cow_entry] *
					     snap_volume->chunk_size + i*io_size,
					     sectors,
					     snap_volume->chunk_data_buffer) ) {
					LOG_ERROR("Error writing chunk %u to original '%s' during rollback. Continuing.\n",
						  snap_volume->next_free_chunk,
						  org_volume->logical_node->name);
				}

				if ( sectors < io_size ) {
					break;
				}
			}
		}

		/* Move on to the next COW table sector if necessary. */
		if ( snap_volume->next_cow_entry ==
		     (EVMS_VSECTOR_SIZE/sizeof(u64)) ) {
			snap_volume->current_cow_sector++;
		} else {
			done = TRUE;
			snap_volume->flags |= EVMS_SNAPSHOT_DISABLED |
					      EVMS_SNAPSHOT_ROLLBACK_COMP;
			snap_volume->flags &= ~EVMS_SNAPSHOT_ROLLBACK;
			org_volume->flags &= ~EVMS_SNAPSHOT_ROLLBACK;
			set_snapshot_flags(snap_volume->logical_node,
					   EVMS_SNAPSHOT_DISABLED |
					   EVMS_SNAPSHOT_ROLLBACK_COMP,
					   EVMS_SNAPSHOT_ROLLBACK);
			LOG_DEFAULT("Rollback complete from snapshot %s\n",
				    snap_volume->exported_node->name);
		}
	}
}

/**
 * snap_proc_read
 *
 * Callback function for the proc-fs entry for each snapshot node.
 * Print out pertinent information about this snapshot. The "data"
 * parameter is a pointer to an EVMS logical node.
 */
static int snap_proc_read(char * page, char ** start, off_t off,
			  int count, int * eof, void * data)
{
	struct evms_logical_node * snap_node = data;
	struct snapshot_volume * snap_volume = snap_node->private;
	int sz = 0;

	PROCPRINT("Snapshot of    : %s\n",	(snap_volume->snapshot_org) ? EVMS_GET_NODE_NAME(snap_volume->snapshot_org->logical_node) : (u8 *)"Unknown");
	PROCPRINT("Size (KB)      : %u\n",	(snap_volume->num_chunks * snap_volume->chunk_size)/2);
	PROCPRINT("Chunk Size (KB): %u\n",	(snap_volume->chunk_size)/2);
	PROCPRINT("Writeable      : %s\n",	(snap_volume->flags & EVMS_SNAPSHOT_WRITEABLE) ? "Yes" : "No");
	PROCPRINT("Usage          : %u%%\n",	(snap_volume->next_free_chunk * 100) / snap_volume->num_chunks);
	PROCPRINT("Status         : %s\n", 	(snap_volume->flags & EVMS_SNAPSHOT_FULL) ? "Full / Disabled" : (snap_volume->flags & EVMS_SNAPSHOT_DISABLED) ? "Disabled" : "Active");
#ifdef SNAPSHOT_DEBUG
	PROCPRINT("Next free chunk: %u\n",	snap_volume->next_free_chunk);
	PROCPRINT("COW Writes     : %u\n", 	atomic_read(&snap_volume->cow_table_writes));
	PROCPRINT("COW Overlaps   : %u\n", 	atomic_read(&snap_volume->cow_table_overlaps));
#endif

out:
	*start = page + off;
	sz -= off;
	if (sz < 0)
		sz = 0;
	return sz > count ? count : sz;
}

/**
 * snapshot_init
 */
int __init snapshot_init(void)
{
	struct proc_dir_entry * pde;

	/* Register a directory in proc-fs. */
	pde = evms_cs_get_evms_proc_dir();
	if (pde) {
		snap_proc = create_proc_entry("snapshot", S_IFDIR, pde);
	}

	/* Register with EVMS. */
	return evms_cs_register_plugin(&plugin_header);
}

/**
 * snapshot_exit
 */
void __exit snapshot_exit(void)
{
	struct proc_dir_entry * pde;

	/* Unregister the directory in proc-fs. */
	pde = evms_cs_get_evms_proc_dir();
	if (pde) {
		remove_proc_entry("snapshot", pde);
	}
	
	evms_cs_unregister_plugin(&plugin_header);
}

module_init(snapshot_init);
module_exit(snapshot_exit);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

