/*
 * kernel/lvm-mp.c
 *
 * Copyright (C) 2001, Stefan Bader, IBM Development, Germany
 *
 *
 * LVM driver is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * LVM driver is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU CC; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

/*
 * $Id$
 * 
 * $Log$
 */
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
 
#include <linux/sem.h>
#include <linux/spinlock.h>
 
#include <linux/fs.h>
#include <linux/blk.h>
#include <linux/blkdev.h>

#include <linux/lvm.h>
#include "lvm-internal.h"

/*
 * Until there is some way to cooperativly use bh->b_private together
 * with ext3...
 */
#define LVM_MP_PRIVATE_PRIVATE

#ifdef CONFIG_ARCH_S390
#define CONFIG_S390DBF
#endif

#ifdef CONFIG_S390DBF
#include <asm/debug.h>
/*
 * S390 debug facility
 */
static debug_info_t *debug_info;
 
#define DBF_MSG_TYPE_MAP        1
#define DBF_MSG_TYPE_STARTIO    2
#define DBF_MSG_TYPE_ENDIO      3
#define DBF_MSG_TYPE_REQUEST    4
#define DBF_MSG_TYPE_ERROR      5

/*-----------------------------------------------------------------------------
 * All output to the S/390 debug facillity is done in fixed sized records.
 * So a message is defined as a struct where the id shows the type of the
 * message and the data area is union of the various private structures.
 * This way the length of a message is automatically the longest of all
 * private structs.
 *-----------------------------------------------------------------------------
 */
struct dbf_msg {
  unsigned int id;
  union {
    struct {
      kdev_t        srcdev;
      unsigned long srcsect;
      unsigned int  tgtpv;
      unsigned int  tgtpath;
      unsigned long tgtsect;
    } map;
    struct {
      uint32_t      vg_number;
      uint32_t      pv_number;
      uint32_t      pv_path;
      kdev_t        path_dev;
      int           path_pending_io;
    } start_io;
    struct {
      uint32_t      vg_number;
      uint32_t      pv_number;
      uint32_t      pv_path;
      kdev_t        path_dev;
      int           path_pending_io;
    } end_io;
    struct {
      uint32_t      vg_number;
      uint32_t      lv_number;
      unsigned long offset;
      unsigned long length;
    } request;
    struct {
      char         *text;
      int           errno;
      unsigned long line;
    } error;
  } data;
};
#endif

/*-----------------------------------------------------------------------------
 * Private end_io data structure
 * It would be much easier to store a pointer to the PV or the path.
 * But there may be the chance that this pointer is invalid because
 * the LVM layout changed during IO (maybe only a slight one...).
 *-----------------------------------------------------------------------------
 */
typedef struct {
	bh_end_io_t *old_end_io;
	void        *old_private;

	uint32_t     vg_number;
	uint32_t     pv_number;
	uint32_t     pv_path;
	kdev_t       path_dev;
	int          skiplist[PV_MAX_PATHS];
	int          rw;
} lvm_mp_private_t;

/*-----------------------------------------------------------------------------
 * This list contains all requests that failed and should be retried
 * on a different path.
 * The spinlock protects access to the queue.
 *-----------------------------------------------------------------------------
 */
static struct buffer_head *lvm_mp_fail_queue     = NULL;
static spinlock_t          lvm_mp_failqueue_lock = SPIN_LOCK_UNLOCKED;

/*-----------------------------------------------------------------------------
 * lvm_mp_failed_req signals entries in the retry queue (or teardown)
 * lvm_mp_thread_state_change is used to synchronize start/stop of the thread
 *-----------------------------------------------------------------------------
 */
static DECLARE_MUTEX_LOCKED(lvm_mp_failed_req);
static DECLARE_MUTEX_LOCKED(lvm_mp_thread_state_change);

/*=============================================================================
 * Internal function prototypes
 *=============================================================================
 */
#ifdef CONFIG_S390DBF
static int dbf_format_lvm_msg(
  debug_info_t      *id,
  struct debug_view *view,
  char              *out,
  const char        *in
);
#endif
static int lvm_mp_reroute_thread(
	void *data
);
static int lvm_mp_select_path(
	struct buffer_head *bh,
	int                 rw,
	pv_t               *pv,
	unsigned int       *pathno,
	int                *skiplist
);
static void lvm_mp_end_io(
	struct buffer_head *bh,
	int uptodate
);

/*=============================================================================
 * Function: dbf_format_lvm_msg (internal, optional)
 *
 * Takes one debug message entry and formats that into human readable text.
 *
 * Arguments:
 *   id   : pointer to id (not used)
 *   view : pointer to debug view structure (not used)
 *   in   : pointer to message data
 *   out  : pointer to write area
 *
 * Returns:
 *   The number of bytes written to the output area
 *=============================================================================
 */
#ifdef CONFIG_S390DBF
static int dbf_format_lvm_msg(
  debug_info_t      *id,
  struct debug_view *view,
  char              *out,
  const char        *in
) {
  struct dbf_msg *msg = (struct dbf_msg *) in;
  int             rc  = 0;
 
  switch(msg->id) {
    case DBF_MSG_TYPE_MAP:
      rc += sprintf(out, "MAP: RDEV(%3u:%-3u)+%-16lu -> (PV(%3u)->PATH(%2u))"
                    "+%-16lu\n",
                    MAJOR(msg->data.map.srcdev),
                    MINOR(msg->data.map.srcdev),
                    msg->data.map.srcsect, msg->data.map.tgtpv,
                    msg->data.map.tgtpath, msg->data.map.tgtsect);
      break;
    case DBF_MSG_TYPE_STARTIO:
      rc += sprintf(out, "START_IO: VG(%u)->PV(%u)->PATH(%u)->PENDING"
                    "(%3u:%-3u) = %i\n",
                    msg->data.start_io.vg_number,
                    msg->data.start_io.pv_number,
                    msg->data.start_io.pv_path,
                    MAJOR(msg->data.start_io.path_dev),
                    MINOR(msg->data.start_io.path_dev),
                    msg->data.start_io.path_pending_io);
      break;
    case DBF_MSG_TYPE_ENDIO:
      rc += sprintf(out, "END_IO: VG(%u)->PV(%u)->PATH(%u)->PENDING"
                    "(%3u:%-3u) = %i\n",
                    msg->data.end_io.vg_number,
                    msg->data.end_io.pv_number,
                    msg->data.end_io.pv_path,
                    MAJOR(msg->data.end_io.path_dev),
                    MINOR(msg->data.end_io.path_dev),
                    msg->data.end_io.path_pending_io);
      break;
    case DBF_MSG_TYPE_REQUEST:
      rc += sprintf(out, "LVM_REQUEST: VG(%u)->PV(%u)@%lu+%lu\n",
                    msg->data.request.vg_number,
                    msg->data.request.lv_number,
                    msg->data.request.offset,
                    msg->data.request.length);
      break;
    case DBF_MSG_TYPE_ERROR:
      rc += sprintf(out, "LVM_ERROR: %s at line %lu(%i)\n",
                    msg->data.error.text,
                    msg->data.error.line,
                    msg->data.error.errno);
      break;
    default:
      rc += sprintf(out, "Unknown message id(%d)\n", msg->id);
  }
  return(rc);
}

/*-----------------------------------------------------------------------------
 * A new debug view is registered by using this structure. It defines an own
 * formatter for the debug data.
 *-----------------------------------------------------------------------------
 */
struct debug_view lvm_msg_view = {
  "lvm_msg",
  NULL,
  &debug_dflt_header_fn,
  &dbf_format_lvm_msg,
  NULL,
  NULL
};
#endif

/*=============================================================================
 * Function: lvm_mp_get_from_private (internal)
 *
 * This function verifies the contents of a private structure and returns
 * pointers to the physical volume and/or the last used path.
 *
 * Arguments:
 *   priv     : pointer to the private structure to read from
 *   pv_ptr   : address of the pointer to store the used PV in (ignore the
 *              pointer and return nothing if this address is NULL)
 *   path_ptr : address of the pointer to store the used path in (ignored
 *              if NULL)
 *
 * Returns:
 *   0 : when the verification fails (this will also set the returned
 *       pv and path to NULL)
 *   1 : on success
 */
static inline int lvm_mp_get_from_private(
	lvm_mp_private_t  *priv,
	pv_t             **pv_ptr,
	pv_path_t        **path_ptr
) {
	/*
	 * Paranoia checks... (could it be possible that the setup changed
	 * while doing IO?)
	 */
	if(vg[priv->vg_number] != NULL) {
		pv_t *this_pv = (vg[priv->vg_number])->pv[priv->pv_number];

		if(pv_ptr != NULL)
			*pv_ptr = this_pv;

		if(this_pv != NULL) {
			pv_path_t *this_path =
				&this_pv->pv_path[priv->pv_path];

			if(path_ptr != NULL)
				*path_ptr = this_path;

			if(this_path->path_dev == priv->path_dev) {
				/* All is well return success */
				return(1);
			} else {
				printk(
					KERN_ALERT
					"lvm-mp: device %u:%u != %u:%u\n",
					MAJOR(this_path->path_dev),
					MINOR(this_path->path_dev),
					MAJOR(priv->path_dev),
					MINOR(priv->path_dev)
				);
			} /* endif(path_dev != saved_dev) */
		} else {
			printk(KERN_ALERT "lvm-mp: pv is NULL\n");
		} /* endif(pv_ptr == NULL) */
	} else {
		printk(KERN_ALERT "lvm-mp: vg is NULL\n");
	} /* endif(vg_ptr == NULL) */

	if(pv_ptr != NULL)
		*pv_ptr = NULL;
	if(path_ptr != NULL)
		*path_ptr = NULL;
	return(0);
}


/*=============================================================================
 * Function: lvm_mp_add_failed_bh (internal)
 *
 * Add a failed (with lvm_mp_private!) buffer_head to the retry queue.
 *
 * Arguments:
 *   bh : pointer to the failed buffer_head
 *
 * Returns:
 *   0 : on failure (should never happen, now)
 *   1 : on success
 *=============================================================================
 */
static int lvm_mp_add_failed_bh(struct buffer_head *bh) {
	unsigned long flags;
	
	spin_lock_irqsave(&lvm_mp_failqueue_lock, flags);
	bh->b_reqnext     = lvm_mp_fail_queue;
	lvm_mp_fail_queue = bh;
	spin_unlock_irqrestore(&lvm_mp_failqueue_lock, flags);
	up(&lvm_mp_failed_req);

	return(1);
}

/*=============================================================================
 * Function: lvm_mp_get_failed_bh (internal)
 *
 * Get the next (oldest) buffer_head out of the retry queue and return its
 * pointer.
 *
 * Returns:
 *   Pointer to the (FIFO) next buffer_head in the retry queue or NULL if
 *   the queue is empty.
 *=============================================================================
 */
static struct buffer_head *lvm_mp_get_failed_bh(void) {
	unsigned long        flags;
	struct buffer_head **prev_ptr;
	struct buffer_head  *bh      = NULL;

	spin_lock_irqsave(&lvm_mp_failqueue_lock, flags);
	if(*(prev_ptr = &lvm_mp_fail_queue) != NULL) {
		while((*prev_ptr)->b_reqnext != NULL)
			prev_ptr = &(*prev_ptr)->b_reqnext;

		bh        = (*prev_ptr);
		*prev_ptr = NULL;
	}
	spin_unlock_irqrestore(&lvm_mp_failqueue_lock, flags);

	return(bh);
}

/*=============================================================================
 * Function: vlm_mp_can_merge (internal)
 *
 * Tests whether a given buffer could be merged into the request queue of
 * the b_rdev.
 *=============================================================================
 */
#ifdef LVM_MP_OPTIMIZE_MERGE
static int lvm_mp_can_merge(
	struct buffer_head *bh,
	int                 rw
) {
	request_queue_t  *q   = blk_get_queue(bh->b_rdev);
	struct request   *req = NULL;
	int               rc  = 0;
	struct list_head *head;
	int               max;
	unsigned long     flags;

	/*
	 * Is there a queue for the device and has it been inititalized.
	 * If not I assume it's another mapper. As it could be possible
	 * that the layer below could merge. So just feed it a bit
	 * (MAX_SECTORS).
	 */
	if(!q || !(q->queue_head.next))
		return(1);

	head = &q->queue_head;
	spin_lock_irqsave(&io_request_lock, flags);

	if(!list_empty(head)) {
		if(q->head_active && !q->plugged)
			head = head->next;

		if(!max_sectors[MAJOR(bh->b_rdev)])
			max = MAX_SECTORS;
		else
			max = max_sectors[MAJOR(bh->b_rdev)][MINOR(bh->b_rdev)];

		if(rw == READA)
			rw = READ;

		/*
		 * This function is defined in drivers/block/elevator.c.
		 * Basically it just checks whether a bh is adjacent to
		 * an existing request. The merge could still fail based
		 * on an elevator specific decision. But I think this
		 * shouldn't hurt us too much here.
		 * FIXME: shouldn't rely on elevator_noop not to change
		 *        anything.
		 */
		if(
			elevator_noop_merge(q, &req, head, bh, rw, max)
				!=
			ELEVATOR_NO_MERGE
		)
			rc = 1;
	}
	spin_unlock_irqrestore(&io_request_lock, flags);

	return(rc);
}
#endif

/*=============================================================================
 * Function: lvm_mp_select_path (internal)
 *
 * Select one of the available paths to use next. The selected path should be
 * one of the enabled paths that has the least number of unfinished requests.
 * If there are two with the same number of pending requests either the one
 * with the higher weight or the first in the array is used.
 *
 * NOTE:
 *   Local interrupts must be disabled before calling this function.
 *   On successful completion the selected path is still locked and must be
 *   unlocked by the caller. 
 *
 * Arguments:
 *   bh       : pointer to the IO request structure
 *   rw       : IO operation to be performed (READ, READA, WRITE)
 *   pv       : pointer to the physical device structure to which the I/O
 *              should be done.
 *   pathno   : gives the number of the path to start the search and
 *              returns the number of the selected path
 *   skiplist : point to an array of booleans that tell which paths to skip
 *              regardless of their state (or NULL should all be used)
 *
 * Returns:
 *   0 : no valid path left
 *   1 : a valid path has been found (selected path is locked!)
 *=============================================================================
 */
static int lvm_mp_select_path(
	struct buffer_head *bh,
	int                 rw,
	pv_t               *pv,
	unsigned int       *pathno,
	int                *skiplist
) {
	int           rc      = 0;
#ifdef LVM_MP_OPTIMIZE_MERGE
	int           mergeok;
#endif
	unsigned int  path;
	unsigned int  i;
	pv_path_t    *cur_ptr;

	spin_lock(&pv->pv_lock);
	if(pv->pv_path_kept++ < 32) { /* FIXME: this should be changeable */
		bh->b_rdev = pv->pv_path[*pathno].path_dev;
#ifdef LVM_MP_OPTIMIZE_MERGE
		if(lvm_mp_can_merge(bh, rw)) {
#endif
			spin_lock(&pv->pv_path_lock[*pathno]);
			if(pv->pv_path[*pathno].path_state == lvm_path_enabled) {
				spin_unlock(&pv->pv_lock);
				return(1);
			}
			spin_unlock(&pv->pv_path_lock[*pathno]);
#ifdef LVM_MP_OPTIMIZE_MERGE
		}
#endif
	}

	pv->pv_path_kept = 0;
	spin_unlock(&pv->pv_lock);

	path       = *pathno;	
	cur_ptr    = &pv->pv_path[*pathno];

	if(cur_ptr->path_state == lvm_path_enabled) {
		rc = 1;
	
		if(skiplist != NULL)
			if(skiplist[path])
				rc = 0;
	}

lvm_mp_select_path_again:
	for(i=1; i < pv->pv_paths; i++) {
		path = (path + 1) % pv->pv_paths;

		/* Skip previously selected paths (if applicable) */
		if(skiplist != NULL)
			if(skiplist[path])
				continue;

		/*
		 * The path is unlocked here by purpose. Although there is a
		 * slight chance, that the information read is invalidated
		 * there is a much bigger danger of deadlock when
		 * lvm_mp_can_merge tries to grab the io_request_lock while
		 * the path is locked because lvm_mp_end_io is most time
		 * called with that one beeing locked already!
		 */
		bh->b_rdev = pv->pv_path[path].path_dev;

#ifdef LVM_MP_OPTIMIZE_MERGE
		mergeok    = lvm_mp_can_merge(bh, rw);
#endif
		
		spin_lock(&pv->pv_path_lock[path]);
		/*
		 * If the current paths state is 'waiting for retry' then
		 * decrement the fail counter (in this case skip counter).
		 * In case the counter reaches 0 change the state to 'do retry'
		 * and use this path once. Furter requests will skip this path
		 * until the one request has either failed (state will be
		 * 'waiting for retry' again) or succeeded (path will be
		 * re-enabled).
		 */
		if(
			pv->pv_path[path].path_state
				==
			lvm_path_disabled_wait_retry
		) {
			if(--pv->pv_path[path].path_fail_count == 0) {
				pv->pv_path[path].path_state =
					lvm_path_disabled_retry;
				*pathno = path;
				return(1);
			} else
				goto lvm_mp_select_path_next;
		} else if(pv->pv_path[path].path_state != lvm_path_enabled) {
			/* Ignore paths that are not 'enabled' */
			goto lvm_mp_select_path_next;
		}

#ifdef LVM_MP_OPTIMIZE_MERGE
		/*
		 * If the request can be merged to the request queue of this
		 * path then stop searching and use it.
		 */
		if(mergeok) {
			*pathno = path;
			return(1);
		}
#endif

		/*
		 * Is the currently selected path valid? Otherwise this is the
		 * first valid path we found. So take it.
		 */
		if(cur_ptr->path_state == lvm_path_enabled) {
			/*
			 * If paths have different weights, prefer the ones
			 * with the lesser value.
			 */
			if(pv->pv_path[path].path_weight > cur_ptr->path_weight)
				goto lvm_mp_select_path_next;

			/*
			 * If this path already has more pending request than
			 * the last one found check the next one.
		 	 */
			if(
				(pv->pv_path[path].path_pending_io)
					>
				cur_ptr->path_pending_io
			)
				goto lvm_mp_select_path_next;

#if 0
			/*
			 * If the pending count is equal to the number found
			 * before, take a closer look. Do not change the path
			 * if the weight is less or equal.
			 */
			if(
				pv->pv_path[path].path_pending_io
					==
				cur_ptr->path_pending_io
			)
				if(
					pv->pv_path[path].path_weight
						<=
					cur_ptr->path_weight
				)
					goto lvm_mp_select_path_next;
#endif
		}

		*pathno  = path;
		cur_ptr  = &pv->pv_path[path];
		rc       = 1;

lvm_mp_select_path_next:
		spin_unlock(&pv->pv_path_lock[path]);
	}

	/*
	 * There seems to be a valid path. But since we would get into trouble
	 * with the io_request lock (merge test) do some recheck here.
	 * (The reason for the trouble is the end_io function which is called
	 * with the io_request lock held. Then it tries to lock its path which
	 * may be locked from this function. And then the merge function is
	 * called here that also tries to get the io_request lock...)
	 */
	if(rc == 1)
	{
		spin_lock(&pv->pv_path_lock[*pathno]);
		if(cur_ptr->path_state != lvm_path_enabled) {
			rc = 0;
			spin_unlock(&pv->pv_path_lock[*pathno]);
			goto lvm_mp_select_path_again;
		}
		bh->b_rdev = pv->pv_path[*pathno].path_dev;
	}

	return(rc);
}

/*=============================================================================
 * Function: lvm_mp_reroute_thread (internal, kernel-thread)
 *
 * After beeing started, this thread waits for entries in the retry queue
 * (signaled through the lvm_mp_failed_req semaphore). Then it tries to
 * find an alternate path for each request and re-submits the request on
 * that path. If there is no path found the failure is finally reported
 * to the upper layer.
 * To stop the thread the semophore can be up'ed without any requests to
 * re-submit.
 * This thread is used to shift workload away from the end_io function
 * which supposedly should complete fast because IRQs are very likely
 * still disabled and we call generic_make_request here.
 *=============================================================================
 */
static int lvm_mp_reroute_thread(void *data) {
	struct buffer_head  *bh      = NULL;
	lvm_mp_private_t    *private = NULL;
	pv_t                *pv      = NULL;
	pv_path_t           *path    = NULL;

	daemonize();
	exit_files(current);

	sprintf(current->comm, "lvm-mpd");

	spin_lock_irq(&current->sigmask_lock);
	sigfillset(&current->blocked);
	/* Oder */
	/* current->exit_signal = SIGCHLD; */
	/* siginitsetinv(&current->blocked, sigmask(SIGKILL)); */
	flush_signals(current);
	spin_unlock_irq(&current->sigmask_lock);

	/*
	 * This thread should run at a priority equal or higher than bdflush,
	 * so buffers can be rerouted even if memory is tight
	 */
	current->policy = SCHED_OTHER;
	set_user_nice(current, -20);

	up(&lvm_mp_thread_state_change);
	for(;;) {
		down_interruptible(&lvm_mp_failed_req);

		/*
		 * To signal STOP to the thread lvm_mp_failed_req is upped
		 * without failed requests on the queue.
		 */
		bh = lvm_mp_get_failed_bh();
		if(!bh) break;

#ifdef LVM_MP_PRIVATE_PRIVATE
		private = (lvm_mp_private_t *) bh->b_mp_private;
#else
		private = (lvm_mp_private_t *) bh->b_private;
#endif
		if(!private) {
			printk(
				KERN_ERR
				"lvm: lvm_mp_private was NULL in reschedule!\n"
			);
			continue;
		}

		/* reschedule */
		if(lvm_mp_get_from_private(private, &pv, &path)) {
			/*
			 * If the failing path is still enabled then increment
			 * and test the fail counter (done here because this
			 * thread works sequentially). All requests that end
			 * here have failed (not uptodate).
			 */
			local_irq_disable();
			spin_lock_irq(&pv->pv_path_lock[private->pv_path]);

			if(path->path_state == lvm_path_disabled_retry) {
				path->path_fail_count = PATH_SKIP_ON_FAIL;
				path->path_state = lvm_path_disabled_wait_retry;
			} else if(path->path_state == lvm_path_enabled) {
				if(path->path_fail_count++ < PATH_MAX_FAIL) {
					printk(
						KERN_ERR
						"lvm-mp: vg#%d:pv#%d:path#%d "
						"failure #%d at offset %lu\n",
						private->vg_number,
						private->pv_number,
						private->pv_path,
						path->path_fail_count,
						bh->b_rsector
					);
				} else {
					printk(
						KERN_ERR
						"lvm-mp: vg#%d:pv#%d:path#%d "
						"disabled\n",
						private->vg_number,
						private->pv_number,
						private->pv_path
					);
					if(path->path_pending_io > 0) {
						printk(
							KERN_ERR
							"lvm-mp: "
							"vg#%d:pv#%d:path#%d "
							"path has requests to "
							"finish\n",
							private->vg_number,
							private->pv_number,
							private->pv_path
						);
						path->path_state = lvm_path_disabled_finish_requests;
					} else {
						printk(
							KERN_ERR
							"lvm_mp: "
							"vg#%d:pv#%d:path#%d "
							"path waits for "
							"retry\n",
							private->vg_number,
							private->pv_number,
							private->pv_path
						);
						path->path_fail_count = PATH_SKIP_ON_FAIL;
						path->path_state      = lvm_path_disabled_wait_retry;
					}
				}
			}
			spin_unlock(&pv->pv_path_lock[private->pv_path]);

			/*
			 * Make sure lvm_mp_select_path will not choose the
			 * path that was used before (it may not be disabled,
			 * yet).
			 */
			private->skiplist[private->pv_path] = 1;
			private->pv_path =
				(private->pv_path + 1) % pv->pv_paths;

			if(lvm_mp_select_path(bh, private->rw, pv,
				&private->pv_path, private->skiplist)) {
				/*
				 * To resubmit the request we have to update
				 * the private structure (used path, path
				 * device and pending requests) as well as the
				 * device to use (b_rdev) in the buffer_head.
				 */
				path = &pv->pv_path[private->pv_path];
				path->path_pending_io++;
				private->path_dev =
				bh->b_rdev        = path->path_dev;
				spin_unlock_irq(
					&pv->pv_path_lock[private->pv_path]
				);
				printk(
					KERN_ERR
					"lvm-mp: vg#%d:pv#%d:path#%d "
					"resubmitting request at offset=%lu\n",
					private->vg_number,
					private->pv_number,
					private->pv_path,
					bh->b_rsector
				);
				generic_make_request(private->rw, bh);
			} else /* no (more) valid paths */ {
				local_irq_enable();
				printk(
					KERN_ERR
					"lvm-mp: vg#%d:pv#%d:path#%d retry "
					"failed at offset %lu\n",
					private->vg_number,
					private->pv_number,
					private->pv_path,
					bh->b_rsector
				);
				bh->b_end_io  = private->old_end_io;
#ifdef LVM_MP_PRIVATE_PRIVATE
				bh->b_mp_private = NULL;
#else
				bh->b_private = private->old_private;
#endif
				kfree(private);
				bh->b_end_io(bh, 0);
			} /* endif(select path) */
		}
			
	} /* endfor(thread loop) */
	up(&lvm_mp_thread_state_change);

	return(0);
}

/*=============================================================================
 * Function: lvm_mp_end_io (internal, callback)
 *
 * This function is called when IO completes. Beside of keeping track on
 * certain internal stuff it restores the original pointers and calls the
 * old end_io function.
 *
 * Arguments:
 *   bh       : pointer to the buffer_head structure
 *   uptodate : flag whether io operation was successful (1) or not (0)
 *=============================================================================
 */
static void lvm_mp_end_io(struct buffer_head *bh, int uptodate) {
	unsigned long     flags;
	lvm_mp_private_t *lvm_private;
	pv_t             *this_pv;
	pv_path_t        *this_path;
#ifdef CONFIG_S390DBF
	struct dbf_msg   msg;
#endif

#ifdef LVM_MP_PRIVATE_PRIVATE
	lvm_private = bh->b_mp_private;
#else
	lvm_private = bh->b_private;
#endif

	if(lvm_mp_get_from_private(lvm_private, &this_pv, &this_path)) {
		/* All checks passed. Real work starts here... */
		spin_lock_irqsave(&this_pv->pv_path_lock[lvm_private->pv_path], flags);

		/*
		 * Reduce the number of unfinished requests. If the path has
		 * been disabled and all previous requests have finished then
		 * set the state to skip a certain number of requests before
		 * trying again to use this path.
		 */
		if(--this_path->path_pending_io == 0)
			if(
				this_path->path_state
					==
				lvm_path_disabled_finish_requests
			) {
				this_path->path_fail_count = PATH_SKIP_ON_FAIL;
				this_path->path_state =
					lvm_path_disabled_wait_retry;
			}

#ifdef CONFIG_S390DBF
		msg.id                          = DBF_MSG_TYPE_ENDIO;
		msg.data.end_io.vg_number       = lvm_private->vg_number;
		msg.data.end_io.pv_number       = lvm_private->pv_number;
		msg.data.end_io.pv_path         = lvm_private->pv_path;
		msg.data.end_io.path_dev        = lvm_private->path_dev;
		msg.data.end_io.path_pending_io = this_path->path_pending_io;
		debug_event(debug_info, 5, &msg, sizeof(msg));
#endif
		spin_unlock_irqrestore(
			&this_pv->pv_path_lock[lvm_private->pv_path],
			flags
		);

		/*
		 * Check whether I/O was successful. Set the counters
		 * appropriatly
		 */
		if(!uptodate) {
			/*
			 * Reschedule request here (changing the state of
			 * the path is done by the thread since that will
			 * sequentially).
			 */
			lvm_mp_add_failed_bh(bh);
			return;
		} /* endif(!uptodate) */ else {
			/*-----------------------------------------------------
			 * On a successfull retry the path will be enabled.
			 *-----------------------------------------------------
			 */
			if(this_path->path_state == lvm_path_disabled_retry) {
				this_path->path_fail_count = 0;
				this_path->path_state      = lvm_path_enabled;
				printk(
					KERN_WARNING
					"lvm-mp: vg#%d:pv#%d:path#%d "
					"re-enabled\n",
					lvm_private->vg_number,
					lvm_private->pv_number,
					lvm_private->pv_path
				);
			}
		} /* endif(uptodate) */
	} /* endif(lvm_private successfully verified) */

	bh->b_end_io  = lvm_private->old_end_io;
#ifdef LVM_MP_PRIVATE_PRIVATE
	bh->b_mp_private = NULL;
#else
	bh->b_private = lvm_private->old_private;
#endif
	kfree(lvm_private);
	bh->b_end_io(bh, uptodate);
}

/*=============================================================================
 * Function: lvm_mp_map (public)
 *
 * Changes the real device lvm remaps an IO request to one of the available
 * paths. This also registers lvm_mp_end_io as the function to be called as
 * soon as the device driver completed its work.
 *
 * Arguments:
 *   bh   : this is the request to be remapped (the values of b_rdev,
 *          b_end_io and b_private are changed)
 *   vgno : volume group to use (this is an index into the array of volume
 *          groups I need because minor->vgno is not global)
 *   pvno : physical volume to work with (index into the array of PVs of
 *          the VG)
 *          NOTICE: the number of paths has to be >1 !!!
 *   rw   : type of the request (as used in generic_make_request) needed
 *          to re-do the request on another path in case of failure
 *
 * Returns:
 *   0    : on failure (no valid path was found)
 *   1    : on success
 *=============================================================================
 */
int lvm_mp_map(struct buffer_head *bh, uint32_t vgno, uint32_t pvno, int rw)
{
	pv_t             *pv          = vg[vgno]->pv[pvno];
	lvm_mp_private_t *lvm_private;
	unsigned int      pathno;
#ifdef CONFIG_S390DBF
	struct dbf_msg msg;
#endif

	lvm_private = kmalloc(sizeof(lvm_mp_private_t), GFP_KERNEL);
	local_irq_disable();

	pathno = pv->pv_current_path;
	if(!lvm_mp_select_path(bh, rw, pv, &pathno, NULL)) {
		local_irq_enable();
		printk(KERN_ERR "lvm-mp: path mapping failed!\n");
		if(lvm_private != NULL)
			kfree(lvm_private);
		return(0);
	}
	pv->pv_current_path = pathno;

#ifdef CONFIG_S390DBF
	/* Make a record of the mapping */
	msg.id               = DBF_MSG_TYPE_MAP;
	msg.data.map.srcdev  = bh->b_dev;
	msg.data.map.srcsect = bh->b_rsector;
	msg.data.map.tgtpv   = pv->pv_number-1;
	msg.data.map.tgtpath = pathno;
	msg.data.map.tgtsect = bh->b_rsector;
	debug_event(debug_info, 5, &msg, sizeof(msg));
#endif

	if(lvm_private != NULL) {
		/*
		 * If we got some memory for out private structure, then
		 * fill in the values and update the pending IO counter.
		 */
		memset(lvm_private, 0, sizeof(lvm_mp_private_t));

		lvm_private->old_end_io  = bh->b_end_io;
#ifdef LVM_MP_PRIVATE_PRIVATE
		bh->b_mp_private         = lvm_private;
#else
		lvm_private->old_private = bh->b_private;
		bh->b_private            = lvm_private;
#endif
		bh->b_end_io             = lvm_mp_end_io;
		lvm_private->vg_number   = vgno;
		lvm_private->pv_number   = pvno;
		lvm_private->pv_path     = pathno;
		lvm_private->path_dev    = bh->b_rdev;
		lvm_private->rw          = rw;

		pv->pv_path[pathno].path_pending_io++;

#ifdef CONFIG_S390DBF
		/*
		 * Put together the the information for the debug message.
		 * FIXME: Should be optimized (especially the pending_io
		 *        part)!!!
		 */
		msg.id                            = DBF_MSG_TYPE_STARTIO;
		msg.data.start_io.vg_number       = lvm_private->vg_number;
		msg.data.start_io.pv_number       = lvm_private->pv_number;
		msg.data.start_io.pv_path         = lvm_private->pv_path;
		msg.data.start_io.path_dev        = lvm_private->path_dev;
		msg.data.start_io.path_pending_io =
			pv->pv_path[pathno].path_pending_io;
		debug_event(debug_info, 5, &msg, sizeof(msg));
#endif
	} /* endif(GotPrivateStructure) */

	spin_unlock_irq(&pv->pv_path_lock[pathno]);
	return(1);
} /* lvm_mp_map() */


/*=============================================================================
 * Function: lvm_mp_init (public)
 *
 * This function has to be called from the lvm init code. It optionally
 * register to the debug facility and starts up the thread that will be
 * re-submitting failed request.
 *=============================================================================
 */
void lvm_mp_init(void) {
#ifdef CONFIG_S390DBF
  debug_info = debug_register("lvm", 2, 2, sizeof(struct dbf_msg));
  debug_register_view(debug_info, &debug_hex_ascii_view);
  debug_register_view(debug_info, &lvm_msg_view);
#endif

	kernel_thread(lvm_mp_reroute_thread, NULL, CLONE_FS|CLONE_FILES);
	down(&lvm_mp_thread_state_change);
}

/*=============================================================================
 * Function: lvm_mp_cleanup (public)
 *
 * This must be called when removing the lvm (module). It stops the helper
 * thread and (optionally) unregisters from the debug facility.
 *=============================================================================
 */
void lvm_mp_cleanup(void) {
	/*
	 * Stop the requeue-thread.
	 */
	up(&lvm_mp_failed_req);
	down(&lvm_mp_thread_state_change);

#ifdef CONFIG_S390DBF
  debug_unregister(debug_info);
#endif
}
