/*
 * highmem.h: virtual kernel memory mappings for high memory
 *
 * Used in CONFIG_HIGHMEM systems for memory pages which
 * are not addressable by direct kernel virtual addresses.
 *
 * Copyright (C) 1999 Gerhard Wichert, Siemens AG
 *		      Gerhard.Wichert@pdb.siemens.de
 *
 *
 * Redesigned the x86 32-bit VM architecture to deal with 
 * up to 16 Terabyte physical memory. With current x86 CPUs
 * we now support up to 64 Gigabytes physical RAM.
 *
 * Copyright (C) 1999 Ingo Molnar <mingo@redhat.com>
 */

#ifndef _ASM_HIGHMEM_H
#define _ASM_HIGHMEM_H

#ifdef __KERNEL__

#include <linux/config.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <asm/kmap_types.h>
#include <asm/pgtable.h>

#ifdef CONFIG_DEBUG_HIGHMEM
#define HIGHMEM_DEBUG 1
#else
#define HIGHMEM_DEBUG 0
#endif

/* declarations for highmem.c */
extern unsigned long highstart_pfn, highend_pfn;

extern pte_t *kmap_pte;
extern pgprot_t kmap_prot;
extern pte_t *pkmap_page_table;

extern void kmap_init(void) __init;

enum km_serie_type {
	KM_SERIE_DEFAULT,
	KM_SERIE_PAGETABLE,
	KM_NR_SERIES,
};

#define LAST_PKMAP 1024
#define PKMAP_SIZE ((LAST_PKMAP*KM_NR_SERIES) << PAGE_SHIFT)
#define PKMAP_BASE (FIXADDR_START - PKMAP_SIZE - PAGE_SIZE) /* left a page in between */
#define PKMAP_NR(virt)  (((virt)-PKMAP_BASE) >> PAGE_SHIFT)
#define PKMAP_ADDR(nr)  (PKMAP_BASE + ((nr) << PAGE_SHIFT))

extern void * FASTCALL(kmap_high(struct page *page, int serie, int nonblocking));
extern void FASTCALL(kunmap_high(void * vaddr, int serie));

#define kmap(page) kmap_serie(page, KM_SERIE_DEFAULT, 0)
#define kmap_nonblock(page) kmap_serie(page, KM_SERIE_DEFAULT, 1)
#define kmap_pagetable(page) kmap_serie(page, KM_SERIE_PAGETABLE, 0)

static inline void *kmap_serie(struct page *page, int serie, int nonblocking)
{
	if (in_interrupt())
		out_of_line_bug();
	if (page < highmem_start_page)
		return page_address(page);
	return kmap_high(page, serie, nonblocking);
}

static inline void kunmap(struct page *page)
{
	if (in_interrupt())
		out_of_line_bug();
	if (page < highmem_start_page)
		return;
	kunmap_high(page->virtual, KM_SERIE_DEFAULT);
}

static inline void kunmap_vaddr(void *kvaddr)
{
	if (in_interrupt())
		out_of_line_bug();
	if ((unsigned long) kvaddr < PKMAP_BASE)
		return;
	kunmap_high(kvaddr, KM_SERIE_DEFAULT);
}

/*
 * The use of kmap_atomic/kunmap_atomic is discouraged - kmap/kunmap
 * gives a more generic (and caching) interface. But kmap_atomic can
 * be used in IRQ contexts, so in some (very limited) cases we need
 * it.
 */
static inline void *kmap_atomic(struct page *page, enum km_type type)
{
	enum fixed_addresses idx;
	unsigned long vaddr;

	if (page < highmem_start_page)
		return page_address(page);

	idx = type + KM_TYPE_NR*smp_processor_id();
	vaddr = __fix_to_virt(FIX_KMAP_BEGIN + idx);
#if HIGHMEM_DEBUG
	if (!pte_none(*(kmap_pte-idx)))
		out_of_line_bug();
#endif
	set_pte(kmap_pte-idx, mk_pte(page, kmap_prot));
	__flush_tlb_one(vaddr);

	return (void*) vaddr;
}

static inline void kunmap_atomic(void *kvaddr, enum km_type type)
{
#if HIGHMEM_DEBUG
	unsigned long vaddr = (unsigned long) kvaddr;
	enum fixed_addresses idx = type + KM_TYPE_NR*smp_processor_id();

	if (vaddr < FIXADDR_START) // FIXME
		return;

	if (vaddr != __fix_to_virt(FIX_KMAP_BEGIN+idx))
		out_of_line_bug();

	/*
	 * force other mappings to Oops if they'll try to access
	 * this pte without first remap it
	 */
	pte_clear(kmap_pte-idx);
	__flush_tlb_one(vaddr);
#endif
}

static inline void kunmap_pagetable(void *kvaddr)
{
	/* direct map */
	if ((unsigned long) kvaddr < PKMAP_BASE)
		return;
	/* atomic kmap */
	if ((unsigned long) kvaddr >= FIXADDR_START) {
		kvaddr = (void *) ((unsigned long) kvaddr & PAGE_MASK);
		kunmap_atomic(kvaddr, KM_USER0);
		return;
	}
	if (in_interrupt())
		out_of_line_bug();
	kunmap_high(kvaddr, KM_SERIE_PAGETABLE);
}

#endif /* __KERNEL__ */

#endif /* _ASM_HIGHMEM_H */
