/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/* linux/include/linux/evms/evms_bbr_k.h
 *
 * Kernel header file for Bad Block Relocation (BBR) Feature
 *
 * BBR feature is designed to remap I/O write failures to another safe location
 * on disk. Note that most disk drives have BBR built into them, this means
 * that our software BBR will be only activated when all hardware BBR
 * replacement sectors have been used.
 */

#ifndef __EVMS_BBR_K__
#define __EVMS_BBR_K__ 1

#define EVMS_BBR_VERSION_MAJOR			1
#define EVMS_BBR_VERSION_MINOR			1
#define EVMS_BBR_VERSION_PATCHLEVEL		2

#define EVMS_BBR_COMMON_SERVICES_MAJOR		0
#define EVMS_BBR_COMMON_SERVICES_MINOR		6
#define EVMS_BBR_COMMON_SERVICES_PATCHLEVEL	0

#define EVMS_BBR_FEATURE_ID			6
#define EVMS_BBR_SIGNATURE			0x42627246 /* BbrF */
#define EVMS_BBR_TABLE_SIGNATURE		0x42627254 /* BbrT */

#define EVMS_BBR_ENTRIES_PER_SECT		31
#define BBR_POOL_NAME_LENGTH			20
#define BBR_STOP_REMAP				(1<<0)
#define BBR_BH_USE_EVMS_CALLBACK		(1<<0)

/* BBR direct ioctl commands.
 *
 * BBR_GET_INFO_CMD:	Return total number of sectors that are currently
 *			remapped for the specified BBR object.
 * BBR_STOP_REMAP_CMD:	Stop remapping. Do not remap any new sectors or even
 *			honor any existing remaps for the specified BBR object
 *			until the next rediscover command is received.
 * BBR_SECTOR_IO_CMD:	Process an I/O from the engine directly through the
 *			specified BBR object in the kernel.
 */
#define BBR_GET_INFO_CMD	1
#define BBR_STOP_REMAP_CMD	2
#define BBR_SECTOR_IO_CMD	3

/**
 * struct evms_bbr_table_entry
 * @bad_sect:		LBA of bad location.
 * @replacement_sect:	LBA of new location.
 *
 * Structure to describe one BBR remap.
 */
struct evms_bbr_table_entry {
	u64 bad_sect;
	u64 replacement_sect;
};

/**
 * struct evms_bbr_table
 * @signature:		Signature on each BBR table sector.
 * @crc:		CRC for this table sector.
 * @sequence_number:	Used to resolve conflicts when primary and secondary
 *			tables do not match.
 * @in_use_cnt:		Number of in-use table entries.
 * @entries:		Actual table of remaps.
 *
 * Structure to describe each sector of the metadata table. Each sector in this
 * table can describe 31 remapped sectors.
 */
struct evms_bbr_table {
	u32				signature;
	u32				crc;
	u32				sequence_number;
	u32				in_use_cnt;
	struct evms_bbr_table_entry	entries[EVMS_BBR_ENTRIES_PER_SECT];
};

/**
 * struct evms_bbr_metadata
 * @signature:			0	EVMS_BBR_SIGNATURE
 * @crc:			4
 * @block_size:			8	Block size in bytes.
 * @flags:			12	Global flags used by BBR.
 * @sequence_number:		16
 * @start_sect_bbr_table:	24	LBA of start of BBR table.
 * @nr_sects_bbr_table:		32	Number of sectors in the BBR table.
 * @start_replacement_sect:	40	LBA of start of replacement sectors.
 * @nr_replacement_blks:	48	Number of replacement sectors.
 * @pads:			56
 *
 * On-disk metadata identifying an object as a BBR object.
 */
struct evms_bbr_metadata {
	u32	signature;
	u32	crc;
	u32	block_size;
	u32	flags;
	u64	sequence_number;
	u64	start_sect_bbr_table;
	u64	nr_sects_bbr_table;
	u64	start_replacement_sect;
	u64	nr_replacement_blks;
	u8	pads[456];
};

/**
 * struct evms_notify_bbr
 * @object_name:	Input - Name of BBR object from feature header.
 * @count:		Output - Number of remapped sectors.
 * @start_sect:		Input - Start sector for sector_io.
 * @nr_sect:		Input - Number of sectors for sector_io.
 * @buffer:		Input/Output - Pointer to data buffer for sector_io.
 * @rw:			Input - READ or WRITE for sector_io.
 */
struct evms_notify_bbr {
	u8	object_name[EVMS_VOLUME_NAME_SIZE+1];
	u64	count;
	u64	start_sect;
	u64	nr_sect;
	u8	* buffer;
	s32	rw;
};

/**
 * struct bbr_runtime_remap
 *
 * Node in the binary tree used to keep track of remaps.
 */
struct bbr_runtime_remap {
	struct evms_bbr_table_entry	remap;
	struct bbr_runtime_remap	* left;
	struct bbr_runtime_remap	* right;
};

/**
 * struct bbr_private
 * @next:			List of all bbr_private structures.
 * @node:			Output node.
 * @source:			Consumed node.
 * @bbr_table:			Copy of metadata table.
 * @lba_table1:			LBA of primary BBR table.
 * @lba_table2:			LBA of secondary BBR table.
 * @nr_sects_bbr_table:		Size of each BBR table.
 * @nr_replacement_blks:	Number of replacement sectors.
 * @start_replacement_sect:	LBA of start of replacement sectors.
 * @blksize_in_sects:		Size of each sector.
 * @in_use_replacement_blks:	Current number of remaps.
 * @remap_root:			Binary tree containing all remaps.
 * @bbr_id_lock:		Lock for the binary tree.
 * @flag:			BBR_STOP_REMAP
 */
struct bbr_private {
	struct bbr_private		* next;
	struct evms_logical_node	* node;
	struct evms_logical_node	* source;
	struct evms_bbr_table		* bbr_table;
	u64				lba_table1;
	u64				lba_table2;
	u64				nr_sects_bbr_table;
	u64				nr_replacement_blks;
	u64				start_replacement_sect;
	u32				blksize_in_sects;
	atomic_t			in_use_replacement_blks;
	struct bbr_runtime_remap	* remap_root;
	spinlock_t			bbr_id_lock;
	u32				flag;
};

/**
 * struct bbr_io_buffer
 * @bbr_io_list:	Thread's list of bbr_io_buf's.
 * @bbr_id:		Object for this request.
 * @bh:			Original buffer_head.
 * @org_end_io:		Saved callback address from original buffer_head.
 * @org_private:	Saved private data address from original buffer_head.
 * @org_rsector:	Saved sector value from original buffer_head.
 * @org_dev:		Saved b_rdev field from original buffer_head.
 * @complete:		Completion structure used by init_io.
 * @rw:			READ or WRITE.
 * @rc:			Return code from bbr_io_handler.
 *
 * Structure used to track each write request.
 */
struct bbr_io_buffer {
	struct list_head	bbr_io_list;
	struct bbr_private	* bbr_id;
	struct buffer_head	* org_bh;
	struct completion	* complete;
	s32			rw;
	s32			rc;
};

#ifdef EVMS_BBR_DEBUG
static void print_meta_data(struct evms_bbr_metadata * md);
static void print_bbr_table_sector(struct evms_bbr_table * bbr_table);
static void print_remap_list(struct bbr_private * bbr_id);
#define BBR_DEBUG_PRINT_META_DATA(md) print_meta_data(md)
#define BBR_DEBUG_PRINT_TABLE_SECTOR(table) print_bbr_table_sector(table)
#define BBR_DEBUG_PRINT_REMAP_LIST(bbr_id) print_remap_list(bbr_id)
#else
#define BBR_DEBUG_PRINT_META_DATA(md)
#define BBR_DEBUG_PRINT_TABLE_SECTOR(table)
#define BBR_DEBUG_PRINT_REMAP_LIST(bbr_id)
#endif

#endif
