/*
 *
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: linux/include/linux/evms_os2.h
 */

/*
 * Change History:
 *
 */

/*
 * Description:  This module defines the disk structures used by the OS/2
 *               Logical Volume Manager, including that of the Master
 *               Boot Record (MBR) and Extended Boot Records (EBR).
 *
 * Notes: LVM Drive Letter Assignment Tables (DLA_Tables) appear on the
 *        last sector of each track containing a valid MBR or EBR.  Since
 *        partitions must be track aligned, any track containing an MBR or
 *        EBR will be almost all empty sectors.  We will grab the last
 *        of these empty sectors for our DLT_Tables.
 *
 */

#ifndef OS2LVM_INCLUDED__
#define OS2LVM_INCLUDED__

/* The following define the values used to indicate that a partition table entry is for an EBR, not a partition. */
#define EBR_BOOT_INDICATOR     0
#define EBR_FORMAT_INDICATOR   5

/* The following define is used as the default Format_Indicator for new non-primary partitions. */
#define NEW_LOGICAL_DRIVE_FORMAT_INDICATOR   0x6

/* The following define is used as the default Format_Indicator for a new non-active primary partitions. */
#define NEW_PRIMARY_PARTITION_FORMAT_INDICATOR   0x16

/* The following define is used as the default Format_Indicator for a new active primary partition. */
#define NEW_ACTIVE_PRIMARY_PARTITION_FORMAT_INDICATOR  0x06

/* The following define is used to hold the value of the Boot_Indicator for active partitions. */
#define ACTIVE_PARTITION   0x80

/* Define the size of a Partition Name.  Partition Names are user defined names given to a partition. */
#define PARTITION_NAME_SIZE  20

/* Define the size of a volume name.  Volume Names are user defined names given to a volume. */
#define VOLUME_NAME_SIZE  20

/* Define the size of a disk name.  Disk Names are user defined names given to physical disk drives in the system. */
#define DISK_NAME_SIZE    20

/* The name of the filesystem in use on a partition.  This name may be up to 12 ( + NULL terminator) characters long. */
#define FILESYSTEM_NAME_SIZE 20

/* The comment field is reserved but is not currently used.  This is for future expansion and use. */
#define COMMENT_SIZE 81

/* Define the minimum number of sectors to reserve on the disk for Boot Manager. */
#define BOOT_MANAGER_SIZE     2048

#define OS2_BYTES_PER_SECTOR  512
#define OS2_SECTOR_SHIFT      9

/*--------------------------------------------------
 * Type definitions
 --------------------------------------------------*/

/* The following definitions define the drive letter assignment table used by LVM.
   For each partition table on the disk, there will be a drive letter assignment table in the last sector
   of the track containing the partition table. */

/* NOTE: DLA stands for Drive Letter Assignment. */

#define DLA_TABLE_SIGNATURE1  0x424D5202L
#define DLA_TABLE_SIGNATURE2  0x44464D50L

struct dla_entry {	/* DE */
	u32 Volume_Serial_Number;	/* The serial number of the volume that this partition belongs to. */
	u32 partition_serial;	/* The serial number of this partition. */
	u32 Partition_Size;	/* The size of the partition, in sectors. */
	u32 Partition_Start;	/* The starting sector of the partition. */
	unsigned char On_Boot_Manager_Menu;	/* Set to TRUE if this volume/partition is on the Boot Manager Menu. */
	unsigned char Installable;	/* Set to TRUE if this volume is the one to install the operating system on. */
	char Drive_Letter;	/* The drive letter assigned to the partition. */
	unsigned char Reserved;
	char Volume_Name[VOLUME_NAME_SIZE];	/* The name assigned to the volume by the user. */
	char Partition_Name[PARTITION_NAME_SIZE];	/* The name assigned to the partition. */
};

struct dla_table_sector {	/* DTS */
	u32 DLA_Signature1;	/* The magic signature (part 1) of a Drive Letter Assignment Table. */
	u32 DLA_Signature2;	/* The magic signature (part 2) of a Drive Letter Assignment Table. */
	u32 DLA_CRC;	/* The 32 bit CRC for this sector.  Calculated assuming that this field and all unused space in the sector is 0. */
	u32 Disk_Serial_Number;	/* The serial number assigned to this disk. */
	u32 Boot_Disk_Serial_Number;	/* The serial number of the disk used to boot the system.  This is for conflict resolution when multiple volumes
						   want the same drive letter.  Since LVM.EXE will not let this situation happen, the only way to get this situation
						   is for the disk to have been altered by something other than LVM.EXE, or if a disk drive has been moved from one
						   machine to another.  If the drive has been moved, then it should have a different Boot_Disk_Serial_Number.  Thus,
						   we can tell which disk drive is the "foreign" drive and therefore reject its claim for the drive letter in question.
						   If we find that all of the claimaints have the same Boot_Disk_Serial_Number, then we must assign drive letters on
						   a first come, first serve basis. */
	u32 Install_Flags;	/* Used by the Install program. */
	u32 Cylinders;
	u32 Heads_Per_Cylinder;
	u32 Sectors_Per_Track;
	char Disk_Name[DISK_NAME_SIZE];	/* The name assigned to the disk containing this sector. */
	unsigned char Reboot;	/* For use by Install.  Used to keep track of reboots initiated by install. */
	unsigned char Reserved[3];	/* Alignment. */
	struct dla_entry DLA_Array[4];	/* These are the four entries which correspond to the entries in the partition table. */
};

/* The following definitions define the LVM signature sector which will appear as the last sector in an LVM partition. */

#define  OS2LVM_PRIMARY_SIGNATURE   0x4A435332L
#define  OS2LVM_SECONDARY_SIGNATURE 0x4252444BL

#define  CURRENT_OS2LVM_MAJOR_VERSION_NUMBER   2	/* Define as appropriate. */
#define  CURRENT_OS2LVM_MINOR_VERSION_NUMBER   0	/* Define as appropriate. */

/* The following definitions limit the number of LVM features that can be applied to a volume, as well as defining a "NULL" feature for use in feature table entries that are not being used. */
#define  OS2LVM_MAX_FEATURES_PER_VOLUME  10	/* The maximum number of LVM features that can be applied to a volume. */
#define  OS2LVM_NULL_FEATURE              0	/* No feature.  Used in all unused entries of the feature array in the LVM Signature sector. */

/* The following structure is used to hold the location of the feature specific data for LVM features. */
typedef struct _LVM_Feature_Data {	/* LFD */
	u32 Feature_ID;	/* The ID of the feature. */
	u32 Location_Of_Primary_Feature_Data;	/* The u32 of the starting sector of the private data for this feature. */
	u32 Location_Of_Secondary_Feature_Data;	/* The u32 of the starting sector of the backup copy of the private data for this feature. */
	u32 Feature_Data_Size;	/* The number of sectors used by this feature for its private data. */
	u16 Feature_Major_Version_Number;	/* The integer portion of the version number of this feature. */
	u16 Feature_Minor_Version_Number;	/* The decimal portion of the version number of this feature. */
	unsigned char Feature_Active;	/* TRUE if this feature is active on this partition/volume, FALSE otherwise. */
	unsigned char Reserved[3];	/* Alignment. */
} LVM_Feature_Data;

/* The following structure defines the LVM Signature Sector.  This is the last sector of every partition which is part of an LVM volume.  It gives vital
   information about the version of LVM used to create the LVM volume that it is a part of, as well as which LVM features (BBR, drive linking, etc.) are
   active on the volume that this partition is a part of.                                                                                                   */
typedef struct _LVM_Signature_Sector {	/* LSS */
	u32 LVM_Signature1;	/* The first part of the magic LVM signature. */
	u32 LVM_Signature2;	/* The second part of the magic LVM signature. */
	u32 Signature_Sector_CRC;	/* 32 bit CRC for this sector.  Calculated using 0 for this field. */
	u32 partition_serial;	/* The LVM assigned serial number for this partition.  */
	u32 Partition_Start;	/* u32 of the first sector of this partition. */
	u32 Partition_End;	/* u32 of the last sector of this partition. */
	u32 Partition_Sector_Count;	/* The number of sectors in this partition. */
	u32 LVM_Reserved_Sector_Count;	/* The number of sectors reserved for use by LVM. */
	u32 Partition_Size_To_Report_To_User;	/* The size of the partition as the user sees it - i.e. (the actual size of the partition - LVM reserved sectors) rounded to a track boundary. */
	u32 Boot_Disk_Serial_Number;	/* The serial number of the boot disk for the system.  If the system contains Boot Manager, then this is the serial number of the disk containing the active copy of Boot Manager. */
	u32 Volume_Serial_Number;	/* The serial number of the volume that this partition belongs to. */
	u32 Fake_EBR_Location;	/* The location, on disk, of a Fake EBR, if one has been allocated. */
	u16 LVM_Major_Version_Number;	/* Major version number of the LVM that created this partition. */
	u16 LVM_Minor_Version_Number;	/* Minor version number of the LVM that created this partition. */
	char Partition_Name[PARTITION_NAME_SIZE];	/* User defined partition name. */
	char Volume_Name[VOLUME_NAME_SIZE];	/* The name of the volume that this partition belongs to. */
	LVM_Feature_Data LVM_Feature_Array[OS2LVM_MAX_FEATURES_PER_VOLUME];	/* The feature array.  This indicates which LVM features, if any, are active on this volume
										   and what order they should be applied in.                                                  */
	char Drive_Letter;	/* The drive letter assigned to the volume that this partition is part of. */
	unsigned char Fake_EBR_Allocated;	/* If TRUE, then a fake EBR has been allocated. */
	char Comment[COMMENT_SIZE];	/* User comment. */
	char Disk_Name[DISK_NAME_SIZE];	/* Added to allow BBR to report the name of a disk when bad sectors are encountered on that disk. */
	u32 Sequence_Number;	/* This indicates the order that partitions within a volume are used.  This number is 1 based.  A 0 here indicates that the volume was made by LVM Ver. 1. */
	u32 Next_Aggregate_Number;	/* Used during volume creation and expansion when creating unique names for aggregates. */
	/* The remainder of the sector is reserved for future use and should be all zero or else the CRC will not come out correctly. */
} LVM_Signature_Sector;

/* The following definitions define the format of a partition table and the Master Boot Record (MBR). */
typedef struct _Partition_Record {	/* PR */
	unsigned char Boot_Indicator;	/* 80h = active partition. */
	unsigned char Starting_Head;
	unsigned char Starting_Sector;	/* Bits 0-5 are the sector.  Bits 6 and 7 are the high order bits of the starting cylinder. */
	unsigned char Starting_Cylinder;	/* The cylinder number is a 10 bit value.  The high order bits of the 10 bit value come from bits 6 & 7 of the Starting_Sector field. */
	unsigned char Format_Indicator;	/* An indicator of the format/operation system on this partition. */
	unsigned char Ending_Head;
	unsigned char Ending_Sector;
	unsigned char Ending_Cylinder;
	u32 Sector_Offset;	/* The number of sectors on the disk which are prior to the start of this partition. */
	u32 Sector_Count;	/* The number of sectors in this partition. */
} Partition_Record;

typedef struct _Master_Boot_Record {	/* MBR */
	unsigned char Reserved[446];
	Partition_Record Partition_Table[4];
	u16 Signature;	/* AA55h in this field indicates that this is a valid partition table/MBR. */
} Master_Boot_Record;

typedef Master_Boot_Record Extended_Boot_Record;

/* The following definition covers the Boot Manager Alias Table in the EBR.

   The Alias Table in the EBR has 2 entries in it, although only the first one is actually used.  */
#define ALIAS_NAME_SIZE  8
typedef struct _AliasTableEntry {	/* ATE */
	unsigned char On_Boot_Manager_Menu;
	char Name[ALIAS_NAME_SIZE];
} AliasTableEntry;

#define ALIAS_TABLE_OFFSET  0x18A

/* XLATOFF */
/* The following text is used for the Boot Manager Alias for items that were placed on the Boot Manager Menu by FDISK and
   which have since been migrated to the new LVM format.  This text is put into the Name field of an AliasTableEntry so
   that, if FDISK ( or another program which understands the old Boot Manager Menu format) is run, it will display
   something for those partitions/volumes which are on the Boot Manager Menu.

   NOTE: This text must be exactly ALIAS_NAME_SIZE characters in length!                                                     */
#define ALIAS_TABLE_ENTRY_MIGRATION_TEXT       "--> LVM "
#define ALIAS_TABLE_ENTRY_MIGRATION_TEXT2      "--> LVM*"

/* XLATON */

/* The following is the signature used for an Master Boot Record, an Extended Boot Record, and a Boot Sector. */
#define MBR_EBR_SIGNATURE  0xAA55

/* The following list of definitions defines the values of interest for the Format_Indicator in a Partition_Record. */
#define EBR_INDICATOR                          0x5
#define WINDOZE_EBR_INDICATOR                  0xF
#define UNUSED_INDICATOR                       0x0
#define IFS_INDICATOR                          0x7
#define FAT12_INDICATOR                        0x1
#define FAT16_SMALL_PARTITION_INDICATOR        0x4
#define FAT16_LARGE_PARTITION_INDICATOR        0x6
#define BOOT_MANAGER_HIDDEN_PARTITION_FLAG     0x10
#define LVM_PARTITION_INDICATOR                0x35
#define BOOT_MANAGER_INDICATOR                 0x0A

/* The following is the signature used in the Boot Sector for Boot Manager. */
#define OS2LVM_BOOT_MANAGER_SIGNATURE       "APJ&WN"

/* The following is used for determining the synthetic geometry reported for Volumes employing drive linking. */
#define OS2LVM_SYNTHETIC_SECTORS_PER_TRACK  63

/*--------------------------------------------------
 * Declares for Drive Linking feature:
 *--------------------------------------------------*/

/* The following defines uniquely identify Drive Linking. */
#define DRIVE_LINKING_FEATURE_ID     100
#define DRIVE_LINKING_MAJOR_VERSION  1
#define DRIVE_LINKING_MINOR_VERSION  0

/* The following definitions are used for the disk structures supporting drive linking. */

#define LINK_TABLE_MASTER_SIGNATURE  0x434E4157L
#define LINK_TABLE_SIGNATURE         0X4D4D5652L

#define MAXIMUM_LINKS   246

#define DRIVE_LINKING_RESERVED_SECTOR_COUNT 4

#define LINKS_IN_FIRST_SECTOR 60

#define LINKS_IN_NEXT_SECTOR  62

struct drive_link {
	u32 drive_serial;
	u32 partition_serial;
};

struct link_table_first_sector {
	u32 Link_Table_Signature;	/* Use the LINK_TABLE_MASTER_SIGNATURE here. */
	u32 Link_Table_CRC;
	u32 Sequence_Number;	/* Used to resolve conflicts when the primary and secondary tables do not match. */
	u32 Links_In_Use;
	struct drive_link Link_Table[LINKS_IN_FIRST_SECTOR];
};

struct link_table_sector {
	u32 Link_Table_Signature;	/* Use LINK_TABLE_SIGNATURE here. */
	u32 Link_Table_CRC;
	u32 Sequence_Number;	/* Used to resolve conflicts when the primary and secondary tables do not match. */
	struct drive_link Link_Table[LINKS_IN_NEXT_SECTOR];
};

/*--------------------------------------------------
 * Declares for Bad Block Relocation feature:
 *--------------------------------------------------*/

/* The following definition is the numeric ID for Bad Block Relocation.  */
#define BBR_FEATURE_ID  101

#define BBR_FEATURE_MAJOR_VERSION       0x0001
#define BBR_FEATURE_MINOR_VERSION       0x0000

/* The following definitions are used for the disk structures supporting bad block relocation. */

/* NOTE: BBR stands for Bad Block Relocation. */

#define BBR_TABLE_MASTER_SIGNATURE  0x00726D62
#define BBR_TABLE_SIGNATURE         0x01726276

struct bbr_table_entry {
	u32 BadSector;
	u32 ReplacementSector;
};

typedef struct _LVM_BBR_Table_First_Sector {
	u32 Signature;	/* Signature for the first sector of the BBR Table. Use BBR_TABLE_MASTER_SIGNATURE here. */
	u32 CRC;		/* CRC for this sector. */
	u32 Sequence_Number;	/* Used to resolve conflicts when the primary and secondary tables do not match. */
	u32 Table_Size;	/* The number of BBR_Table_Entries in the BBR Table. */
	u32 Table_Entries_In_Use;	/* The number of BBR Table entries which are in use. */
	u32 Sectors_Per_Table;	/* The number of LVM_BBR_Table_Sectors used to hold the BBR Table. */
	u32 First_Replacement_Sector;	/* The location of the first replacement sector. */
	u32 Last_Replacement_Sector;	/* The location of the last replacement sector. */
	u32 Replacement_Sector_Count;	/* The number of replacement sectors. */
	u32 Flags;	/* Flags global to the Bad Block Relocation Feature. */
} LVM_BBR_Table_First_Sector;

/*  Flags for LVM_BBR_Table_First_Sector  */
#define BBR_Flag_Write_Verify    0x00000001	/* Indicate convert Write I/O to Write/Verify */

#define BBR_TABLE_ENTRIES_PER_SECTOR   62

typedef struct _LVM_BBR_Table_Sector {
	u32 Signature;	/* Signature for a sector of the BBR_Table which is not the first sector of the BBR Table. Use BBR_TABLE_SIGNATURE here. */
	u32 CRC;		/* CRC for this sector of the BBR Table. */
	u32 Sequence_Number;	/* Used to resolve conflicts when the primary and secondary tables do not match. */
	struct bbr_table_entry BBR_Table[BBR_TABLE_ENTRIES_PER_SECTOR];
	u32 reserved1;	/* for block alignment */
} LVM_BBR_Table_Sector;

//
// Combined structure to hold entire BBR feature data as it exists on disk.
typedef struct _LVM_BBR_Feature {
	LVM_BBR_Table_First_Sector control;
	char reserved1[OS2_BYTES_PER_SECTOR -
		       sizeof (LVM_BBR_Table_First_Sector)];
	LVM_BBR_Table_Sector remap[1];
}
LVM_BBR_Feature;

/* The following defines establish the minimum and maximum number of replacement sectors which can be allocated for
   Bad Block Relocation.  Otherwise, 1 replacement sector per MB of disk space is allocated.                          */
#define BBR_FLOOR    62
#define BBR_LIMIT  4096

// In-memory Meta Data for Bad Block Relocation
// In-memory Meta Data for Drive Linking
struct os2_dl_entry {
	u64 start_sector;
	u64 sector_count;
	u64 dl_lsn1;		/* LSN of first on-disk copy of drive linking data. */
	u64 dl_lsn2;		/* LSN of the second on-disk copy of drive linking data. */
	char *link_data;
	u32 partition_serial;
	u64 bbr_lsn1;	/* LSN of the first on-disk copy of the BBR data. */
	u64 bbr_lsn2;	/* LSN of the second on-disk copy of the BBR data. */
	u32 bbr_feature_size;	/* # of sectors of BBR data. */
	u32 bbr_is_active;
	struct semaphore bbr_table_lock;	/* Used to serialize writers */
	unsigned int guard1;	/* Lamport's Theorem for mutual exclusion */
	char *bbr_data;
	unsigned int guard2;	/* Lamport's Theorem for mutual exclusion */
	struct evms_logical_node *link_partition;
	struct os2_dl_entry *next;
};

// In-memory Meta Data for each OS/2 LVM Volume:
typedef struct os2_volume_runtime_entry_s {
	int complete;
	u32 Export_Needed;
	u64 size_in_sectors;
	u32 Volume_Serial_Number;
	u32 drive_link_count;
	struct os2_dl_entry *drive_link;
	struct evms_logical_node *next_os2lvm_node;
} os2_volume_runtime_entry_t;

#endif
