/*
 * linux/mii.h: definitions for MII-compatible transceivers
 * Originally drivers/net/sunhme.h.
 *
 * Copyright (C) 1996, 1999, 2001 David S. Miller (davem@redhat.com)
 * 
 * 2002-08-08: symbols for LXT970 MII transceiver added (<gernot@hillier.de>)
 *
 */

#ifndef __LINUX_MII_H__
#define __LINUX_MII_H__

#include <linux/types.h>

/* Generic MII registers (0,1 mandatory, 2-7 optional). */
#define MII_BMCR            0x00        /* Basic mode control register */
#define MII_BMSR            0x01        /* Basic mode status register  */
#define MII_PHYSID1         0x02        /* PHYS ID 1                   */
#define MII_PHYSID2         0x03        /* PHYS ID 2                   */
#define MII_ADVERTISE       0x04        /* Advertisement control reg   */
#define MII_LPA             0x05        /* Link partner ability reg    */
#define MII_EXPANSION       0x06        /* Auto-Negotiation Expansion  */
#define MII_NEXTPAGE	    0x07        /* Auto-Negotiation Next Page  */
/* 0x07 - 0x0F: reserved */

/* vendor specific MII registers. */
#define MII_DCOUNTER        0x12        /* Disconnect counter          */
#define MII_FCSCOUNTER      0x13        /* False carrier counter       */
#define MII_NWAYTEST        0x14        /* N-way auto-neg test reg     */
#define MII_RERRCOUNTER     0x15        /* Receive error counter       */
#define MII_SREVISION       0x16        /* Silicon revision            */
#define MII_RESV1           0x17        /* Reserved...                 */
#define MII_LBRERROR        0x18        /* Lpback, rx, bypass error    */
#define MII_PHYADDR         0x19        /* PHY address                 */
#define MII_RESV2           0x1a        /* Reserved...                 */
#define MII_TPISTATUS       0x1b        /* TPI status for 10mbps       */
#define MII_NCONFIG         0x1c        /* Network interface config    */

/* vendor specific MII registers for Level One LXT970 transceiver (see LXT970 Data Sheet, Revision 1.1 of May 1997). */
#define MII_LXT970_MIRROR   0x10	/* Mirror Register	       */
#define MII_LXT970_INTENAB  0x11	/* Interrupt Enable Register   */
#define MII_LXT970_INTSTAT  0x12	/* Interrupt Status Register   */
#define MII_LXT970_CONFIG   0x13	/* Config Register	       */
#define MII_LXT970_CHIPSTAT 0x14	/* Chip Status Register	       */
 

/* Basic mode control register. */
#define BMCR_RESV               0x007f  /* Unused...                   */
#define BMCR_CTST               0x0080  /* Collision test              */
#define BMCR_FULLDPLX           0x0100  /* Full duplex                 */
#define BMCR_ANRESTART          0x0200  /* Auto negotiation restart    */
#define BMCR_ISOLATE            0x0400  /* Disconnect DP83840 from MII */
#define BMCR_PDOWN              0x0800  /* Powerdown the DP83840       */
#define BMCR_ANENABLE           0x1000  /* Enable auto negotiation     */
#define BMCR_SPEED100           0x2000  /* Select 100Mbps              */
#define BMCR_LOOPBACK           0x4000  /* TXD loopback bits           */
#define BMCR_RESET              0x8000  /* Reset the DP83840           */

/* Basic mode status register. */
#define BMSR_ERCAP              0x0001  /* Ext-reg capability          */
#define BMSR_JCD                0x0002  /* Jabber detected             */
#define BMSR_LSTATUS            0x0004  /* Link status                 */
#define BMSR_ANEGCAPABLE        0x0008  /* Able to do auto-negotiation */
#define BMSR_RFAULT             0x0010  /* Remote fault detected       */
#define BMSR_ANEGCOMPLETE       0x0020  /* Auto-negotiation complete   */
#define BMSR_MFPREAMBLESUP      0x0040  /* MF with suppressed preamble */
#define BMSR_MSCONFIGFAULT      0x0080  /* Master-Slave config fault   */
#define BMSR_RESV               0x0100  /* Unused...                   */
#define BMSR_100BASE2HALF       0x0200  /* Can do 100BASET2, half-dupl */
#define BMSR_100BASE2FULL       0x0400  /* Can do 100BASET2, full-dupl */
#define BMSR_10HALF             0x0800  /* Can do 10mbps, half-duplex  */
#define BMSR_10FULL             0x1000  /* Can do 10mbps, full-duplex  */
#define BMSR_100HALF            0x2000  /* Can do 100mbps, half-duplex */
#define BMSR_100FULL            0x4000  /* Can do 100mbps, full-duplex */
#define BMSR_100BASE4           0x8000  /* Can do 100mbps, 4k packets  */

/* Advertisement control register. */
#define ADVERTISE_SLCT          0x001f  /* Selector bits               */
#define ADVERTISE_CSMA          0x0001  /* Only selector supported     */
#define ADVERTISE_10HALF        0x0020  /* Try for 10mbps half-duplex  */
#define ADVERTISE_10FULL        0x0040  /* Try for 10mbps full-duplex  */
#define ADVERTISE_100HALF       0x0080  /* Try for 100mbps half-duplex */
#define ADVERTISE_100FULL       0x0100  /* Try for 100mbps full-duplex */
#define ADVERTISE_100BASE4      0x0200  /* Try for 100mbps 4k packets  */
#define ADVERTISE_RESV          0x1c00  /* Unused...                   */
#define ADVERTISE_RFAULT        0x2000  /* Say we can detect faults    */
#define ADVERTISE_LPACK         0x4000  /* Ack link partners response  */
#define ADVERTISE_NPAGE         0x8000  /* Next page bit               */

#define ADVERTISE_FULL (ADVERTISE_100FULL | ADVERTISE_10FULL | \
			ADVERTISE_CSMA)
#define ADVERTISE_ALL (ADVERTISE_10HALF | ADVERTISE_10FULL | \
                       ADVERTISE_100HALF | ADVERTISE_100FULL)

/* Link partner ability register. */
#define LPA_SLCT                0x001f  /* Same as advertise selector  */
#define LPA_10HALF              0x0020  /* Can do 10mbps half-duplex   */
#define LPA_10FULL              0x0040  /* Can do 10mbps full-duplex   */
#define LPA_100HALF             0x0080  /* Can do 100mbps half-duplex  */
#define LPA_100FULL             0x0100  /* Can do 100mbps full-duplex  */
#define LPA_100BASE4            0x0200  /* Can do 100mbps 4k packets   */
#define LPA_RESV                0x1c00  /* Unused...                   */
#define LPA_RFAULT              0x2000  /* Link partner faulted        */
#define LPA_LPACK               0x4000  /* Link partner acked us       */
#define LPA_NPAGE               0x8000  /* Next page bit               */

#define LPA_DUPLEX		(LPA_10FULL | LPA_100FULL)
#define LPA_100			(LPA_100FULL | LPA_100HALF | LPA_100BASE4)

/* Expansion register for auto-negotiation. */
#define EXPANSION_NWAY          0x0001  /* Can do N-way auto-nego      */
#define EXPANSION_LCWP          0x0002  /* Got new RX page code word   */
#define EXPANSION_ENABLENPAGE   0x0004  /* This enables npage words    */
#define EXPANSION_NPCAPABLE     0x0008  /* Link partner supports npage */
#define EXPANSION_MFAULTS       0x0010  /* Multiple faults detected    */
#define EXPANSION_RESV          0xffe0  /* Unused...                   */

/* N-way test register. */
#define NWAYTEST_RESV1          0x00ff  /* Unused...                   */
#define NWAYTEST_LOOPBACK       0x0100  /* Enable loopback for N-way   */
#define NWAYTEST_RESV2          0xfe00  /* Unused...                   */

/* LXT970 Interrupt Enable Register. */
#define LXT970_INTENAB_TINT	0x0001  /* force MDINT		       */
#define LXT970_INTENAB_INTEN    0x0002  /* enable interrupts	       */
#define LXT970_INTENAB_RESV	0xfffc  /* Unused...		       */

/* LXT970 Interrupt Status Register. */
#define LXT970_INTSTAT_RESV	0x3fff  /* Unused...		       */
#define LXT970_INTSTAT_XTALOK	0x4000	/* XTAL circuit available      */ 
#define LXT970_INTSTAT_MINT	0x8000  /* MII interrupt pending       */

/* LXT970 Configuration Register. */
#define LXT970_CONFIG_TRANSDISC 0x0001  /* disconnect TP transmitter   */
#define LXT970_CONFIG_RESV1	0x0002	/* Unused...		       */
#define LXT970_CONFIG_100BASEFX 0x0004  /* enable fiber interface      */
#define LXT970_CONFIG_NOSCRAM   0x0008  /* bypass scramblers           */
#define LXT970_CONFIG_5BSYMBOL  0x0010  /* 5-bit Symbol Mode	       */
#define LXT970_CONFIG_TXCLOCK	0x0020  /* special TX clock mode       */
#define LXT970_CONFIG_LEDC_COLL 0x0000  /* LEDC indicates collision    */
#define LXT970_CONFIG_LEDC_OFF  0x0040  /* LEDC off                    */
#define LXT970_CONFIG_LEDC_ACT  0x0080  /* LEDC indicates activity     */
#define LXT970_CONFIG_LEDC_ON   0x00c0  /* LEDC continuously on        */
#define LXT970_CONFIG_NOLINKTST 0x0100  /* disable 10BaseT link test   */
#define LXT970_CONFIG_NOJABBER  0x0200  /* disable 10BaseT jabber      */
#define LXT970_CONFIG_SQE	0x0400  /* enable 10BaseT SQE	       */
#define LXT970_CONFIG_NOTPLOOPB 0x0800  /* disable 10BaseT TP loopback */
#define LXT970_CONFIG_MDIOINT   0x1000  /* enable MDIO interrupt       */
#define LXT970_CONFIG_REPEATER 	0x2000  /* enable repeater mode	       */
#define LXT970_CONFIG_TXTEST 	0x4000  /* enable 100BaseT transmit tst*/
#define LXT970_CONFIG_RESV2	0x8000  /* Unused...		       */

/* LXT970 Chip Status Register. */
#define LXT970_CHIPSTAT_NOPLL	0x0001	/* 100BaseTX/FX recv not locked*/
#define LXT970_CHIPSTAT_RESV1   0x0002  /* Unused...		       */
#define LXT970_CHIPSTAT_LVFAULT 0x0004  /* low voltage fault on VCC    */
#define LXT970_CHIPSTAT_RESV2	0x0008  /* Unused...		       */
#define LXT970_CHIPSTAT_MLT3ERR 0x0010  /* MLT3 encoding error 	       */
#define LXT970_CHIPSTAT_SYMBERR 0x0020  /* Symbol Error detected       */
#define LXT970_CHIPSTAT_STREAML 0x0040  /* scrambler/decoder locked    */ 
#define LXT970_CHIPSTAT_RESV3	0x0080	/* Unused...		       */
#define LXT970_CHIPSTAT_PAGERCV 0x0100  /* duplicate of EXPANSION_LCWP */
#define LXT970_CHIPSTAT_ANCOMPL 0x0200  /* dupl of BMSR_ANEGCOMPLETE   */
#define LXT970_CHIPSTAT_RESV4   0x0400  /* Unused...		       */
#define LXT970_CHIPSTAT_100MBPS 0x0800  /* 100 Mbps operation	       */
#define LXT970_CHIPSTAT_FULLDPL 0x1000  /* full duplex operation       */
#define LXT970_CHIPSTAT_LINKOK  0x2000  /* link is up (dynamic)        */
#define LXT970_CHIPSTAT_RESV5  	0xc000  /* Unused...		       */


struct mii_if_info {
	int phy_id;
	int advertising;
	int phy_id_mask;
	int reg_num_mask;

	unsigned int full_duplex : 1;	/* is full duplex? */
	unsigned int force_media : 1;	/* is autoneg. disabled? */

	struct net_device *dev;
	int (*mdio_read) (struct net_device *dev, int phy_id, int location);
	void (*mdio_write) (struct net_device *dev, int phy_id, int location, int val);
};

struct ethtool_cmd;
struct mii_ioctl_data;

extern int mii_link_ok (struct mii_if_info *mii);
extern int mii_nway_restart (struct mii_if_info *mii);
extern int mii_ethtool_gset(struct mii_if_info *mii, struct ethtool_cmd *ecmd);
extern int mii_ethtool_sset(struct mii_if_info *mii, struct ethtool_cmd *ecmd);
extern void mii_check_link (struct mii_if_info *mii);
extern unsigned int mii_check_media (struct mii_if_info *mii,
				     unsigned int ok_to_print,
				     unsigned int init_media);
extern int generic_mii_ioctl(struct mii_if_info *mii_if,
                      	     struct mii_ioctl_data *mii_data, int cmd,
			     unsigned int *duplex_changed);



/* This structure is used in all SIOCxMIIxxx ioctl calls */
struct mii_ioctl_data {
	u16		phy_id;
	u16		reg_num;
	u16		val_in;
	u16		val_out;
};


/**
 * mii_nway_result
 * @negotiated: value of MII ANAR and'd with ANLPAR
 *
 * Given a set of MII abilities, check each bit and returns the
 * currently supported media, in the priority order defined by
 * IEEE 802.3u.  We use LPA_xxx constants but note this is not the
 * value of LPA solely, as described above.
 *
 * The one exception to IEEE 802.3u is that 100baseT4 is placed
 * between 100T-full and 100T-half.  If your phy does not support
 * 100T4 this is fine.  If your phy places 100T4 elsewhere in the
 * priority order, you will need to roll your own function.
 */
static inline unsigned int mii_nway_result (unsigned int negotiated)
{
	unsigned int ret;

	if (negotiated & LPA_100FULL)
		ret = LPA_100FULL;
	else if (negotiated & LPA_100BASE4)
		ret = LPA_100BASE4;
	else if (negotiated & LPA_100HALF)
		ret = LPA_100HALF;
	else if (negotiated & LPA_10FULL)
		ret = LPA_10FULL;
	else
		ret = LPA_10HALF;

	return ret;
}

/**
 * mii_duplex
 * @duplex_lock: Non-zero if duplex is locked at full
 * @negotiated: value of MII ANAR and'd with ANLPAR
 *
 * A small helper function for a common case.  Returns one
 * if the media is operating or locked at full duplex, and
 * returns zero otherwise.
 */
static inline unsigned int mii_duplex (unsigned int duplex_lock,
				       unsigned int negotiated)
{
	if (duplex_lock)
		return 1;
	if (mii_nway_result(negotiated) & LPA_DUPLEX)
		return 1;
	return 0;
}


#endif /* __LINUX_MII_H__ */
